@@ -1,5 +1,5 @@
 /* cooledit.c - main file of cooledit
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,6 +20,7 @@
 #include <config.h>
 #include <stdio.h>
 #include <stdlib.h>
+
 #include <signal.h>
 #include <sys/types.h>
 #if HAVE_SYS_WAIT_H
@@ -58,12 +59,20 @@
 #include <ndir.h>
 #endif
 #endif
+
+#ifdef HAVE_LANGINFO_H
+#include <langinfo.h>
+#endif
+
 #include "mad.h"
 
+extern struct look *look;
+
 void do_mail (CWidget *edit);
 
 int start_width = 80;
 int start_height = 30;
+char *option_look = "gtk";
 
 /* error handler */
 #ifdef DEBUG
@@ -130,18 +139,7 @@ static RETSIGTYPE quit_handler (int x)
 }
 
 void set_to_kill (pid_t p);
-
-static RETSIGTYPE child_handler (int x)
-{
-    pid_t p;
-
-    p = waitpid (0, 0, WNOHANG | WUNTRACED);
-    set_to_kill (p);
-    debug_check (p);
-
-    signal (SIGCHLD, child_handler);
-    handler_return;
-}
+void shell_output_set_to_kill (pid_t p);
 
 static void set_signals (void)
 {
@@ -156,7 +154,6 @@ static void set_signals (void)
     signal (SIGINT, SIG_DFL);
     signal (SIGTERM, SIG_DFL);
 #endif
-    signal (SIGCHLD, child_handler);
 }
 
 /* }}} signal handling */
@@ -207,9 +204,11 @@ void get_next_hint_message (void)
     char label[128];
     static char *hints[] =
     {
+/* HINTSTART */
 /* The following are hints that go on the title bar: eg "Cooledit - Hint: Undo key-for-key with Ctrl-Backspace" */
 	gettext_noop("To drag and drop, highlight text, then click and drag from within the selection"),
 	gettext_noop("Dragging with the right mouse button will move text"),
+	gettext_noop("Goto http://www.obsidian.co.za/ - Obsidian Systems - Linux IT services"),
 	gettext_noop("Dragging with the left mouse button will copy text"),
 	gettext_noop("Use Shift-Insert to copy text from other applications"),
 	gettext_noop("Use Alt-Insert to bring up a history of selected text"),
@@ -225,6 +224,7 @@ void get_next_hint_message (void)
 	gettext_noop("Undo key-for-key with Ctrl-Backspace"),
 	gettext_noop("Drag function prototypes and other info from the man page browser"),
 	gettext_noop("Double click on words in the man page browser to search for new man pages"),
+	gettext_noop("Goto http://www.obsidian.co.za/ - Obsidian Systems - Linux IT services"),
 	gettext_noop("Search for [Options] in  ~/.cedit/.cooledit.ini  and change settings"),
 	gettext_noop("Edit the example Scripts to see how to create your own"),
 	gettext_noop("Double click on gcc error messages after running Make or Cc"),
@@ -240,6 +240,7 @@ void get_next_hint_message (void)
 	gettext_noop("Read all these hint messages in editor/cooledit.c in the source distribution"),
 	gettext_noop("See the file INTERNATIONAL for composing international characters"),
 	gettext_noop("Edit/move at high speed with the many Ctrl/Meta-Arrow combinations"),
+	gettext_noop("Goto http://www.obsidian.co.za/ - Obsidian Systems - Linux IT services"),
 	gettext_noop("Use drag-and-drop to and from input widgets"),
 	gettext_noop("Open a file by dragging its name from the 'File browser' onto the background"),
 	gettext_noop("Use Shift-Ins to paste into input widgets"),
@@ -259,13 +260,28 @@ void get_next_hint_message (void)
 	gettext_noop("Try your hand at adding new Cooledit features with Python"),
 	gettext_noop("Mark locations in files with the book mark commands in the Edit menu"),
 	gettext_noop("Debug programs with Meta-F5"),
+	gettext_noop("Goto http://www.obsidian.co.za/ - Obsidian Systems - Linux IT services"),
 	gettext_noop("Execute explicit debugger commands with Meta-F1"),
 	gettext_noop("Turn off spell checking from the Options menu"),
 	gettext_noop("File files containing regular expressions with Ctrl-Meta-F"),
 	gettext_noop("Delete debugger variables with Del in the variable dialog"),
+	gettext_noop("Add debugger watches with Ins in the variable dialog"),
 	gettext_noop("Make a variable watch point by marking it with Ins in the variable dialog"),
-	gettext_noop("Change the default size of the undo stack in ~/.cedit/.cooledit.ini")
+	gettext_noop("Change the default size of the undo stack in ~/.cedit/.cooledit.ini"),
+	gettext_noop("Bookmark all matching strings with the `Bookmarks' switch in the `Search' dialog"),
+	gettext_noop("Create a printable postscript manual with   man -t cooledit > cooledit.ps"),
+	gettext_noop("Read the Python online tutorials: http://www.python.org/"),
+	gettext_noop("Enable \"Smooth scrolling\" option if you have a fast graphics card"),
+	gettext_noop("Start internal XTerms with \"Terminal\" in the command menu"),
+	gettext_noop("Use the coolproject command to start projects"),
+	gettext_noop("Goto http://www.obsidian.co.za/ - Obsidian Systems - Linux IT services"),
+	gettext_noop("Display Unicode text - See FAQ"),
+	gettext_noop("All colours can have their RGB values set - see COLOURS in the man page"),
+	gettext_noop("Try different looks with the \"--look\" command line option"),
+	gettext_noop("Press Escape to enter shell commands"),
+	gettext_noop("Select text then use Escape to pipe through a shell command"),
     };
+/* HINTEND */
 
     if (i < 0) {
 	n = sizeof (hints) / sizeof (char *);
@@ -302,24 +318,34 @@ void edit_change_directory (void)
 int load_setup (const char *file);
 int save_setup (const char *file);
 
-static char *command_line_files[] =
-{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+static char **command_line_files;
 
 char *editor_options_file = 0;
 extern char *option_preferred_visual;
 extern int option_force_own_colormap;
 extern int option_force_default_colormap;
+extern int option_invert_colors;
+extern int option_invert_crome;
+extern int option_invert_red_green;
+extern int option_invert_green_blue;
+extern int option_invert_red_blue;
 
 int option_suppress_load_files = 0;
 int option_suppress_load_files_cmdline = 0;
 int option_suppress_load_options = 0;
-int option_save_setup_on_exit = 0;
+int option_save_setup_on_exit = 1;
 int option_hint_messages = 120;
 int option_cursor_blink_rate = 7;
 int option_pull_down_window_list = 1;
 int option_toolbar = 1;
 int option_minimal_main_window = 0;
+extern int option_utf_interpretation;
+extern int option_no_font_set;
 extern char *option_geometry;
+extern int option_use_xim;
+int option_aa_font = -1;
+extern int option_rgb_order;
+extern int option_interchar_spacing;
 
 int option_new_window_ask_for_file = 1;
 
@@ -330,6 +356,7 @@ char *option_foreground_red = 0;
 char *option_foreground_green = 0;
 char *option_foreground_blue = 0;
 char *option_font = 0;
+char *option_widget_font = 0;
 
 static int option_command_line_doesnt_override = 0;
 static int get_help = 0;
@@ -356,11 +383,13 @@ void usage (void)
 	    "-m, --minimal-main-window                main window border only\n" \
 	    "-lines <n>                               number of text lines\n" \
 	    "-columns <n>                             number of text columns\n" \
+	    "--edit-bg <nn>                           <nn> 0-26 (rather edit .cooledit.ini)\n" \
 	    "-bg, --background-color <color>          eg blue for solid blue, default: igloo\n" \
 	    "-R, --foreground-red <value>             red component, default: 0.9\n" \
 	    "-G, --foreground-green <value>           green component, default: 1.1\n" \
 	    "-B, --foreground-blue <value>            blue component, default: 1.4\n" \
 	    "-f, -fn, -font <font-name>               use   cooledit -font h   for help\n" \
+	    "--widget-font <font-name>                font of widgets and controls\n" \
 	    "-S, --suppress-load-files                don't load saved desktop\n" \
 	    "-U, --suppress-load-options              don't load saved options\n" \
 	    "-E, -no-override                         command line doesn't override init file\n" \
@@ -381,6 +410,22 @@ void usage (void)
 	    "-no-half, --no-fake-half-tabs            turn off half tabbing\n" \
 	    "-toolbar                                 edit windows have a toolbar\n" \
 	    "-no-toolbar                              disabled\n" \
+	    "--no-xim                                 disable XIM support\n" \
+	    "-utf8, --utf8-interpretation             interpret UTF-8 encoding\n" \
+	    "-no-fontset, --no-fontset                use raw fonts\n" \
+	    "--anti-aliasing                          depreciated, use <fontname>/3\n" \
+	    "--red-first                              R-G-B LCD sub-pixel order aliasing\n" \
+	    "--blue-first                             B-G-R LCD sub-pixel order aliasing\n" \
+	    "--interchar-spacing <n>                  extra spacing between chars in pixels\n" \
+	    "                                         (anti-aliased fonts only)\n" \
+	    "--interwidget-spacing <n>                spacing between widgets in pixels\n" \
+	    "--look [gtk|cool|next]                   look emulation. defaults to gtk\n" \
+	    "--[no-]bright-invert                     invert the brightness of all colours\n" \
+            "                                         (useful for white backgrounds)\n" \
+	    "--[no-]chromin-invert                    invert the chrominance of all colours\n" \
+	    "--[no-]swap-red-green                    swap red with green\n" \
+	    "--[no-]swap-green-blue                   swap green with blue\n" \
+	    "--[no-]swap-red-blue                     swap red with blue\n" \
 	    "-A, -save-setup                          save setup on exit\n" \
 	    "-P, -no-save-setup                       don't save setup on exit\n" \
 	    "-W, --whole-chars-search <chars>         characters that constitute a whole word\n" \
@@ -418,7 +463,7 @@ struct prog_options {
 
 struct prog_options cooledit_options[] =
 {
-    {' ', "", "", ARG_STRINGS, 0, command_line_files, 0},
+    {' ', "", "", ARG_STRINGS, 0, 0, 0},
     {0, "-bg", "--background-color", ARG_STRING, &option_background_color, 0, 0},
 #ifdef GUESS_VISUAL
     {0, "-vis", "--visual", ARG_STRING, &option_preferred_visual, 0, 0},
@@ -431,6 +476,7 @@ struct prog_options cooledit_options[] =
     {'G', "", "--foreground-green", ARG_STRING, &option_foreground_green, 0, 0},
     {'B', "", "--foreground-blue", ARG_STRING, &option_foreground_blue, 0, 0},
     {'f', "-fn", "-font", ARG_STRING, &option_font, 0, 0},
+    {0, "", "--widget-font", ARG_STRING, &option_widget_font, 0, 0},
     {'S', "", "--suppress-load-files", ARG_SET, 0, 0, &option_suppress_load_files_cmdline},
     {'U', "", "--suppress-load-options", ARG_SET, 0, 0, &option_suppress_load_options},
     {'E', "-no-override", "", ARG_SET, 0, 0, &option_command_line_doesnt_override},
@@ -439,7 +485,6 @@ struct prog_options cooledit_options[] =
     {0, "-wordwrap", "--word-wrap", ARG_INT, 0, 0, &option_word_wrap_line_length},
     {0, "-typew", "--type-writer", ARG_SET, 0, 0, &option_typewriter_wrap},
     {0, "-autop", "--auto-paragraph", ARG_SET, 0, 0, &option_auto_para_formatting},
-    {0, "-noi", "--no-international-characters", ARG_CLEAR, 0, 0, &option_international_characters},
     {'t', "-tab", "--tab-spacing", ARG_INT, 0, 0, &option_tab_spacing},
     {'s', "-space", "--space-filled-tabs", ARG_SET, 0, 0, &option_fill_tabs_with_spaces},
     {0, "-nospace", "--no-space-filled-tabs", ARG_CLEAR, 0, 0, &option_fill_tabs_with_spaces},
@@ -462,6 +507,27 @@ struct prog_options cooledit_options[] =
     {0, "-lines", "", ARG_INT, 0, 0, &start_height},
     {0, "-columns", "", ARG_INT, 0, 0, &start_width},
     {'m', "--minimal-main-window", "", ARG_SET, 0, 0, &option_minimal_main_window},
+    {0, "", "--no-xim", ARG_CLEAR, 0, 0, &option_use_xim},
+    {0, "-utf8", "--utf8-interpretation", ARG_SET, 0, 0, &option_utf_interpretation},
+    {0, "-no-fontset", "--no-fontset", ARG_SET, 0, 0, &option_no_font_set},
+    {0, "", "--red-first", ARG_CLEAR, 0, 0, &option_rgb_order},
+    {0, "", "--blue-first", ARG_SET, 0, 0, &option_rgb_order},
+    {0, "", "--anti-aliasing", ARG_SET, 0, 0, &option_aa_font},
+    {0, "", "--no-anti-aliasing", ARG_CLEAR, 0, 0, &option_aa_font},
+    {0, "", "--interchar-spacing", ARG_INT, 0, 0, &option_interchar_spacing},
+    {0, "", "--edit-bg", ARG_INT, 0, 0, &option_editor_bg_normal},
+    {0, "", "--interwidget-spacing", ARG_INT, 0, 0, &option_interwidget_spacing},
+    {0, "", "--look", ARG_STRING, &option_look, 0, 0},
+    {0, "", "--bright-invert", ARG_SET, 0, 0, &option_invert_colors},
+    {0, "", "--no-bright-invert", ARG_CLEAR, 0, 0, &option_invert_colors},
+    {0, "", "--chromin-invert", ARG_SET, 0, 0, &option_invert_crome},
+    {0, "", "--no-chromin-invert", ARG_CLEAR, 0, 0, &option_invert_crome},
+    {0, "", "--swap-red-green", ARG_SET, 0, 0, &option_invert_red_green},
+    {0, "", "--no-swap-red-green", ARG_CLEAR, 0, 0, &option_invert_red_green},
+    {0, "", "--swap-green-blue", ARG_SET, 0, 0, &option_invert_green_blue},
+    {0, "", "--no-swap-green-blue", ARG_CLEAR, 0, 0, &option_invert_green_blue},
+    {0, "", "--swap-red-blue", ARG_SET, 0, 0, &option_invert_red_blue},
+    {0, "", "--no-swap-red-blue", ARG_CLEAR, 0, 0, &option_invert_red_blue},
     {0, 0, 0, 0, 0, 0, 0}
 };
 
@@ -557,7 +623,7 @@ int get_file_and_line_from_error (char *message, int *line_number, char **new_fi
 	  try_new:
 	    memset (m, 0, MAX_PATH_LEN);
 	    if (!strncmp (message, "./", 2)) {
-		getcwd (m, MAX_PATH_LEN - 1);
+		get_current_wd (m, MAX_PATH_LEN - 1);
 		c = strchr (message, ':');
 		if (!c)
 		    c = message + strlen (message);
@@ -580,7 +646,7 @@ int get_file_and_line_from_error (char *message, int *line_number, char **new_fi
 		}
 		strncpy (m, message, (unsigned long) c - (unsigned long) message);
 	    }
-	    c = canonicalize_pathname (m);
+	    c = pathdup (m);
 	    for (i = 0; i < last_edit; i++)
 		if (edit[i]->editor->filename)
 		    if (*(edit[i]->editor->filename)) {
@@ -663,8 +729,10 @@ static int bookmarks_select_callback (CWidget * w, XEvent * x, CEvent * c)
     if (c->double_click || (c->command == CK_Enter && !c->handled)) {
 	int width;
 	char *q;
+	CPushFont ("editor", 0);
 	width = w->options & TEXTBOX_WRAP ? (w->width - TEXTBOX_BDR) / FONT_MEAN_WIDTH : 32000;
 	q = strline (w->text, strmovelines (w->text, w->current, w->cursor - w->firstline, width));
+	CPopFont ();
 	CDestroyWidget ("bookmarks");
 	edit[current_edit]->editor->force |= REDRAW_COMPLETELY;
 	goto_error (q, 1);
@@ -692,6 +760,7 @@ void goto_file_dialog (char *heading, char *tool_hint, char *text)
     win = CDrawHeadedDialog ("bookmarks", main_window, 20, 20, heading);
     CIdent ("bookmarks")->position = WINDOW_ALWAYS_RAISED;
     CGetHintPos (&x, &y);
+    CPushFont ("editor", 0);
     w = CDrawTextbox ("bookmarks.text", win, x, y, 70 * FONT_MEAN_WIDTH + EDIT_FRAME_W, 20 * FONT_PIX_PER_LINE + EDIT_FRAME_H, 0, 0, text, TEXTBOX_NO_STRDUP);
     if (tool_hint)
 	CSetToolHint ("bookmarks.text", tool_hint);
@@ -702,6 +771,7 @@ void goto_file_dialog (char *heading, char *tool_hint, char *text)
     CCentre ("bookmarks.done");
     CSetSizeHintPos ("bookmarks");
     CSetWindowResizable ("bookmarks", FONT_MEAN_WIDTH * 15, FONT_PIX_PER_LINE * 15, 1600, 1200);
+    CPopFont ();
     CMapDialog ("bookmarks");
     CAddCallback ("bookmarks.text", bookmarks_select_callback);
     CAddCallback ("bookmarks.done", bookmarks_done_callback);
@@ -891,7 +961,9 @@ int new_editor (int number, int x, int y, int columns, int lines, char *f, char
 	    d = catstrs (current_dir, "/", 0);
     }
     q = catstrs ("editor", itoa (edit_count), 0);
+    CPushFont ("editor", 0);
     win = CDrawDialog (p = catstrs ("wineditor", itoa (edit_count), 0), main_window, x, y);
+    CPopFont ();
 
     CGetHintPos (&xe, &ye);
     if (option_toolbar) {
@@ -915,17 +987,21 @@ int new_editor (int number, int x, int y, int columns, int lines, char *f, char
     }
     edit_set_syntax_change_callback (change_syntax_type);
     reset_hint_pos (0, 0);
+    CPushFont ("editor", 0);
     new_edit = CDrawEditor (q, win, xe, ye, columns * FONT_MEAN_WIDTH,
 			    lines * FONT_PIX_PER_LINE, t, f, d, EDITOR_HORIZ_SCROLL, size);
     if (!new_edit) {
 	CDestroyWidget (catstrs ("wineditor", itoa (edit_count), 0));
+	CPopFont ();
 	return 0;
     }
     new_edit->position |= (POSITION_WIDTH | POSITION_HEIGHT);
     new_edit->editor->modified = modified;
 
     CSetSizeHintPos (p);
+/* font height needed for resizing granularity - well, i guest thats why they call it hacking */
     CSetWindowResizable (p, 150, 150, 1600, 1200);
+    CPopFont ();
     CMapDialog (p);
     CAddBeforeCallback (p, draw_vert_text);
 
@@ -981,21 +1057,28 @@ void get_main_window_size (unsigned int *width, unsigned int *height)
 
 /* {{{  'Window' menu call backs */
 
-/* returns -1 if not, number of editor if yes */
-int file_is_open (char *p)
+extern int (*edit_file_is_open) (char *);
+
+/* returns non-zero if alrady open */
+static int file_is_open (char *p)
 {
-    char s[MAX_PATH_LEN];
-    int r = -1, i;
-    p = canonicalize_pathname (p);
-    for (i = 0; i < last_edit && r >= 0; i++) {
-	char *t;
-	strcpy (s, edit[i]->editor->dir);
-	strcat (s, "/");
-	strcat (s, edit[i]->editor->filename);
-	t = canonicalize_pathname (s);
-	if (!strcmp (t, p))
-	    r = i;
-	free (t);
+    char s[MAX_PATH_LEN], *t;
+    int r = 0, i;
+    p = pathdup (p);
+    for (i = 0; i < last_edit && !r; i++)
+	if (edit[i]->editor->dir && edit[i]->editor->filename && i != current_edit) {
+	    strcpy (s, edit[i]->editor->dir);
+	    strcat (s, "/");
+	    strcat (s, edit[i]->editor->filename);
+	    t = pathdup (s);
+	    if (!strcmp (t, p))
+		r = 1;
+	    free (t);
+	}
+    if (r) {
+	CPushFont ("widget", 0);
+	r = CQueryDialog (0, 0, 0, _ (" Warning "), catstrs (_ (" This file is already open. Open another? "), "\n", p, 0), _ ("Yes"), _ ("No"), 0);
+	CPopFont ();
     }
     free (p);
     return r;
@@ -1005,16 +1088,21 @@ static int height_offset;
 
 void fit_into_main_window (int *lines, int *columns, int x, int y)
 {
-    unsigned int width, height;
+    unsigned int width, height, f;
     get_main_window_size (&width, &height);
-    while (*columns * FONT_MEAN_WIDTH + 25 + EDIT_FRAME_W + 4 + 2 + 20 + WIDGET_SPACING * 2> width - x)
+    CPushFont ("widget", 0);
+    f = FONT_PIX_PER_LINE;
+    CPopFont ();
+    CPushFont ("editor", 0);
+    while (*columns * FONT_MEAN_WIDTH + 25 + EDIT_FRAME_W + 4 + 2 + 20 + WIDGET_SPACING * 2 > width - x)
 	(*columns)--;
     if (*columns * FONT_MEAN_WIDTH < 150)
 	(*columns) = 150 / FONT_MEAN_WIDTH;
-    while ((*lines + 1) * FONT_PIX_PER_LINE + EDIT_FRAME_H + WIDGET_SPACING * 3 + 8 + TEXT_RELIEF * 2 + 3 + 3 + 3 > height - y)
+    while (*lines * FONT_PIX_PER_LINE + f + EDIT_FRAME_H + WIDGET_SPACING * 3 + 8 + TEXT_RELIEF * 2 + 3 + 3 + 3 + 2 > height - y)
 	(*lines)--;
     if (*lines * FONT_PIX_PER_LINE < 250)
 	*lines = 250 / FONT_PIX_PER_LINE;
+    CPopFont ();
 }
 
 /* returns non-zero on error */
@@ -1026,13 +1114,9 @@ int new_file_callback (char *full_file_name,...)
     if (!full_file_name || full_file_name == NEW_WINDOW_FROM_TEXT) {
 	if (last_edit)
 	    /* copy the current directory: */
-	    d = (char *) strdup (edit[current_edit]->editor->dir);
+	    d = (char *) pathdup (edit[current_edit]->editor->dir);
 	else
-	    d = (char *) strdup (home_dir);
-    } else {
-	if (file_is_open (full_file_name) >= 0)
-	    if (CQueryDialog (main_window, 20, 20, _ (" Open "), _ (" A file by this name is already open, \n Continue open? "), _ (" Continue "), _ (" Cancel "), 0))
-		return 0;
+	    d = (char *) pathdup (home_dir);
     }
 
     CGetWindowPosition (edit[current_edit]->parentid, main_window, &x, &y);
@@ -1117,6 +1201,7 @@ void exit_app (unsigned long save)
     if (option_save_setup_on_exit)
 	save_setup (editor_options_file);
     save_options_section (editor_options_file, "[Input Histories]", p = get_all_lists ());
+    rxvtlib_shutall ();
     free (p);
     free_all_scripts ();
     complete_command (0);
@@ -1132,6 +1217,8 @@ void exit_app (unsigned long save)
     free_all_lists ();
     load_setup (0);
     catstrs_clean ();
+    edit_free_cache_lines ();
+    postscript_clean ();
 #if HAVE_MAD
 /* NLS ? */
     fprintf (stderr, "cooledit: calling mad_finalise(),\n");
@@ -1171,6 +1258,7 @@ void menu_browse_cmd (unsigned long ignored)
 }
 
 extern char *init_font;
+extern char *init_widget_font;
 extern char *init_bg_color;
 
 void run_main_callback (unsigned long ignored)
@@ -1181,7 +1269,7 @@ void run_main_callback (unsigned long ignored)
     switch (fork()) {
 	case 0:
 	    set_signal_handlers_to_default ();
-	    execlp (argv_nought, argv_nought, "-Smf", init_font, "-lines", lines, "-columns", columns, 0);
+	    execlp (argv_nought, argv_nought, "-Smf", init_font, "--widget-font", init_widget_font, "-lines", lines, "-columns", columns, 0);
 	    exit (0);
 	case -1:
 	    CErrorDialog (0, 0, 0, _(" Run 'cooledit' "), get_sys_error (_(" Error trying to fork process ")));
@@ -1343,6 +1431,7 @@ int read_config (void)
 	if (i < 3)
 	    break;
 	if (i != 8) {
+	    fprintf (stderr, _ ("cooledit: error in initialisation file %s: line %d\n"), editor_options_file, num_read + 1);
 	    CErrorDialog (main_window, 20, 20, _ (" Load Config "), _ (" Error in initialisation file %s: line %d "), editor_options_file, num_read + 1);
 	    free (r);
 	    return 1;
@@ -1530,23 +1619,6 @@ static void open_drop_file (unsigned char *data, int size, Atom type)
 /* }} */
 #endif
 
-char *get_full_filename (const char *f)
-{
-    char *s, *p;
-
-    if (*f == '/')
-	return (char *) f;
-
-    s = malloc (2048);
-    if (getcwd (s, 2000)) {
-	p = (char *) strdup (catstrs (s, "/", f, 0));
-	free (s);
-	return p;
-    }
-    free (s);
-    return 0;
-}
-
 int editors_modified (void)
 {
     int i, r = 0;
@@ -1565,10 +1637,12 @@ static void cooledit_init (void)
     cooledit_startup.geometry = option_geometry;
     cooledit_startup.display = option_display;
     cooledit_startup.font = option_font;
+    cooledit_startup.widget_font = option_widget_font;
     cooledit_startup.bg = option_background_color;
     cooledit_startup.fg_red = option_foreground_red;
     cooledit_startup.fg_green = option_foreground_green;
     cooledit_startup.fg_blue = option_foreground_blue;
+    cooledit_startup.look = option_look;
 
     if (option_verbose)
 	cooledit_startup.options = CINIT_OPTION_VERBOSE;
@@ -1605,22 +1679,28 @@ static int similar_size_to_main_window (CWidget * w)
 static void maximise_window (char *ident)
 {
     CWidget *w;
-    int lines, columns;
+    int lines, columns, f;
     unsigned int wm, hm;
     char x[34];
     strcpy (x, "win");
     strcat (x, ident);
     get_main_window_size (&wm, &hm);
+    CPushFont ("widget", 0);
+    f = FONT_PIX_PER_LINE;
+    CPopFont ();
+    CPushFont ("editor", 0);
     lines = hm / FONT_PIX_PER_LINE;
     columns = wm / FONT_MEAN_WIDTH;
     fit_into_main_window (&lines, &columns, 0, height_offset);
     CSetWidgetSize (x, columns * FONT_MEAN_WIDTH + 25 + EDIT_FRAME_W + 4 + 2 + 20 + WIDGET_SPACING * 2,
-		    (lines + 1) * FONT_PIX_PER_LINE + EDIT_FRAME_H + WIDGET_SPACING * 3 + 8 + TEXT_RELIEF * 2 + 3 + 3 + 3);
+		    lines * FONT_PIX_PER_LINE + f + EDIT_FRAME_H + WIDGET_SPACING * 3 + 8 + TEXT_RELIEF * 2 + 3 + 3 + 3 + 2);
     w = CIdent (x);
     CSetWidgetPosition (x, (wm - w->width) / 2, height_offset + (hm - height_offset - w->height) / 2);
+    CPopFont ();
 }
 
 void debug_key_command (int command);
+void edit_insert_shell_output (WEdit * edit);
 
 /* ----main_loop()--------------------------------------------------------- */
 
@@ -1649,6 +1729,17 @@ static void main_loop (void)
 		break;
 	    if (cooledit_cevent.kind == C_EDITOR_WIDGET || cooledit_cevent.kind == C_MENU_BUTTON_WIDGET || !last_edit) {
 		switch ((int) cooledit_cevent.command) {
+		case CK_Insert_Unicode: {
+			long c;
+			if ((c = CUnicodeDialog (main_window, 20, 20, 0)) >= 0) {
+			    edit_insert_wide (edit[current_edit]->editor, c);
+			    edit[current_edit]->editor->force |= REDRAW_PAGE;
+			}
+		    }
+		    break;
+		case CK_Terminal:
+		    rxvt_start (CRoot, 0, 0);
+		    break;
 		case CK_Complete:
 		    complete_command (edit[current_edit]);
 		    break;
@@ -1680,8 +1771,12 @@ static void main_loop (void)
 		    else if (is_focus_prev_key (cooledit_cevent.key, cooledit_cevent.command, cooledit_xevent.xkey.state) ||
 			cooledit_cevent.command == CK_Cycle) {
 			window_cycle_callback (0);
-			if (option_pull_down_window_list)
+			if (option_pull_down_window_list) {
 			    CPullDown (CIdent ("menu.wlist"));
+			    XSync (CDisplay, 0);
+			}
+		    } else if (cooledit_cevent.command == CK_Cancel) {
+			edit_insert_shell_output (edit[current_edit]->editor);
 		    } else {
 			CPullUp (CIdent ("menu.wlist"));
 		    }
@@ -1854,41 +1949,68 @@ static struct mouse_funcs main_mouse_funcs =
     mime_majors
 };
 
+extern int (*global_callback) (XEvent *x);
+int rxvt_event (XEvent * xevent);
+
 
 /* ----main()-------------------------------------------------------------- */
 int main (int argc, char **argv)
 {
     Window buttonwin;
     CWidget *w;
+    char *s;
     int x, y, setsize = 0;
     char *example_fonts[] =
     {
-	"-misc-fixed-bold-r-normal--13-120-75-75-c-80-iso8859-1",
-	"-adobe-courier-medium-r-normal--13-120-75-75-m-60-iso8859-1",
-	"-*-times-medium-r-*--14-*-*-*-p-*-iso8859-1",
-	"-*-helvetica-bold-r-*--14-*-*-*-p-*-iso8859-1",
-	"-*-charter-bold-r-*--14-*-*-*-p-*-iso8859-1",
-	"-*-times-medium-r-*--20-*-*-*-p-*-iso8859-1",
-	"-*-helvetica-bold-r-*--20-*-*-*-p-*-iso8859-1",
-	"-*-charter-bold-r-*--20-*-*-*-p-*-iso8859-1",
-	"-*-*-bold-*-normal--10-100-75-75-c-80-iso8859-1",
-	"-*-*-bold-r-normal--12-120-75-75-c-80-iso8859-1",
-	"-*-*-bold-r-normal-*-15-*-*-*-c-*-iso8859-1"
+	"-misc-fixed-bold-r-normal--13-120-75-75-c-80",
+	"-adobe-courier-medium-r-normal--13-120-75-75-m-60",
+	"-*-times-medium-r-*--14-*-*-*-p-*",
+	"-*-helvetica-bold-r-*--14-*-*-*-p-*",
+	"-*-charter-bold-r-*--14-*-*-*-p-*",
+	"-*-times-medium-r-*--20-*-*-*-p-*",
+	"-*-helvetica-bold-r-*--20-*-*-*-p-*",
+	"-*-charter-bold-r-*--20-*-*-*-p-*",
+	"-*-*-bold-*-normal--10-100-75-75-c-80",
+	"-*-*-bold-r-normal--12-120-75-75-c-80",
+	"-*-*-bold-r-normal-*-15-*-*-*-c-*",
+	"-winfonts-arial-bold-r-*-*-45-*-*-*-*-*-iso8859-1"
     };
-    int n;
+    int n, qv;
 
-    setlocale (LC_CTYPE, "");
+    command_line_files = (char **) malloc ((argc + 1) * sizeof (char *));
+    memset (command_line_files, 0, (argc + 1) * sizeof (char *));
+    cooledit_options[0].strs = command_line_files;
+
+    edit_file_is_open = file_is_open;
+
+    argv_nought = (char *) strdup (argv[0]);
+
+
+    if (!setlocale (LC_CTYPE, ""))
+	fprintf (stderr, _ ("%s: cannot set locale, see setlocale(3), current locale is %s\n"), argv_nought, setlocale (LC_MESSAGES, 0));
     setlocale (LC_TIME, "");
     setlocale (LC_MESSAGES, "");
     bindtextdomain (PACKAGE, LOCALEDIR);
     textdomain (PACKAGE);
 
+    if ((s = getenv ("LC_ALL")) || (s = getenv ("LC_CTYPE")) || (s = getenv ("LANG"))) {
+	if (strstr (s, "UTF-8"))
+	    option_utf_interpretation = 1;
+    }
+#if defined (HAVE_NL_LANGINFO) && defined (HAVE_LANGINFO_H) && defined (CODESET)
+    if (!strcmp (nl_langinfo (CODESET), "UTF-8"))
+	option_utf_interpretation = 1;
+#endif
+
     n = sizeof (example_fonts) / sizeof (char *);
 
-    argv_nought = (char *) strdup (argv[0]);
-
     process_command_line (argc, argv);
 
+    if (option_aa_font != -1) {
+	fprintf (stderr, "cooledit: --[no-]anti-aliasing is depreciated, use <fontname>/3 instead\n");
+	exit (1);
+    }
+
     if (option_font)
 	if (!strcmp (option_font, "?") || !strcmp (option_font, "h")
 	 || !strcmp (option_font, "-?") || !strcmp (option_font, "-h")) {
@@ -1901,10 +2023,13 @@ int main (int argc, char **argv)
 		    "cooledit -font 1-%d\n", n);
 	    exit (1);
 	}
+
     get_home_dir ();
+
     if (!editor_options_file)
 	editor_options_file = (char *) strdup (catstrs (DEFAULT_INI_FILE, 0));
 
+
     if (!option_suppress_load_options) {
 	char *p;
 	load_setup (editor_options_file);
@@ -1912,13 +2037,21 @@ int main (int argc, char **argv)
 	if (p)
 	    free (p);
     }
+
     if (option_minimal_main_window)
 	option_geometry = 0;
+
     if (!option_command_line_doesnt_override)
 	process_command_line (argc, argv);
+
     if (option_font)
 	if (!strcmp (option_font, "default"))
-	    option_font = (char *) strdup ("8x13bold");
+	    option_font = (char *) strdup ("-*-fixed-bold-r-*--13-*-*-*-*-80-*");
+
+    if (option_widget_font)
+	if (!strcmp (option_widget_font, "default"))
+	    option_widget_font = (char *) strdup ("-*-helvetica-bold-r-*--13-*-*-*-*-*-*");
+
     if (option_font)
 	if (strspn (option_font, "0123456789") == strlen (option_font)) {
 	    int i;
@@ -1984,6 +2117,7 @@ int main (int argc, char **argv)
 
     edit_set_user_key_function (user_defined_key);
     edit_set_user_command (custom_keys);
+
 #ifdef HAVE_PYTHON
 /* this must be done before loading files */
     coolpython_init (argc, argv);
@@ -1992,7 +2126,6 @@ int main (int argc, char **argv)
 
     if (command_line_files[0]) {
 	int i;
-	char *q;
 	long l = 0, num_read = 0;
 	for (i = 0; command_line_files[i]; i++) {
 	    if (command_line_files[i][0] == '+') {
@@ -2001,11 +2134,7 @@ int main (int argc, char **argv)
 		    l = 0;
 		continue;
 	    }
-	    q = get_full_filename (command_line_files[i]);
-#if 0
-	    fit_into_main_window (&initial_lines, &initial_columns, 0, height_offset);
-#endif
-	    if (new_editor (num_read, 0, height_offset, initial_columns, initial_lines, command_line_files[i], 0)) {
+	    else if (new_editor (num_read, 0, height_offset, initial_columns, initial_lines, command_line_files[i], 0)) {
 		edit_move_display (edit[num_read]->editor, l);
 		edit_move_to_line (edit[num_read]->editor, l);
 		num_read++;
@@ -2046,7 +2175,9 @@ int main (int argc, char **argv)
     if (setsize || !(w->options & WINDOW_USER_SIZE))	/* this means that the size was specified by the user in 'geometry' */
 	CSetWidgetSize ("cooledit", extents_width, extents_height);
 
+    CPushFont ("editor", 0);
     CSetWindowResizable ("cooledit", 300, 300, 20000, 20000);
+    CPopFont ();
 /* Toolhint */
     CSetToolHint ("cooledit", _ ("Drop a filename or text onto this space"));
 
@@ -2064,7 +2195,30 @@ int main (int argc, char **argv)
 	else
 	    w->position &= ~(POSITION_HEIGHT | POSITION_WIDTH);
     }
+
+#if 0
+    {
+	Window win;
+	int xe, ye, t;
+	win = CDrawDialog ("rxvtwin", main_window, 0, 0);
+	CGetHintPos (&xe, &ye);
+	w = (CWidget *) CDrawRxvt ("rxvt", win, xe, ye, 0);
+	w->position |= POSITION_WIDTH | POSITION_HEIGHT;
+	CSetSizeHintPos ("rxvtwin");
+/* font height needed for resizing granularity - well, i guest thats why they call it hacking */
+	CPushFont ("editor", 0);
+	t = option_text_line_spacing;
+	option_text_line_spacing = 0;
+	CSetWindowResizable ("rxvtwin", 150, 150, 1600, 1200);
+	option_text_line_spacing = t;
+	CPopFont ();
+	CMapDialog ("rxvtwin");
+    }
+#endif
+
+    global_callback = rxvt_event;
+
     main_loop ();
-    return 0;
+    return r;
 }
 
