@@ -1,5 +1,5 @@
 /* debug.c - interactive debugger
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 #include <stdio.h>
 #include <my_string.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <signal.h>
 #include <sys/types.h>
 #if HAVE_SYS_WAIT_H
@@ -59,6 +60,8 @@
 #include "find.h"
 #include "mad.h"
 
+extern struct look *look;
+
 #define DEBUG_TIME_OUT		15
 #define DEBUG_BREAKPOINT_COLOR	18
 #define DEBUG_CURRENT_COLOR	6
@@ -80,7 +83,9 @@
 #define ACTION_LASTVAR		15
 #define ACTION_WATCH		16
 
-#define GDB_VERSION		"GNU gdb 4.17"
+#define GDB_VERSION1		"GNU gdb 4.17"
+#define GDB_VERSION2		"GNU gdb 4.18"
+#define GDB_VERSION3		"GNU gdb 5.0"
 
 static void debug_read_callback (int fd, fd_set * reading, fd_set * writing, fd_set * error, void *data);
 static void debug_write_callback (int fd, fd_set * reading, fd_set * writing, fd_set * error, void *data);
@@ -98,7 +103,10 @@ typedef struct struct_debug {
     char *get_line;
     char *get_source;
     char *get_backtrace;
-    char *query;
+    struct debug_query {
+	char *query;
+	char *response;
+    } query[10];
     char *progname;
     char *debugger;
     char *args;
@@ -124,8 +132,8 @@ typedef struct struct_debug {
     char *file;
     char *condition;
     int show_output, stop_at_main, show_on_stdout;
-    char *xterm_cmd;
-    pid_t xterm_pid, ttyname_pid;
+    int ignore_SIGTTOU;
+    pid_t xterm_pid;
     int break_point_line;
     char *break_point_editor;
 } Debug;
@@ -175,7 +183,8 @@ static void xdebug_cursor_bookmark_flush (void)
 {
     int i;
     for (i = 0; i < last_edit; i++)
-	book_mark_flush (edit[i]->editor, DEBUG_CURRENT_COLOR);
+	if (edit[i]->editor)
+	    book_mark_flush (edit[i]->editor, DEBUG_CURRENT_COLOR);
 }
 
 static void xdebug_append_command (Debug * d, char *s, int action, char *editor, int line)
@@ -317,6 +326,7 @@ static void xdebug_display_variable (Debug * d, char *message, int action)
 	    int x, y;
 	    win = CDrawMainWindow ("debug_vars.win", _ ("Variables"));
 	    CGetHintPos (&x, &y);
+	    CPushFont ("editor", 0);
 	    w = CDrawFieldedTextbox ("debug_vars", win, x, y, FONT_MEAN_WIDTH * 50 + 7,
 		FONT_PIX_PER_LINE * 15 + 6, 0, 0, xdebug_get_line, 0, d);
 	    w->hook = d;
@@ -328,6 +338,7 @@ static void xdebug_display_variable (Debug * d, char *message, int action)
 	    CMapDialog ("debug_vars.win");
 	    CSetWindowResizable ("debug_vars.win", FONT_MEAN_WIDTH * 20, FONT_PIX_PER_LINE * 10, 1600, 1200);
 	    CSetToolHint ("debug_vars", _ ("Use Del to delete from this list. Use Ins to\nmark variable as a watchpoint. This will cause\ndebugger to break if the variable changes."));
+	    CPopFont ();
 	}
     }
 }
@@ -401,7 +412,7 @@ void debug_callback (void)
 	    CSetColor (color_widget (4));
 	    XFillArc (CDisplay, debug_session.w, CGC, debug_session.x, debug_session.y, debug_session.r * 2, debug_session.r * 2, (animation * 487) % (360 * 64) - 180 * 64, 90 * 64);
 	    XFillArc (CDisplay, debug_session.w, CGC, debug_session.x, debug_session.y, debug_session.r * 2, debug_session.r * 2, (animation * 487 + 180 * 64) % (360 * 64) - 180 * 64, 90 * 64);
-	    CSetColor (color_widget (12));
+	    CSetColor (color_widget (13));
 	    XFillArc (CDisplay, debug_session.w, CGC, debug_session.x, debug_session.y, debug_session.r * 2, debug_session.r * 2, (animation * 487 + 90 * 64) % (360 * 64) - 180 * 64, 90 * 64);
 	    XFillArc (CDisplay, debug_session.w, CGC, debug_session.x, debug_session.y, debug_session.r * 2, debug_session.r * 2, (animation * 487 + 270 * 64) % (360 * 64) - 180 * 64, 90 * 64);
 	    refresh = 1;
@@ -446,7 +457,7 @@ static void xdebug_check_watch_point (Debug * d, char *m)
 	CSetTextboxPos (w, TEXT_SET_CURSOR_LINE, atoi (p + 12));
 	CRedrawFieldedTextbox ("debug_vars", 1);
     }
-    r = strchr ((char *) pool_start (d->pool), ':');
+    r = strchr (p, ':');
     if (!r)
 	return;
     r += 2;
@@ -485,20 +496,28 @@ static void debug_read_callback (int fd, fd_set * reading, fd_set * writing, fd_
     char buf[1025];
     int c, action, n;
     d = &debug_session;
+    if (CChildExitted (d->pid, 0) || (!rxvt_have_pid (d->xterm_pid) && d->show_output)) {
+	debug_finish (0);
+	return;
+    }
     do {
 	c = read (d->out, buf, 1024);
     } while (c < 0 && errno == EINTR);
     if (c <= 0)
 	return;
     buf[c] = '\0';
-    if (from_end (buf, d->query, 1)) {
-	p = strchr (buf, '\n');
-	if (!p)
-	    buf[0] = '\0';
-	else
-	    p[1] = '\0';
-	while (write (d->in, "y\n", 2) < 0 && errno == EINTR);
+    for (n = 0; d->query[n].query; n++) {
+	if (from_end (buf, d->query[n].query, 1)) {
+	    p = strrchr (buf, '\n');
+	    if (!p)
+		buf[0] = '\0';
+	    else
+		p[1] = '\0';
+	    while (write (d->in, d->query[n].response, strlen (d->query[n].response)) < 0 && errno == EINTR);
+	    break;
+	}
     }
+    n = 0;
     if (!d->pool)
 	d->pool = pool_init ();
     pool_printf (d->pool, "%s", buf);
@@ -529,8 +548,11 @@ static void debug_read_callback (int fd, fd_set * reading, fd_set * writing, fd_
 	    debug_finish (0L);
 	break;
     case ACTION_VERSION:
-	if (!strstr ((char *) pool_start (d->pool), GDB_VERSION))
-	    debug_error1 (d, catstrs (GDB_VERSION, _ (" expected. You seem to have a different debugger \n or version. If things don't work properly, this is why."), 0));
+/* 4.18 doesn't seem to do next's properly */
+	if (!strstr ((char *) pool_start (d->pool), GDB_VERSION1) && !strstr ((char *) pool_start (d->pool), GDB_VERSION3))
+	    debug_error1 (d, catstrs (GDB_VERSION1 " or " GDB_VERSION3, _ (" expected. You seem to have a different debugger \n or version. If things don't work properly, this is why."), 0));
+	if (!d->pool)	/* process was terminated */
+	    break;
 	if (from_end ((char *) pool_start (d->pool), "no debugging symbols found", 6)) {
 	    debug_error1 (d, _ ("The debugger returned \"no debugging symbols found\" implying that \n the executable has been stripped, or was not compiled with \n the \"-g\" option for debugging"));
 	    debug_finish (0L);
@@ -589,8 +611,11 @@ static void debug_read_callback (int fd, fd_set * reading, fd_set * writing, fd_
 	    w = CIdent (d->break_point_editor);
 	    if (p) {
 		if (w) {
-		    book_mark_insert (w->editor, d->break_point_line - 1, DEBUG_BREAKPOINT_COLOR);
-		    w->editor->force |= REDRAW_COMPLETELY;
+		    if (!book_mark_query_color (w->editor, d->break_point_line - 1, DEBUG_BREAKPOINT_COLOR)) {
+			book_mark_insert (w->editor, d->break_point_line - 1, DEBUG_BREAKPOINT_COLOR);
+			w->editor->force |= REDRAW_COMPLETELY;
+			edit_render_keypress (edit[current_edit]->editor);
+		    }
 		}
 	    } else {
 		p = from_start ((char *) pool_start (d->pool), "Deleted breakpoint", 2);
@@ -601,11 +626,14 @@ static void debug_read_callback (int fd, fd_set * reading, fd_set * writing, fd_
 		if (p && w) {
 		    while (!book_mark_clear (w->editor, d->break_point_line - 1, DEBUG_BREAKPOINT_COLOR));
 		    w->editor->force |= REDRAW_COMPLETELY;
+		    edit_render_keypress (w->editor);
 		}
 	    }
 	}
 	if (!p) {
 	    debug_error1 (d, (char *) pool_start (d->pool));
+	    if (!d->pool)
+		break;
 	    xdebug_flush_commands (d);
 	}
 	if (p)
@@ -668,6 +696,14 @@ static void debug_read_callback (int fd, fd_set * reading, fd_set * writing, fd_
 	for (p = (char *) pool_start (d->pool); p; p++) {
 	    if (!*p)
 		break;
+            if (!strncmp (p, "0x", 2)) {
+/* {like: */
+/* 0x8060bd0       2110    return -1; } */
+                p += 2;
+                p += strspn (p, "0123456789abcdef");
+                while (*p == ' ' || *p == '\t')
+                    p++;
+            }
 	    if (*p >= '0' && *p <= '9')
 		d->line = atoi (p);
 	    p = strchr (p, '\n');
@@ -678,6 +714,8 @@ static void debug_read_callback (int fd, fd_set * reading, fd_set * writing, fd_
 	p = from_end ((char *) pool_start (d->pool), "\nProgram exited normally", 3);
 	if (!p)
 	    p = from_end ((char *) pool_start (d->pool), "\nProgram exited with code", 7);
+	if (!p)
+	    p = from_end ((char *) pool_start (d->pool), "No such process", 7);
 	if (p) {
 	    xdebug_cursor_bookmark_flush ();
 	    d->child = 0;
@@ -720,6 +758,10 @@ static void debug_write_callback (int fd, fd_set * reading, fd_set * writing, fd
     Debug *d;
     int c;
     d = &debug_session;
+    if (CChildExitted (d->pid, 0) || (!rxvt_have_pid (d->xterm_pid) && d->show_output)) {
+	debug_finish (0);
+	return;
+    }
     if (d->show_on_stdout) {
 	printf ("%s", d->command[0].command);
 	fflush (stdout);
@@ -753,72 +795,59 @@ static int xdebug_run_program (Debug * d)
     int i = 0;
     struct stat s;
     xdebug_cursor_bookmark_flush ();
-    if (!ttyname (0))
-	d->show_output = 1;
     if (stat (d->progname, &s)) {
-	debug_error2 (d, _ ("Error trying to stat executable: check that the filename and \n current directory are correct"), strerror (errno));
+	debug_error2 (d,
+		      _
+		      ("Error trying to stat executable: check that the filename and \n current directory are correct"),
+		      strerror (errno));
 	return 1;
-    } else {
 	if (!(s.st_mode & S_IXUSR)) {
-	    debug_error2 (d, _ ("Your program does not seem to have execute permissions"), d->progname);
+	    debug_error2 (d, _("Your program does not seem to have execute permissions"),
+			  d->progname);
 	    return 1;
 	}
     }
     arg[i++] = d->debugger;
     if (d->show_output) {
-	int tries, p, l, success = 0;
-	char tty[100], *f;
-	if (!d->xterm_cmd)
-	    d->xterm_cmd = (char *) strdup ("xterm");
-	f = (char *) tempnam (0, "cetty");
-	mknod (f, S_IFIFO | S_IRUSR | S_IWUSR, 0);
-	if (!(d->xterm_pid = fork ())) {
-	    execlp ("/bin/sh", "/bin/sh", "-c", catstrs (d->xterm_cmd, " -T ", d->progname, " -e ttyname_stop ", f, 0), 0);
-	    exit (1);
-	}
-	do {
-	    p = open (f, O_RDONLY);
-	} while (p < 0 && errno == EINTR);
-	if (p >= 0)
-	    for (tries = 0; tries < 6; tries++) {
-		if ((l = read (p, tty, 80)) > 5) {
-		    char *r;
-		    tty[l] = '\0';
-		    r = strchr (tty, ',');
-		    if (r) {
-			*r++ = '\0';
-			if (atoi (r))
-			    d->ttyname_pid = atoi (r);
-			success = 1;
-			arg[i++] = catstrs ("-tty=", tty, 0);
-		    }
-		    break;
-		}
-		CSleep (1.0);
-	    }
-	close (p);
-	unlink (f);
-	free (f);
-	if (!success) {
-	    kill (debug_session.xterm_pid, SIGTERM);
-	    debug_session.xterm_pid = 0;
-	    debug_error1 (d, _ ("Could not get xterm pseudo terminal"));
+	char *p, *gargv[2] = {0, 0};
+	void *rxvt;
+        gargv[0] = d->progname;
+	rxvt = (void *) rxvt_start (CRoot, gargv, 1);
+	if (!rxvt) {
+	    d->xterm_pid = 0;
+	    d->pid = 0;
+	    debug_error1 (d, _("Could not open an rxvt"));
+	    xdebug_flush_commands (d);
 	    return 1;
 	}
+	p = (char *) strdup ("-tty=                    ");
+	rxvt_get_tty_name (rxvt, p + 5);
+	d->xterm_pid = (pid_t) rxvt_get_pid (rxvt);
+	arg[i++] = p;
+	arg[i++] = d->progname;
+	arg[i] = 0;
+	d->pid = triple_pipe_open (&d->in, &d->out, 0, 1, d->debugger, arg);
+	free (p);
+    } else {
+	char *p;
+	p = (char *) strdup ("-tty=                    ");
+	arg[i++] = p;
+	arg[i++] = d->progname;
+	arg[i] = 0;
+	d->pid = open_under_pty (&d->in, &d->out, p + 5, d->debugger, arg);
+	free (p);
     }
-    arg[i++] = d->progname;
-    arg[i] = 0;
-    d->pid = triple_pipe_open (&d->in, &d->out, 0, 1, d->debugger, arg);
-    d->child = 0;
     if (d->pid <= 0) {
 	d->pid = 0;
-	kill (debug_session.ttyname_pid, SIGTERM);
-	kill (debug_session.xterm_pid, SIGTERM);
-	debug_session.xterm_pid = debug_session.ttyname_pid = 0;
-	debug_error2 (d, _ ("Error executing gdb"), strerror (errno));
+	if (d->xterm_pid)
+	    kill (d->xterm_pid, SIGTERM);
+	d->xterm_pid = 0;
+	d->pid = 0;
+	debug_error1 (d, _("Could not open gdb"));
 	xdebug_flush_commands (d);
 	return 1;
     }
+    d->child = 0;
     d->action = ACTION_VERSION;
     CAddWatch (d->out, debug_read_callback, WATCH_READING, 0);
     return 0;
@@ -841,8 +870,10 @@ static void xdebug_run (Debug * d)
     xdebug_append_command (d, "clear\n", ACTION_BREAKPOINT, 0, 0);
 /* programs get this signal when they ioctl the terminal given to them by gdb */
 /* this prevents them quiting when using the `show output' option */
-    sprintf (s, "call signal (%d, (void *) %lu)\n", SIGTTOU, (unsigned long) SIG_IGN);
-    xdebug_append_command (d, s, ACTION_PROMPT, 0, 0);
+    if (d->ignore_SIGTTOU) {
+	sprintf (s, "call signal (%d, (void *) %lu)\n", SIGTTOU, (unsigned long) SIG_IGN);
+	xdebug_append_command (d, s, ACTION_PROMPT, 0, 0);
+    }
     xdebug_append_command (d, "info program\n", ACTION_INFO_PROGRAM, 0, 0);
     if (!d->stop_at_main)
 	xdebug_append_command (d, "c\n", ACTION_RUNNING, 0, 0);
@@ -916,14 +947,12 @@ static int debug_set_info (unsigned long x)
     {
 	gettext_noop (""),
 	gettext_noop (""),
-	"xterm -fn 8x13bold -fg white -bg black",
 	0
     };
     char *input_labels[10] =
     {
 	gettext_noop ("Program name"),
-	gettext_noop ("Enter program arguments"),
-	gettext_noop ("Terminal"),
+	gettext_noop ("Enter program arguments (space for none)"),
 	0
     };
     char *check_labels[10] =
@@ -931,27 +960,27 @@ static int debug_set_info (unsigned long x)
 	gettext_noop ("Show program output"),
 	gettext_noop ("Stop at main() function"),
 	gettext_noop ("Show on stdout"),
+	gettext_noop ("Ignore SIGTTOU"),
 	0
     };
     char *check_tool_hints[10] =
     {
-	gettext_noop ("This will open the xterm for the program given under \"Terminal\".\nIf you started Cooledit without a controlling tty, this must be set"),
+	gettext_noop ("This will cause the program to be opened udner an rxvt terminal so that its output can be viewed."),
 	gettext_noop ("This will cause the debugger to break at the first\nline of code in the program when issuing a `Start'."),
 	gettext_noop ("This will cause all debugger commands to show at\nthe terminal where you started cooledit."),
+	gettext_noop ("Run `call signal()' inside gdb so that SIGTTOU signals are ignored. Without this gdb sometimes exits."),
 	0
     };
     char *input_names[10] =
     {
 	gettext_noop ("debugprogname"),
-	gettext_noop ("debugargs"),
-	gettext_noop ("debugxterm"),
+	gettext_noop ("debugargs+"),
 	0
     };
     char *input_tool_hint[10] =
     {
 	gettext_noop ("The executable that you would like to debug - this must be compiled with the -g option"),
 	gettext_noop ("Arguments to be passed to the executable - can be empty"),
-	gettext_noop ("The programs stdout etc. will be redirected to this program for viewing"),
 	0
     };
     int *checks_values_result[10];
@@ -959,14 +988,12 @@ static int debug_set_info (unsigned long x)
     int r;
     inputs_result[0] = &inputs[0];
     inputs_result[1] = &inputs[1];
-    inputs_result[2] = &inputs[2];
-    inputs_result[3] = 0;
-    if (!ttyname (0))
-	debug_session.show_output = 1;
+    inputs_result[2] = 0;
     checks_values_result[0] = &debug_session.show_output;
     checks_values_result[1] = &debug_session.stop_at_main;
     checks_values_result[2] = &debug_session.show_on_stdout;
-    checks_values_result[3] = 0;
+    checks_values_result[3] = &debug_session.ignore_SIGTTOU;
+    checks_values_result[4] = 0;
     r = CInputsWithOptions (0, 0, 0, _ (" Debug : Program Data "), inputs_result, input_labels, input_names, input_tool_hint, checks_values_result, check_labels, check_tool_hints, INPUTS_WITH_OPTIONS_BROWSE_LOAD_1, 60);
     if (r)
 	return 1;
@@ -976,9 +1003,6 @@ static int debug_set_info (unsigned long x)
     if (debug_session.args)
 	free (debug_session.args);
     debug_session.args = inputs[1];
-    if (debug_session.xterm_cmd)
-	free (debug_session.xterm_cmd);
-    debug_session.xterm_cmd = inputs[2];
     if (debug_session.pid)
 	debug_error1 (&debug_session, _ (" The debugger is currently running. You will have \n to kill the debugger for these changes to take effect. "));
     return 0;
@@ -1073,10 +1097,6 @@ static void debug_finish (unsigned long x)
 	CRemoveWatch (debug_session.out, debug_read_callback, WATCH_READING);
 	close (debug_session.in);
 	close (debug_session.out);
-	if (debug_session.ttyname_pid) {
-	    kill (debug_session.ttyname_pid, SIGTERM);
-	    debug_session.ttyname_pid = 0;
-	}
 	if (debug_session.xterm_pid) {
 	    kill (debug_session.xterm_pid, SIGTERM);
 	    debug_session.xterm_pid = 0;
@@ -1088,12 +1108,6 @@ static void debug_finish (unsigned long x)
     }
 }
 
-void debug_check (pid_t p)
-{
-    if (p == debug_session.xterm_pid)
-	debug_finish (0);
-}
-
 static void debug_restart (unsigned long x)
 {
     if (debug_session.action) {
@@ -1156,11 +1170,15 @@ static void debug_break_point (unsigned long x)
     if (debug_session.pid)
 	xdebug_break_point (&debug_session, cmd, edit[current_edit]->ident, edit[current_edit]->editor->curs_line + 1);
     else {
-	if (!strcmp (cmd, "clear"))
+	if (!strcmp (cmd, "clear")) {
 	    while (!book_mark_clear (edit[current_edit]->editor, edit[current_edit]->editor->curs_line, DEBUG_BREAKPOINT_COLOR));
-	else
+	    edit[current_edit]->editor->force |= REDRAW_COMPLETELY;
+	    edit_render_keypress (edit[current_edit]->editor);
+	} else {
 	    book_mark_insert (edit[current_edit]->editor, edit[current_edit]->editor->curs_line, DEBUG_BREAKPOINT_COLOR);
-	edit[current_edit]->editor->force |= REDRAW_COMPLETELY;
+	    edit[current_edit]->editor->force |= REDRAW_COMPLETELY;
+	    edit_render_keypress (edit[current_edit]->editor);
+	}
     }
 }
 
@@ -1217,9 +1235,17 @@ void debug_init (void)
     debug_session.get_line = "info line\n";
     debug_session.get_source = "info source\n";
     debug_session.get_backtrace = "backtrace\n";
-    debug_session.query = "( y or n )";
-    debug_session.debugger = "gdb";
+    debug_session.query[0].query = "( y or n )";
+    debug_session.query[0].response = "y\n";
+    debug_session.query[1].query = "(y or n)";
+    debug_session.query[1].response = "y\n";
+    debug_session.query[2].query = "<return> to quit---";
+    debug_session.query[2].response = "\n";
+    debug_session.query[3].query = 0;
+    debug_session.query[3].response = 0;
+    debug_session.debugger = "cooledit-gdb";
     debug_session.stop_at_main = 1;
+    debug_session.ignore_SIGTTOU = 1;
     xdebug_init_animation ();
 }
 
@@ -1232,8 +1258,6 @@ void debug_shut (void)
 	free (debug_session.progname);
     if (debug_session.args)
 	free (debug_session.args);
-    if (debug_session.xterm_cmd)
-	free (debug_session.xterm_cmd);
     if (debug_session.file)
 	free (debug_session.file);
     if (debug_session.condition)
@@ -1251,22 +1275,22 @@ void debug_menus (Window parent, Window main_window, int x, int y)
 {
     CDrawMenuButton ("menu.debugmenu", parent, main_window, x, y, AUTO_SIZE, 16,
 			 _ (" Debug "),
-			 _ ("Start\tM-F5"), '~', debug_restart, 0L,
-			 _ ("Stop\tC-c"), '~', debug_break, 0L,
-			 _ ("Attach"), '~', debug_attach, 0L,
-			 _ ("Detach"), '~', debug_detach, 0L,
-			 _ ("Toggle breakpoint\tM-F2"), '~', debug_break_point, 0L,
-			 _ ("Insert conditional breakpoint"), '~', debug_conditional, 0L,
-			 _ ("Clear all breakpoints"), '~', debug_clear_breakpoints, 0L,
-			 _ ("Next\tM-F9"), '~', debug_step, (unsigned long) "n\n",
-			 _ ("Step\tM-F8"), '~', debug_step, (unsigned long) "s\n",
-			 _ ("Back trace"), '~', debug_show_output, (unsigned long) "backtrace\n",
-			 _ ("Continue\tM-F4"), '~', debug_command, (unsigned long) "c\n",
-			 _ ("Continue until cursor\tM-F3"), '~', debug_until, 0L,
-			 _ ("Kill debugger"), '~', debug_finish, 0L,
-			 _ ("Enter command...\tM-F1"), '~', debug_show_output, 0L,
-			 _ ("Display variable..."), '~', debug_add_var, 0L,
-			 _ ("Set info..."), '~', debug_set_info, 0L
+			 _ ("Start\tM-F5"), (int) '~', debug_restart, 0L,
+			 _ ("Stop\tC-c"), (int) '~', debug_break, 0L,
+			 _ ("Attach"), (int) '~', debug_attach, 0L,
+			 _ ("Detach"), (int) '~', debug_detach, 0L,
+			 _ ("Toggle breakpoint\tM-F2"), (int) '~', debug_break_point, 0L,
+			 _ ("Insert conditional breakpoint"), (int) '~', debug_conditional, 0L,
+			 _ ("Clear all breakpoints"), (int) '~', debug_clear_breakpoints, 0L,
+			 _ ("Next\tM-F9"), (int) '~', debug_step, (unsigned long) "n\n",
+			 _ ("Step\tM-F8"), (int) '~', debug_step, (unsigned long) "s\n",
+			 _ ("List back trace"), (int) '~', debug_show_output, (unsigned long) "backtrace\n",
+			 _ ("Continue\tM-F4"), (int) '~', debug_command, (unsigned long) "c\n",
+			 _ ("Continue until cursor\tM-F3"), (int) '~', debug_until, 0L,
+			 _ ("Kill debugger"), (int) '~', debug_finish, 0L,
+			 _ ("Enter command...\tM-F1"), (int) '~', debug_show_output, 0L,
+			 _ ("Display variable..."), (int) '~', debug_add_var, 0L,
+			 _ ("Set info..."), (int) '~', debug_set_info, 0L
 	);
     CSetToolHint ("menu.debugmenu", _ ("Interactive graphical debugger"));
 }
@@ -1334,42 +1358,4 @@ void debug_key_command (int command)
 
 
 
-#if 0
-    if (!ttyname (0)) {
-	int success = 0, tty_fd;
-	char *a1, *a2, t[] = "/dev/ttyXX", p[] = "/dev/ptyXX";
-	struct group *g;
-	int gid;
-	if ((g = getgrnam ("tty")) != NULL)
-	    gid = g->gr_gid;
-	else
-	    gid = getgid ();
-	for (a1 = "pqrstuvwxyzabcde"; *a1; a1++) {
-	    p[8] = t[8] = *a1;
-	    for (a2 = "0123456789abcdef"; *a2; a2++) {
-		p[9] = t[9] = *a2;
-		tty_fd = open (p, O_RDWR, 0);
-		if (tty_fd < 0 && errno == ENOENT)
-		    goto break_out;
-		if (tty_fd < 0)
-		    continue;
-		close (0);
-		close (1);
-		close (2);
-		dup (0);
-		dup (1);
-		dup (2);
-		chown (t, getuid (), gid);
-		chmod (t, S_IRUSR | S_IWUSR | S_IWGRP);
-		success = 1;
-		goto break_out;
-	    }
-	}
-      break_out:
-	if (!success) {
-	    debug_error2 (d, _ ("Cooledit is not attached to a terminal, run cooledit from a terminal to use gdb\nor set to display in an xterm from `Set data...'"), strerror (errno));
-	    return 1;
-	}
-    }
-#endif
 
