@@ -1,5 +1,5 @@
 /* editoptions.c - save and load init file sections and do learn keys
-   Copyright (C) 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -137,10 +137,15 @@ static struct key_list klist[] =
     {gettext_noop("Ctags"), CK_Ctags, 0, 0, 0, 0, 0, 0},
     {gettext_noop("Complete"), CK_Complete, 0, 0, 0, 0, 0, 0},
     {gettext_noop("Paragraph_Format"), CK_Paragraph_Format, 0, 0, 0, 0, 0, 0},
+    {gettext_noop("Paragraph_Indent_Mode"), CK_Paragraph_Indent_Mode, 0, 0, 0, 0, 0, 0},
 #ifdef HAVE_PYTHON
     {gettext_noop("Type_Load_Python"), CK_Type_Load_Python, 0, 0, 0, 0, 0, 0},
 #endif
     {gettext_noop("Util"), CK_Util, 0, 0, 0, 0, 0, 0},
+    {gettext_noop("Match_Bracket"), CK_Match_Bracket, 0, 0, 0, 0, 0, 0},
+    {gettext_noop("Terminal"), CK_Terminal, 0, 0, 0, 0, 0, 0},
+    {gettext_noop("Terminal_App"), CK_Terminal_App, 0, 0, 0, 0, 0, 0},
+    {gettext_noop("Insert Unicode"), CK_Insert_Unicode, 0, 0, 0, 0, 0, 0},
     {"\t \t", 0, 0, 0, 0, 0, 0, 0},
     {gettext_noop("\t.A.P.P.L.I.C.A.T.I.O.N .C.O.N.T.R.O.L\t"), 0, 0, 0, 0, 0, 0, 0},
     {gettext_noop("Cancel"), CK_Cancel, 0, 0, 0, 0, 0, 0},
@@ -302,7 +307,7 @@ int save_options_section (const char *file, const char *section, const char *tex
 	result = catstrs (section, "\n", text, "\n", t, 0);
     }
     free (t);
-    return savefile (file, result, strlen (result), 0644);
+    return savefile (file, result, strlen (result), 0600);
 }
 
 
@@ -397,7 +402,6 @@ int save_user_defined_keys (struct key_list k_list[], const char *file)
    in a text box widget. Return 0 on error. Result must be free'd.
  */
 
-extern unsigned char per_char[256];
 
 char **get_key_text (void *data, int line, int *num_fields, int *tagged)
 {
@@ -415,7 +419,7 @@ char **get_key_text (void *data, int line, int *num_fields, int *tagged)
     if (!get_klist[line].key_name[0])
 	return 0;
     if (!key_s[0])
-	sprintf (key_s, "\f%c", (unsigned char) XTextWidth (CFontStruct, "99999", 5));
+	sprintf (key_s, "\f%c", (unsigned char) CImageStringWidth ("99999"));
 
     if (!i)
 	for (i = 0; get_klist[i].key_name[0]; i++) {
@@ -546,9 +550,11 @@ CWidget *Cdrawlearnkeys (Window parent, int x, int y, int columns, int lines)
 {
     CWidget *w;
 
+    CPushFont ("editor", 0);
     w = CDrawFieldedTextbox ("_learnkeysbox", parent, x, y,
 	    AUTO_WIDTH, lines * FONT_PIX_PER_LINE + 6,
 			     0, 0, get_key_text, TEXTBOX_NO_KEYS, (void *) klist);
+    CPopFont ();
 /* Tool hint */
     CSetToolHint ("_learnkeysbox", _("Click on an editing action and the press the key to bind it to"));
 
