@@ -1,5 +1,5 @@
 /* options - global options that may be set on the command line
-   Copyright (C) 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -46,6 +46,25 @@ extern int option_edit_right_extreme;
 extern int option_edit_left_extreme;
 extern int option_edit_top_extreme;
 extern int option_edit_bottom_extreme;
+extern int option_typing_replaces_selection;
+extern int option_latin2;
+extern int option_utf_interpretation;
+extern int option_reverse_hebrew;
+extern int option_rgb_order;
+extern int option_interchar_spacing;
+extern int option_file_browser_width;
+extern int option_file_browser_height;
+extern int option_shell_command_line_sticky;
+extern int option_shell_command_line_pty;
+extern int option_invert_colors;
+extern int option_invert_crome;
+
+extern int option_invert_red_green;
+extern int option_invert_green_blue;
+extern int option_invert_red_blue;
+
+extern int last_unichar_left;
+extern int last_unichar_right;
 
 extern int option_xor_cursor;
 extern int option_flashing_cursor;
@@ -58,6 +77,8 @@ extern int option_middle_button_pastes;
 extern int option_syntax_highlighting;
 extern int option_auto_spellcheck;
 
+extern int option_smooth_scrolling;
+
 extern int option_new_window_ask_for_file;
 
 #ifdef HAVE_DND
@@ -69,12 +90,44 @@ extern int option_interpret_numlock;
 extern char *editor_options_file;
 extern char *option_preferred_visual;
 
+
+extern int option_color_0;
+extern int option_color_1;
+extern int option_color_2;
+extern int option_color_3;
+extern int option_color_4;
+extern int option_color_5;
+extern int option_color_6;
+extern int option_color_7;
+extern int option_color_8;
+extern int option_color_9;
+extern int option_color_10;
+extern int option_color_11;
+extern int option_color_12;
+extern int option_color_13;
+extern int option_color_14;
+extern int option_color_15;
+extern int option_color_16;
+extern int option_color_17;
+extern int option_color_18;
+extern int option_color_19;
+extern int option_color_20;
+extern int option_color_21;
+extern int option_color_22;
+extern int option_color_23;
+extern int option_color_24;
+extern int option_color_25;
+extern int option_color_26;
+
+
 static struct {
     char *name;
     int *value;
     char *prompt;
-#define TYPE_ON_OFF	1
-#define TYPE_VALUE	2
+#define TYPE_ON_OFF		1
+#define TYPE_VALUE		2
+#define TYPE_HIDDEN_VALUE	3
+#define TYPE_HIDDEN_HEX_VALUE	4
     int type;
 } integer_options [] = {
 /* The following are check box labels */
@@ -106,7 +159,9 @@ static struct {
 	{"option_dnd_version", &option_dnd_version, 0, 0},
 #endif
 	{"option_max_undo", &option_max_undo, 0, 0},
+#if 0
 	{"option_interwidget_spacing", &option_interwidget_spacing, 0, 0},
+#endif
 	{"option_toolbar", &option_toolbar, gettext_noop(" Toolbar on edit windows "), TYPE_ON_OFF},
 	{"option_interpret_numlock", &option_interpret_numlock, gettext_noop(" Interpret Num-Lock "), TYPE_ON_OFF},
 	{"option_long_whitespace", &option_long_whitespace, gettext_noop(" Whitespace is doubled "), TYPE_ON_OFF},
@@ -120,6 +175,51 @@ static struct {
 #endif
 	{"option_middle_button_pastes", &option_middle_button_pastes, gettext_noop(" Mouse middle button pastes "), TYPE_ON_OFF},
 	{"option_new_window_ask_for_file", &option_new_window_ask_for_file, gettext_noop(" New windows ask for file "), TYPE_ON_OFF},
+	{"option_smooth_scrolling", &option_smooth_scrolling, gettext_noop(" Smooth scrolling "), TYPE_ON_OFF},
+	{"option_typing_replaces_selection", &option_typing_replaces_selection, gettext_noop(" Typing replaces selection "), TYPE_ON_OFF},
+	{"option_latin2", &option_latin2, gettext_noop(" Latin 2 composing "), TYPE_ON_OFF},
+	{"option_utf_interpretation", &option_utf_interpretation, gettext_noop(" UTF8 Interpretation "), TYPE_ON_OFF},
+	{"option_reverse_hebrew", &option_reverse_hebrew, gettext_noop(" Reverse Hebrew (Exprmntl!) "), TYPE_ON_OFF},
+	{"last_unichar_left", &last_unichar_left, 0, TYPE_HIDDEN_VALUE},
+	{"last_unichar_right", &last_unichar_right, 0, TYPE_HIDDEN_VALUE},
+	{"option_rgb_order", &option_rgb_order, 0, TYPE_HIDDEN_VALUE},
+	{"option_interchar_spacing", &option_interchar_spacing, 0, TYPE_HIDDEN_VALUE},
+	{"option_file_browser_width", &option_file_browser_width, 0, TYPE_HIDDEN_VALUE},
+	{"option_file_browser_height", &option_file_browser_height, 0, TYPE_HIDDEN_VALUE},
+	{"option_shell_command_line_sticky", &option_shell_command_line_sticky, 0, TYPE_HIDDEN_VALUE},
+	{"option_shell_command_line_pty", &option_shell_command_line_pty, 0, TYPE_HIDDEN_VALUE},
+	{"option_invert_colors", &option_invert_colors, 0, TYPE_HIDDEN_VALUE},
+	{"option_invert_crome", &option_invert_crome, 0, TYPE_HIDDEN_VALUE},
+	{"option_invert_red_green", &option_invert_red_green, 0, TYPE_HIDDEN_VALUE},
+	{"option_invert_green_blue", &option_invert_green_blue, 0, TYPE_HIDDEN_VALUE},
+	{"option_invert_red_blue", &option_invert_red_blue, 0, TYPE_HIDDEN_VALUE},
+	{"option_color_0", &option_color_0, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_1", &option_color_1, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_2", &option_color_2, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_3", &option_color_3, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_4", &option_color_4, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_5", &option_color_5, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_6", &option_color_6, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_7", &option_color_7, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_8", &option_color_8, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_9", &option_color_9, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_10", &option_color_10, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_11", &option_color_11, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_12", &option_color_12, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_13", &option_color_13, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_14", &option_color_14, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_15", &option_color_15, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_16", &option_color_16, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_17", &option_color_17, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_18", &option_color_18, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_19", &option_color_19, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_20", &option_color_20, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_21", &option_color_21, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_22", &option_color_22, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_23", &option_color_23, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_24", &option_color_24, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_25", &option_color_25, 0, TYPE_HIDDEN_HEX_VALUE},
+	{"option_color_26", &option_color_26, 0, TYPE_HIDDEN_HEX_VALUE},
 	{"option_syntax_highlighting", &option_syntax_highlighting, 0, 0},
 	{"option_auto_spellcheck", &option_auto_spellcheck, 0, 0},
 	{0, 0}
@@ -132,23 +232,30 @@ extern char *option_foreground_red;
 extern char *option_foreground_green;
 extern char *option_foreground_blue;
 extern char *option_font;
+extern char *option_widget_font;
 
 extern char *option_man_cmdline;
 extern char *option_alternate_dictionary;
 
+extern char *option_look;
+
 static struct {
     char *name;
     char **value;
 } string_options [] = {
+	{"option_look", &option_look},
 	{"option_whole_chars_search", &option_whole_chars_search},
 	{"option_chars_move_whole_word", &option_chars_move_whole_word},
+#if 0   /* mmmmmmh it shouldn't remember the display */
 	{"option_display", &option_display},
+#endif
 	{"option_geometry", &option_geometry},
 	{"option_background_color", &option_background_color},
 	{"option_foreground_red", &option_foreground_red},
 	{"option_foreground_green", &option_foreground_green},
 	{"option_foreground_blue", &option_foreground_blue},
 	{"option_font", &option_font},
+	{"option_widget_font", &option_widget_font},
 	{"option_backup_ext", &option_backup_ext},
 	{"option_man_cmdline", &option_man_cmdline},
 	{"option_preferred_visual", &option_preferred_visual}, 
@@ -229,7 +336,7 @@ int load_setup (const char *file)
 		l = strnlen (p, l);
 		if (p[l] && strchr ("\t =", p[l])) {
 		    if (!strncasecmp (p, integer_options[i].name, l)) {
-			*(integer_options[i].value) = atoi (p + l + strspn (p + l, " =\t"));
+			*(integer_options[i].value) = strtol (p + l + strspn (p + l, " =\t"), (char **) 0, 0);
 			break;
 		    }
 		}
@@ -305,7 +412,10 @@ int save_setup (const char *file)
 	}
     }
     for (i = 0; integer_options[i].name; i++) {
-	sprintf (p, "%s = %d\n%n", integer_options[i].name, *integer_options[i].value, &r);
+	if (integer_options[i].type == TYPE_HIDDEN_HEX_VALUE)
+	    sprintf (p, "%s = 0x%X\n%n", integer_options[i].name, *integer_options[i].value, &r);
+	else
+	    sprintf (p, "%s = %d\n%n", integer_options[i].name, *integer_options[i].value, &r);
 	p += r;
     }
     for (i = 0; color_options[i].name; i++) {
