@@ -1,5 +1,5 @@
 /* shell.c - user defined shell commands 
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -42,6 +42,8 @@
 #undef gettext_noop
 #define gettext_noop(x) x
 
+extern struct look *look;
+
 int save_options_section (const char *file, const char *section, const char *text);
 extern char *editor_options_file;
 extern char *init_font;
@@ -183,7 +185,7 @@ struct shell_cmd default_scripts[] =
 	'~',
 	XK_a,
 	Mod1Mask | ControlMask,
-	" Enter command : ",
+	gettext_noop (" Enter command : "),
 	SHELL_OPTION_REQUEST_ARGUMENTS |
 	SHELL_OPTION_RUN_IN_BACKGROUND,
 	0,
@@ -196,7 +198,7 @@ struct shell_cmd default_scripts[] =
 	'~',
 	XK_e,
 	Mod1Mask | ControlMask,
-	" Enter command : ",
+	gettext_noop (" Enter command : "),
 	SHELL_OPTION_REQUEST_ARGUMENTS |
 	SHELL_OPTION_RUN_IN_BACKGROUND,
 	0,
@@ -204,16 +206,18 @@ struct shell_cmd default_scripts[] =
 	XTERM_CMD " -e %a\n"
     },
     {
-	" Xterm ",
-	"Xterm\tC-M-t",
+	" Translate ",
+	"Translate...\tC-M-t",
 	'~',
 	XK_t,
 	Mod1Mask | ControlMask,
-	"",
-	SHELL_OPTION_RUN_IN_BACKGROUND,
+	gettext_noop (" Enter args for tr (See `man tr' for help): "),
+	SHELL_OPTION_SAVE_BLOCK | SHELL_OPTION_REQUEST_ARGUMENTS |
+	SHELL_OPTION_DELETE_BLOCK | SHELL_OPTION_INSERT_STDOUT |
+	SHELL_OPTION_DISPLAY_ERROR_FILE | SHELL_OPTION_CHECK_ERROR_FILE,
 	0,
 	"#!/bin/sh\n" \
-	XTERM_CMD "\n"
+	"cat %b | tr %a 2>%e\n"
     },
     {
 	gettext_noop (" C Run "),
@@ -262,7 +266,7 @@ static struct running_shell {
 static void kill_process (pid_t p)
 {
     if (p)
-	kill (p, SIGKILL);
+	kill (p, SIGTERM);
 }
 
 
@@ -375,7 +379,9 @@ static int goto_file_callback (CWidget * w, XEvent * x, CEvent * c)
     if (c->double_click || (c->command == CK_Enter && !c->handled)) {
 	int width;
 	char *q;
+	CPushFont ("editor", 0);
 	width = w->options & TEXTBOX_WRAP ? (w->width - TEXTBOX_BDR) / FONT_MEAN_WIDTH : 32000;
+	CPopFont ();
 	q = strline (w->text, strmovelines (w->text, w->current, w->cursor - w->firstline, width));
 	goto_error (q, 1);
     }
@@ -426,7 +432,6 @@ static int display_file_callback (CWidget * w, XEvent * x, CEvent * c)
 
 
 #define CHUNK 8192
-extern char per_char[256];
 
 static void shell_pool_update (int fd, fd_set * reading, fd_set * writing, fd_set * error, void *data)
 {
@@ -454,16 +459,14 @@ static void shell_pool_update (int fd, fd_set * reading, fd_set * writing, fd_se
 	    pool_advance (r->shell_pool, CHUNK + 1);
 	    do_redraw = 1;	/* must redraw cause pool has changed */
 	}
-	do {
-	    c = read (fd, pool_current (r->shell_pool), CHUNK);
-	} while (c < 0 && errno == EINTR);
+	while ((c = read (fd, pool_current (r->shell_pool), CHUNK)) == -1 && errno == EINTR);
 /* translate unreadables */
 	for (j = 0, p = (unsigned char *) pool_current (r->shell_pool); j < c; j++, p++) {
 	    if (*p == '\t')
 		*p = ' ';
 	    else if (*p == '\n')
 		*p = '\n';
-	    else if (!per_char[*p])
+	    else if (!FONT_PER_CHAR(*p))
 		*p = '?';
 	}
 	if (c <= 0)
@@ -477,6 +480,7 @@ static void shell_pool_update (int fd, fd_set * reading, fd_set * writing, fd_se
     if (count || do_redraw) {
 /* we do a manual appending of the text for optimization purposes, instead of CRedrawTextbox */
 	w->text = (char *) pool_start (r->shell_pool);
+	CPushFont ("editor", 0);
 	w->numlines += strcountlines (w->text + w->textlength, 0, 2000000000, w->options & TEXTBOX_WRAP ? (w->width - TEXTBOX_BDR) / FONT_MEAN_WIDTH : 32000);
 	w->textlength = pool_length (r->shell_pool);
 /* only redraw if we crossed a newline */
@@ -486,7 +490,9 @@ static void shell_pool_update (int fd, fd_set * reading, fd_set * writing, fd_se
 		if (w->numlines > w->firstline + (w->height / FONT_PIX_PER_LINE - 1))
 		    CSetTextboxPos (w, TEXT_SET_LINE, w->numlines - (w->height / FONT_PIX_PER_LINE - 1));
 	}
+	CPopFont ();
     }
+    r->killme |= CChildExitted (r->shell_pid, 0);
     if (!count && r->killme) {
 	pool_break (r->shell_pool);
 	r->shell_pool = 0;
@@ -518,6 +524,7 @@ static void shell_display_output (char *text, const char *heading, int (*select_
 	CWidget *w;
 	win = CDrawMainWindow (nm (i, "shelldisplaytext", 0, 0), heading);
 	CGetHintPos (&x, &y);
+	CPushFont ("editor", 0);
 	running_shell[i].w = w = CDrawTextbox (nm (i, "shelldisplaytext", "text", 0), win, x, y, 80 * FONT_MEAN_WIDTH + 7, 25 * FONT_PIX_PER_LINE + 6, 0, 0, text, TEXTBOX_WRAP | TEXTBOX_NO_STRDUP);
 /* Toolhint */
 	CSetToolHint (nm (i, "shelldisplaytext", "text", 0), _ ("Double click on file:line type messages to goto the\nfile and line number.  Note that the file will not\nauto-load unless there is a full path in the message."));
@@ -530,6 +537,7 @@ static void shell_display_output (char *text, const char *heading, int (*select_
 	CCentre (nm (i, "shelldisplaytext", "done", 0));
 	CSetSizeHintPos (nm (i, "shelldisplaytext", 0, 0));
 	CSetWindowResizable (nm (i, "shelldisplaytext", 0, 0), FONT_MEAN_WIDTH * 15, FONT_PIX_PER_LINE * 15, 1600, 1200);	/* minimum and maximum sizes */
+	CPopFont ();
 	CMapDialog (nm (i, "shelldisplaytext", 0, 0));
 	CAddCallback (nm (i, "shelldisplaytext", "text", 0), select_line_callback);
 	CAddCallback (nm (i, "shelldisplaytext", "done", 0), display_file_callback);
@@ -555,7 +563,7 @@ int execute_background_display_output (char *title, char *s, char *name)
     int i;
     i = restart_shell (0, name, 0);
     argv[0] = hme_i (SCRIPT_FILE, i);
-    savefile (argv[0], s, strlen (s), 0777);
+    savefile (argv[0], s, strlen (s), 0700);
     if ((p = triple_pipe_open (0, &running_shell[i].shell_pipe, 0, 1, argv[0], argv)) < 0)
 	return 1;
     running_shell[i].shell_pid = p;
@@ -581,7 +589,7 @@ static int execute_background_shell (struct shell_cmd *s, char *script, char *na
 
     i = restart_shell (0, name, 0);
     argv[0] = hme_i (SCRIPT_FILE, i);
-    savefile (argv[0], script, strlen (script), 0777);
+    savefile (argv[0], script, strlen (script), 0700);
 
     if (s->options & SHELL_OPTION_DISPLAY_STDOUT_CONTINUOUS) {
 	if ((p = triple_pipe_open (0, &running_shell[i].shell_pipe, 0,
@@ -640,7 +648,7 @@ static char *execute_foreground_shell (struct shell_cmd *s, char *script)
     char *t;
 
     argv[0] = hme (SCRIPT_FILE);
-    savefile (argv[0], script, strlen (script), 0777);
+    savefile (argv[0], script, strlen (script), 0700);
     if (s->options & SHELL_OPTION_INSERT_STDOUT) {
 	if ((p = triple_pipe_open (0, &shell_foreground_pipe, 0, (s->options & SHELL_OPTION_INSERT_STDERR) ? 1 : 0, hme (SCRIPT_FILE), argv)) < 0)
 	    return 0;
@@ -1099,7 +1107,7 @@ int edit_scripts_dialog (Window parent, int x, int y, struct shell_cmd *s)
 	CSetToolHint (catstrs (options[i].w->ident, ".label", 0), _ (options[i].toolhint));
 	CGetHintPos (0, &y);
     }
-    x = xs + CStringWidth (_ (options[3].name)) + 32 + WIDGET_SPACING * 2 + 10;
+    x = xs + CImageStringWidth (_ (options[3].name)) + 32 + WIDGET_SPACING * 2 + 10;
     y = ys;
     for (i = NUM_OPTS / 2; i < NUM_OPTS; i++) {
 	options[i].w = CDrawSwitch (catstrs ("shellopt.", itoa (i), 0), win, x, y, (s->options & options[i].flag) ? 1 : 0, _ (options[i].name), 0);
@@ -1107,8 +1115,10 @@ int edit_scripts_dialog (Window parent, int x, int y, struct shell_cmd *s)
 	CSetToolHint (catstrs (options[i].w->ident, ".label", 0), _ (options[i].toolhint));
 	CGetHintPos (0, &y);
     }
+    CPushFont ("editor", 0);
     w = CDrawEditor ("shellopt.edit", win, xs, y,
 		     77 * FONT_MEAN_WIDTH, 10 * FONT_PIX_PER_LINE, s->script, 0, 0, EDITOR_NO_TEXT | EDITOR_NO_FILE, strlen (s->script));
+    CPopFont ();
 /* Toolhint */
     CSetToolHint ("shellopt.edit", _ ("Enter your shell script here"));
 
@@ -1135,10 +1145,20 @@ int edit_scripts_dialog (Window parent, int x, int y, struct shell_cmd *s)
 	}
 	if (!strcmp (cw.ident, "shellopt.ok")) {
 	    XEvent *p;
+#ifdef USE_XIM
+	    XIC ic = 0;
+#endif
 	    p = CRawkeyQuery ((CIdent ("shellopt.ok"))->mainid, 20, 20, _ (" Script Edit "), _ (" Press the key combination to envoke the script : "));
 	    if (!p)
 		continue;
+#ifdef USE_XIM
+	    ic = CIC;
+	    CIC = 0;
 	    s->key = CKeySym (p);
+	    CIC = ic;
+#else
+	    s->key = CKeySym (p);
+#endif
 	    s->keyboard_state = p->xkey.state;
 	    s->options = 0;
 	    for (i = 0; i < NUM_OPTS; i++)
@@ -1252,3 +1272,163 @@ void new_script_cmd (unsigned long ignored)
     update_script_menu_items ();
 }
 
+static CWidget *CDrawMiniSwitch (const char *identifier, Window parent, int x, int y, int d, int on)
+{
+    CWidget *w;
+    w = CSetupWidget (identifier, parent, x, y, d, d, C_SWITCH_WIDGET, INPUT_BUTTON, COLOR_FLAT, 1);
+    w->fg = COLOR_BLACK;
+    w->bg = COLOR_FLAT;
+    w->keypressed = on;
+    w->render = render_switch;
+    w->options |= WIDGET_TAKES_FOCUS_RING;
+    return w;
+}
+
+pid_t open_under_pty (int *in, int *out, char *line, const char *file, char *const argv[]);
+
+int option_shell_command_line_sticky = 0;
+int option_shell_command_line_pty = 0;
+void shell_output_add_job (WEdit * edit, int in, int out, pid_t pid, char *name, int close_on_error);
+
+static struct shell_job *get_job (WEdit * edit, int n)
+{
+    struct shell_job *j;
+    int i;
+    for (i = 0, j = edit->jobs; j && i < n; j = j->next, i++);
+    return j;
+}
+
+static char *list_jobs_get_line (void *data, int line)
+{
+    struct shell_job *j;
+    j = get_job ((WEdit *) data, line);
+    if (j)
+	return j->name;
+    return "";
+}
+
+static int list_jobs (void)
+{
+    struct shell_job *j;
+    int i, c, n;
+    WEdit *e;
+    e = edit[current_edit]->editor;
+    c = max (20, e->num_widget_columns - 5);
+    for (n = 0, j = e->jobs; j; j = j->next, n++);
+    i = CListboxDialog (edit[current_edit]->mainid, 20, 20, c, 10, 0, 0, 0, n, list_jobs_get_line, (void *) e);
+    if (i >= 0) {
+	j = get_job (e, i);
+	if (j)
+	    shell_output_kill_job (e, j->pid);
+    }
+    return 0;
+}
+
+void edit_insert_shell_output (WEdit * edit)
+{
+    char id[33], q[1024], *p;
+    CWidget *w, *v, *i, *b, *h, *c;
+    int tolong = 0, done = 0;
+    CState s;
+    while (!done) {
+	strcpy (id, CIdentOf (edit->widget));
+	strcat (id, ".text");
+	w = CIdent (id);
+	if (!w)
+	    return;
+	CBackupState (&s);
+	CDisable ("*");
+	p = getenv ("PWD");
+	get_current_wd (q, 1023);
+	q[1023] = '\0';
+	p = q;
+	CPushFont ("widget", 0);
+	while (*p && CImageStringWidth (p) > CWidthOf (w) * 2 / 3 - 20) {
+	    tolong = 1;
+	    p++;
+	}
+	CPopFont ();
+	h = CDrawButton ("status_button", edit->widget->parentid, CXof (w), CYof (w), CHeightOf (w), CHeightOf (w), 0);
+	CSetToolHint (h->ident, _("Click for list of running jobs."));
+	b =
+	    CDrawMiniSwitch ("status_switch", edit->widget->parentid,
+			     CXof (w) + CWidthOf (h), CYof (w), CHeightOf (w), option_shell_command_line_sticky);
+	CSetToolHint (b->ident, _("If not depressed then input line will close on Enter."));
+	c =
+	    CDrawMiniSwitch ("status_switch2", edit->widget->parentid,
+			     CXof (w) + CWidthOf (h) + CWidthOf (b), CYof (w), CHeightOf (w),
+			     option_shell_command_line_pty);
+	CSetToolHint (c->ident, _("If depressed then opens under a tty."));
+	v =
+	    CDrawText ("status_prompt", edit->widget->parentid,
+		       CXof (w) + CWidthOf (b) + CWidthOf (c) + CWidthOf (h), CYof (w), "[%s%s]#", tolong ? "..." : "",
+		       p);
+	CSetToolHint (v->ident,
+		      _("Current directory. This is the global current directory as set from the Command menu."));
+	i =
+	    CDrawTextInput ("status_input", edit->widget->parentid,
+			    CXof (w) + CWidthOf (b) + CWidthOf (c) + CWidthOf (v) + CWidthOf (h), CYof (w),
+			    CWidthOf (edit->widget) - CWidthOf (v) - CWidthOf (h) - CWidthOf (b) - CWidthOf (c),
+			    AUTO_HEIGHT, 32768, TEXTINPUT_LAST_INPUT);
+	CSetToolHint (i->ident,
+		      _
+		      ("Enter shell command. Input will be piped to this command from\nselected text. Output will be inserted at editor cursor."));
+	CFocus (i);
+	edit->force |= REDRAW_PAGE;
+	edit_render_keypress (edit);
+	edit_push_action (edit, KEY_PRESS + edit->start_display);
+	for (;;) {
+	    XEvent xev;
+	    CEvent cev;
+	    CNextEvent (&xev, &cev);
+	    option_shell_command_line_pty = c->keypressed;
+	    if (xev.type == KeyPress && cev.command == CK_Enter && cev.kind == C_TEXTINPUT_WIDGET) {
+		char *t;
+		for (t = i->text; *t && isspace (*t); t++);
+		if (!strncmp (t, "cd ", 3)) {
+		    for (t = t + 3; *t && isspace (*t); t++);
+		    if (change_directory (t) < 0)
+			CErrorDialog (main_window, 20, 20,
+				      _(" Change directory "), get_sys_error (_(" Error return from chdir. ")));
+		    done = !b->keypressed;
+		} else {
+		    int in, out;
+		    pid_t pid;
+		    char *arg[5];
+		    char line[80];
+		    arg[0] = "/bin/sh";
+		    arg[1] = "-c";
+		    arg[2] = i->text;
+		    arg[3] = 0;
+		    if (option_shell_command_line_pty) {
+			pid = open_under_pty (&in, &out, line, "/bin/sh", arg);
+		    } else {
+			pid = triple_pipe_open (&in, &out, 0, 1, "/bin/sh", arg);
+		    }
+		    shell_output_add_job (edit, in, out, pid, i->text, !option_shell_command_line_pty);
+		    done = !b->keypressed;
+		}
+		break;
+	    }
+	    if (xev.type == KeyPress && cev.command == CK_Cancel) {
+		done = 1;
+		break;
+	    }
+	    if (xev.type == ButtonPress && cev.window != v->winid && cev.window != i->winid && cev.window != b->winid
+		&& cev.window != c->winid && cev.window != h->winid) {
+		done = 1;
+		break;
+	    }
+	    if (!strcmp (cev.ident, "status_button"))
+		list_jobs ();
+	}
+	option_shell_command_line_sticky = b->keypressed;
+	CDestroyWidget ("status_prompt");
+	CDestroyWidget ("status_input");
+	CDestroyWidget ("status_button");
+	CDestroyWidget ("status_switch");
+	CDestroyWidget ("status_switch2");
+	CRestoreState (&s);
+    }
+}
+
