@@ -1,5 +1,5 @@
 /* coolicons.c - example usage of the coolwidget X API
-   Copyright (C) 1997, 1998 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -48,6 +48,7 @@ char *gather_icon_options_for_config (void);
 void edit_tri_cursor (Window win);
 void free_icon_elements (CPowerIcon * icon);
 void free_all_lists (void);
+void extra_loop_stuff (XEvent * xevent);
 
 #define SYSTEM_ICONS LIBDIR "/coolicon.config"
 #define E_DATA LIBDIR "/e.data"
@@ -86,7 +87,7 @@ extern char *init_font;
 char *option_display = 0;
 
 /* font from the command line */
-char *option_font = "-*-helvetica-bold-r-*--12-*-*-*-p-*-iso8859-1";
+char *option_font = "-adobe-helvetica-medium-r-normal--*-120-*-*-*-*-*-*";
 
 /* config file */
 char *current_config_file = 0;
@@ -172,21 +173,9 @@ static void process_command_line (int argc, char **argv)
 
 /* }}} command-line options */
 
-
-static RETSIGTYPE child_handler (int x)
-{
-    waitpid (0, 0, WNOHANG | WUNTRACED);
-
-    waitpid (0, 0, WNOHANG | WUNTRACED);
-
-    signal (SIGCHLD, child_handler);
-    handler_return;
-}
-
 static void set_signals (void)
 {
     signal (SIGPIPE, SIG_IGN);
-    signal (SIGCHLD, child_handler);
 }
 
 static int icon_width;
@@ -239,18 +228,21 @@ void arrange_icons (void)
     }
 }
 
+static long raise_lower_all_callback (CWidget * w, long r)
+{
+    if (w->kind == C_ICON_WIDGET) {
+	if (r)
+	    XRaiseWindow (CDisplay, w->winid);
+	else
+	    XLowerWindow (CDisplay, w->winid);
+    }
+    return 0;
+}
+
 void raise_lower_all (int r)
 {
-    int i = 0;
     XUngrabPointer (CDisplay, CurrentTime);
-    while (last_widget > i++)
-	if (CIndex (i) != 0)
-	    if ((CIndex (i))->kind == C_ICON_WIDGET) {
-		if (r)
-		    XRaiseWindow (CDisplay, (CIndex (i))->winid);
-		else
-		    XLowerWindow (CDisplay, (CIndex (i))->winid);
-	    }
+    for_all_widgets ((void *) raise_lower_all_callback, (void *) r, 0);
 }
 
 void edit_change_directory (void)
@@ -307,8 +299,10 @@ int edit_dialog (CPowerIcon * icon, char *heading)
     CTextSize (&w, &h, _(" Prompt before executing : "));
     w += WIDGET_SPACING * 2;
     CDrawText ("edit.ttitle", win, x, y, _(" Icon title : "));
+    CPushFont ("editor", 0);
     edit0 = CDrawEditor ("edit.title", win, x + w, y,
 			 FONT_MEAN_WIDTH * 41, 2 * FONT_PIX_PER_LINE, icon->title ? icon->title : "", 0, 0, EDITOR_NO_TEXT | EDITOR_NO_FILE | EDITOR_NO_SCROLL, strlen (icon->title ? icon->title : ""));
+    CPopFont ();
     CGetHintPos (0, &y);
 
     CDrawText ("edit.tfilename", win, x, y, _(" Icon XPM file-name : "));
@@ -351,7 +345,7 @@ int edit_dialog (CPowerIcon * icon, char *heading)
     CDrawSwitch ("edit.promptdrop", win, x2, y, icon->options & ICON_OPTION_PROMPT_DROP, _ (" Show prompt "), 2);
     CGetHintPos (0, &y);
 
-    CDrawText ("edit.tmimetypes", win, x, y, _(" Comma separated list of major Top-Level MIME types : "));
+    CDrawText ("edit.tmimetypes", win, x, y, _(" A prioritised comma separated list of MIME types : "));
     CGetHintPos (0, &y);
 
     mime_majors = 0;
@@ -364,12 +358,16 @@ int edit_dialog (CPowerIcon * icon, char *heading)
 
     CDrawText ("edit.tdropscript", win, x, y, _(" Script for receiving a drop : "));
     CGetHintPos (0, &y2);
+    CPushFont ("editor", 0);
     edit1 = CDrawEditor ("edit.dropscript", win, x, y2,
 			 40 * FONT_MEAN_WIDTH, 10 * FONT_PIX_PER_LINE, icon->drop_script ? icon->drop_script : "#!/bin/sh\n\n", 0, 0, EDITOR_NO_TEXT | EDITOR_NO_FILE, strlen (icon->drop_script ? icon->drop_script : "#!/bin/sh\n\n"));
+    CPopFont ();
     CGetHintPos (&half, 0);
     CDrawText ("edit.tdclick", win, x + half, y, _(" Script for double click : "));
+    CPushFont ("editor", 0);
     edit2 = CDrawEditor ("edit.dclick", win, x + half, y2,
 			 40 * FONT_MEAN_WIDTH, 10 * FONT_PIX_PER_LINE, icon->double_click_script ? icon->double_click_script : "#!/bin/sh\n\n", 0, 0, EDITOR_NO_TEXT | EDITOR_NO_FILE, strlen (icon->double_click_script ? icon->double_click_script : "#!/bin/sh\n\n"));
+    CPopFont ();
     CGetHintPos (0, &y);
 
     get_hint_limits (&x, &y);
@@ -383,7 +381,9 @@ int edit_dialog (CPowerIcon * icon, char *heading)
     CMapDialog ("edit");
 
     for (;;) {
-	CNextEvent (0, &cwevent);
+	XEvent xevent;
+	CNextEvent (&xevent, &cwevent);
+	extra_loop_stuff (&xevent);
 	if (!CIdent ("edit"))
 	    break;
 	if (!strcmp (cwevent.ident, "edit.cancel"))
@@ -763,6 +763,30 @@ char *load_config (char *file, long *l)
 
 extern int option_auto_spellcheck;
 
+void extra_loop_stuff (XEvent * xevent)
+{
+    if (xevent->type == AlarmEvent) {
+	if (cursor_revert_timeout <= 0 && cursor_revert_window) {
+	    edit_tri_cursor (cursor_revert_window);
+	    cursor_revert_window = 0;
+	} else {
+	    cursor_revert_timeout--;
+	}
+    }
+    if (CDndClass->stage == XDND_DROP_STAGE_IDLE && drop.data) {
+	CWidget *w;
+	w = CIdent (drop.ident);
+	if (w)
+	    execute_icon_drop (w, (CPowerIcon *) w->user,
+			       (char *) drop.data, drop.size,
+			       drop.atom_name);
+	free (drop.data);
+	free (drop.ident);
+	free (drop.atom_name);
+	memset (&drop, 0, sizeof (drop));
+    }
+}
+
 int main (int argc, char **argv)
 {
     char *config;
@@ -789,11 +813,15 @@ int main (int argc, char **argv)
     memset (&coolicon_startup, 0, sizeof (coolicon_startup));
     coolicon_startup.name = argv[0];
     coolicon_startup.display = option_display;
-    coolicon_startup.font = option_font;
+    coolicon_startup.font = "7x13";
+    coolicon_startup.widget_font = option_font;
     if (option_wait_for_display)
 	coolicon_startup.options |= CINIT_OPTION_WAIT_FOR_DISPLAY;
 
     CInitialise (&coolicon_startup);
+    CPushFont ("icon", "-*-helvetica-bold-r-*--12-*-*-*-*-*-*");
+    CPushFont ("editor", 0);
+    CPushFont ("widget", 0);
 
 /* create main window that will not be displayed (needed because the
    library does some important initialisation with the first window) */
@@ -823,24 +851,7 @@ int main (int argc, char **argv)
 /* Run the application. */
     while (!done) {
 	CNextEvent (&xevent, &cwevent);
-	if (xevent.type == AlarmEvent) {
-	    if (cursor_revert_timeout <= 0 && cursor_revert_window) {
-		edit_tri_cursor (cursor_revert_window);
-		cursor_revert_window = 0;
-	    } else {
-		cursor_revert_timeout--;
-	    }
-	}
-	if (CDndClass->stage == XDND_DROP_STAGE_IDLE && drop.data) {
-	    CWidget *w;
-	    w = CIdent (drop.ident);
-	    if (w)
-		execute_icon_drop (w, (CPowerIcon *) w->user, (char *) drop.data, drop.size, drop.atom_name);
-	    free (drop.data);
-	    free (drop.ident);
-	    free (drop.atom_name);
-	    memset (&drop, 0, sizeof (drop));
-	}
+	extra_loop_stuff (&xevent);
     }
     return 0;
 }
