@@ -1,5 +1,5 @@
 /* iconmanager.c - runs stuff when icons are clicked on etc.
-   Copyright (C) 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -301,25 +301,7 @@ void icon_menu (char *ident)
 	return;
     if (!w->droppedmenu)
 	return;
-
-    CPullDown (w->droppedmenu);
-    XGrabPointer (CDisplay, w->droppedmenu->droppedmenu->winid, False,
-		  ButtonPressMask | ButtonReleaseMask | ButtonMotionMask,
-		  GrabModeSync, GrabModeAsync, CRoot,
-		  CGetCursorID (2), CurrentTime);
-    for (;;) {
-	XAllowEvents (CDisplay, SyncPointer, CurrentTime);
-	CNextEvent (&xevent, &cwevent);
-	if (xevent.type == ButtonRelease)
-	    break;
-    }
-    XUngrabPointer (CDisplay, CurrentTime);
-    w = CIdent (ident);
-    if (!w)
-	return;
-    if (!w->droppedmenu)
-	return;
-    CPullUp (w->droppedmenu);
+    CMenuSelectionDialog (w->droppedmenu);
 }
 
 #ifndef HAVE_DND
@@ -476,6 +458,7 @@ CWidget *CDrawPowerIcon (CPowerIcon * p)
 					_ ("Save icons"), 0, save_icons, 0
 	    );
     }
+    w->droppedmenu->options &= ~MENU_AUTO_PULL_UP;
     return w;
 }
 
