@@ -1,6 +1,6 @@
 /* coolnext.c - CNextEvent function and support, and various event
 		handlers and utils
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,9 +27,7 @@
 #include <stdlib.h>
 #include <stdarg.h>
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xresource.h>
+#include <X11/Intrinsic.h>
 #include "lkeysym.h"
 
 #include "stringtools.h"
@@ -45,6 +43,8 @@
 
 #include "mad.h"
 
+
+
 int last_region = 0;
 
 
@@ -59,7 +59,6 @@ extern Atom DndProtocol, OldDndProtocol;
 void add_to_focus_stack (Window w);
 void selection_clear (void); /* from editwidget.c */
 
-static Window window_is_resizing = 0;
 int option_mouse_double_click = 300;
 int option_middle_button_pastes = 1;
 
@@ -113,24 +112,30 @@ void resolve_button (XEvent * xevent, CEvent * cwevent)
     window = xevent->xany.window;
 }
 
-/*sends all windows that are marked top_bottom = 1 to the bottom */
-void CLowerWindows ()
+static long lower_window_callback (CWidget * w)
 {
-    int j = 0;
-    while (last_widget > j++)
-	if (CIndex (j) != NULL)
-	    if (CIndex (j)->position & WINDOW_ALWAYS_LOWERED)
-		XLowerWindow (CDisplay, CIndex (j)->winid);
+    if (w->position & WINDOW_ALWAYS_LOWERED)
+	XLowerWindow (CDisplay, w->winid);
+    return 0;
+}
+
+/*sends all windows that are marked top_bottom = 1 to the bottom */
+void CLowerWindows (void)
+{
+    for_all_widgets ((void *) lower_window_callback, 0, 0);
+}
+
+static long raise_window_callback (CWidget * w)
+{
+    if (w->position & WINDOW_ALWAYS_RAISED)
+	XRaiseWindow (CDisplay, w->winid);
+    return 0;
 }
 
 /*sends all windows that are marked top_bottom = 2 to the top */
-void CRaiseWindows ()
+void CRaiseWindows (void)
 {
-    int j = 0;
-    while (last_widget > j++)
-	if (CIndex (j) != NULL)
-	    if (CIndex (j)->position & WINDOW_ALWAYS_RAISED)
-		XRaiseWindow (CDisplay, CIndex (j)->winid);
+    for_all_widgets ((void *) raise_window_callback, 0, 0);
 }
 
 
@@ -325,7 +330,7 @@ int CCheckWindowEvent (Window w, long event_mask, XEvent ** event_return)
 int CKeyPending (void)
 {
     XSync (CDisplay, 0);
-    return CCheckWindowEvent (0, KeyPressMask | KeyReleaseMask, 0);
+    return CCheckWindowEvent (0, KeyPressMask, 0);
 }
 
 /* send an expose event via the internal queue */
@@ -460,30 +465,12 @@ static void pop_all_regions (Window w)
 
 /* {{{ key conversion utilities */
 
-/* just get a keysym (one of eg XK_... in keysymdef.h) x_lat must have space for four bytes */
-static KeySym key_sym_xlat (XEvent * e, char *x_lat)
-{
-    KeySym key;
-    if (e->type == KeyPress || e->type == KeyRelease) {
-	XLookupString (&(e->xkey), x_lat, 4, &key, NULL);
-	return key;
-    } else
-	return 0;
-}
+/* xim.c */
+KeySym key_sym_xlat (XEvent * ev, char *x_lat);
 
 KeySym CKeySym (XEvent * e)
 {
-    char xlat[4];
-    return key_sym_xlat (e, xlat);
-}
-
-static unsigned long toggle_bit (unsigned long x, unsigned long mask)
-{
-    unsigned long m = -1;
-    if (x & mask)
-	return x & (m - mask);
-    else
-	return x | mask;
+    return key_sym_xlat (e, 0);
 }
 
 int mod_type_key (KeySym x)
@@ -543,17 +530,31 @@ short CKeySymMod (XEvent * ev)
 {
     KeySym p;
     XEvent e;
+#ifdef USE_XIM
+    XIC ic;
+#endif
     int state;
     if (!ev)
 	return 0;
     e = *ev;
     state = e.xkey.state;
     e.xkey.state = 0;		/* want the raw key */
+#ifdef USE_XIM
+    ic = CIC;
+/* turn off IC 'cos we want raw translation */
+    CIC = 0;
     p = CKeySym (&e);
-    if (p && !mod_type_key(p)) {
-	p = toggle_bit (p, 0x1000 * ((state & ShiftMask) != 0));
-	p = toggle_bit (p, 0x2000 * ((state & ControlMask) != 0));
-	p = toggle_bit (p, 0x4000 * ((state & MyAltMask) != 0));
+    CIC = ic;
+#else
+    p = CKeySym (&e);
+#endif
+    if (p && !mod_type_key (p)) {
+	if (state & ShiftMask)
+	    p ^= 0x1000;
+	if (state & ControlMask)
+	    p ^= 0x2000;
+	if (state & MyAltMask)
+	    p ^= 0x4000;
 	p &= 0x7FFF;
     } else
 	p = 0;
@@ -655,7 +656,7 @@ static int CCheckButtonHotKey (XEvent * xevent, CEvent * cwevent)
 	    w = CFindFirstDescendent (xevent->xany.window);
 	else if (!w->takes_focus)
 	    w = CChildFocus (w);
-	p = w;
+	p = w = CNextFocus (w);
 	do {
 	    if (!w)
 		return 0;
@@ -669,22 +670,21 @@ static int CCheckButtonHotKey (XEvent * xevent, CEvent * cwevent)
     return 0;
 }
 
+static long check_hotkey_callback (CWidget * w, long k)
+{
+    if (w->takes_focus && !w->disabled)
+	if (match_hotkey (w->hotkey, k)) {
+	    click_on_widget (w);
+	    return 1;
+	}
+    return 0;
+}
+
 /* checks all widgets for a hotkey if alt is pressed */
 static int CCheckGlobalHotKey (XEvent * xevent, CEvent * cwevent)
 {
-    if (xevent->type == KeyPress && (xevent->xkey.state & MyAltMask) && !(xevent->xkey.state & ControlMask)) {
-	int i = last_widget;
-	KeySym k;
-	CWidget *w;
-	k = CKeySym (xevent);
-	while (--i > 0)
-	    if ((w = CIndex (i)) != NULL)
-		if (w->takes_focus && !w->disabled)
-		    if (match_hotkey (w->hotkey, k)) {
-			click_on_widget (w);
-			return 1;
-		    }
-    }
+    if (xevent->type == KeyPress && (xevent->xkey.state & MyAltMask) && !(xevent->xkey.state & ControlMask))
+	return for_all_widgets ((void *) check_hotkey_callback, (void *) CKeySym (xevent), 0);
     return 0;
 }
 
@@ -779,6 +779,8 @@ static int set_mapped (Window win, int i)
 }
 
 int compose_key_pressed = 0;
+int compose_key_which = 0;
+
 static unsigned int key_board_state = 0;
 
 unsigned int CGetKeyBoardState (void)
@@ -786,15 +788,41 @@ unsigned int CGetKeyBoardState (void)
     return key_board_state;
 }
 
+extern int option_latin2;
+
 static void set_compose_key (XEvent * xevent, int type)
 {
     KeySym compose_key;
     compose_key = CKeySym (xevent);
-    if (compose_key == MyComposeKey || compose_key == XK_Multi_key) {
-	compose_key_pressed = (type == KeyPress);
+
+    if (!option_latin2) {
+	if (compose_key == MyComposeKey || compose_key == XK_Multi_key) {
+	    compose_key_pressed = (type == KeyPress);
+	} else {
+	    if (!(xevent->xkey.state & ControlMask))
+		compose_key_pressed = 0;
+	}
     } else {
-	if (!(xevent->xkey.state & ControlMask))
-	    compose_key_pressed = 0;
+#if defined(XK_dead_belowdot) && defined(XK_dead_acute)
+	if (compose_key >= XK_dead_acute &&
+	    compose_key <= XK_dead_belowdot) {
+	    if (compose_key_pressed)
+		return;
+	    compose_key_pressed = 1;
+	    compose_key_which = compose_key;
+	} else
+#endif
+	{
+	    if (!(xevent->xkey.state & ControlMask))
+		if (compose_key != XK_Shift_L &&
+		    compose_key != XK_Shift_R &&
+		    compose_key != XK_Alt_L &&
+		    compose_key != XK_Alt_R &&
+		    compose_key != XK_Mode_switch) {
+		    compose_key_pressed = 0;
+		    compose_key_which = 0;
+		}
+	}
     }
 }
 
@@ -822,8 +850,8 @@ static void render_text_ordinary (Window win, int x, int y, char *q)
     for (;;) {
 	if (!(p = strchr (q, '\n')))
 	    p = q + strlen (q);
-	XDrawString (CDisplay, win, CGC, FONT_OFFSET_X + x, FONT_OFFSET_Y + h + y, \
-		     q, (unsigned long) p - (unsigned long) q);
+	CImageText (win, FONT_OFFSET_X + x, FONT_OFFSET_Y + h + y, q,
+		    (unsigned long) p - (unsigned long) q);
 	h += FONT_PIX_PER_LINE;
 	if (!*p)
 	    break;
@@ -836,6 +864,7 @@ static int eh_toolhint (CWidget * w, XEvent * xevent, CEvent * cwevent)
     if (xevent->type == Expose)
 	if (!xevent->xexpose.count && w->label) {
 	    CSetColor (color_widget (4));
+	    CSetBackgroundColor (color_widget (14));
 	    render_text_ordinary (w->winid, 2, 2, w->label);
 	    CSetColor (color_widget (0));
 	    XDrawRectangle (CDisplay, w->winid, CGC, 0, 0, w->width - 1, w->height - 1);
@@ -893,89 +922,115 @@ void CSetToolHint (const char *ident, char *text)
 #define FD_SETSIZE 256
 #endif
 
-struct file_des_watch {
+static struct file_des_watch {
     int fd, how;
     void (*callback) (int, fd_set *, fd_set *, fd_set *, void *);
     struct file_des_watch *next;
     void *data;
-} *watch_table = 0;
+} *watch_table[FD_SETSIZE];
+
+static int watch_table_last = 0;
 
 /* returns non-zero if table is full */
-int CAddWatch (int fd, void (*callback) (int, fd_set *, fd_set *, fd_set *, void *), int how, void *data)
+int CAddWatch (int fd, void (*callback) (int, fd_set *, fd_set *, fd_set *, void *), int how,
+	       void *data)
 {
-    struct file_des_watch *l;
     int i;
-    if (!callback || fd < 0 || !how)
+    if (!callback || fd < 0 || !how) {
+	fprintf (stderr, "bad args to CAddWatch??");
 	return 1;
-    for (i = 0, l = watch_table; l; l = l->next, i++) {
-	if (l->callback == callback && l->fd == fd) {
-	    l->how |= how;
+    }
+    for (i = 0; i < watch_table_last; i++) {
+	if (!watch_table[i])
+	    continue;
+	if (watch_table[i]->callback == callback && watch_table[i]->fd == fd) {
+	    watch_table[i]->how |= how;
 	    return 0;
 	}
     }
-    if (i >= FD_SETSIZE)
+    for (i = 0; i < watch_table_last && watch_table[i]; i++);
+    if (i >= FD_SETSIZE) {
+	fprintf (stderr, "watch table overflow??");
 	return 1;
-    l = malloc (sizeof (*l));
-    l->next = watch_table;
-    l->callback = callback;
-    l->how = how;
-    l->fd = fd;
-    l->data = data;
-    watch_table = l;
+    }
+    watch_table[i] = malloc (sizeof (struct file_des_watch));
+    watch_table[i]->callback = callback;
+    watch_table[i]->how = how;
+    watch_table[i]->fd = fd;
+    watch_table[i]->data = data;
+    if (watch_table_last < i + 1)
+	watch_table_last = i + 1;
     return 0;
 }
 
 void CRemoveWatch (int fd, void (*callback) (int, fd_set *, fd_set *, fd_set *, void *), int how)
 {
-    struct file_des_watch *l, *p = 0;
-    for (l = watch_table; l; l = l->next) {
-	if (l->callback == callback && l->fd == fd) {
-	    l->how &= ~how;
-	    if (!l->how) {
-		if (p)
-		    p->next = l->next;
-		else
-		    watch_table = l->next;
-		free (l);
-	    }
+    int i;
+    for (i = 0; i < watch_table_last; i++) {
+	if (!watch_table[i])
+	    continue;
+	if (watch_table[i]->callback == callback && watch_table[i]->fd == fd) {
+	    watch_table[i]->how &= ~how;
+	    if (!watch_table[i]->how)
+		goto do_free;
 	    return;
 	}
-	p = l;
+    }
+    return;
+  do_free:
+    free (watch_table[i]);
+    watch_table[i] = 0;
+    for (;;) {
+	if (!watch_table_last)
+	    break;
+	if (watch_table[watch_table_last - 1])
+	    break;
+	watch_table_last--;
     }
 }
 
 void remove_all_watch (void)
 {
-    struct file_des_watch *l, *p;
-    for (l = watch_table; l; l = p) {
-	p = l->next;
-	free (l);
+    int i;
+    for (i = 0; i < watch_table_last; i++) {
+	if (!watch_table[i])
+	    continue;
+	free (watch_table[i]);
+	watch_table[i] = 0;
     }
-    watch_table = 0;
+    watch_table_last = 0;
+}
+
+static int CIdle = 0;
+
+int CIsIdle (void)
+{
+    return CIdle;
 }
 
 static int run_watches (void)
 {
-    struct file_des_watch *l, *p;
-    int r, n = 0;
+    int r, n = 0, i, found_watch;
     fd_set reading, writing, error;
     FD_ZERO (&reading);
     FD_ZERO (&writing);
     FD_ZERO (&error);
     FD_SET (ConnectionNumber (CDisplay), &reading);
     n = max (n, ConnectionNumber (CDisplay));
-    for (l = watch_table; l; l = l->next) {
-	if (l->how & WATCH_READING) {
-	    FD_SET (l->fd, &reading);
-	    n = max (n, l->fd);
+    for (i = 0; i < watch_table_last; i++) {
+	if (!watch_table[i])
+	    continue;
+	if (watch_table[i]->how & WATCH_READING) {
+	    FD_SET (watch_table[i]->fd, &reading);
+	    n = max (n, watch_table[i]->fd);
 	}
-	if (l->how & WATCH_WRITING) {
-	    FD_SET (l->fd, &writing);
-	    n = max (n, l->fd);
+	if (watch_table[i]->how & WATCH_WRITING) {
+	    FD_SET (watch_table[i]->fd, &writing);
+	    n = max (n, watch_table[i]->fd);
 	}
-	if (l->how & WATCH_ERROR) {
-	    FD_SET (l->fd, &error);
-	    n = max (n, l->fd);
+	if (watch_table[i]->how & WATCH_ERROR) {
+	    FD_SET (watch_table[i]->fd, &error);
+	    n = max (n, watch_table[i]->fd);
 	}
     }
     if (!n)
@@ -983,14 +1038,42 @@ static int run_watches (void)
     r = select (n + 1, &reading, &writing, &error, 0);
     if (r <= 0)
 	return 0;
-    for (l = watch_table; l; l = p) {
-	p = l->next;	/* callback may remove list entry */
-	if (FD_ISSET (l->fd, &reading) || FD_ISSET (l->fd, &writing) || FD_ISSET (l->fd, &error))
-	    (*l->callback) (l->fd, &reading, &writing, &error, l->data);
+    if (FD_ISSET (ConnectionNumber (CDisplay), &reading)) {
+	CIdle = 0;
+	r = 1;
+    } else {
+	CIdle = 1;
+	r = 0;
     }
-    if (FD_ISSET (ConnectionNumber (CDisplay), &reading))
-	return 1;
-    return 0;
+/* callbacks could add or remove watches, so we restart the loop after each callback */
+    do {
+	found_watch = 0;
+	for (i = 0; i < watch_table_last; i++) {
+	    int fd;
+	    if (!watch_table[i])
+		continue;
+	    fd = watch_table[i]->fd;
+	    if (FD_ISSET (fd, &reading) && watch_table[i]->how & WATCH_READING) {
+		(*watch_table[i]->callback) (fd, &reading, &writing, &error, watch_table[i]->data);
+		FD_CLR (fd, &reading);
+		found_watch = 1;
+		break;
+	    }
+	    if (FD_ISSET (fd, &writing) && watch_table[i]->how & WATCH_WRITING) {
+		(*watch_table[i]->callback) (fd, &reading, &writing, &error, watch_table[i]->data);
+		FD_CLR (fd, &writing);
+		found_watch = 1;
+		break;
+	    }
+	    if (FD_ISSET (fd, &error) && watch_table[i]->how & WATCH_ERROR) {
+		(*watch_table[i]->callback) (fd, &reading, &writing, &error, watch_table[i]->data);
+		FD_CLR (fd, &error);
+		found_watch = 1;
+		break;
+	    }
+	}
+    } while (found_watch);
+    return r;
 }
 
 /* }}} fd selection */
@@ -1023,6 +1106,8 @@ extern Atom ATOM_WM_PROTOCOLS, ATOM_WM_DELETE_WINDOW, ATOM_WM_TAKE_FOCUS;
    was handled, 0 otherwise.
  */
 
+int (*global_callback) (XEvent *x) = 0;
+extern int menu_grabbed;
 
 /* xevent or cwevent or both my be passed as NULL */
 void CNextEvent (XEvent * xevent, CEvent * cwevent)
@@ -1048,6 +1133,7 @@ void CNextEvent (XEvent * xevent, CEvent * cwevent)
     static int button_repeat_count = 0;
     static Window toolhint_window = 0;
     static int toolhint_count = 0, x_toolhint, y_toolhint;
+    int next_event = 0;
 
     if (!xevent)
 	xevent = &private_xevent;
@@ -1059,25 +1145,35 @@ void CNextEvent (XEvent * xevent, CEvent * cwevent)
     while (!pop_event (xevent)) {	/* first check our own events, if none of our own coming, _then_ we look at the server */
 	if (QLength (CDisplay)) {
 	    XNextEvent (CDisplay, xevent);
+	    next_event = 1;
 	    break;
 	} else if (run_watches ()) {
 	    XNextEvent (CDisplay, xevent);
+	    next_event = 1;
 	    break;
 	}
     }
 
-/*
-   --> Waiting for an alarm signal with pause() ensures that a TickEvent is
-   sent before we continue. This prevented XNextEvent blocking, because we
-   check for TickEvents first with pop_event. I don't know if it helps with
-   XCheckIfEvent though, but I'm leaving it in anyway.
- */
-
     memset (cwevent, 0, sizeof (CEvent));
     memset (no_ident, 0, 33);
     cwevent->text = no_ident;
     cwevent->ident = no_ident;
 
+/* used for rxvt stuff */
+    if (global_callback)
+	if ((*global_callback) (xevent))
+	    return;
+
+    if (next_event && xevent->xany.type != KeyRelease) {
+	if (XFilterEvent(xevent, CGetICFocus ())) {
+	    xevent->xany.type = 0;
+	    xevent->xany.window = 0;
+	    cwevent->type = 0;
+	    cwevent->kind = 0;
+	    return;
+	}
+    }
+
     win = xevent->xany.window;
     type = xevent->type;
 
@@ -1089,7 +1185,7 @@ void CNextEvent (XEvent * xevent, CEvent * cwevent)
 	    if (button_repeat.type == ButtonRepeat)
 		if (!(button_repeat_count % (ALRM_PER_SECOND / 25)))
 		    CSendEvent (&button_repeat);
-	if (toolhint_window && option_toolhint_milliseconds) {
+	if (toolhint_window && option_toolhint_milliseconds && !menu_grabbed) {
 	    toolhint_count++;
 	    if (toolhint_count > (ALRM_PER_SECOND * option_toolhint_milliseconds) / 1000) {
 		show_toolhint (toolhint_window, x_toolhint, y_toolhint);
@@ -1100,7 +1196,10 @@ void CNextEvent (XEvent * xevent, CEvent * cwevent)
 	return;
     case AlarmEvent:
 	{
+	    static int last_root_x_return = 0, last_root_y_return = 0;
+	    static Window last_top = 0;
 	    int a;
+	    CWidget *pulled_menu;
 	    xevent->type = cwevent->type = AlarmEvent;
 	    toggle_cursor ();
 	    for (a = 0; a < 33; a++) {
@@ -1353,17 +1452,23 @@ int inbounds (int x, int y, int x1, int y1, int x2, int y2);
 /*-----------------------------------------------------------------------*/
 int eh_button (CWidget * w, XEvent * xevent, CEvent * cwevent)
 {
+    static Window last_win = 0;
     switch (xevent->type) {
     case MotionNotify:
 	break;
     case ButtonPress:
-	w->options &= (0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT);
-	w->options |= BUTTON_PRESSED;
-	CFocus (w);
-	(*w->render) (w);
+	last_win = xevent->xbutton.window;
+	if (xevent->xbutton.button == Button1 || xevent->xbutton.button == Button2
+	    || xevent->xbutton.button == Button3) {
+	    w->options &= (0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT);
+	    w->options |= BUTTON_PRESSED;
+	    CFocus (w);
+	    (*w->render) (w);
+	}
 	break;
     case KeyPress:
-	if ((cwevent->command != CK_Enter || w->kind == C_SWITCH_WIDGET) && cwevent->key != XK_space)
+	if ((cwevent->command != CK_Enter || w->kind == C_SWITCH_WIDGET)
+	    && cwevent->key != XK_space)
 	    break;
 	w->options &= (0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT);
 	w->options |= BUTTON_PRESSED;
@@ -1377,20 +1482,24 @@ int eh_button (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	(*w->render) (w);
 	break;
     case ButtonRelease:
-	w->options &= (0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT);
-	w->options |= BUTTON_HIGHLIGHT;
-	if (inbounds (xevent->xbutton.x, xevent->xbutton.y, 0, 0, w->width, w->height)) {
-	    if (w->kind == C_SWITCH_WIDGET)
-		toggle_radio_button (w);
-	    cwevent->ident = w->ident;	/* return the event */
+	last_win = 0;
+	if (xevent->xbutton.button == Button1 || xevent->xbutton.button == Button2
+	    || xevent->xbutton.button == Button3) {
+	    w->options &= (0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT);
+	    w->options |= BUTTON_HIGHLIGHT;
+	    if (inbounds (xevent->xbutton.x, xevent->xbutton.y, 0, 0, w->width, w->height)) {
+		if (w->kind == C_SWITCH_WIDGET)
+		    toggle_radio_button (w);
+		cwevent->ident = w->ident;	/* return the event */
+		(*w->render) (w);
+		return 1;
+	    }
 	    (*w->render) (w);
-	    return 1;
 	}
-	(*w->render) (w);
 	return 0;
     case EnterNotify:
-	w->options &= (0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT);
-	w->options |= BUTTON_HIGHLIGHT;
+	w->options &= ~(BUTTON_PRESSED | BUTTON_HIGHLIGHT);
+	w->options |= BUTTON_HIGHLIGHT | (last_win == xevent->xbutton.window ? BUTTON_PRESSED : 0);
 	(*w->render) (w);
 	break;
     case Expose:
@@ -1418,94 +1527,12 @@ int eh_bitmap (CWidget * w, XEvent * xevent, CEvent * cwevent)
     return 0;
 }
 
-
+extern struct look *look;
 
 /*-----------------------------------------------------------------------*/
 int eh_window (CWidget * w, XEvent * xevent, CEvent * cwevent)
 {
-    static int windowx, windowy;
-    static int wx = 0, wy = 0;
-    static int wwidth = 0, wheight = 0;
-    static int allowwindowmove = 0;
-    static int allowwindowresize = 0;
-
-    switch (xevent->type) {
-    case ClientMessage:
-	if (!w->disabled)
-	    cwevent->ident = w->ident;
-	break;
-    case Expose:
-	if (!xevent->xexpose.count)
-	    render_window (w);
-	break;
-    case ButtonRelease:
-	strcpy (cwevent->ident, w->ident);
-	window_is_resizing = 0;
-	resolve_button (xevent, cwevent);
-	allowwindowmove = 0;
-	allowwindowresize = 0;
-	break;
-    case ButtonPress:
-	strcpy (cwevent->ident, w->ident);
-	resolve_button (xevent, cwevent);
-	if (cwevent->double_click == 1) {
-	    CWidget *c = CChildFocus (w);
-	    if (c)
-		CFocus (c);
-	}
-	if (cwevent->button == Button1 && !(w->position & WINDOW_ALWAYS_LOWERED)) {
-	    XRaiseWindow (CDisplay, w->winid);
-	    CRaiseWindows ();
-	} else if (cwevent->button == Button2 && !(w->position & WINDOW_ALWAYS_RAISED)) {
-	    XLowerWindow (CDisplay, w->winid);
-	    CLowerWindows ();
-	}
-	windowx = xevent->xbutton.x_root - w->x;
-	windowy = xevent->xbutton.y_root - w->y;
-	wx = xevent->xbutton.x;
-	wy = xevent->xbutton.y;
-	wwidth = w->width;
-	wheight = w->height;
-	if (wx + wy > w->width + w->height - 33 && w->position & WINDOW_RESIZABLE)
-	    allowwindowresize = 1;
-	else
-	    allowwindowmove = 1;
-	break;
-    case MotionNotify:
-	resolve_button (xevent, cwevent);
-	if (!(w->position & WINDOW_UNMOVEABLE) && allowwindowmove && (cwevent->state & (Button1Mask | Button2Mask))) {
-	    w->x = xevent->xmotion.x_root - windowx;
-	    w->y = xevent->xmotion.y_root - windowy;
-	    if (w->x + xevent->xmotion.x < 2)
-		w->x = -wx + 2;
-	    if (w->y + xevent->xmotion.y < 2)
-		w->y = -wy + 2;
-	    XMoveWindow (CDisplay, w->winid, w->x, w->y);
-	}
-	if ((w->position & WINDOW_RESIZABLE) && allowwindowresize && (cwevent->state & (Button1Mask | Button2Mask))) {
-	    int wi, he;
-	    window_is_resizing = w->winid;
-	    wi = wwidth + xevent->xmotion.x_root - windowx - w->x;
-	    he = wheight + xevent->xmotion.y_root - windowy - w->y;
-
-/* this is actually for the edit windows, and needs to be generalized */
-	    if (wi < w->mark1)
-		wi = w->mark1;
-	    if (he < w->mark2)
-		he = w->mark2;
-
-	    wi -= w->firstcolumn;
-	    wi -= wi % w->textlength;
-	    wi += w->firstcolumn;
-	    he -= w->firstline;
-	    he -= he % w->numlines;
-	    he += w->firstline;
-	    w->position &= ~WINDOW_MAXIMISED;
-	    CSetSize (w, wi, he);
-	}
-	break;
-    }
-    return 0;
+    return (*look->window_handler) (w, xevent, cwevent);
 }
 
 /*-----------------------------------------------------------------------*/
@@ -1601,6 +1628,7 @@ int eh_bwimage (CWidget * w, XEvent * xevent, CEvent * cwevent)
 extern int eh_textbox (CWidget * w, XEvent * xevent, CEvent * cwevent);
 extern int eh_textinput (CWidget * w, XEvent * xevent, CEvent * cwevent);
 extern int eh_scrollbar (CWidget * w, XEvent * xevent, CEvent * cwevent);
+extern int eh_unicode (CWidget * w, XEvent * xevent, CEvent * cwevent);
 
 int (*default_event_handler (int i)) (CWidget *, XEvent *, CEvent *) {
     switch (i) {
@@ -1639,6 +1667,8 @@ int (*default_event_handler (int i)) (CWidget *, XEvent *, CEvent *) {
 #endif
     case C_EDITOR_WIDGET:
 	return eh_editor;
+    case C_UNICODE_WIDGET:
+	return eh_unicode;
     }
     return (int (*) (CWidget *, XEvent *, CEvent *)) 0;
 }
