@@ -1,5 +1,5 @@
 /* coolwidget.c - routines for simple widgets. Widget setup and destruction
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,6 +25,8 @@
 
 #include "mad.h"
 
+extern struct look *look;
+
 /* call this for fatal errors */
 void CError (const char *fmt,...)
 {
@@ -96,18 +98,23 @@ int allocate_color (char *color)
 struct cursor_state {
     int x, y, h, w;
     Window window;
+    GC gc;
+    XFontSet font_set;
+    XFontStruct *font_struct;
     int state;
     int type;
-    char chr;
+    wchar_t chr;
     unsigned long bg, fg;
+    int style;
+    int font_x, font_y;
 };
 
 struct cursor_state CursorState =
-{0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
 
 void render_cursor (struct cursor_state c);
 
-void set_cursor_position (Window win, int x, int y, int w, int h, int type, int chr, unsigned long bg, unsigned long fg)
+void set_cursor_position (Window win, int x, int y, int w, int h, int type, wchar_t chr, unsigned long bg, unsigned long fg, int style)
 {
     if (win == CGetFocus ()) {
 	CursorState.x = x;
@@ -115,10 +122,16 @@ void set_cursor_position (Window win, int x, int y, int w, int h, int type, int
 	CursorState.h = h;
 	CursorState.w = w;
 	CursorState.window = win;
+	CursorState.gc = CGC;
+	CursorState.font_set = current_font->font_set;
+	CursorState.font_struct = current_font->font_struct;
 	CursorState.type = type;
 	CursorState.chr = chr;
+	CursorState.style = style;
 	CursorState.bg = bg;
 	CursorState.fg = fg;
+	CursorState.font_x = FONT_OFFSET_X;
+	CursorState.font_y = FONT_OFFSET_Y;
 	render_cursor (CursorState);
     } else {
 	if (!(win | h | w))
@@ -137,42 +150,40 @@ void render_cursor (struct cursor_state c)
     if (c.type == CURSOR_TYPE_EDITOR) {
 	if (CursorState.window != CGetFocus ())
 	    return;
+	CPushFont ("editor", 0);
 	if (!option_xor_cursor) {
 	    if (c.state || !option_flashing_cursor)
 		CSetColor (option_cursor_color);
 	    else
 		CSetColor (c.bg);
-	    CLine (c.window, c.x, c.y + FONT_OVERHEAD, c.x, c.y + c.h - 1);
-	    CLine (c.window, c.x + 1, c.y + FONT_OVERHEAD, c.x + 1, c.y + c.h - 1);
-	    CLine (c.window, c.x + 2, c.y + FONT_OVERHEAD, c.x + c.w - 1, c.y + FONT_OVERHEAD);
-	    CLine (c.window, c.x + 2, c.y + FONT_OVERHEAD + 1, c.x + c.w - 1, c.y + FONT_OVERHEAD + 1);
+	    if (c.style & MOD_REVERSE) {
+		CLine (c.window, c.x + c.w - 1, c.y + FONT_OVERHEAD, c.x + c.w - 1, c.y + c.h - 1);
+		CLine (c.window, c.x + c.w - 2, c.y + FONT_OVERHEAD, c.x + c.w - 2, c.y + c.h - 1);
+	    } else {
+		CLine (c.window, c.x, c.y + FONT_OVERHEAD, c.x, c.y + c.h - 1);
+		CLine (c.window, c.x + 1, c.y + FONT_OVERHEAD, c.x + 1, c.y + c.h - 1);
+	    }
+	    CLine (c.window, c.x, c.y + FONT_OVERHEAD, c.x + c.w - 1, c.y + FONT_OVERHEAD);
+	    CLine (c.window, c.x, c.y + FONT_OVERHEAD + 1, c.x + c.w - 1, c.y + FONT_OVERHEAD + 1);
 	}
 	if (!c.state && option_flashing_cursor) {
-	    CSetBackgroundColor (c.bg);
-	    CSetColor (c.fg);
-	    XDrawImageString (CDisplay, c.window, CGC, c.x + FONT_OFFSET_X, c.y + FONT_OFFSET_Y, &(c.chr), 1);
+	    XSetBackground (CDisplay, c.gc, c.bg);
+	    XSetForeground (CDisplay, c.gc, c.fg);
+	    CImageTextWC (c.window, c.x + c.font_x, c.y + c.font_y, 0, &(c.chr), 1);
 	} else {
 	    if (option_xor_cursor) {
-		CSetBackgroundColor (c.fg);
-		CSetColor (c.bg);
-		XDrawImageString (CDisplay, c.window, CGC, c.x + FONT_OFFSET_X, c.y + FONT_OFFSET_Y, &(c.chr), 1);
+		XSetBackground (CDisplay, c.gc, c.fg);
+		XSetForeground (CDisplay, c.gc, c.bg);
+		CImageTextWC (c.window, c.x + c.font_x, c.y + c.font_y, 0, &(c.chr), 1);
 	    }
 	}
+	CPopFont ();
     } else {
 	if (CursorState.window != CGetFocus ()) {
 	    CSetColor (COLOR_FLAT);
 	    CLine (c.window, c.x, c.y, c.x, c.y + c.h - 6);
-	    return;
 	} else {
-	    if (CursorState.state) {
-		CSetColor (COLOR_FLAT);
-		CLine (c.window, c.x, c.y, c.x, c.y + c.h - 6);
-		render_bevel (c.window, c.x - 1, c.y - 1, c.x + 1, c.y + c.h - 5, 1, 0);
-	    } else {
-		CSetColor (COLOR_FLAT);
-		CLine (c.window, c.x, c.y, c.x, c.y + c.h - 6);		/*cursor bar flat */
-		render_bevel (c.window, c.x - 1, c.y - 1, c.x + 1, c.y + c.h - 5, 1, -1);	/*cursor bar bevel */
-	    }
+	    render_bevel (c.window, c.x - 1, c.y - 1, c.x, c.y + c.h - 5, 1, CursorState.state ? 0 : 1);
 	}
     }
 }
@@ -184,18 +195,24 @@ void CSetCursorColor (unsigned long c)
 }
 
 /* this is called from CNextEvent if an alarm event comes */
-void toggle_cursor ()
+void toggle_cursor (void)
 {
     CursorState.state = 1 - CursorState.state;
     render_cursor (CursorState);
 }
 
-void set_cursor_visible ()
+void set_cursor_visible (void)
 {
     CursorState.state = 1;
     render_cursor (CursorState);
 }
 
+void set_cursor_invisible (void)
+{
+    CursorState.state = 0;
+    render_cursor (CursorState);
+}
+
 
 
 /*
@@ -438,6 +455,12 @@ CWidget *CSetupWidget (const char *identifier, Window parent, int x, int y,
     switch ((*w)->kind) {
     case C_WINDOW_WIDGET:	/* window widgets must only be mapped 
 				   once all there children are drawn */
+#ifdef USE_XIM
+	if (CIM) {
+	    create_input_context (*w, get_input_style ());
+	    set_status_position (*w);
+	}
+#endif
 	break;
     default:
 	XMapWindow (CDisplay, newwin);	/* shows the window */
@@ -551,11 +574,7 @@ Window CDrawHeadedDialog (const char *identifier, Window parent, int x, int y, c
 	if (!CFirstWindow) {	/* create the GC the first time round and
 				   CFirstWindow. when the window is closed,
 				   the app gets a QuitApplication event */
-	    XGCValues gcv;
 	    CFirstWindow = win;
-	    gcv.font = CFontStruct->fid;
-	    gcv.foreground = COLOR_BLACK;
-	    CGC = XCreateGC (CDisplay, win, (GCFont | GCForeground), &gcv);
 
 /* these options tell CSetWindowSizeHints() to give those hints to the WM: */
 	    if (bitmask & (XValue | YValue))
@@ -577,8 +596,10 @@ Window CDrawHeadedDialog (const char *identifier, Window parent, int x, int y, c
 	win = CDrawDialog (identifier, parent, x, y);
 	(CDrawText (catstrs (identifier, ".header", 0), win, WIDGET_SPACING, WIDGET_SPACING + 2, label))->position |= POSITION_CENTRE;
 	CGetHintPos (&x, &y);
+#ifndef NEXT_LOOK_X
 	(CDrawBar (win, WIDGET_SPACING, y, 10))->position |= POSITION_FILL;
 	CGetHintPos (&x, &y);
+#endif
 	reset_hint_pos (WIDGET_SPACING + 2, y);
     }
     return win;
@@ -614,34 +635,6 @@ Window CGetWMWindow (Window win)
     return 0;
 }
 
-#if 0
-
-/* I tested this procedure with kwm, fvwm95, fvwm, twm, olwm, without problems */
-void CResizeWMWindow (char *ident, int width, int height)
-{
-    int width_old, height_old, d;
-    int wm_width, wm_height;
-    CWidget *w;
-    Window root, wm_win;
-
-    wm_win = CGetWMWindow (ident);
-    if (!wm_win)
-	return;
-    w = CIdent (ident);
-
-    width_old = w->width;
-    height_old = w->height;
-
-    configure_children (w, width, height);
-    XResizeWindow (CDisplay, w->winid, width, height);
-
-    XGetGeometry (CDisplay, w->winid, &root, &d, &d, (unsigned int *) &wm_width, (unsigned int *) &wm_height, (unsigned int *) &d, (unsigned int *) &d);
-    XResizeWindow (CDisplay, wm_win, wm_width - width_old + width, wm_height - height_old + height);
-    XFlush (CDisplay);
-}
-
-#endif
-
 /* I tested this procedure with kwm, fvwm95, fvwm, twm, olwm, without problems */
 void CRaiseWMWindow (char *ident)
 {
@@ -731,19 +724,26 @@ CWidget *CDrawButton (const char *identifier, Window parent, int x, int y,
 {
     CWidget *wdt;
     int w, h;
+    CPushFont ("widget", 0);
     if (width == AUTO_WIDTH || height == AUTO_HEIGHT)
 	CTextSize (&w, &h, label);
     if (width == AUTO_WIDTH)
 	width = w + 4 + BUTTON_RELIEF * 2;
-    if (height == AUTO_HEIGHT)
+    if (height == AUTO_HEIGHT) {
 	height = h + 4 + BUTTON_RELIEF * 2;
+#ifdef NEXT_LOOK_X
+	height++ ;
+#endif	
+    }	
     wdt = CSetupWidget (identifier, parent, x, y,
 	    width, height, C_BUTTON_WIDGET, INPUT_BUTTON, COLOR_FLAT, 1);
-    wdt->label = (char *) strdup (label);
+    if (label)
+	wdt->label = (char *) strdup (label);
     wdt->hotkey = find_hotkey (wdt);
     wdt->render = render_button;
     wdt->options |= WIDGET_TAKES_FOCUS_RING | WIDGET_HOTKEY_ACTIVATES;
     set_hint_pos (x + width + WIDGET_SPACING, y + height + WIDGET_SPACING);
+    CPopFont ();
     return wdt;
 }
 
@@ -788,7 +788,7 @@ void CTextSize (int *w, int *h, const char *str)
 	if (!(p = strchr (q, '\n')))
 	    p = q + strlen (q);
 	*h += FONT_PIX_PER_LINE;
-	*w = max (XTextWidth (CFontStruct, q, (unsigned long) p - (unsigned long) q), *w);
+	*w = max (CImageTextWidth (q, (unsigned long) p - (unsigned long) q), *w);
 	if (!*p)
 	    break;
 	q = p + 1;
@@ -807,6 +807,7 @@ CWidget *CDrawText (const char *identifier, Window parent, int x, int y, const c
     str = vsprintf_alloc (fmt, pa);
     va_end (pa);
 
+    CPushFont ("widget", 0);
     CTextSize (&w, &h, str);
     w += TEXT_RELIEF * 2 + 2;
     h += TEXT_RELIEF * 2 + 2;
@@ -815,6 +816,7 @@ CWidget *CDrawText (const char *identifier, Window parent, int x, int y, const c
     wdt->text = (char *) strdup (str);
     free (str);
     set_hint_pos (x + w + WIDGET_SPACING, y + h + WIDGET_SPACING);
+    CPopFont ();
     return wdt;
 }
 
@@ -875,7 +877,7 @@ int free_single_widget (int i)
 		if (CIndex (i)->winid == XGetSelectionOwner (CDisplay, XA_PRIMARY))
 		    XSetSelectionOwner (CDisplay, XA_PRIMARY, CFirstWindow, CurrentTime);
 	    if (CursorState.window == CIndex (i)->winid)
-		set_cursor_position (0, 0, 0, 0, 0, 0, 0, 0, 0);
+		set_cursor_position (0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
 	    XUnmapWindow (CDisplay, CIndex (i)->winid);
 	    XDestroyWindow (CDisplay, CIndex (i)->winid);
 	    if (CFirstWindow == CIndex (i)->winid)
@@ -905,6 +907,10 @@ int free_single_widget (int i)
 	    XFreePixmap (CDisplay, CIndex (i)->pixmap);
 	    CIndex (i)->pixmap = 0;
 	}
+	if (CIndex (i)->pixmap_mask) {
+	    XFreePixmap (CDisplay, CIndex (i)->pixmap_mask);
+	    CIndex (i)->pixmap_mask = 0;
+	}
 	if (CIndex (i)->graphic)
 	    free (CIndex (i)->graphic);
 	if (CIndex (i)->tab)
@@ -952,6 +958,18 @@ int find_last_child_of (Window win)
     return 0;
 }
 
+/* int for_all_widgets (int (callback *) (CWidget *, void *, void *), void *data1, void *data2) */
+long for_all_widgets (void *call_back, void *data1, void *data2)
+{
+    long (*callback) (CWidget *, void *, void *) = (long (*)(CWidget *, void *, void *)) call_back;
+    int i = last_widget;
+    while (--i > 0)
+	if (CIndex (i) != NULL)
+	    if ((*callback) (CIndex (i), data1, data2))
+		return 1;
+    return 0;
+}
+
 int widget_of_window (Window win)
 {
     int i = 0;
@@ -984,6 +1002,20 @@ int find_previous_child_of (Window win, Window child)
     return 0;
 }
 
+CWidget *CDialogOfWindow (Window window)
+{
+    for (;;) {
+	CWidget *w;
+	w = CWidgetOfWindow (window);
+	if (!w)
+	    break;
+	if (w->kind == C_WINDOW_WIDGET)
+	    return w;
+	window = w->parentid;
+    }
+    return 0;
+}
+
 Window CFindParentMainWindow (Window parent)
 {
     int i;
@@ -1046,9 +1078,8 @@ void CShutdown (void)
     edit_replace_cmd (0, 0);
     edit_search_cmd (0, 0);
     free_selections ();
-    XFreeGC (CDisplay, CGC);
-    XUnloadFont (CDisplay, CFontStruct->fid);
     mouse_shut ();
+    CFreeAllFonts ();
     XCloseDisplay (CDisplay);
 }
 
@@ -1058,16 +1089,9 @@ void drawstring_xy (Window win, int x, int y, const char *text)
 	return;
     if (!*text)
 	return;
-    XDrawString (CDisplay, win,
-		 CGC,
-		 FONT_OFFSET_X + x,
-		 FONT_OFFSET_Y + y,
-		 text,
-		 strlen (text));
+    CImageString (win, FONT_OFFSET_X + x, FONT_OFFSET_Y + y, text);
 }
 
-extern unsigned char per_char[256];
-extern unsigned char per_char_descent[256];
 
 char *whereis_hotchar (const char *labl, int hotkey)
 {
@@ -1090,11 +1114,9 @@ void underline_hotkey (Window win, int x, int y, const char *text, int hotkey)
 	return;
     if (!(p = whereis_hotchar (text, hotkey)))
 	return;
-    x += XTextWidth (CFontStruct, text, (unsigned long) p - (unsigned long) text);
-    y += FONT_BASE_LINE + per_char_descent[hotkey] + 1;
-    CLine (win, x, y, x + per_char[hotkey], y);
-    CLine (win, x - 1, y + 1, x + per_char[hotkey] / 2, y + 1);
-    CLine (win, x - 1, y + 2, x + per_char[hotkey] / 4 - 1, y + 2);
+    x += CImageTextWidth (text, (unsigned long) p - (unsigned long) text);
+    y += FONT_BASE_LINE + FONT_PER_CHAR_DESCENT(hotkey) + 1;
+    (*look->draw_hotkey_understroke) (win, x, y, hotkey);
 }
 
 void drawstring_xy_hotkey (Window win, int x, int y, const char *text, int hotkey)
@@ -1105,46 +1127,30 @@ void drawstring_xy_hotkey (Window win, int x, int y, const char *text, int hotke
 
 void render_button (CWidget * wdt)
 {
-    int w = wdt->width, h = wdt->height;
-    int x = 0, y = 0;
-
-    Window win = wdt->winid;
-
-    if (wdt->disabled)
-	goto disabled;
-    if (wdt->options & BUTTON_PRESSED) {
-	render_bevel (win, x, y, x + w - 1, y + h - 1, 2, 1);
-    } else if (wdt->options & BUTTON_HIGHLIGHT) {
-	CSetColor (COLOR_FLAT);
-	XDrawRectangle (CDisplay, win, CGC, x + 1, y + 1, w - 3, h - 3);
-	render_bevel (win, x, y, x + w - 1, y + h - 1, 1, 0);
-    } else {
-      disabled:
-	render_bevel (win, x, y, x + w - 1, y + h - 1, 2, 0);
-    }
-
-    if (!wdt->label)
-	return;
-    if (!(*(wdt->label)))
-	return;
-    CSetColor (COLOR_BLACK);
-    drawstring_xy_hotkey (win, x + 2 + BUTTON_RELIEF, y + 2 + BUTTON_RELIEF, wdt->label, wdt->hotkey);
+    (*look->render_button) (wdt);
 }
 
 void render_bar (CWidget * wdt)
 {
-    int w = wdt->width, h = wdt->height;
-
-    Window win = wdt->winid;
-
-    CSetColor (COLOR_FLAT);
-    CLine (win, 1, 1, w - 2, 1);
-    render_bevel (win, 0, 0, w - 1, h - 1, 1, 1);
+    (*look->render_bar) (wdt);
 }
 
+#if 0
+
+#ifdef NEXT_LOOK_X
+#define FB (TEXT_RELIEF)
+#else
+#define FB (TEXT_RELIEF + 1)
+#endif
+#define FS (TEXT_RELIEF + 1)
+
+#else
+
 #define FB (TEXT_RELIEF + 1)
 #define FS (TEXT_RELIEF + 1)
 
+#endif
+
 /* this is a zero flicker routine */
 void render_status (CWidget * wdt, int expose)
 {
@@ -1157,13 +1163,14 @@ void render_status (CWidget * wdt, int expose)
     int w = CWidthOf (wdt);
     int l, x, x1 = 0, color = 0;
     char *p;
+    CPushFont ("widget", 0);
     q = wdt->text;
     p = lasttext;
     x = TEXT_RELIEF + 1;	/* bevel is 1 */
     if (lastwin == win && !expose) {
 	for (; *p && *q && *p == *q; p++, q++) {
 	    if (*q >= ' ') {
-		x += XTextWidth (CFontStruct, q, 1);
+		x += CImageTextWidth (q, 1);
 	    } else {
 		if (*q == '\034') {
 		    x1 = x;
@@ -1177,13 +1184,13 @@ void render_status (CWidget * wdt, int expose)
     }
     for (l = x, r = q; *r; r++)
 	if (*r >= ' ')
-	    l += XTextWidth (CFontStruct, r, 1);
+	    l += CImageTextWidth (r, 1);
 	else if (*r == '\035')
 	    l += FS;
     if (lastwin == win && !expose) {
 	for (last_width = x, r = p; *r; r++)
 	    if (*r >= ' ')
-		last_width += XTextWidth (CFontStruct, r, 1);
+		last_width += CImageTextWidth (r, 1);
 	    else if (*r == '\035')
 		last_width += FS;
     }
@@ -1195,14 +1202,24 @@ void render_status (CWidget * wdt, int expose)
     CSetBackgroundColor (COLOR_FLAT);
     for (p = q;; p++) {
 	if (*p < ' ') {
-	    XDrawImageString (CDisplay, win, CGC, FONT_OFFSET_X + x, FONT_OFFSET_Y + TEXT_RELIEF + 1, q, (unsigned long) p - (unsigned long) q);
-	    x += XTextWidth (CFontStruct, q, (unsigned long) p - (unsigned long) q);
+	    CImageText (win, FONT_OFFSET_X + x, FONT_OFFSET_Y + TEXT_RELIEF + 1, q, (unsigned long) p - (unsigned long) q);
+	    x += CImageTextWidth (q, (unsigned long) p - (unsigned long) q);
 	    if (*p == '\035') {
 		XClearArea (CDisplay, win, x, TEXT_RELIEF + 1, x + FS, FONT_PIX_PER_LINE, 0);
 		if (x - x1 + FB + FB - 2 > 0) {
 		    render_bevel (win, x1 - FB, 0, x + FB - 1, h - 1, 1, 1);
+#if 0
+#ifdef NEXT_LOOK_X
+		    XClearArea (CDisplay, win, x1 - FB + 2, 2, x - x1 + FB + FB - 3, TEXT_RELIEF, 0);
+		    XClearArea (CDisplay, win, x1 - FB + 2, h - TEXT_RELIEF - 1, x - x1 + FB + FB - 3, TEXT_RELIEF, 0);
+#else
 		    XClearArea (CDisplay, win, x1 - FB + 1, 1, x - x1 + FB + FB - 2, TEXT_RELIEF, 0);
 		    XClearArea (CDisplay, win, x1 - FB + 1, h - TEXT_RELIEF - 1, x - x1 + FB + FB - 2, TEXT_RELIEF, 0);
+#endif
+#else
+		    XClearArea (CDisplay, win, x1 - FB + 1, 1, x - x1 + FB + FB - 2, TEXT_RELIEF, 0);
+		    XClearArea (CDisplay, win, x1 - FB + 1, h - TEXT_RELIEF - 1, x - x1 + FB + FB - 2, TEXT_RELIEF, 0);
+#endif
 		}
 		x1 = x;
 		x += FS;
@@ -1221,79 +1238,18 @@ void render_status (CWidget * wdt, int expose)
     }
     lastwin = win;
     strncpy (lasttext, wdt->text, 1023);
+    CPopFont ();
+    return;
 }
 
 void render_text (CWidget * wdt)
 {
-    Window win = wdt->winid;
-    char text[1024], *p, *q;
-    int hot, y, w = wdt->width, center = 0;
-
-    CSetColor (COLOR_FLAT);
-    CRectangle (win, 1, 1, w - 2, wdt->height - 2);
-    CSetColor (COLOR_BLACK);
-
-    hot = wdt->hotkey;		/* a letter that needs underlining */
-    y = 1;			/* bevel */
-    q = wdt->text;
-
-    for (;;) {
-	p = strchr (q, '\n');
-	if (!p) {	/* last line */
-	    if (wdt->options & TEXT_CENTRED)
-		center = (wdt->width - (TEXT_RELIEF + 1) * 2 - XTextWidth (CFontStruct, q, strlen (q))) / 2;
-	    drawstring_xy_hotkey (win, TEXT_RELIEF + 1 + center,
-		 TEXT_RELIEF + y, q, hot);
-	    break;
-	} else {
-	    int l;
-	    l = min (1023, (unsigned long) p - (unsigned long) q);
-	    memcpy (text, q, l);
-	    text[l] = 0;
-	    if (wdt->options & TEXT_CENTRED)
-		center = (wdt->width - (TEXT_RELIEF + 1) * 2 - XTextWidth (CFontStruct, q, l)) / 2;
-	    drawstring_xy_hotkey (win, TEXT_RELIEF + 1 + center,
-		 TEXT_RELIEF + y, text, hot);
-	}
-	y += FONT_PIX_PER_LINE;
-	hot = 0;	/* only for first line */
-	q = p + 1;	/* next line */
-    }
-    render_bevel (win, 0, 0, w - 1, wdt->height - 1, 1, 1);
+    (*look->render_text) (wdt);
 }
 
 void render_window (CWidget * wdt)
 {
-    int w = wdt->width, h = wdt->height;
-
-    Window win = wdt->winid;
-
-    if (wdt->options & WINDOW_NO_BORDER)
-	return;
-
-    if (wdt->position & WINDOW_RESIZABLE) {
-	CSetColor (color_widget (13));
-	CLine (win, w - 4, h - 31, w - 31, h - 4);
-	CLine (win, w - 4, h - 21, w - 21, h - 4);
-	CLine (win, w - 4, h - 11, w - 11, h - 4);
-
-	CLine (win, w - 4, h - 32, w - 32, h - 4);
-	CLine (win, w - 4, h - 22, w - 22, h - 4);
-	CLine (win, w - 4, h - 12, w - 12, h - 4);
-
-	CSetColor (color_widget (3));
-	CLine (win, w - 4, h - 27, w - 27, h - 4);
-	CLine (win, w - 4, h - 17, w - 17, h - 4);
-	CLine (win, w - 4, h -  7, w -  7, h - 4);
-
-	CLine (win, w - 4, h - 28, w - 28, h - 4);
-	CLine (win, w - 4, h - 18, w - 18, h - 4);
-	CLine (win, w - 4, h -  8, w -  8, h - 4);
-    }
-    render_bevel (win, 0, 0, w - 1, h - 1, 2, 0);
-    if (CRoot != wdt->parentid)
-	if (win == CGetFocus ())
-	    render_bevel (win, 4, 4, w - 5, h - 5, 3, 1);
+    (*look->render_window) (wdt);
 }
 
 void render_progress (CWidget * wdt)
@@ -1315,7 +1271,6 @@ void render_progress (CWidget * wdt)
     render_bevel (win, 0, 0, w - 1, h - 1, 2, 1);
 }
 
-
 void render_sunken (CWidget * wdt)
 {
     int w = wdt->width, h = wdt->height;
@@ -1323,88 +1278,14 @@ void render_sunken (CWidget * wdt)
     render_bevel (win, 0, 0, w - 1, h - 1, 2, 1);
 }
 
-void render_sunken_bevel (Window win, int x1, int y1, int x2, int y2, int thick, int sunken)
-{
-    int i;
-
-    if ((sunken & 2)) {
-	CSetColor (COLOR_FLAT);
-	CRectangle (win, x1 + thick, y1 + thick, x2 - x1 - 2 * thick + 1, y2 - y1 - 2 * thick + 1);
-    }
-
-    i = /* thick - 1 */ 0;
-
-    CSetColor (color_widget (11));
-    CLine (win, x1 + i, y2 - i, x2 - 1 - i, y2 - i);
-    CLine (win, x2 - i, y1 + i, x2 - i, y2 - i - 1);
-
-    CSetColor (color_widget (7));
-    CLine (win, x1, y1, x1, y2 - 1);
-    CLine (win, x1, y1, x2 - 1, y1);
-
-    if (thick > 1) {
-	CSetColor (color_widget (4));
-	for (i = 1; i < thick; i++) {
-	    CLine (win, x1 + i + 1, y1 + i, x2 - 1 - i, y1 + i);
-	    CLine (win, x1 + i, y1 + i, x1 + i, y2 - 1 - i);
-	}
-	CSetColor (color_widget (13));
-	for (i = 1; i < thick; i++) {
-	    CLine (win, x2 - i, y1 + i, x2 - i, y2 - i - 1);
-	    CLine (win, x1 + i, y2 - i, x2 - i - 1, y2 - i);
-	}
-    }
-    CSetColor (color_widget (14));
-    for (i = 0; i < thick; i++)
-	XDrawPoint (CDisplay, win, CGC, x2 - i, y2 - i);
-}
-
-
-void render_raised_bevel (Window win, int x1, int y1, int x2, int y2, int thick, int sunken)
-{
-    int i;
-
-    if ((sunken & 2)) {
-	CSetColor (COLOR_FLAT);
-	CRectangle (win, x1 + thick, y1 + thick, x2 - x1 - 2 * thick + 1, y2 - y1 - 2 * thick + 1);
-    }
-
-    i = thick - 1;
-
-    CSetColor (color_widget (7));
-    CLine (win, x1 + i, y2 - i, x2 - i - 1, y2 - i);
-    CLine (win, x2 - i, y1 + i, x2 - i, y2 - i);
-
-    CSetColor (color_widget (12));
-    CLine (win, x1 + i, y1 + i, x1 + i, y2 - 1 - i);
-    CLine (win, x1 + 1 + i, y1 + i, x2 - 1 - i, y1 + i);
-
-    if (thick > 1) {
-	CSetColor (color_widget (11));
-	for (i = 0; i < thick - 1; i++) {
-	    CLine (win, x1 + i + 1, y1 + i, x2 - 1 - i, y1 + i);
-	    CLine (win, x1 + i, y1 + i + 1, x1 + i, y2 - 1 - i);
-	}
-	CSetColor (color_widget (4));
-	for (i = 0; i < thick - 1; i++) {
-	    CLine (win, x2 - i, y1 + i, x2 - i, y2 - i);
-	    CLine (win, x1 + i, y2 - i, x2 - i - 1, y2 - i);
-	}
-    }
-    CSetColor (color_widget (15));
-    for (i = 0; i < thick; i++)
-	XDrawPoint (CDisplay, win, CGC, x1 + i, y1 + i);
-}
-
-
 void render_bevel (Window win, int x1, int y1, int x2, int y2, int thick, int sunken)
 {
     if (option_low_bandwidth)
 	return;
     if (sunken & 1)
-	render_sunken_bevel (win, x1, y1, x2, y2, thick, sunken);
-    else 
-	render_raised_bevel (win, x1, y1, x2, y2, thick, sunken);
+	(*look->render_sunken_bevel) (win, x1, y1, x2, y2, thick, sunken);
+    else
+	(*look->render_raised_bevel) (win, x1, y1, x2, y2, thick, sunken);
     CSetColor (COLOR_BLACK);
 }
 
@@ -1453,7 +1334,7 @@ void configure_children (CWidget * wt, int w, int h)
 	    if (wdt->position & POSITION_CENTRE)
 		set_widget_position (wdt, (w - wdt->width) / 2, wdt->y);
 	    if (wdt->position & POSITION_FILL)
-		CSetSize (wdt, w - (WIDGET_SPACING + 2) - wdt->x, wdt->height);
+		CSetSize (wdt, w - (WIDGET_SPACING + WINDOW_EXTRA_SPACING) - wdt->x, wdt->height);
 	    if (wdt->position & POSITION_RIGHT)
 		set_widget_position (wdt, wdt->x + w - wt->width, wdt->y);
 	    if (wdt->position & POSITION_WIDTH)
@@ -1487,6 +1368,10 @@ void CSetSize (CWidget * wt, int w, int h)
 
     if (wt->kind == C_WINDOW_WIDGET)
 	configure_children (wt, w, h);
+#if 0
+    else if (wt->kind == C_RXVT_WIDGET)
+	rxvt_resize_window (wt->rxvt, w, h);
+#endif
 
 /* redraw right and bottom borders */
     w_min = min (wt->width, w);
@@ -1500,6 +1385,9 @@ void CSetSize (CWidget * wt, int w, int h)
     if (wt->parentid == CRoot && wt->mapped)	/* afterstep doesn't like us to change the size of a mapped main window */
 	return;
     XResizeWindow (CDisplay, wt->winid, w, h);
+#ifdef USE_XIM
+    set_status_position (wt);
+#endif
 }
 
 void CSetWidgetSize (const char *ident, int w, int h)
@@ -1531,33 +1419,16 @@ void CSetSizeHintPos (const char *ident)
     CWidget *w;
     get_hint_limits (&x, &y);
     w = CIdent (ident);
-    x += 2;
-    y += 2;
+    x += WINDOW_EXTRA_SPACING;
+    y += WINDOW_EXTRA_SPACING;
+    if (!(w->options & WINDOW_NO_BORDER))
+	y += (*look->get_window_resize_bar_thickness) ();
     XResizeWindow (CDisplay, w->winid, x, y);
     w->width = x;
     w->height = y;
     configure_children (w, x, y);
 }
 
-#if 0
-
-void CResizeWMWindowHintPos (const char *ident)
-{
-    int x, y;
-    CWidget *w;
-    get_hint_limits (&x, &y);
-    w = CIdent (ident);
-    if (w->parentid != CRoot)
-/* Not essential to translate */
-	CError (_("CResizeWMWindowHintPos(): trying to resize a non-main window"));
-    x += 2;
-    y += 2;
-    CResizeWMWindow ((char *) ident, x, y);
-    configure_children (w, x, y);
-}
-
-#endif
-
 /* for mapping a main window. other widgets are mapped when created */
 void CMapDialog (const char *ident)
 {
