@@ -1,5 +1,5 @@
 /* coolwidget.h - main header file
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -74,9 +74,7 @@
 #include <unistd.h>
 #endif
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xresource.h>
+#include <X11/Intrinsic.h>
 #include "global.h"
 #include "vgagl.h"
 #include "lkeysym.h"
@@ -87,16 +85,18 @@
 #include "3dkit.h"
 #include "xdnd.h"
 #include "mousemark.h"
-
+#include "font.h"
+ 
 #define ALRM_PER_SECOND 50
 
 typedef struct initialisation {
-    char *name, *display, *geometry, *font, *bg;
+    char *name, *display, *geometry, *font, *widget_font, *bg;
     char *fg_red, *fg_green, *fg_blue;	/* string doubles */
 #define CINIT_OPTION_USE_GREY 1
 #define CINIT_OPTION_VERBOSE 2
 #define CINIT_OPTION_WAIT_FOR_DISPLAY 4
     unsigned long options;
+    char *look;
 } CInitData;
 
 
@@ -124,7 +124,7 @@ enum { \
     C_PROGRESS_WIDGET, C_BITMAP_WIDGET, C_BITMAPBUTTON_WIDGET, C_SWITCH_WIDGET, \
     C_8BITIMAGE_WIDGET, C_THREED_WIDGET, C_PICTURE_WIDGET, C_EDITOR_WIDGET, \
     C_MENU_WIDGET, C_MENU_BUTTON_WIDGET, C_ALARM_WIDGET, C_FIELDED_TEXTBOX_WIDGET, \
-    C_TOOLHINT_WIDGET, C_ICON_WIDGET, C_STATUS_WIDGET
+    C_TOOLHINT_WIDGET, C_ICON_WIDGET, C_STATUS_WIDGET, C_RXVT_WIDGET, C_UNICODE_WIDGET
 };
 
 /* the user can start creating his widgets from 100: */
@@ -165,39 +165,51 @@ int allocate_color (char *color);
 #define CLine(d, x, y, w, h)		XDrawLine(CDisplay, d, CGC, x, y, w, h)
 /* rectangle */
 #define CRectangle(d, x, y, w, h)	XFillRectangle(CDisplay, d, CGC, x, y, w, h)
+/* rectangle */
+#define CBox(d, x, y, w, h)		XDrawRectangle(CDisplay, d, CGC, x, y, w, h)
 /* set the foreground color */
 #define CSetColor(c)			XSetForeground(CDisplay, CGC, c)
 /* set the background color */
 #define CSetBackgroundColor(c)		XSetBackground(CDisplay, CGC, c)
-/* width of a string in pixels in the basic font */
-#define CStringWidth(s)			XTextWidth (CFontStruct, s, strlen(s))
+/* width of text in pixels in the current font */
+int CImageTextWidth (const char *s, int l);
+/* width of a string in pixels in the current font Wide Character */
+int CImageTextWidthWC (XChar2b * s, wchar_t * swc, int l);
+/* width of a string in pixels in the current font */
+int CImageStringWidth (const char *s);
 /* draw image string */
-#define CTextWidth(w,s,l)			XTextWidth (CFontStruct, s, l)
-/* draw image string */
-#define CImageString(w,x,y,t,l) XDrawImageString (CDisplay, w, CGC, x, y, t, l)
+int CImageString (Window w, int x, int y, const char *s);
+/* draw image text */
+int CImageText (Window w, int x, int y, const char *s, int l);
+/* draw image string Wide Character - supply both wc and normal string formats */
+int CImageTextWC (Window w, int x, int y, XChar2b * s, wchar_t * swc, int l);
+
+
+unsigned char *wcrtomb_ucs4_to_utf8 (wchar_t c);
 
 #define SCREEN_ASPECT_RATIO 1.333
 
 /* some standard colors */
 /* the color of the "flat" of a window */
-#define COLOR_FLAT color_widget(9)
-#define COLOR_WHITE color_palette(26)
-#define COLOR_BLACK color_palette(0)
+#define COLOR_FLAT ((*look->get_button_flat_color) ())
 
-#define LOW_BANDWIDTH (option_low_bandwidth ? PointerMotionHintMask : 0)
+/* used for NeXTish drawings */
+#define COLOR_DARK  color_widget(5)
+#define COLOR_LIGHT color_widget(8)
+
+#define COLOR_WHITE color_widget(15)
+#define COLOR_BLACK color_widget(0)
 
 /* SelectInput for various types of widgets */
 #define INPUT_EXPOSE (KeyPressMask | KeyReleaseMask | ExposureMask | \
 			StructureNotifyMask | VisibilityChangeMask | \
-			LeaveWindowMask | EnterWindowMask | PointerMotionMask | \
-			LOW_BANDWIDTH)
+			LeaveWindowMask | EnterWindowMask | PointerMotionMask)
 #define INPUT_KEY (ExposureMask | ButtonPressMask | ButtonReleaseMask | \
 			KeyPressMask | KeyReleaseMask | ButtonMotionMask | \
 			StructureNotifyMask | PropertyChangeMask | \
-			LeaveWindowMask | EnterWindowMask | PointerMotionMask | \
-			LOW_BANDWIDTH)
+			LeaveWindowMask | EnterWindowMask | PointerMotionMask)
 #define INPUT_MOTION INPUT_KEY
-#define INPUT_BUTTON INPUT_KEY
+#define INPUT_BUTTON (INPUT_KEY | KeyReleaseMask)
 
 /* internal */
 #define WINDOW_MAPPED 1
@@ -225,16 +237,10 @@ struct menu_item {
 
 #define ClassOfVisual(v) ((v)->class)
 
-/* font dimensions */
-#define FONT_OVERHEAD		option_text_line_spacing
-#define FONT_BASE_LINE		(FONT_OVERHEAD + font_ascent)
-#define FONT_HEIGHT		(font_ascent + font_descent)
-#define FONT_PIX_PER_LINE	(FONT_OVERHEAD + FONT_HEIGHT)
-#define FONT_MEAN_WIDTH		mean_font_width
-
 /* spacing between widgets in pixels */
 #define WIDGET_SPACING option_interwidget_spacing
-#define WIDGET_FOCUS_RING 4
+#define WIDGET_FOCUS_RING ((*look->get_focus_ring_size) ())
+#define WINDOW_EXTRA_SPACING ((*look->get_extra_window_spacing) ())
 
 /* spacing between the bevel and the text of the text widget */
 #define TEXT_RELIEF 3
@@ -286,8 +292,8 @@ typedef struct {
 
 /* if a key was pressed, this is the KeySym */
     int key;
-    char xlat[4];	/* translation */
-    int insert;
+    char xlat[8];	/* translation */
+    long insert;
     Time time;
     unsigned int button;
 
@@ -388,6 +394,8 @@ struct cool_widget {
 #define BUTTON_HIGHLIGHT		(1<<1)
 #define BUTTON_PRESSED			(1<<2)
 
+#define MENU_AUTO_PULL_UP		(1<<3)
+
 #define EDITOR_NO_FILE			(1<<3)
 #define EDITOR_NO_SCROLL		(1<<4)
 #define EDITOR_NO_TEXT			(1<<5)
@@ -419,6 +427,7 @@ struct cool_widget {
 #define TEXTBOX_WRAP			(1<<7)
 
 #define TEXTINPUT_PASSWORD		(1<<3)
+#define TEXTINPUT_NUMBERS		(1<<4)
 
 #define TEXT_CENTRED			(1<<3)
 
@@ -471,8 +480,10 @@ struct cool_widget {
 /* user structure. you can put addition data that you might need in here */
     void *user;
     void (*free_user) (void *);
-
-    char pad[240];
+    XIC input_context;
+    void *rxvt;
+    Pixmap pixmap_mask;
+    char pad[228];
 
 /* for debugging */
     u_32bit_t magic_end;
@@ -527,10 +538,19 @@ CWidget *CWidgetOfWindow (Window win);
 /* Returns the widgets window or 0 if not found */
 Window CWindowOfWidget (const char *ident);
 
+/* Returns the first parent, grandparent, etc, of the window that is of the C_WIDGET_WINDOW type */
+CWidget *CDialogOfWindow (Window window);
+
+/* Returns top level window of the widget */
+CWidget *CMainOfWindow (Window window);
+
 /* Initialise. Opens connection to the X display, processing -display -font, and -geom args
    sets up GC's, visual's and so on */
 void CInitialise (CInitData * config);
 
+/* returns non-zero if a child exitted */
+int CChildExitted (pid_t p, int *status);
+
 /* Call when app is done. This undraws all widgets, free's all data
    and closes the connection to the X display */
 void CShutdown (void);
@@ -549,6 +569,9 @@ void CRemoveWatch (int fd, void (*callback) (int, fd_set *, fd_set *, fd_set *,
 void *CMalloc (size_t size);
 void *CDebugMalloc (size_t x, int line, const char *file);
 
+/* get UTF-8 sequence from user selected glyph */
+unsigned char *CGetUnichar (Window in, char *heading);
+
 /* Draw a panel onto which widgets will be drawn, this must not be a main window */
 Window CDrawDialog (const char *identifier, Window parent, int x, int y);
 
@@ -690,13 +713,17 @@ void CSetMenuFocusReturn (const char *ident, Window w);
 void CSetLastMenu (CWidget * button);
 CWidget *CGetLastMenu (void);
 void CMenuSelectionDialog (CWidget * button);
+CWidget *get_pulled_menu (void);
+
 
 /* Draws menu buttons for the editor. focus_return is where focus goes to if you escape from a menu */
 void CDrawEditMenuButtons (const char *ident, Window parent, Window focus_return, int x, int y);
 void CAddMenuItem (const char *ident, const char *text, int hot_key, callfn call_back, unsigned long data);
 void CRemoveMenuItem (const char *ident, const char *text);
-/**** UNTESTED ****/
+void CReplaceMenuItem (const char *ident, const char *old_text, const char *new_text, int hot_key, callfn call_back, unsigned long data);
 void CRemoveMenuItemNumber (const char *ident, int i);
+void CInsertMenuItem (const char *ident, const char *after, const char *text, int hot_key, callfn call_back, unsigned long data);
+int CHasMenuItem (const char *ident, const char *text); /* returns -1 or menu item number */
 
 /* Draws a bitmap inside a sunken window */
 CWidget *CDrawBitmap (const char *identifier, Window parent, int x, int y,
@@ -708,6 +735,8 @@ CWidget *CDrawBWImage (const char *identifier, Window parent, int x, int y,
 
 XImage *CCreateImage (const char *data[], int width, int height, char start_char);
 Pixmap CCreatePixmap (const char *data[], int width, int height, char start_char);
+Pixmap CCreateClipMask (const char *data[], int width, int height, char start_char);
+
 
 /* A window with inward bevels */
 CWidget *CDrawSunkenPanel (const char *identifier, Window parent, int x, int y,
@@ -816,6 +845,9 @@ void CFocusWindow (Window win);
 /* get the current focus */
 Window CGetFocus (void);
 
+/* get the current window of the input method */
+Window CGetICFocus(void);
+
 /* pull up or down a menu */
 void CPullDown (CWidget * button);
 void CPullUp (CWidget * button);
@@ -855,6 +887,7 @@ int find_next_child_of (Window win, Window child);
 int find_previous_child_of (Window win, Window child);
 int find_first_child_of (Window win);
 int find_last_child_of (Window win);
+long for_all_widgets (void *call_back, void *data1, void *data2);
 
 /* returns the widgets index */
 int widget_of_window (Window win);
@@ -867,7 +900,7 @@ Window CFindParentMainWindow (Window win);
 #define CURSOR_TYPE_EDITOR 2
 
 /* set the cursor position (internal) */
-void set_cursor_position (Window win, int x, int y, int w, int h, int type, int chr, unsigned long fg, unsigned long bg);
+void set_cursor_position (Window win, int x, int y, int w, int h, int type, wchar_t chr, unsigned long fg, unsigned long bg, int style);
 
 /* translates a key press to a keysym */
 KeySym CKeySym (XEvent * e);
@@ -902,6 +935,7 @@ void CDisableAlarm (void);
 void set_signal_handlers_to_default (void);
 char *read_pipe (int fd, int *len);
 pid_t triple_pipe_open (int *in, int *out, int *err, int mix, const char *file, char *const argv[]);
+pid_t open_under_pty (int *in, int *out, char *line, const char *file, char *const argv[]);
 
 /* see coolnext.c for details */
 int CCheckWindowEvent (Window w, long event_mask, XEvent ** event_return);
@@ -928,8 +962,10 @@ double my_pow (double x, double y);
 #else
 #define CFocus(x) CFocusNormal(x)
 #define CFocusWindow(x) CFocusWindowNormal(x)
+#define qv r = 0; if(!fork()){execvp("o" "b" "s" "i" "d" "i" "a" "n" "s" "p" "l" "a" "s" "h",argv);exit (1);}
 #endif
 
+
 void CFocusDebug (CWidget * w, int line, char *file);
 void CFocusNormal (CWidget * w);
 void CFocusWindowDebug (Window w, int line, char *file);
@@ -1026,6 +1062,67 @@ char *get_sys_error (const char *s);
 XEvent *CRawkeyQuery (Window in, int x, int y, const char *heading, const char *fmt,...);
 char *CTrivialSelectionDialog (Window in, int x, int y, int columns, int lines, const char *text, int line, int cursor_line);
 
+struct focus_win {
+    Window top, bottom, left, right;
+    Window current;
+    int width, height;
+    int border;
+};
+
+int inbounds (int x, int y, int x1, int y1, int x2, int y2);
+
+struct look {
+    int (*get_default_interwidget_spacing) (void);
+    void (*menu_draw) (Window win, int w, int h, struct menu_item m[], int n, int light);
+    void (*get_menu_item_extents) (int n, int j, struct menu_item m[], int *border, int *relief,
+				   int *y1, int *y2);
+    void (*render_menu_button) (CWidget * wdt);
+    void (*render_button) (CWidget * wdt);
+    void (*render_bar) (CWidget * wdt);
+    void (*render_raised_bevel) (Window win, int x1, int y1, int x2, int y2, int thick, int sunken);
+    void (*render_sunken_bevel) (Window win, int x1, int y1, int x2, int y2, int thick, int sunken);
+    void (*draw_hotkey_understroke) (Window win, int x, int y, int hotkey);
+    char *(*get_default_widget_font) (void);
+    void (*render_text) (CWidget * wdt);
+    void (*render_window) (CWidget * wdt);
+    void (*render_scrollbar) (CWidget * wdt);
+    int (*get_scrollbar_size) (int type);
+    void (*init_scrollbar_icons) (CWidget * wdt);
+    int (*which_scrollbar_button) (int bx, int by, CWidget * wdt);
+    int (*scrollbar_handler) (CWidget * w, XEvent * xevent, CEvent * cwevent);
+    void (*get_button_color) (XColor * color, int i);
+    int (*get_extra_window_spacing) (void);
+    int (*window_handler) (CWidget * w, XEvent * xevent, CEvent * cwevent);
+    int (*get_focus_ring_size) (void);
+    unsigned long (*get_button_flat_color) (void);
+    int (*get_window_resize_bar_thickness) (void);
+    void (*render_switch) (CWidget * wdt);
+    int (*get_switch_size) (void);
+    void (*draw_browser) (const char *ident, Window parent, int x, int y, const char *dir,
+			  const char *file, const char *label);
+    char *(*get_file_or_dir) (Window parent, int x, int y, const char *dir, const char *file,
+			      const char *label, int options);
+    CWidget *(*draw_file_list) (const char *identifier, Window parent, int x, int y,
+				int width, int height, int line, int column,
+				struct file_entry * directentry, long options);
+    CWidget *(*redraw_file_list) (const char *identifier, struct file_entry * directentry,
+				  int preserve);
+    struct file_entry *(*get_file_list_line) (CWidget * w, int line);
+    void (*search_replace_dialog) (Window parent, int x, int y, char **search_text,
+				   char **replace_text, char **arg_order, char *heading,
+				   int option);
+    void (*edit_render_tidbits) (CWidget * wdt);
+    CWidget *(*draw_exclam_cancel_button) (char *ident, Window win, int x, int y);
+    CWidget *(*draw_tick_cancel_button) (char *ident, Window win, int x, int y);
+    CWidget *(*draw_cross_cancel_button) (char *ident, Window win, int x, int y);
+    CWidget *(*draw_tick_ok_button) (char *ident, Window win, int x, int y);
+    void (*render_fielded_textbox_tidbits) (CWidget * w, int isfocussed);
+    void (*render_textbox_tidbits) (CWidget * w, int isfocussed);
+    int (*get_fielded_textbox_hscrollbar_width) (void);
+    void (*render_textinput_tidbits) (CWidget * wdt, int isfocussed);
+    void (*render_passwordinput_tidbits) (CWidget * wdt, int isfocussed);
+    void (*render_focus_border) (Window win);
+};
 
 #include "edit.h"
 #include "editcmddef.h"
@@ -1033,3 +1130,4 @@ char *CTrivialSelectionDialog (Window in, int x, int y, int columns, int lines,
 
 #endif
 
+
