@@ -1,5 +1,5 @@
 /* dialog.c - draws various useful dialog boxes
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -26,9 +26,7 @@
 #include <unistd.h>
 #endif
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xresource.h>
+#include <X11/Intrinsic.h>
 #include "lkeysym.h"
 
 #include "stringtools.h"
@@ -45,6 +43,7 @@
 #define MID_X 20
 #define MID_Y 20
 
+extern struct look *look;
 
 /* Yuriy Elkin: (from mc/src/util.c) */
 char *get_sys_error (const char *s)
@@ -84,12 +83,20 @@ Window find_mapped_window (Window w)
 
 void CErrorDialog (Window in, int x, int y, const char *heading, const char *fmt,...)
 {
+    static int inside = 0;
     va_list pa;
     char *str;
     Window win;
     CEvent cwevent;
     CState s;
 
+    if (inside)
+        return;
+
+    inside = 1;
+
+    CPushFont ("widget", 0);
+
     va_start (pa, fmt);
     str = vsprintf_alloc (fmt, pa);
     va_end (pa);
@@ -106,7 +113,7 @@ void CErrorDialog (Window in, int x, int y, const char *heading, const char *fmt
     (CDrawText ("", win, x, y, "%s", str))->position = POSITION_CENTRE;
     free (str);
     CGetHintPos (0, &y);
-    (CDrawPixmapButton ("_clickhere", win, -50, y, PIXMAP_BUTTON_EXCLAMATION))->position = POSITION_CENTRE;
+    ((*look->draw_exclam_cancel_button) ("_clickhere", win, -50, y))->position = POSITION_CENTRE;
     CIdent ("_error")->position = WINDOW_UNMOVEABLE | WINDOW_ALWAYS_RAISED;
     CSetSizeHintPos ("_error");
     CMapDialog ("_error");
@@ -116,9 +123,13 @@ void CErrorDialog (Window in, int x, int y, const char *heading, const char *fmt
 	if (!CIdent ("_error"))
 	    break;
     } while (strcmp (cwevent.ident, "_clickhere") && cwevent.command != CK_Cancel);
+
+    CPopFont ();
     
     CDestroyWidget ("_error");
     CRestoreState (&s);
+
+    inside = 0;
 }
 
 void CMessageDialog (Window in, int x, int y, unsigned long options, const char *heading, const char *fmt,...)
@@ -129,6 +140,8 @@ void CMessageDialog (Window in, int x, int y, unsigned long options, const char
     CEvent cwevent;
     CState s;
 
+    CPushFont ("widget", 0);
+
     va_start (pa, fmt);
     str = vsprintf_alloc (fmt, pa);
     va_end (pa);
@@ -142,7 +155,7 @@ void CMessageDialog (Window in, int x, int y, unsigned long options, const char
     (CDrawText ("", win, x, y, "%s", str))->options = options;
     free (str);
     CGetHintPos (0, &y);
-    CDrawPixmapButton ("_clickhere", win, -50, y, PIXMAP_BUTTON_TICK);
+    ((*look->draw_tick_cancel_button) ("_clickhere", win, -50, y))->position = POSITION_CENTRE;
     CCentre ("_clickhere");
     CIdent ("_error")->position = WINDOW_UNMOVEABLE | WINDOW_ALWAYS_RAISED;
     CSetSizeHintPos ("_error");
@@ -153,6 +166,9 @@ void CMessageDialog (Window in, int x, int y, unsigned long options, const char
 	if (!CIdent ("_error"))
 	    break;
     } while (strcmp (cwevent.ident, "_clickhere") && cwevent.command != CK_Cancel && cwevent.command != CK_Enter);
+
+    CPopFont ();
+
     CDestroyWidget ("_error");
     CRestoreState (&s);
 }
@@ -165,9 +181,11 @@ void CTextboxMessageDialog (Window in, int x, int y, int columns, int lines, con
     CState s;
     int width, height;
 
+    CPushFont ("editor", 0);
     CTextSize (&width, &height, text);
     width = min (columns * FONT_MEAN_WIDTH, width) + 1 + 6;
     height = min (lines * FONT_PIX_PER_LINE, height) + 1 + 6;
+    CPopFont ();
 
     if (!in) {
 	x = MID_X;
@@ -181,7 +199,7 @@ void CTextboxMessageDialog (Window in, int x, int y, int columns, int lines, con
     CGetHintPos (&x, &y);
     CDrawTextbox ("_textmessbox", win, x, y, width, height, line, 0, text, 0);
     CGetHintPos (0, &y);
-    CDrawPixmapButton ("_clickhere", win, -50, y, PIXMAP_BUTTON_TICK);
+    ((*look->draw_tick_cancel_button) ("_clickhere", win, -50, y))->position = POSITION_CENTRE;
     CCentre ("_clickhere");
     CIdent ("_error")->position = WINDOW_UNMOVEABLE | WINDOW_ALWAYS_RAISED;
     CSetSizeHintPos ("_error");
@@ -217,8 +235,10 @@ int CListboxDialog (Window in, int x, int y, int columns, int lines,
     int width, height, len, i;
     char *text, *p;
 
+    CPushFont ("editor", 0);
     width = columns * FONT_MEAN_WIDTH + 1 + 6;
     height = lines * FONT_PIX_PER_LINE + 1 + 6;
+    CPopFont ();
 
     if (!in) {
 	x = MID_X;
@@ -232,13 +252,15 @@ int CListboxDialog (Window in, int x, int y, int columns, int lines,
     for (i = 0; i < num_lines; i++)
 	len += strlen ((*get_line) (data, i)) + 1;
     p = text = CMalloc (len + 1);
+    p[0] = '\0';
     for (i = 0; i < num_lines; i++) {
 	strcpy (p, (*get_line) (data, i));
 	p += strlen (p);
 	*p++ = '\n';
     }
     i = -1;
-    *(--p) = 0;
+    if ((unsigned long) p > (unsigned long) text)
+	*(--p) = 0;
     if (heading)
 	win = CDrawHeadedDialog ("_error", in, x, y, heading);
     else
@@ -247,7 +269,7 @@ int CListboxDialog (Window in, int x, int y, int columns, int lines,
     (CDrawTextbox ("_textmessbox", win, x, y, width, height, start_line, 0, text, TEXTBOX_MAN_PAGE))->cursor = cursor_line;
     CGetHintPos (0, &y);
     if (heading) {
-	CDrawPixmapButton ("_clickhere", win, -50, y, PIXMAP_BUTTON_CROSS);
+	((*look->draw_cross_cancel_button) ("_clickhere", win, -50, y))->position = POSITION_CENTRE;
 	CCentre ("_clickhere");
     }
     CIdent ("_error")->position = WINDOW_ALWAYS_RAISED;
@@ -295,8 +317,10 @@ char *CTrivialSelectionDialog (Window in, int x, int y, int columns, int lines,
     char *p = 0;
     CWidget *w;
 
+    CPushFont ("editor", 0);
     width = columns * FONT_MEAN_WIDTH + 1 + 6;
     height = lines * FONT_PIX_PER_LINE + 1 + 6;
+    CPopFont ();
 
     CBackupState (&s);
     CDisable ("*");
@@ -356,7 +380,7 @@ void CFatalErrorDialog (int x, int y, const char *fmt,...)
 	CDrawText ("fatalerror.text", win, x, y, "%s", str);
 	CCentre ("fatalerror.text");
 	CGetHintPos (0, &y);
-	CDrawPixmapButton ("clickhere", win, -50, y, PIXMAP_BUTTON_CROSS);
+	((*look->draw_cross_cancel_button) ("clickhere", win, -50, y))->position = POSITION_CENTRE;
 	CCentre ("clickhere");
 	CIdent ("fatalerror")->position = WINDOW_UNMOVEABLE | WINDOW_ALWAYS_RAISED;
 	CSetSizeHintPos ("fatalerror");
@@ -403,7 +427,7 @@ XEvent *CRawkeyQuery (Window in, int x, int y, const char *heading, const char *
     free (str);
     CDrawTextInput ("_inputdialog.input", win, x, y, (FONT_MEAN_WIDTH) * 6, AUTO_HEIGHT, 256, "");
     CGetHintPos (0, &y);
-    CDrawPixmapButton ("_inputdialog.crosshere", win, -50, y, PIXMAP_BUTTON_CROSS);
+    ((*look->draw_cross_cancel_button) ("_inputdialog.crosshere", win, -50, y))->position = POSITION_CENTRE;
     CCentre ("_inputdialog.crosshere");
     CSetSizeHintPos ("_inputdialog");
     CMapDialog ("_inputdialog");
@@ -441,7 +465,7 @@ char *CInputDialog (const char *ident, Window in, int x, int y, int min_width, c
     Window win;
     CEvent cwevent;
     CState s;
-    char inp_name[256], *b = 0;
+    char inp_name[256];
     int browse = 0, xf, yf;
 
     min_width &= ~INPUT_DIALOG_BROWSE_MASK;
@@ -476,8 +500,8 @@ char *CInputDialog (const char *ident, Window in, int x, int y, int min_width, c
 	w += (CDrawButton ("_inputdialog.browse", win, x, y, AUTO_SIZE, _ (" Browse... ")))->width + WIDGET_SPACING;
     }
     CGetHintPos (0, &y);
-    CDrawPixmapButton ("_inputdialog.clickhere", win, (w + 16) / 4 - 22, y, PIXMAP_BUTTON_TICK);
-    CDrawPixmapButton ("_inputdialog.crosshere", win, 3 * (w + 16) / 4 - 22, y, PIXMAP_BUTTON_CROSS);
+    (*look->draw_tick_ok_button) ("_inputdialog.clickhere", win, (w + 16) / 4 - 22, y);
+    (*look->draw_cross_cancel_button) ("_inputdialog.crosshere", win, 3 * (w + 16) / 4 - 22, y);
     CSetSizeHintPos ("_inputdialog");
     CMapDialog ("_inputdialog");
     CFocus (CIdent (inp_name));
@@ -490,7 +514,7 @@ char *CInputDialog (const char *ident, Window in, int x, int y, int min_width, c
 	if (cwevent.command == CK_Enter)
 	    break;
 	if (!strcmp (cwevent.ident, "_inputdialog.browse")) {
-	    char *f;
+	    char *f = 0;
 /* dialog title */
 	    switch (browse) {
 	    case INPUT_DIALOG_BROWSE_DIR:
