@@ -1,6 +1,6 @@
 /* editor low level data handling and cursor fundamentals.
 
-   Copyright (C) 1996, 1997 the Free Software Foundation
+   Copyright (C) 1996-2000 the Free Software Foundation
    
    Authors: 1996, 1997 Paul Sheer
 
@@ -73,7 +73,9 @@
    Returns '\n' if out of bounds.
  */
 
-#ifndef NO_INLINE_GETBYTE
+static int push_action_disabled = 0;
+
+#ifdef NO_INLINE_GETBYTE
 
 int edit_get_byte (WEdit * edit, long byte_index)
 {
@@ -103,6 +105,25 @@ char *edit_get_buffer_as_text (WEdit * e)
     return t;
 }
 
+
+/* Note on CRLF->LF translation: */
+
+#if MY_O_TEXT
+#error MY_O_TEXT is depreciated. CR_LF_TRANSLATION must be defined which does CR-LF translation internally. See note in source.
+#endif
+
+/* 
+   The edit_open_file (previously edit_load_file) function uses
+   init_dynamic_edit_buffers to load a file. This is unnecessary
+   since you can just as well fopen the file and insert the
+   characters one by one. The real reason for
+   init_dynamic_edit_buffers (besides allocating the buffers) is
+   as an optimisation - it uses raw block reads and inserts large
+   chunks at a time. It is hence extremely fast at loading files.
+   Where we might not want to use it is if we were doing
+   CRLF->LF translation or if we were reading from a pipe.
+ */
+
 /* Initialisation routines */
 
 /* returns 1 on error */
@@ -110,14 +131,6 @@ char *edit_get_buffer_as_text (WEdit * e)
 /* cursor set to start of file */
 int init_dynamic_edit_buffers (WEdit * edit, const char *filename, const char *text)
 {
- 
-#if defined CR_LF_TRANSLATION
-    /* Variables needed for safe handling of Translation from Microsoft CR/LF EOL to
-       Unix Style LF EOL - Franco */
-    long bytes_wanted,bytes_read,bytes_missing;
-    char *p;
-#endif
- 
     long buf;
     int j, file = -1, buf2;
 
@@ -127,9 +140,9 @@ int init_dynamic_edit_buffers (WEdit * edit, const char *filename, const char *t
     }
 
     if (filename)
-	if ((file = open ((char *) filename, O_RDONLY | MY_O_TEXT)) == -1) {
+	if ((file = open ((char *) filename, O_RDONLY)) == -1) {
 /* The file-name is printed after the ':' */
-	    edit_error_dialog (_(" Error "), get_sys_error (catstrs (_(" Failed trying to open file for reading: "), filename, " ", 0)));
+	    edit_error_dialog (_ (" Error "), get_sys_error (catstrs (_ (" Failed trying to open file for reading: "), filename, " ", 0)));
 	    return 1;
 	}
     edit->curs2 = edit->last_byte;
@@ -138,59 +151,18 @@ int init_dynamic_edit_buffers (WEdit * edit, const char *filename, const char *t
 
     edit->buffers2[buf2] = CMalloc (EDIT_BUF_SIZE);
 
-/*
-_read returns the number of bytes read, 
-which may be less than count if there are fewer than count bytes left in the file 
-or if the file was opened in text mode, 
-in which case each carriage returnlinefeed (CR-LF) pair is replaced 
-with a single linefeed character. Only the single linefeed character is counted 
-in the return value. The replacement does not affect the file pointer.
-
-_eof returns 1 if the current position is end of file, or 0 if it is not. 
-A return value of -1 indicates an error; in this case, errno is set to EBADF, 
-which indicates an invalid file handle.
-*/
-    if (filename){
-
-#if defined CR_LF_TRANSLATION
-	bytes_wanted=edit->curs2 & M_EDIT_BUF_SIZE;
-	p = (char *) edit->buffers2[buf2] + EDIT_BUF_SIZE - (edit->curs2 & M_EDIT_BUF_SIZE);
-	bytes_read = read (file, p , edit->curs2 & M_EDIT_BUF_SIZE);
-	bytes_missing = bytes_wanted - bytes_read ;
-	while(bytes_missing ){
-		p += bytes_read;
-		bytes_read = read(file,p,bytes_missing);
-		if(bytes_read <= 0) break;
-		bytes_missing -= bytes_read ;
-	}
-#else
+    if (filename) {
 	read (file, (char *) edit->buffers2[buf2] + EDIT_BUF_SIZE - (edit->curs2 & M_EDIT_BUF_SIZE), edit->curs2 & M_EDIT_BUF_SIZE);
-#endif
-    }
-    else {
+    } else {
 	memcpy (edit->buffers2[buf2] + EDIT_BUF_SIZE - (edit->curs2 & M_EDIT_BUF_SIZE), text, edit->curs2 & M_EDIT_BUF_SIZE);
 	text += edit->curs2 & M_EDIT_BUF_SIZE;
     }
 
     for (buf = buf2 - 1; buf >= 0; buf--) {
 	edit->buffers2[buf] = CMalloc (EDIT_BUF_SIZE);
- 	if (filename){
-#if defined CR_LF_TRANSLATION
-		bytes_wanted = EDIT_BUF_SIZE;
-		p = (char *) edit->buffers2[buf];
-		bytes_read = read (file, p, EDIT_BUF_SIZE);
-		bytes_missing = bytes_wanted - bytes_read ;
-		while(bytes_missing ){
-			p += bytes_read;
-			bytes_read = read(file,p,bytes_missing);
-			if(bytes_read <= 0) break;
-			bytes_missing -= bytes_read ;
-		}
-#else
+	if (filename) {
 	    read (file, (char *) edit->buffers2[buf], EDIT_BUF_SIZE);
-#endif
- 	}
-	else {
+	} else {
 	    memcpy (edit->buffers2[buf], text, EDIT_BUF_SIZE);
 	    text += EDIT_BUF_SIZE;
 	}
@@ -199,81 +171,285 @@ which indicates an invalid file handle.
     edit->curs1 = 0;
     if (file != -1)
 	close (file);
+    return 0;
+}
 
+/* detecting an error on save is easy: just check if every byte has been written. */
+/* detecting an error on read, is not so easy 'cos there is not way to tell
+   whether you read everything or not. */
+/* FIXME: add proper `triple_pipe_open' to read, write and check errors. */
+static struct edit_filters {
+    char *read, *write, *extension;
+} all_filters[] = {
+
+    {
+	"bzip2 -cd %s", "bzip2 > %s", ".bz2"
+    },
+    {
+	"gzip -cd %s", "gzip > %s", ".gz"
+    },
+    {
+	"compress -cd %s", "compress > %s", ".Z"
+    }
+};
+
+static int edit_find_filter (const char *filename)
+{
+    int i, l;
+    if (!filename)
+	return -1;
+    l = strlen (filename);
+    for (i = 0; i < sizeof (all_filters) / sizeof (struct edit_filters); i++) {
+	int e;
+	e = strlen (all_filters[i].extension);
+	if (l > e)
+	    if (!strcmp (all_filters[i].extension, filename + l - e))
+		return i;
+    }
+    return -1;
+}
+
+char *edit_get_filter (const char *filename)
+{
+    int i, l;
+    char *p;
+    i = edit_find_filter (filename);
+    if (i < 0)
+	return 0;
+    l = strlen (filename);
+    p = malloc (strlen (all_filters[i].read) + l + 2);
+    sprintf (p, all_filters[i].read, filename);
+    return p;
+}
+
+char *edit_get_write_filter (char *writename, char *filename)
+{
+    int i, l;
+    char *p;
+    i = edit_find_filter (filename);
+    if (i < 0)
+	return 0;
+    l = strlen (writename);
+    p = malloc (strlen (all_filters[i].write) + l + 2);
+    sprintf (p, all_filters[i].write, writename);
+    return p;
+}
+
+#ifdef CR_LF_TRANSLATION
+/* reads into buffer, replace \r\n with \n */
+long edit_insert_stream (WEdit * edit, FILE * f)
+{
+    int a = -1, b;
+    long i;
+    while ((b = fgetc (f)) >= 0) {
+	if (a == '\r' && b == '\n') {
+	    edit_insert (edit, '\n');
+	    i++;
+	    a = -1;
+	    continue;
+	} else if (a >= 0) {
+	    edit_insert (edit, a);
+	    i++;
+	}
+	a = b;
+    }
+    if (a >= 0)
+	edit_insert (edit, a);
+    return i;
+}
+/* writes buffer, replaces, replace \n with \r\n */
+long edit_write_stream (WEdit * edit, FILE * f)
+{
+    long i;
+    int c;
+    for (i = 0; i < edit->last_byte; i++) {
+	c = edit_get_byte (edit, i);
+	if (c == '\n') {
+	    if (fputc ('\r', f) < 0)
+		break;
+	    if (fputc ('\n', f) < 0)
+		break;
+	} else {
+	    if (fputc (c, f) < 0)
+		break;
+	}
+    }
+    return i;
+}
+#else
+
+long edit_insert_stream (WEdit * edit, int fd)
+{
+    int len;
+    long total = 0;
+    for (;;) {
+	char *p, *q;
+	len = 8192;
+	q = p = read_pipe (fd, &len);
+	if (!len) {
+	    free (p);
+	    return total;
+	}
+	total += len;
+	while (len--)
+	    edit_insert (edit, *p++);
+	free (q);
+    }
+    return total;
+}
+
+long edit_write_stream (WEdit * edit, FILE * f)
+{
+    long i;
+    for (i = 0; i < edit->last_byte; i++) {
+	int r;
+	while ((r = fputc (edit_get_byte (edit, i), f)) == -1 && errno == EINTR);
+	if (r < 0)
+	    break;
+    }
+    return i;
+}
+#endif
+
+#define TEMP_BUF_LEN 1024
+
+/* inserts a file at the cursor, returns 1 on success */
+int edit_insert_file (WEdit * edit, const char *filename)
+{
+    char *p;
+    if ((p = edit_get_filter (filename))) {
+	long current = edit->curs1;
+	int f, g;
+	char *a[8];
+	a[0] = "/bin/sh";
+	a[1] = "-c";
+	a[2] = p;
+	a[3] = 0;
+	if (triple_pipe_open (0, &f, &g, 0, "sh", a) > 0) {
+	    edit_insert_stream (edit, f);
+	    edit_cursor_move (edit, current - edit->curs1);
+	    free (p);
+	    p = read_pipe (g, 0);
+	    if (strlen (p)) {
+		edit_error_dialog (_ (" Error "), catstrs (_ (" Error reading from pipe: "), p, " ", 0));
+		free (p);
+		close (f);
+		close (g);
+		return 0;
+	    }
+	    close (f);
+	    close (g);
+	} else {
+	    edit_error_dialog (_ (" Error "), get_sys_error (catstrs (_ (" Failed trying to open pipe for reading: "), p, " ", 0)));
+	    free (p);
+	    return 0;
+	}
+	free (p);
+#ifdef CR_LF_TRANSLATION
+    } else {
+	FILE *f;
+	long current = edit->curs1;
+	f = fopen (filename, "r");
+	if (f) {
+	    edit_insert_stream (edit, f);
+	    edit_cursor_move (edit, current - edit->curs1);
+	    if (fclose (f)) {
+		edit_error_dialog (_ (" Error "), get_sys_error (catstrs (_ (" Error reading file: "), filename, " ", 0)));
+		return 0;
+	    }
+	} else {
+	    edit_error_dialog (_ (" Error "), get_sys_error (catstrs (_ (" Failed trying to open file for reading: "), filename, " ", 0)));
+	    return 0;
+	}
+#else
+    } else {
+	int i, file, blocklen;
+	long current = edit->curs1;
+	unsigned char *buf;
+	if ((file = open ((char *) filename, O_RDONLY)) == -1)
+	    return 0;
+	buf = malloc (TEMP_BUF_LEN);
+	while ((blocklen = read (file, (char *) buf, TEMP_BUF_LEN)) > 0) {
+	    for (i = 0; i < blocklen; i++)
+		edit_insert (edit, buf[i]);
+	}
+	edit_cursor_move (edit, current - edit->curs1);
+	free (buf);
+	close (file);
+	if (blocklen)
+	    return 0;
+#endif
+    }
+    return 1;
+}
+
+static int check_file_access (WEdit *edit, const char *filename, struct stat *st)
+{
+    int file;
+#if defined(MIDNIGHT) || defined(GTK)
+    if ((file = open ((char *) filename, O_RDONLY)) < 0) {
+	close (creat ((char *) filename, 0666));
+	if ((file = open ((char *) filename, O_RDONLY)) < 0) {
+	    edit_error_dialog (_ (" Error "), get_sys_error (catstrs (" Fail trying to open the file, ", filename, ", for reading ", 0)));
+	    return 2;
+	}
+    }
+#else
+    if ((file = open ((char *) filename, O_RDONLY)) < 0) {
+	edit_error_dialog (_ (" Error "), get_sys_error (catstrs (_ (" Failed trying to open file for reading: "), filename, " ", 0)));
+	return 1;
+    }
+#endif
+    if (stat ((char *) filename, st) < 0) {
+	close (file);
+/* The file-name is printed after the ':' */
+	edit_error_dialog (_ (" Error "), get_sys_error (catstrs (_ (" Cannot get size/permissions info on file: "), filename, " ", 0)));
+	return 1;
+    }
+    if (S_ISDIR (st->st_mode) || S_ISSOCK (st->st_mode)
+	|| S_ISFIFO (st->st_mode)) {
+	close (file);
+/* The file-name is printed after the ':' */
+	edit_error_dialog (_ (" Error "), catstrs (_ (" Not an ordinary file: "), filename, " ", 0));
+	return 1;
+    }
+    if (st->st_size >= SIZE_LIMIT) {
+	close (file);
+/* The file-name is printed after the ':' */
+	edit_error_dialog (_ (" Error "), catstrs (_ (" File is too large: "), \
+						   filename, _ (" \n Increase edit.h:MAXBUF and recompile the editor. "), 0));
+	return 1;
+    }
+    close (file);
     return 0;
 }
 
 /* returns 1 on error */
-int edit_load_file (WEdit * edit, const char *filename, const char *text, unsigned long text_size)
+int edit_open_file (WEdit * edit, const char *filename, const char *text, unsigned long text_size)
 {
-    struct stat s;
-    int file;
-
-/* VARS for Lastbyte calculation in TEXT mode FRANCO */
-#if defined CR_LF_TRANSLATION
-    char tmp_buf[1024];
-    long real_size,bytes_read;
-#endif
-
+    struct stat st;
     if (text) {
 	edit->last_byte = text_size;
-	filename = NULL;
+	filename = 0;
     } else {
+	int r;
+	r = check_file_access (edit, filename, &st);
 #if defined(MIDNIGHT) || defined(GTK)
-	if ((file = open ((char *) filename, O_RDONLY | MY_O_TEXT )) < 0)
-	{
-		close(creat((char *) filename, 0666));
-		if ((file = open ((char *) filename, O_RDONLY | MY_O_TEXT )) < 0) {
-			edit_error_dialog (_(" Error "), get_sys_error (catstrs (" Fail trying to open the file, ", filename, ", for reading ", 0)));
-			return 1;
-		}
-		edit->delete_file = 1;
-	}
+	if (r == 2)
+	    return edit->delete_file = 1;
+#endif
+	if (r)
+	    return 1;
+	edit->stat = st;
+#ifndef CR_LF_TRANSLATION
+	edit->last_byte = st.st_size;
 #else
-	if ((file = open ((char *) filename, O_RDONLY)) < 0) {
-	    edit_error_dialog (_(" Error "), get_sys_error (catstrs (_(" Failed trying to open file for reading: "), filename, " ", 0)));
-	    return 1;
-	}
+/* going to read the file into the buffer later byte by byte */
+	edit->last_byte = 0;
+	filename = 0;
+	text = "";
 #endif
-	if (stat ((char *) filename, &s) < 0) {
-	    close (file);
-/* The file-name is printed after the ':' */
-	    edit_error_dialog (_(" Error "), get_sys_error (catstrs (_(" Cannot get size/permissions info on file: "), filename, " ", 0)));
-	    return 1;
-	}
-	if (S_ISDIR (s.st_mode) || S_ISSOCK (s.st_mode)
-	    || S_ISFIFO (s.st_mode)) {
-	    close (file);
-/* The file-name is printed after the ':' */
-	    edit_error_dialog (_(" Error "), catstrs (_(" Not an ordinary file: "), filename, " ", 0));
-	    return 1;
-	}
-	if (s.st_size >= SIZE_LIMIT) {
-	    close (file);
-/* The file-name is printed after the ':' */
-	    edit_error_dialog (_(" Error "), catstrs (_(" File is too large: "), \
-		filename, _(" \n Increase edit.h:MAXBUF and recompile the editor. "), 0));
-	    return 1;
-	}
-
-/* Lastbyte calculation in TEXT mode FRANCO */
-#if defined CR_LF_TRANSLATION
-	if(file && (!text)){
-		real_size=0;
-		tmp_buf[sizeof (tmp_buf) - 1] = 0;
-		while((bytes_read = read(file,tmp_buf,1024)) > 0){
-			real_size += bytes_read;
-		}
-		s.st_size = real_size;
-	}
-
-#endif
-
-	close (file);
-	edit->last_byte = s.st_size;
-	edit->stat = s;
     }
-
     return init_dynamic_edit_buffers (edit, filename, text);
 }
 
@@ -282,7 +458,6 @@ int edit_load_file (WEdit * edit, const char *filename, const char *text, unsign
 #else
 int space_width;
 extern int option_long_whitespace;
-extern unsigned char per_char[256];
 
 void edit_set_space_width (int s)
 {
@@ -291,26 +466,25 @@ void edit_set_space_width (int s)
 
 #endif
 
+int (*edit_file_is_open) (char *) = 0;
+
 /* fills in the edit struct. returns 0 on fail. Pass edit as NULL for this */
 WEdit *edit_init (WEdit * edit, int lines, int columns, const char *filename, const char *text, const char *dir, unsigned long text_size)
 {
     char *f;
     int to_free = 0;
+    int use_filter = 0;
 #ifndef MIDNIGHT
     if (option_long_whitespace)
-	edit_set_space_width (per_char[' '] * 2);
+	edit_set_space_width (FONT_PER_CHAR(' ') * 2);
     else
-	edit_set_space_width (per_char[' ']);
+	edit_set_space_width (FONT_PER_CHAR(' '));
 #endif
     if (!edit) {
 	edit = malloc (sizeof (WEdit));
 	memset (edit, 0, sizeof (WEdit));
 	to_free = 1;
     }
-    if (!edit) {
-	edit_error_dialog (_(" Error "), _(" Error allocating memory "));
-	return 0;
-    }
     memset (&(edit->from_here), 0, (unsigned long) &(edit->to_here) - (unsigned long) &(edit->from_here));
 #ifndef MIDNIGHT
     edit->max_column = columns * FONT_MEAN_WIDTH;
@@ -321,16 +495,36 @@ WEdit *edit_init (WEdit * edit, int lines, int columns, const char *filename, co
     edit->stat.st_uid = getuid ();
     edit->stat.st_gid = getgid ();
     edit->bracket = -1;
+    edit->last_get_mb_rule = -2;
     if (!dir)
 	dir = "";
     f = (char *) filename;
-    if (filename)
+    if (filename) {
 	f = catstrs (dir, filename, 0);
-    if (edit_load_file (edit, f, text, text_size)) {
+	if (edit_file_is_open)
+	    if ((*edit_file_is_open) (f)) {
+		if (to_free)
+		    free (edit);
+		return 0;
+	    }
+    }
+    if (edit_find_filter (f) < 0) {
+#ifdef CR_LF_TRANSLATION
+	use_filter = 1;
+#endif
+	if (edit_open_file (edit, f, text, text_size)) {
 /* edit_load_file already gives an error message */
-	if (to_free)
-	    free (edit);
-	return 0;
+	    if (to_free)
+		free (edit);
+	    return 0;
+	}
+    } else {
+	use_filter = 1;
+	if (edit_open_file (edit, 0, "", 0)) {
+	    if (to_free)
+		free (edit);
+	    return 0;
+	}
     }
     edit->force |= REDRAW_PAGE;
     if (filename) {
@@ -338,18 +532,32 @@ WEdit *edit_init (WEdit * edit, int lines, int columns, const char *filename, co
 	edit_split_filename (edit, (char *) filename);
     } else {
 	edit->filename = (char *) strdup ("");
-	edit->dir = (char *) strdup(dir);
+	edit->dir = (char *) strdup (dir);
     }
     edit->stack_size = START_STACK_SIZE;
     edit->stack_size_mask = START_STACK_SIZE - 1;
     edit->undo_stack = malloc ((edit->stack_size + 10) * sizeof (long));
-    if (!edit->undo_stack) {
-	edit_error_dialog (_(" Error "), _(" Error allocating memory "));
-	if (to_free)
-	    free (edit);
-	return 0;
-    }
     edit->total_lines = edit_count_lines (edit, 0, edit->last_byte);
+    if (use_filter) {
+	struct stat st;
+	push_action_disabled = 1;
+	if (check_file_access (edit, filename, &st)) {
+	    edit_clean (edit);
+	    if (to_free)
+		free (edit);
+	    return 0;
+	}
+	edit->stat = st;
+	if (!edit_insert_file (edit, f)) {
+	    edit_clean (edit);
+	    if (to_free)
+		free (edit);
+	    return 0;
+	}
+/* FIXME: this should be an unmodification() function */
+	push_action_disabled = 0;
+    }
+    edit->modified = 0;
     edit_load_syntax (edit, 0, 0);
     {
 	int fg, bg;
@@ -358,13 +566,13 @@ WEdit *edit_init (WEdit * edit, int lines, int columns, const char *filename, co
     return edit;
 }
 
-
 /* clear the edit struct, freeing everything in it. returns 1 on success */
 int edit_clean (WEdit * edit)
 {
     if (edit) {
 	int j = 0;
 	edit_free_syntax_rules (edit);
+	edit_get_wide_byte (edit, -1);
 	book_mark_flush (edit, -1);
 	for (; j <= MAXBUFF; j++) {
 	    if (edit->buffers1[j] != NULL)
@@ -408,12 +616,20 @@ int edit_renew (WEdit * edit)
 /* returns 1 on success, if returns 0, the edit struct would have been free'd */
 int edit_reload (WEdit * edit, const char *filename, const char *text, const char *dir, unsigned long text_size)
 {
+    WEdit *e;
     int lines = edit->num_widget_lines;
     int columns = edit->num_widget_columns;
-    edit_clean (edit);
-    if (!edit_init (edit, lines, columns, filename, text, dir, text_size)) {
+    e = malloc (sizeof (WEdit));
+    memset (e, 0, sizeof (WEdit));
+    e->widget = edit->widget;
+    e->macro_i = -1;
+    if (!edit_init (e, lines, columns, filename, text, dir, text_size)) {
+	free (e);
 	return 0;
     }
+    edit_clean (edit);
+    memcpy (edit, e, sizeof (WEdit));
+    free (e);
     return 1;
 }
 
@@ -458,8 +674,6 @@ int edit_reload (WEdit * edit, const char *filename, const char *text, const cha
 
 */
 
-static int push_action_disabled = 0;
-
 void edit_push_action (WEdit * edit, long c,...)
 {
     unsigned long sp = edit->stack_pointer;
@@ -581,10 +795,19 @@ long pop_action (WEdit * edit)
 }
 
 /* is called whenever a modification is made by one of the four routines below */
-static inline void edit_modification (WEdit * edit)
+static inline void edit_modification (WEdit * edit, long p)
 {
     edit->caches_valid = 0;
     edit->modified = 1;
+    edit->screen_modified = 1;
+    if (edit->last_get_mb_rule > p - 1) {
+	edit->last_get_mb_rule = p - 1;
+	edit->mb_invalidate = 1;
+    }
+    if (edit->last_get_mb_rule > p - 1) {
+	edit->last_get_rule = p - 1;
+	edit->syntax_invalidate = 1;
+    }
 }
 
 /*
@@ -613,7 +836,7 @@ void edit_insert (WEdit * edit, int c)
 	edit->force |= REDRAW_LINE_ABOVE | REDRAW_AFTER_CURSOR;
     }
 /* tell that we've modified the file */
-    edit_modification (edit);
+    edit_modification (edit, edit->curs1);
 
 /* save the reverse command onto the undo stack */
     edit_push_action (edit, BACKSPACE);
@@ -621,7 +844,6 @@ void edit_insert (WEdit * edit, int c)
 /* update markers */
     edit->mark1 += (edit->mark1 > edit->curs1);
     edit->mark2 += (edit->mark2 > edit->curs1);
-    edit->last_get_rule += (edit->last_get_rule > edit->curs1);
 
 /* add a new buffer if we've reached the end of the last one */
     if (!(edit->curs1 & M_EDIT_BUF_SIZE))
@@ -654,12 +876,11 @@ void edit_insert_ahead (WEdit * edit, int c)
 	edit->total_lines++;
 	edit->force |= REDRAW_AFTER_CURSOR;
     }
-    edit_modification (edit);
+    edit_modification (edit, edit->curs1);
     edit_push_action (edit, DELETE);
 
     edit->mark1 += (edit->mark1 >= edit->curs1);
     edit->mark2 += (edit->mark2 >= edit->curs1);
-    edit->last_get_rule += (edit->last_get_rule >= edit->curs1);
 
     if (!((edit->curs2 + 1) & M_EDIT_BUF_SIZE))
 	edit->buffers2[(edit->curs2 + 1) >> S_EDIT_BUF_SIZE] = malloc (EDIT_BUF_SIZE);
@@ -678,7 +899,6 @@ int edit_delete (WEdit * edit)
 
     edit->mark1 -= (edit->mark1 > edit->curs1);
     edit->mark2 -= (edit->mark2 > edit->curs1);
-    edit->last_get_rule -= (edit->last_get_rule > edit->curs1);
 
     p = edit->buffers2[(edit->curs2 - 1) >> S_EDIT_BUF_SIZE][EDIT_BUF_SIZE - ((edit->curs2 - 1) & M_EDIT_BUF_SIZE) - 1];
 
@@ -701,7 +921,7 @@ int edit_delete (WEdit * edit)
 	if (p == '\n')
 	    edit->start_line--;
     }
-    edit_modification (edit);
+    edit_modification (edit, edit->curs1);
 
     return p;
 }
@@ -715,7 +935,6 @@ int edit_backspace (WEdit * edit)
 
     edit->mark1 -= (edit->mark1 >= edit->curs1);
     edit->mark2 -= (edit->mark2 >= edit->curs1);
-    edit->last_get_rule -= (edit->last_get_rule >= edit->curs1);
 
     p = *(edit->buffers1[(edit->curs1 - 1) >> S_EDIT_BUF_SIZE] + ((edit->curs1 - 1) & M_EDIT_BUF_SIZE));
     if (!((edit->curs1 - 1) & M_EDIT_BUF_SIZE)) {
@@ -739,11 +958,95 @@ int edit_backspace (WEdit * edit)
 	if (p == '\n')
 	    edit->start_line--;
     }
-    edit_modification (edit);
+    edit_modification (edit, edit->curs1);
 
     return p;
 }
 
+int edit_delete_wide (WEdit * edit)
+{
+    struct mb_rule r;
+    r = get_mb_rule (edit, edit->curs1);
+    edit_delete (edit);
+    while (r.end--)
+	edit_delete (edit);
+    return r.ch;
+}
+
+extern int option_utf_interpretation;
+
+void edit_insert_wide (WEdit * edit, wchar_t wc)
+{
+    unsigned char *c;
+    if (!option_utf_interpretation) {
+	edit_insert (edit, wc & 0xFF);
+	return;
+    }
+    c = wcrtomb_ucs4_to_utf8 (wc);
+    if (!*c) {
+	edit_insert (edit, *c);
+    } else {
+	for (; *c; c++)
+	    edit_insert (edit, *c);
+    }
+}
+
+long edit_backspace_wide (WEdit * edit)
+{
+    int i;
+    long c = 0;
+    if (!edit->curs1)
+	return 0;
+    for (i = edit->curs1 - 1; i >= 0; i--)
+	if ((c = edit_get_wide_byte (edit, i)) != -1)
+	    break;
+    while (edit->curs1 > i)
+	edit_backspace (edit);
+    return c;
+}
+
+static int right_of_four_spaces (WEdit *edit);
+void edit_tab_cmd (WEdit * edit);
+
+void edit_backspace_tab (WEdit * edit, int whole_tabs_only)
+{
+    int i;
+    if (whole_tabs_only) {
+	int indent;
+	/* count the number of columns of indentation */
+	indent = edit_move_forward3 (edit, edit_bol (edit, edit->curs1), 0, edit->curs1);
+	for (;;) {
+	    int c;
+	    c = edit_get_byte (edit, edit->curs1 - 1);
+	    if (!isspace (c) || c == '\n')
+		break;
+	    edit_backspace (edit);
+	}
+	while (edit_move_forward3 (edit, edit_bol (edit, edit->curs1), 0, edit->curs1) <
+	       (indent - space_width * (option_fake_half_tabs ? HALF_TAB_SIZE : TAB_SIZE)))
+	    edit_tab_cmd (edit);
+	return;
+    }
+    if (option_fake_half_tabs && right_of_four_spaces (edit)) {
+	for (i = 0; i < HALF_TAB_SIZE; i++)
+	    edit_backspace (edit);
+	return;
+    }
+    edit_backspace (edit);
+}
+
+void edit_wide_char_align (WEdit * edit)
+{
+    while (edit_get_wide_byte (edit, edit->curs1) == -1)
+	edit_cursor_move (edit, 1);
+}
+
+void edit_wide_char_align_left (WEdit * edit)
+{
+    while (edit_get_wide_byte (edit, edit->curs1) == -1)
+	edit_cursor_move (edit, -1);
+}
+
 #ifdef FAST_MOVE_CURSOR
 
 static void memqcpy (WEdit * edit, unsigned char *dest, unsigned char *src, int n)
@@ -1054,8 +1357,10 @@ void edit_scroll_upward (WEdit * edit, unsigned long i)
 	edit->start_display = edit_move_backward (edit, edit->start_display, i);
 	edit->force |= REDRAW_PAGE;
 	edit->force &= (0xfff - REDRAW_CHAR_ONLY);
+#ifndef MIDNIGHT
+#endif
     }
-    edit_update_curs_row(edit);
+    edit_update_curs_row (edit);
 }
 
 
@@ -1071,8 +1376,10 @@ void edit_scroll_downward (WEdit * edit, int i)
 	edit->start_display = edit_move_forward (edit, edit->start_display, i, 0);
 	edit->force |= REDRAW_PAGE;
 	edit->force &= (0xfff - REDRAW_CHAR_ONLY);
+#ifndef MIDNIGHT
+#endif
     }
-    edit_update_curs_row(edit);
+    edit_update_curs_row (edit);
 }
 
 void edit_scroll_right (WEdit * edit, int i)
@@ -1403,6 +1710,7 @@ void edit_get_selection (WEdit * edit)
 	    *p++ = edit_get_byte (edit, start_mark);
 	*p = 0;
     }
+    selection_clear ();
     selection.text = selection_history[current_selection].text;
     selection.len = selection_history[current_selection].len;
 }
@@ -1438,6 +1746,7 @@ static unsigned long my_type_of (int c)
     char *p, *q;
     if (!c)
 	return 0;
+    c &= 0xFF;
     if (c == '!') {
 	if (*option_chars_move_whole_word == '!')
 	    return 2;
@@ -1522,9 +1831,9 @@ static void edit_right_delete_word (WEdit * edit)
     for (;;) {
 	if (edit->curs1 >= edit->last_byte)
 	    break;
-	c1 = edit_delete (edit);
+	c1 = edit_delete_wide (edit);
 	c2 = edit_get_byte (edit, edit->curs1);
-	if ((isspace (c1) == 0) != (isspace (c2) == 0))
+	if ((wc_isspace (c1) == 0) != (isspace (c2) == 0))
 	    break;
 	if (!(my_type_of (c1) & my_type_of (c2)))
 	    break;
@@ -1537,9 +1846,9 @@ static void edit_left_delete_word (WEdit * edit)
     for (;;) {
 	if (edit->curs1 <= 0)
 	    break;
-	c1 = edit_backspace (edit);
+	c1 = edit_backspace_wide (edit);
 	c2 = edit_get_byte (edit, edit->curs1 - 1);
-	if ((isspace (c1) == 0) != (isspace (c2) == 0))
+	if ((wc_isspace (c1) == 0) != (isspace (c2) == 0))
 	    break;
 	if (!(my_type_of (c1) & my_type_of (c2)))
 	    break;
@@ -1647,11 +1956,15 @@ void edit_delete_line (WEdit * edit)
 	edit_insert (edit, '\n');
 }
 
-static void insert_spaces_tab (WEdit * edit)
+static void insert_spaces_tab (WEdit * edit, int half)
 {
-    int i = option_tab_spacing;
-    while (i--)
+    int i;
+    edit_update_curs_col (edit);
+    i = ((edit->curs_col / (option_tab_spacing * space_width / (half + 1))) + 1) * (option_tab_spacing * space_width / (half + 1)) - edit->curs_col;
+    while (i > 0) {
 	edit_insert (edit, ' ');
+	i -= space_width;
+    }
 }
 
 static int is_aligned_on_a_tab (WEdit * edit)
@@ -1729,7 +2042,7 @@ static void edit_double_newline (WEdit * edit)
     edit_insert (edit, '\n');
 }
 
-static void edit_tab_cmd (WEdit * edit)
+void edit_tab_cmd (WEdit * edit)
 {
     int i;
 
@@ -1738,23 +2051,18 @@ static void edit_tab_cmd (WEdit * edit)
 	    /*insert a half tab (usually four spaces) unless there is a
 	       half tab already behind, then delete it and insert a 
 	       full tab. */
-	    if (right_of_four_spaces (edit)) {
+	    if (!option_fill_tabs_with_spaces && right_of_four_spaces (edit)) {
 		for (i = 1; i <= HALF_TAB_SIZE; i++)
 		    edit_backspace (edit);
-		if (option_fill_tabs_with_spaces) {
-		    insert_spaces_tab (edit);
-		} else {
-		    edit_insert (edit, '\t');
-		}
+		edit_insert (edit, '\t');
 	    } else {
-		for (i = 1; i <= HALF_TAB_SIZE; i++)
-		    edit_insert (edit, ' ');
+		insert_spaces_tab (edit, 1);
 	    }
 	    return;
 	}
     }
     if (option_fill_tabs_with_spaces) {
-	insert_spaces_tab (edit);
+	insert_spaces_tab (edit, 0);
     } else {
 	edit_insert (edit, '\t');
     }
@@ -1771,10 +2079,14 @@ static void check_and_wrap_line (WEdit * edit)
     edit_update_curs_col (edit);
 #ifdef MIDNIGHT
     if (edit->curs_col < option_word_wrap_line_length)
-#else
-    if (edit->curs_col < option_word_wrap_line_length * FONT_MEAN_WIDTH)
-#endif
 	return;
+#else
+    CPushFont ("editor", 0);
+    c = FONT_MEAN_WIDTH;
+    CPopFont ();
+    if (edit->curs_col < option_word_wrap_line_length * c)
+	return;
+#endif
     curs = edit->curs1;
     for (;;) {
 	curs--;
@@ -1796,13 +2108,13 @@ static void check_and_wrap_line (WEdit * edit)
 void edit_execute_macro (WEdit * edit, struct macro macro[], int n);
 
 /* either command or char_for_insertion must be passed as -1 */
-int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion);
+int edit_execute_cmd (WEdit * edit, int command, long char_for_insertion);
 
 #ifdef MIDNIGHT
-int edit_translate_key (WEdit * edit, unsigned int x_keycode, long x_key, int x_state, int *cmd, int *ch)
+int edit_translate_key (WEdit * edit, unsigned int x_keycode, long x_key, int x_state, int *cmd, long *ch)
 {
     int command = -1;
-    int char_for_insertion = -1;
+    long char_for_insertion = -1;
 
 #include "edit_key_translator.c"
 
@@ -1823,43 +2135,75 @@ void edit_push_key_press (WEdit * edit)
 }
 
 /* this find the matching bracket in either direction, and sets edit->bracket */
-void edit_find_bracket (WEdit * edit)
+static long edit_get_bracket (WEdit * edit, int in_screen, unsigned long furthest_bracket_search)
+{
+    const char *b = "{}{[][()(", *p;
+    int i = 1, a, inc = -1, c, d, n = 0;
+    unsigned long j = 0;
+    long q;
+    edit_update_curs_row (edit);
+    c = edit_get_byte (edit, edit->curs1);
+    p = strchr (b, c);
+/* no limit */
+    if (!furthest_bracket_search)
+	furthest_bracket_search--;
+/* not on a bracket at all */
+    if (!p)
+	return -1;
+/* the matching bracket */
+    d = p[1];
+/* going left or right? */
+    if (strchr ("{[(", c))
+	inc = 1;
+    for (q = edit->curs1 + inc;; q += inc) {
+/* out of buffer? */
+	if (q >= edit->last_byte || q < 0)
+	    break;
+	a = edit_get_byte (edit, q);
+/* don't want to eat CPU */
+	if (j++ > furthest_bracket_search)
+	    break;
+/* out of screen? */
+	if (in_screen) {
+	    if (q < edit->start_display)
+		break;
+/* count lines if searching downward */
+	    if (inc > 0 && a == '\n')
+		if (n++ >= edit->num_widget_lines - edit->curs_row)	/* out of screen */
+		    break;
+	}
+/* count bracket depth */
+	i += (a == c) - (a == d);
+/* return if bracket depth is zero */
+	if (!i)
+	    return q;
+    }
+/* no match */
+    return -1;
+}
+
+static long last_bracket = -1;
+
+static void edit_find_bracket (WEdit * edit)
 {
     if (option_find_bracket) {
-	const char *b = "{}{[][()(", *p;
-	static int last_bracket = -1;
-	int i = 1, a, inc = -1, c, d, n = 0, j = 0;
-	long q;
-
-	edit->bracket = -1;
-	c = edit_get_byte (edit, edit->curs1);
-	p = strchr (b, c);
-	edit_update_curs_row (edit);
-	if (p) {
-	    d = p[1];
-	    if (strchr ("{[(", c))
-		inc = 1;
-	    for (q = edit->curs1 + inc;; q += inc) {
-		if (q >= edit->last_byte || q < edit->start_display || j++ > 10000)
-		    break;
-		a = edit_get_byte (edit, q);
-		if (inc > 0 && a == '\n')
-		    n++;
-		if (n >= edit->num_widget_lines - edit->curs_row)	/* out of screen */
-		    break;
-		i += (a == c) - (a == d);
-		if (!i) {
-		    edit->bracket = q;
-		    break;
-		}
-	    }
-	}
+	edit->bracket = edit_get_bracket (edit, 1, 10000);
 	if (last_bracket != edit->bracket)
 	    edit->force |= REDRAW_PAGE;
 	last_bracket = edit->bracket;
     }
 }
 
+static void edit_goto_matching_bracket (WEdit *edit)
+{
+    long q;
+    q = edit_get_bracket (edit, 0, 0);
+    if (q < 0)
+	return;
+    edit->bracket = edit->curs1;
+    edit->force |= REDRAW_PAGE;
+    edit_cursor_move (edit, q - edit->curs1);
+}
 
 /* this executes a command as though the user initiated it through a key press. */
 /* callback with WIDGET_KEY as a message calls this after translating the key
@@ -1869,7 +2213,7 @@ void edit_find_bracket (WEdit * edit)
 /* one of command or char_for_insertion must be passed as -1 */
 /* commands are executed, and char_for_insertion is inserted at the cursor */
 /* returns 0 if the command is a macro that was not found, 1 otherwise */
-int edit_execute_key_command (WEdit * edit, int command, int char_for_insertion)
+int edit_execute_key_command (WEdit * edit, int command, long char_for_insertion)
 {
     int r;
     if (command == CK_Begin_Record_Macro) {
@@ -1925,7 +2269,7 @@ void edit_mail_dialog (WEdit * edit);
    Returns 0 if the command is a macro that was not found, 1
    otherwise.
  */
-int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
+int edit_execute_cmd (WEdit * edit, int command, long char_for_insertion)
 {
     int result = 1;
     edit->force |= REDRAW_LINE;
@@ -1956,11 +2300,14 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
     }
 /* An ordinary key press */
     if (char_for_insertion >= 0) {
-	if (edit->overwrite) {
+        long start_mark, end_mark;
+        if (option_typing_replaces_selection && !eval_marks (edit, &start_mark, &end_mark)) {
+            edit_block_delete (edit);
+	} else if (edit->overwrite) {
 	    if (edit_get_byte (edit, edit->curs1) != '\n')
-		edit_delete (edit);
+		edit_delete_wide (edit);
 	}
-	edit_insert (edit, char_for_insertion);
+	edit_insert_wide (edit, char_for_insertion);
 	if (option_auto_para_formatting) {
 	    format_paragraph (edit, 0);
 	    edit->force |= REDRAW_PAGE;
@@ -1997,25 +2344,30 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
 
 /* basic cursor key commands */
     switch (command) {
+        long start_mark, end_mark;
     case CK_BackSpace:
+        if (option_typing_replaces_selection && !eval_marks (edit, &start_mark, &end_mark)) {
+            edit_block_delete (edit);
+            break;
+        }
 	if (option_backspace_through_tabs && is_in_indent (edit)) {
 	    while (edit_get_byte (edit, edit->curs1 - 1) != '\n'
 		   && edit->curs1 > 0)
 		edit_backspace (edit);
 	    break;
 	} else {
-	    if (option_fake_half_tabs) {
-		int i;
-		if (is_in_indent (edit) && right_of_four_spaces (edit)) {
-		    for (i = 0; i < HALF_TAB_SIZE; i++)
-			edit_backspace (edit);
-		    break;
-		}
+	    if (is_in_indent (edit)) {
+		edit_backspace_tab (edit, 0);
+		break;
 	    }
 	}
-	edit_backspace (edit);
+	edit_backspace_wide (edit);
 	break;
     case CK_Delete:
+        if (option_typing_replaces_selection && !eval_marks (edit, &start_mark, &end_mark)) {
+            edit_block_delete (edit);
+            break;
+        }
 	if (option_fake_half_tabs) {
 	    int i;
 	    if (is_in_indent (edit) && left_of_four_spaces (edit)) {
@@ -2024,7 +2376,7 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
 		break;
 	    }
 	}
-	edit_delete (edit);
+	edit_delete_wide (edit);
 	break;
     case CK_Delete_Word_Left:
 	edit_left_delete_word (edit);
@@ -2076,6 +2428,7 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
 	    }
 	}
 	edit_cursor_move (edit, -1);
+	edit_wide_char_align_left (edit);
 	break;
     case CK_Right:
     case CK_Right_Highlight:
@@ -2179,6 +2532,7 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
 	break;
 
     case CK_Toggle_Bookmark:
+	book_mark_clear (edit, edit->curs_line, BOOK_MARK_FOUND_COLOR);
 	if (book_mark_query_color (edit, edit->curs_line, BOOK_MARK_COLOR))
 	    book_mark_clear (edit, edit->curs_line, BOOK_MARK_COLOR);
 	else
@@ -2186,6 +2540,7 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
 	break;
     case CK_Flush_Bookmarks:
 	book_mark_flush (edit, BOOK_MARK_COLOR);
+	book_mark_flush (edit, BOOK_MARK_FOUND_COLOR);
 	edit->force |= REDRAW_PAGE;
 	break;
     case CK_Next_Bookmark:
@@ -2306,8 +2661,16 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
 
     case CK_Date:{
 	    time_t t;
+#ifdef HAVE_STRFTIME
+	    char s[1024];
+#endif
 	    time (&t);
+#ifdef HAVE_STRFTIME
+	    strftime (s, 1024, "%c", localtime (&t));
+	    edit_printf (edit, s);
+#else
 	    edit_printf (edit, ctime (&t));
+#endif
 	    edit->force |= REDRAW_PAGE;
 	    break;
 	}
@@ -2318,9 +2681,16 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
 	format_paragraph (edit, 1);
 	edit->force |= REDRAW_PAGE;
 	break;
+    case CK_Paragraph_Indent_Mode:
+	edit_indent_left_right_paragraph (edit, 1);
+	edit->force |= REDRAW_PAGE;
+	break;
     case CK_Delete_Macro:
 	edit_delete_macro_cmd (edit);
 	break;
+    case CK_Match_Bracket:
+	edit_goto_matching_bracket (edit);
+	break;
 #ifdef MIDNIGHT
     case CK_Sort:
 	edit_sort_cmd (edit);
@@ -2336,6 +2706,8 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
     case CK_Mail:
     case CK_Find_File:
     case CK_Ctags:
+    case CK_Terminal:
+    case CK_Terminal_App:
 #endif
     case CK_Complete:
     case CK_Cancel:
@@ -2345,6 +2717,7 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
     case CK_Save_And_Quit:
     case CK_Check_Save_And_Quit:
     case CK_Run_Another:
+    case CK_Insert_Unicode:
     case CK_Debug_Start:
     case CK_Debug_Stop:
     case CK_Debug_Toggle_Break:
@@ -2392,6 +2765,10 @@ int edit_execute_cmd (WEdit * edit, int command, int char_for_insertion)
     }
 #endif
 
+/* any movement key could end up inside a wide char - fix */
+    if (command / 100 == 6 || (command > 0 && command < 100))
+	edit_wide_char_align (edit);
+
 /* keys which must set the col position, and the search vars */
     switch (command) {
     case CK_Find:
