@@ -1,6 +1,6 @@
 /* editor high level editing commands.
 
-   Copyright (C) 1996, 1997 the Free Software Foundation
+   Copyright (C) 1996-2000 the Free Software Foundation
 
    Authors: 1996, 1997 Paul Sheer
 
@@ -48,6 +48,7 @@ int replace_prompt = 1;
 int replace_whole = 0;
 int replace_case = 0;
 int replace_backwards = 0;
+int search_create_bookmark = 0;
 
 /* queries on a save */
 #ifdef MIDNIGHT
@@ -243,79 +244,122 @@ int my_open (const char *pathname, int flags,...)
 /* returns 0 on error */
 int edit_save_file (WEdit * edit, const char *filename)
 {
-    long buf;
+    char *p;
     long filelen = 0;
-    int file;
-    char *savename = (char *) filename;
-    int this_save_mode;
+    FILE *file;
+    char *savename = 0;
+    int this_save_mode, fd;
 
-    if ((file = open (savename, O_WRONLY)) == -1) {
-	this_save_mode = 0;		/* the file does not exists yet, so no safe save or backup necessary */
+    if (!filename)
+	return 0;
+    if (!*filename)
+	return 0;
+
+    savename = (char *) strdup ((char *) filename);
+
+    if ((fd = open (savename, O_WRONLY)) == -1) {
+	this_save_mode = 0;	/* the file does not exists yet, so no safe save or backup necessary */
     } else {
-	close (file);
+	close (fd);
 	this_save_mode = option_save_mode;
     }
 
     if (this_save_mode > 0) {
-	char *savedir = ".", *slashpos = strrchr (filename, '/');
-	if (slashpos != 0) {
+	char *savedir, *slashpos;
+	savedir = (char *) strdup (".");
+	slashpos = strrchr (filename, '/');
+	if (slashpos) {
+	    free (savedir);
 	    savedir = (char *) strdup (filename);
-	    if (savedir == 0)
-		return 0;
 	    savedir[slashpos - filename + 1] = '\0';
 	}
+	if (savename)
+	    free (savename);
 	savename = (char *) tempnam (savedir, "cooledit");
-	if (slashpos)
-	    free (savedir);
+	free (savedir);
 	if (!savename)
-	    return 0;
-    }
-    if ((file = open (savename, O_CREAT | O_WRONLY | O_TRUNC | MY_O_TEXT, edit->stat.st_mode)) == -1) {
-	if (this_save_mode > 0)
-	    free (savename);
-	return 0;
+	    goto error_save;
     }
+    if ((file = fopen (savename, "w+")) == 0)
+	goto error_save;
+    chmod (savename, edit->stat.st_mode);
     chown (savename, edit->stat.st_uid, edit->stat.st_gid);
-    buf = 0;
-    while (buf <= (edit->curs1 >> S_EDIT_BUF_SIZE) - 1) {
-	filelen += write (file, (char *) edit->buffers1[buf], EDIT_BUF_SIZE);
-	buf++;
-    }
-    filelen += write (file, (char *) edit->buffers1[buf], edit->curs1 & M_EDIT_BUF_SIZE);
 
-    if (edit->curs2) {
-	edit->curs2--;
-	buf = (edit->curs2 >> S_EDIT_BUF_SIZE);
-	filelen += write (file, (char *) edit->buffers2[buf] + EDIT_BUF_SIZE - (edit->curs2 & M_EDIT_BUF_SIZE) - 1, 1 + (edit->curs2 & M_EDIT_BUF_SIZE));
-	buf--;
-	while (buf >= 0) {
-	    filelen += write (file, (char *) edit->buffers2[buf], EDIT_BUF_SIZE);
-	    buf--;
-	}
-	edit->curs2++;
-    }
-    close (file);
-
-    if (filelen == edit->last_byte) {
-	if (this_save_mode == 2) {
-	    if (rename (filename, catstrs (filename, option_backup_ext, 0)) == -1) {	/* catstrs free's automatically */
-		free (savename);
-		return 0;
+/* pipe save */
+    if ((p = (char *) edit_get_write_filter (savename, filename))) {
+	fclose (file);
+	file = (FILE *) popen (p, "w");
+	if (file) {
+	    filelen = edit_write_stream (edit, file);
+#if 1
+	    pclose (file);
+#else
+	    if (pclose (file) > 0) {
+		edit_error_dialog (_ (" Error "), catstrs (_ (" Error writing to pipe: "), p, " ", 0));
+		free (p);
+		goto error_save;
 	    }
+#endif
+	} else {
+	    edit_error_dialog (_ (" Error "), get_sys_error (catstrs (_ (" Failed trying to open pipe for writing: "), p, " ", 0)));
+	    free (p);
+	    goto error_save;
 	}
-	if (this_save_mode > 0) {
-	    if (rename (savename, filename) == -1) {
-		free (savename);
-		return 0;
-	    }
-	    free (savename);
-	}
-	return 1;
+	free (p);
+#ifdef CR_LF_TRANSLATION
+    } else {			/* optimised save */
+	filelen = edit_write_stream (edit, f);
+	fclose (file);
+#else
     } else {
-	if (this_save_mode > 0)
-	    free (savename);
-	return 0;
+	long buf;
+	buf = 0;
+	filelen = edit->last_byte;
+	while (buf <= (edit->curs1 >> S_EDIT_BUF_SIZE) - 1) {
+	    if (fwrite ((char *) edit->buffers1[buf], EDIT_BUF_SIZE, 1, file) != 1) {
+		filelen = -1;
+		break;
+	    }
+	    buf++;
+	}
+	if (fwrite ((char *) edit->buffers1[buf], edit->curs1 & M_EDIT_BUF_SIZE, 1, file) == -1) {
+	    filelen = -1;
+	} else if (edit->curs2) {
+	    edit->curs2--;
+	    buf = (edit->curs2 >> S_EDIT_BUF_SIZE);
+	    if (fwrite ((char *) edit->buffers2[buf] + EDIT_BUF_SIZE - (edit->curs2 & M_EDIT_BUF_SIZE) - 1, 1 + (edit->curs2 & M_EDIT_BUF_SIZE), 1, file) != 1) {
+		filelen = -1;
+	    } else {
+		buf--;
+		while (buf >= 0) {
+		    if (fwrite ((char *) edit->buffers2[buf], EDIT_BUF_SIZE, 1, file) != 1) {
+			filelen = -1;
+			break;
+		    }
+		    buf--;
+		}
+	    }
+	    edit->curs2++;
+	}
+	fclose (file);
+#endif
     }
+
+    if (filelen != edit->last_byte)
+	goto error_save;
+    if (this_save_mode == 2)
+	if (rename (filename, catstrs (filename, option_backup_ext, 0)) == -1)
+	    goto error_save;
+    if (this_save_mode > 0)
+	if (rename (savename, filename) == -1)
+	    goto error_save;
+    if (savename)
+	free (savename);
+    return 1;
+  error_save:
+    if (savename)
+	free (savename);
+    return 0;
 }
 
 #ifdef MIDNIGHT
@@ -451,7 +495,11 @@ static char *canonicalize_pathname (char *p)
 void edit_split_filename (WEdit * edit, char *longname)
 {
     char *exp, *p;
-    exp = canonicalize_pathname (longname);	/* this ensures a full path */
+#if defined(MIDNIGHT) || defined(GTK)
+    exp = canonicalize_pathname (longname);
+#else
+    exp = pathdup (longname);	/* this ensures a full path */
+#endif
     if (edit->filename)
 	free (edit->filename);
     if (edit->dir)
@@ -647,13 +695,13 @@ int edit_delete_macro (WEdit * edit, int k)
 	if (!n || n == EOF)
 	    break;
 	n = 0;
-	while (fscanf (f, "%hd %hd, ", &macro[n].command, &macro[n].ch))
+	while (fscanf (f, "%d %ld, ", &macro[n].command, &macro[n].ch))
 	    n++;
 	fscanf (f, ";\n");
 	if (s != k) {
 	    fprintf (g, _("key '%d 0': "), s);
 	    for (i = 0; i < n; i++)
-		fprintf (g, "%hd %hd, ", macro[i].command, macro[i].ch);
+		fprintf (g, "%d %ld, ", macro[i].command, macro[i].ch);
 	    fprintf (g, ";\n");
 	}
     };
@@ -689,7 +737,7 @@ int edit_save_macro_cmd (WEdit * edit, struct macro macro[], int n)
 	if (f) {
 	    fprintf (f, _("key '%d 0': "), s);
 	    for (i = 0; i < n; i++)
-		fprintf (f, "%hd %hd, ", macro[i].command, macro[i].ch);
+		fprintf (f, "%d %ld, ", macro[i].command, macro[i].ch);
 	    fprintf (f, ";\n");
 	    fclose (f);
 	    if (saved_macros_loaded) {
@@ -749,10 +797,10 @@ int edit_load_macro_cmd (WEdit * edit, struct macro macro[], int *n, int k)
 		saved_macro[i++] = s;
 	    if (!found) {
 		*n = 0;
-		while (*n < MAX_MACRO_LENGTH && 2 == fscanf (f, "%hd %hd, ", &macro[*n].command, &macro[*n].ch))
+		while (*n < MAX_MACRO_LENGTH && 2 == fscanf (f, "%d %ld, ", &macro[*n].command, &macro[*n].ch))
 		    (*n)++;
 	    } else {
-		while (2 == fscanf (f, "%hd %hd, ", &dummy.command, &dummy.ch));
+		while (2 == fscanf (f, "%d %ld, ", &dummy.command, &dummy.ch));
 	    }
 	    fscanf (f, ";\n");
 	    if (s == k)
@@ -822,21 +870,13 @@ int edit_new_cmd (WEdit * edit)
 }
 
 /* returns 1 on error */
-int edit_load_file_from_filename (WEdit *edit, char *exp)
+int edit_load_file_from_filename (WEdit * edit, char *exp)
 {
-    int file;
-    if ((file = open ((char *) exp, O_RDONLY, MY_O_TEXT)) != -1) {
-	close (file);
-	if (!edit_reload (edit, exp, 0, "", 0))
-	    return 1;
-	edit_split_filename (edit, exp);
-	edit->modified = 0;
-	return 0;
-    } else {
-/* Heads the 'Load' file dialog box */
-	edit_error_dialog (_ (" Load "), get_sys_error (_ (" Error trying to open file for reading ")));
-    }
-    return 1;
+    if (!edit_reload (edit, exp, 0, "", 0))
+	return 1;
+    edit_split_filename (edit, exp);
+    edit->modified = 0;
+    return 0;
 }
 
 int edit_load_cmd (WEdit * edit)
@@ -1142,39 +1182,36 @@ int edit_block_delete_cmd (WEdit * edit)
 
 int edit_replace_prompt (WEdit * edit, char *replace_text, int xpos, int ypos)
 {
-    if (replace_prompt) {
-	QuickWidget quick_widgets[] =
-	{
+    QuickWidget quick_widgets[] =
+    {
 /* NLS  for hotkeys? */
-           {quick_button, 63, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_("&Cancel"),
-            0, B_CANCEL, 0, 0, XV_WLAY_DONTCARE, NULL},
-           {quick_button, 50, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_("o&Ne"),
-            0, B_REPLACE_ONE, 0, 0, XV_WLAY_DONTCARE, NULL},
-           {quick_button, 37, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_("al&L"),
-            0, B_REPLACE_ALL, 0, 0, XV_WLAY_DONTCARE, NULL},
-           {quick_button, 21, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_("&Skip"),
-            0, B_SKIP_REPLACE, 0, 0, XV_WLAY_DONTCARE, NULL},
-           {quick_button, 4, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_("&Replace"),
-            0, B_ENTER, 0, 0, XV_WLAY_DONTCARE, NULL},
-           {quick_label, 2, CONFIRM_DLG_WIDTH, 2, CONFIRM_DLG_HEIGTH, 0,
-	     0, 0, 0, XV_WLAY_DONTCARE, 0},
-	    {0}};
+	{quick_button, 63, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_ ("&Cancel"),
+	 0, B_CANCEL, 0, 0, XV_WLAY_DONTCARE, NULL},
+	{quick_button, 50, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_ ("o&Ne"),
+	 0, B_REPLACE_ONE, 0, 0, XV_WLAY_DONTCARE, NULL},
+	{quick_button, 37, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_ ("al&L"),
+	 0, B_REPLACE_ALL, 0, 0, XV_WLAY_DONTCARE, NULL},
+	{quick_button, 21, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_ ("&Skip"),
+	 0, B_SKIP_REPLACE, 0, 0, XV_WLAY_DONTCARE, NULL},
+	{quick_button, 4, CONFIRM_DLG_WIDTH, 3, CONFIRM_DLG_HEIGTH, N_ ("&Replace"),
+	 0, B_ENTER, 0, 0, XV_WLAY_DONTCARE, NULL},
+	{quick_label, 2, CONFIRM_DLG_WIDTH, 2, CONFIRM_DLG_HEIGTH, 0,
+	 0, 0, 0, XV_WLAY_DONTCARE, 0},
+	{0}};
 
-	quick_widgets[4].text = catstrs (_(" Replace with: "), replace_text, 0);
+    quick_widgets[4].text = catstrs (_ (" Replace with: "), replace_text, 0);
 
-	{
-	    QuickDialog Quick_input =
-           {CONFIRM_DLG_WIDTH, CONFIRM_DLG_HEIGTH, 0, 0, N_(" Confirm replace "),
-	     "[Input Line Keys]", "quick_input", 0 /*quick_widgets */ };
+    {
+	QuickDialog Quick_input =
+	{CONFIRM_DLG_WIDTH, CONFIRM_DLG_HEIGTH, 0, 0, N_ (" Confirm replace "),
+	 "[Input Line Keys]", "quick_input", 0 /*quick_widgets */ };
 
-	    Quick_input.widgets = quick_widgets;
+	Quick_input.widgets = quick_widgets;
 
-	    Quick_input.xpos = xpos;
-	    Quick_input.ypos = ypos;
-	    return quick_dialog (&Quick_input);
-	}
-    } else
-	return 0;
+	Quick_input.xpos = xpos;
+	Quick_input.ypos = ypos;
+	return quick_dialog (&Quick_input);
+    }
 }
 
 void edit_replace_dialog (WEdit * edit, char **search_text, char **replace_text, char **arg_order)
@@ -1339,153 +1376,105 @@ extern CWidget *wedit;
 
 #ifndef GTK
 
-void edit_search_replace_dialog (Window parent, int x, int y, char **search_text, char **replace_text, char **arg_order, char *heading, int option)
-{
-    Window win;
-    XEvent xev;
-    CEvent cev;
-    CState s;
-    int xh, yh, h, xb, ys, yc, yb, yr;
-    CWidget *m;
+void edit_backspace_tab (WEdit * edit, int whole_tabs_only);
 
+static inline int my_is_blank (int c)
+{
+    return c == ' ' || c == '\t';
+}
+
+void edit_indent_left_right_paragraph (WEdit * edit)
+{
+    char id[33];
+    int width = 0, lines, c;
+    XEvent xevent;
+    KeySym k;
+    char *r;
+    CState s;
+    CWidget *w, *p = 0, *i;
+    long start_mark, end_mark;
+    strcpy (id, CIdentOf (edit->widget));
+    strcat (id, ".text");
+    w = CIdent (id);
+    if (!w)
+	return;
+    if (eval_marks (edit, &start_mark, &end_mark)) {
+	edit_error_dialog (_(" Error "),
+			   _
+			   (" No text highlighted - highlight text, run command again, then use arrow keys. "));
+	return;
+    }
+    r = 0;
     CBackupState (&s);
     CDisable ("*");
-
-    win = CDrawHeadedDialog ("replace", parent, x, y, heading);
-    CGetHintPos (&xh, &h);
-
-/* NLS hotkey ? */
-    CIdent ("replace")->position = WINDOW_ALWAYS_RAISED;
-/* An input line comes after the ':' */
-    (CDrawText ("replace.t1", win, xh, h, _(" Enter search text : ")))->hotkey = 'E';
-
-    CGetHintPos (0, &yh);
-    (m = CDrawTextInput ("replace.sinp", win, xh, yh, 10, AUTO_HEIGHT, 8192, *search_text))->hotkey = 'E';
-
-    if (replace_text) {
-	CGetHintPos (0, &yh);
-	(CDrawText ("replace.t2", win, xh, yh, _(" Enter replace text : ")))->hotkey = 'n';
-	CGetHintPos (0, &yh);
-	(CDrawTextInput ("replace.rinp", win, xh, yh, 10, AUTO_HEIGHT, 8192, *replace_text))->hotkey = 'n';
-	CSetToolHint ("replace.t2", _("You can enter regexp substrings with %s\n(not \\1, \\2 like sed) then use \"Enter...order\""));
-	CSetToolHint ("replace.rinp", _("You can enter regexp substrings with %s\n(not \\1, \\2 like sed) then use \"Enter...order\""));
-	CGetHintPos (0, &yh);
-	(CDrawText ("replace.t3", win, xh, yh, _(" Enter argument (or substring) order : ")))->hotkey = 'o';
-	CGetHintPos (0, &yh);
-	(CDrawTextInput ("replace.ainp", win, xh, yh, 10, AUTO_HEIGHT, 256, *arg_order))->hotkey = 'o';
-/* Tool hint */
-	CSetToolHint ("replace.ainp", _("Enter the order of replacement of your scanf\nformat specifiers or regexp substrings, eg 3,1,2"));
-	CSetToolHint ("replace.t3", _("Enter the order of replacement of your scanf\nformat specifiers or regexp substrings, eg 3,1,2"));
-    }
-    CGetHintPos (0, &yh);
-    ys = yh;
-/* The following are check boxes */
-    CDrawSwitch ("replace.ww", win, xh, yh, replace_whole, _(" Whole words only "), 0);
-    CGetHintPos (0, &yh);
-    CDrawSwitch ("replace.case", win, xh, yh, replace_case, _(" Case sensitive "), 0);
-    yc = yh;
-    CGetHintPos (0, &yh);
-    CDrawSwitch ("replace.reg", win, xh, yh, replace_regexp, _(" Regular expression "), 1);
-    CSetToolHint ("replace.reg", _("See the regex man page for how\nto compose a regular expression"));
-    CSetToolHint ("replace.reg.label", _("See the regex man page for how\nto compose a regular expression"));
-    yb = yh;
-    CGetHintPos (0, &yh);
-    CGetHintPos (&xb, 0);
-
-    if (option & SEARCH_DIALOG_OPTION_BACKWARDS) {
-	CDrawSwitch ("replace.bkwd", win, xh, yh, replace_backwards, _(" Backwards "), 0);
-/* Tool hint */
-	CSetToolHint ("replace.bkwd", _("Warning: Searching backward can be slow"));
-	CSetToolHint ("replace.bkwd.label", _("Warning: Searching backward can be slow"));
-	yb = yh;
-    }
-    if (replace_text) {
-	if (option & SEARCH_DIALOG_OPTION_BACKWARDS)
-	    yr = yc;
-	else
-	    yr = ys;
-    } else {
-	yr = yb;
-    }
-
-    if (replace_text) {
-	CDrawSwitch ("replace.pr", win, xb, yr, replace_prompt, _(" Prompt on replace "), 0);
-/* Tool hint */
-	CSetToolHint ("replace.pr", _("Ask before making each replacement"));
-	CGetHintPos (0, &yr);
-	CDrawSwitch ("replace.all", win, xb, yr, replace_all, _(" Replace all "), 0);
-	CGetHintPos (0, &yr);
-    }
-    CDrawSwitch ("replace.scanf", win, xb, yr, replace_scanf, _(" Scanf expression "), 1);
-/* Tool hint */
-    CSetToolHint ("replace.scanf", _("Allows entering of a C format string,\nsee the scanf man page"));
-
-    get_hint_limits (&x, &y);
-    CDrawPixmapButton ("replace.ok", win, x - WIDGET_SPACING - TICK_BUTTON_WIDTH, h, PIXMAP_BUTTON_TICK);
-/* Tool hint */
-    CSetToolHint ("replace.ok", _("Begin search, Enter"));
-    CDrawPixmapButton ("replace.cancel", win, x - WIDGET_SPACING - TICK_BUTTON_WIDTH, h + WIDGET_SPACING + TICK_BUTTON_WIDTH, PIXMAP_BUTTON_CROSS);
-/* Tool hint */
-    CSetToolHint ("replace.cancel", _("Abort this dialog, Esc"));
-    CSetSizeHintPos ("replace");
-    CMapDialog ("replace");
-
-    m = CIdent ("replace");
-    CSetWidgetSize ("replace.sinp", m->width - WIDGET_SPACING * 3 - 4 - TICK_BUTTON_WIDTH, (CIdent ("replace.sinp"))->height);
-    if (replace_text) {
-	CSetWidgetSize ("replace.rinp", m->width - WIDGET_SPACING * 3 - 4 - TICK_BUTTON_WIDTH, (CIdent ("replace.rinp"))->height);
-	CSetWidgetSize ("replace.ainp", m->width - WIDGET_SPACING * 3 - 4 - TICK_BUTTON_WIDTH, (CIdent ("replace.ainp"))->height);
-    }
-    CFocus (CIdent ("replace.sinp"));
+    p =
+	CDrawText ("status_prompt", edit->widget->parentid, CXof (w), CYof (w), "%s",
+		   " <---  ---> ");
+    width = CWidthOf (p);
+    i = CDrawTextInput ("status_input", edit->widget->parentid, CXof (w) + width, CYof (w),
+			CWidthOf (edit->widget) - width, AUTO_HEIGHT, 1, "");
+    CFocus (i);
+    edit_set_markers (edit, edit_bol (edit, start_mark), edit_eol (edit, end_mark), -1, -1);
+    edit->force |= REDRAW_PAGE;
+    edit_render_keypress (edit);
+    edit_push_action (edit, KEY_PRESS + edit->start_display);
 
     for (;;) {
+	XEvent xev;
+	CEvent cev;
 	CNextEvent (&xev, &cev);
-	if (!CIdent ("replace")) {
-	    *search_text = 0;
-	    break;
-	}
-	if (!strcmp (cev.ident, "replace.cancel") || cev.command == CK_Cancel) {
-	    *search_text = 0;
-	    break;
-	}
-	if (!strcmp (cev.ident, "replace.reg") || !strcmp (cev.ident, "replace.scanf")) {
-	    if (CIdent ("replace.reg")->keypressed || CIdent ("replace.scanf")->keypressed) {
-		if (!(CIdent ("replace.case")->keypressed)) {
-		    CIdent ("replace.case")->keypressed = 1;
-		    CExpose ("replace.case");
+	if (xev.type == KeyPress) {
+	    if (eval_marks (edit, &start_mark, &end_mark))
+		break;
+	    lines = edit_count_lines (edit, start_mark, end_mark);
+	    k = CKeySym (&xevent);
+	    if (cev.command == CK_Right || cev.command == CK_Tab) {
+		long s;
+		for (c = 0, s = start_mark; c <= lines; s = edit_eol (edit, s) + 1, c++) {
+		    while (my_is_blank (edit_get_byte (edit, s)) && s < edit->last_byte)
+			s++;
+		    edit_cursor_move (edit, s - edit->curs1);
+		    edit_tab_cmd (edit);
+		    s = edit->curs1;
 		}
-	    }
-	}
-	if (!strcmp (cev.ident, "replace.ok") || cev.command == CK_Enter) {
-	    if (replace_text) {
-		replace_all = CIdent ("replace.all")->keypressed;
-		replace_prompt = CIdent ("replace.pr")->keypressed;
-		*replace_text = (char *) strdup (CIdent ("replace.rinp")->text);
-		*arg_order = (char *) strdup (CIdent ("replace.ainp")->text);
-	    }
-	    *search_text = (char *) strdup (CIdent ("replace.sinp")->text);
-	    replace_whole = CIdent ("replace.ww")->keypressed;
-	    replace_case = CIdent ("replace.case")->keypressed;
-	    replace_scanf = CIdent ("replace.scanf")->keypressed;
-	    replace_regexp = CIdent ("replace.reg")->keypressed;
-
-	    if (option & SEARCH_DIALOG_OPTION_BACKWARDS) {
-		replace_backwards = CIdent ("replace.bkwd")->keypressed;
+		edit->force |= REDRAW_PAGE;
+		edit_render_keypress (edit);
+		edit_push_action (edit, KEY_PRESS + edit->start_display);
+	    } else if (cev.command == CK_Left || cev.command == CK_BackSpace) {
+		long s;
+		for (c = 0, s = start_mark; c <= lines; s = edit_eol (edit, s) + 1, c++) {
+		    while (my_is_blank (edit_get_byte (edit, s)) && s < edit->last_byte)
+			s++;
+		    edit_cursor_move (edit, s - edit->curs1);
+		    edit_backspace_tab (edit, 1);
+		    s = edit->curs1;
+		}
+		edit->force |= REDRAW_PAGE;
+		edit_render_keypress (edit);
+		edit_push_action (edit, KEY_PRESS + edit->start_display);
 	    } else {
-		replace_backwards = 0;
+		break;
 	    }
-
-	    break;
 	}
     }
-    CDestroyWidget ("replace");
+    CDestroyWidget ("status_prompt");
+    CDestroyWidget ("status_input");
     CRestoreState (&s);
+    return;
+}
+
+extern struct look *look;
+
+void edit_search_replace_dialog (Window parent, int x, int y, char **search_text,
+				 char **replace_text, char **arg_order, char *heading, int option)
+{
+    (*look->search_replace_dialog) (parent, x, y, search_text, replace_text, arg_order, heading, option);
 }
 
 void edit_search_dialog (WEdit * edit, char **search_text)
 {
 /* Heads the 'Search' dialog box */
-    edit_search_replace_dialog (WIN_MESSAGES, search_text, 0, 0, _(" Search "), SEARCH_DIALOG_OPTION_BACKWARDS);
+    edit_search_replace_dialog (WIN_MESSAGES, search_text, 0, 0, _(" Search "), SEARCH_DIALOG_OPTION_BACKWARDS | SEARCH_DIALOG_OPTION_BOOKMARK);
 }
 
 void edit_replace_dialog (WEdit * edit, char **search_text, char **replace_text, char **arg_order)
@@ -1549,44 +1538,38 @@ void edit_replace_dialog (WEdit * edit, char **search_text, char **replace_text,
 
 int edit_replace_prompt (WEdit * edit, char *replace_text, int xpos, int ypos)
 {
-    if (replace_prompt) {
-	char *s;
-	s = gtk_dialog_cauldron (
-				    "Replace", GTK_CAULDRON_TOPLEVEL | GTK_CAULDRON_GRAB,
-		  " ( (Replace with:)d %Ld )xf / ( %Bxfrq || %Bxfq || %Bxfq || %Bxfq || %Bxfgq )f",
-				    replace_text,
-				    "Replace", "Skip", "Replace all", "Replace one",
-				    GNOME_STOCK_BUTTON_CANCEL
-	    );
-	if (s == GTK_CAULDRON_ESCAPE || !s || s == GNOME_STOCK_BUTTON_CANCEL)
-	    return B_CANCEL;
-	if (!strcmp (s, "Replace all"))
-	    return B_REPLACE_ALL;
-	if (!strcmp (s, "Replace one"))
-	    return B_REPLACE_ONE;
-	if (!strcmp (s, "Skip"))
-	    return B_SKIP_REPLACE;
-	if (!strcmp (s, "Replace"))
-	    return B_ENTER;
-    }
-    return 0;
+    char *s;
+    s = gtk_dialog_cauldron (
+		    "Replace", GTK_CAULDRON_TOPLEVEL | GTK_CAULDRON_GRAB,
+				" ( (Replace with:)d %Ld )xf / ( %Bxfrq || %Bxfq || %Bxfq || %Bxfq || %Bxfgq )f",
+				replace_text,
+			 "Replace", "Skip", "Replace all", "Replace one",
+				GNOME_STOCK_BUTTON_CANCEL
+	);
+    if (s == GTK_CAULDRON_ESCAPE || !s || s == GNOME_STOCK_BUTTON_CANCEL)
+	return B_CANCEL;
+    if (!strcmp (s, "Replace all"))
+	return B_REPLACE_ALL;
+    if (!strcmp (s, "Replace one"))
+	return B_REPLACE_ONE;
+    if (!strcmp (s, "Skip"))
+	return B_SKIP_REPLACE;
+    if (!strcmp (s, "Replace"))
+	return B_ENTER;
 }
 
 #else
 
 int edit_replace_prompt (WEdit * edit, char *replace_text, int xpos, int ypos)
 {
-    if (replace_prompt) {
-	int q, x[] =
-	{
-	    B_CANCEL, B_ENTER, B_SKIP_REPLACE, B_REPLACE_ALL, B_REPLACE_ONE, B_CANCEL
-	};
-	q = CQueryDialog (WIN_MESSAGES + (edit->curs_line < 8 ? edit->num_widget_lines / 2 * FONT_PIX_PER_LINE + CYof (edit->widget) : 0),
-			  _ (" Replace "), catstrs (_ (" Replace with: "), replace_text, 0), _ ("Replace"), _ ("Skip"), _ ("Replace all"), _ ("Replace one"), _ ("Cancel"), 0);
-	edit->force |= REDRAW_COMPLETELY;
-	return x[q + 1];
-    }
-    return 0;
+    int q, x[] =
+    {
+	B_CANCEL, B_ENTER, B_SKIP_REPLACE, B_REPLACE_ALL, B_REPLACE_ONE, B_CANCEL
+    };
+    q = CQueryDialog (WIN_MESSAGES + (edit->curs_line < 8 ? edit->num_widget_lines / 2 * FONT_PIX_PER_LINE + CYof (edit->widget) : 0),
+		      _ (" Replace "), catstrs (_ (" Replace with: "), replace_text, 0), _ ("Replace"), _ ("Skip"), _ ("Replace all"), _ ("Replace one"), _ ("Cancel"), 0);
+    edit->force |= REDRAW_COMPLETELY;
+    return x[q + 1];
 }
 
 #endif
@@ -1959,9 +1942,10 @@ void edit_replace_cmd (WEdit * edit, int again)
     char *exp3 = "";
     int replace_yes;
     int replace_continue;
+    int treplace_prompt = 0;
     int i = 0;
     long times_replaced = 0, last_search;
-    char fin_string[32];
+    char fin_string[64];
     int argord[NUM_REPL_ARGS];
 
     if (!edit) {
@@ -1996,6 +1980,7 @@ void edit_replace_cmd (WEdit * edit, int again)
     } else {
 	edit_push_action (edit, KEY_PRESS + edit->start_display);
 	edit_replace_dialog (edit, &exp1, &exp2, &exp3);
+	treplace_prompt = replace_prompt;
     }
 
     if (!exp1 || !*exp1) {
@@ -2064,7 +2049,7 @@ void edit_replace_cmd (WEdit * edit, int again)
 
 	    replace_yes = 1;
 
-	    if (replace_prompt) {
+	    if (treplace_prompt) {
 		int l;
 		l = edit->curs_row - edit->num_widget_lines / 3;
 		if (l > 0)
@@ -2087,7 +2072,7 @@ void edit_replace_cmd (WEdit * edit, int again)
 		    replace_yes = 0;
 		    break;
 		case B_REPLACE_ALL:
-		    replace_prompt = 0;
+		    treplace_prompt = 0;
 		    replace_continue = 1;
 		    break;
 		case B_REPLACE_ONE:
@@ -2181,7 +2166,6 @@ void edit_search_cmd (WEdit * edit, int again)
 	}
 	return;
     }
-
     exp = old ? old : exp;
     if (again) {		/*ctrl-hotkey for search again. */
 	if (!old)
@@ -2199,31 +2183,60 @@ void edit_search_cmd (WEdit * edit, int again)
 		free (old);
 	    old = (char *) strdup (exp);
 
-	    if (edit->found_len && edit->search_start == edit->found_start + 1 && replace_backwards)
-		edit->search_start--;
-
-	    if (edit->found_len && edit->search_start == edit->found_start - 1 && !replace_backwards)
-		edit->search_start++;
-
-	    edit->search_start = edit_find (edit->search_start, (unsigned char *) exp, &len, edit->last_byte,
-		   (int (*)(void *, long)) edit_get_byte, (void *) edit, 0);
-
-	    if (edit->search_start >= 0) {
-		edit->found_start = edit->search_start;
-		edit->found_len = len;
-
-		edit_cursor_move (edit, edit->search_start - edit->curs1);
-		edit_scroll_screen_over_cursor (edit);
-		if (replace_backwards)
-		    edit->search_start--;
-		else
-		    edit->search_start++;
-	    } else if (edit->search_start == -3) {
-		edit->search_start = edit->curs1;
-		regexp_error (edit);
+	    if (search_create_bookmark) {
+		int found = 0, books = 0;
+		int l = 0, l_last = -1;
+		long p, q = 0;
+		for (;;) {
+		    p = edit_find (q, (unsigned char *) exp, &len, edit->last_byte,
+				   (int (*)(void *, long)) edit_get_byte, (void *) edit, 0);
+		    if (p < 0)
+			break;
+		    found++;
+		    l += edit_count_lines (edit, q, p);
+		    if (l != l_last) {
+			book_mark_insert (edit, l, BOOK_MARK_FOUND_COLOR);
+			books++;
+		    }
+		    l_last = l;
+		    q = p + 1;
+		}
+		if (found) {
+		    char fin_string[64];
+/* in response to number of bookmarks added because of string being found %d times */
+		    sprintf (fin_string, _ (" %d finds made, %d bookmarks added "), found, books);
+		    edit_message_dialog (_ (" Search "), fin_string);
+		} else {
+		    edit_error_dialog (_ (" Search "), _ (" Search string not found. "));
+		}
 	    } else {
-		edit->search_start = edit->curs1;
-		edit_error_dialog (_(" Search "), _(" Search string not found. "));
+
+		if (edit->found_len && edit->search_start == edit->found_start + 1 && replace_backwards)
+		    edit->search_start--;
+
+		if (edit->found_len && edit->search_start == edit->found_start - 1 && !replace_backwards)
+		    edit->search_start++;
+
+		edit->search_start = edit_find (edit->search_start, (unsigned char *) exp, &len, edit->last_byte,
+		(int (*)(void *, long)) edit_get_byte, (void *) edit, 0);
+
+		if (edit->search_start >= 0) {
+		    edit->found_start = edit->search_start;
+		    edit->found_len = len;
+
+		    edit_cursor_move (edit, edit->search_start - edit->curs1);
+		    edit_scroll_screen_over_cursor (edit);
+		    if (replace_backwards)
+			edit->search_start--;
+		    else
+			edit->search_start++;
+		} else if (edit->search_start == -3) {
+		    edit->search_start = edit->curs1;
+		    regexp_error (edit);
+		} else {
+		    edit->search_start = edit->curs1;
+		    edit_error_dialog (_ (" Search "), _ (" Search string not found. "));
+		}
 	    }
 	}
 	free (exp);
@@ -2665,29 +2678,6 @@ int edit_save_block_cmd (WEdit * edit)
 }
 
 
-/* inserts a file at the cursor, returns 1 on success */
-int edit_insert_file (WEdit * edit, const char *filename)
-{
-    int i, file, blocklen;
-    long current = edit->curs1;
-    unsigned char *buf;
-
-    if ((file = open ((char *) filename, O_RDONLY)) == -1)
-	return 0;
-    buf = malloc (TEMP_BUF_LEN);
-    while ((blocklen = read (file, (char *) buf, TEMP_BUF_LEN)) > 0) {
-	for (i = 0; i < blocklen; i++)
-	    edit_insert (edit, buf[i]);
-    }
-    edit_cursor_move (edit, current - edit->curs1);
-    free (buf);
-    close (file);
-    if (blocklen)
-	return 0;
-    return 1;
-}
-
-
 /* returns 1 on success */
 int edit_insert_file_cmd (WEdit * edit)
 {
@@ -2827,7 +2817,7 @@ int edit_print_string (WEdit * e, const char *s)
 {
     int i = 0;
     while (s[i])
-	edit_execute_cmd (e, -1, s[i++]);
+	edit_execute_cmd (e, -1, (unsigned char) s[i++]);
     e->force |= REDRAW_COMPLETELY;
     edit_update_screen (e);
     return i;
