@@ -1,6 +1,6 @@
 /* editor menu definitions and initialisation
 
-   Copyright (C) 1996 the Free Software Foundation
+   Copyright (C) 1996-2000 the Free Software Foundation
 
    Authors: 1996, 1997 Paul Sheer
 
@@ -75,7 +75,7 @@ void edit_about_cmd ()
 		      "\n"
 		      "                Cooledit  v3.11.5\n"
 		      "\n"
-		      " Copyright (C) 1996 the Free Software Foundation\n"
+		      " Copyright (C) 1996-2000 the Free Software Foundation\n"
 		      "\n"
 		      "       A user friendly text editor written\n"
 		      "           for the Midnight Commander.\n"
@@ -113,6 +113,7 @@ void menu_beginning_cmd (void)		{ menu_cmd (CK_Beginning_Of_Text); }
 void menu_end_cmd (void)		{ menu_cmd (CK_End_Of_Text); }
 void menu_refresh_cmd (void)		{ menu_cmd (CK_Refresh); }
 void menu_goto_line (void)		{ menu_cmd (CK_Goto); }
+void menu_goto_bracket (void)		{ menu_cmd (CK_Match_Bracket); }
 void menu_lit_cmd (void)		{ menu_key (XCTRL ('q')); }
 void menu_format_paragraph (void)	{ menu_cmd (CK_Paragraph_Format); }
 void edit_options_dialog (void);
@@ -201,6 +202,7 @@ static menu_entry SearReplMenuEmacs[] =
 static menu_entry CmdMenu[] =
 {
     {' ', N_("&Goto line...            M-l"), 'G', menu_goto_line},
+    {' ', N_("goto matching &Bracket   M-b"), 'B', menu_goto_bracket},
     {' ', "", ' ', 0},
     {' ', N_("insert &Literal...       C-q"), 'L', menu_lit_cmd},
     {' ', "", ' ', 0},
@@ -223,6 +225,7 @@ static menu_entry CmdMenu[] =
 static menu_entry CmdMenuEmacs[] =
 {
     {' ', N_("&Goto line...            M-l"), 'G', menu_goto_line},
+    {' ', N_("goto matching &Bracket   M-b"), 'B', menu_goto_bracket},
     {' ', "", ' ', 0},
     {' ', N_("insert &Literal...       C-q"), 'L', menu_lit_cmd},
     {' ', "", ' ', 0},
@@ -375,7 +378,8 @@ void CEditMenuCommand (int i)
 
 static void menu_key (KeySym i, int state)
 {
-    int cmd, ch;
+    int cmd;
+    long ch;
     if (edit_translate_key (0, i, state, &cmd, &ch)) {
 	if (cmd > 0)
 	    menu_cmd (cmd);
@@ -394,14 +398,14 @@ void CDrawEditMenuButtons (const char *ident, Window parent, Window focus_return
     CDrawMenuButton (catstrs (ident, ".filemenu", 0), parent, focus_return, x, y, AUTO_WIDTH, AUTO_HEIGHT, 8,
 /* The following are menu options. Do not change the key bindings (eg. C-o) and preserve '\t' */
 		     _(" File "),
-		     _("Open...\tC-o"), '~', menu_cmd, (unsigned long) CK_Load,
-		     _("New\tC-n"), '~', menu_cmd, (unsigned long) CK_New,
-		     "", ' ', (void *) 0, 0L,
-		     _("Save\tF2"), '~', menu_cmd, (unsigned long) CK_Save,
-		     _("Save as...\tF12"), '~', menu_cmd, (unsigned long) CK_Save_As,
-		     "", ' ', (void *) 0, 0L,
-		     _("Insert file...\tF15"), '~', menu_cmd, (unsigned long) CK_Insert_File,
-		     _("Copy to file...\tC-f"), '~', menu_cmd, (unsigned long) CK_Save_Block
+		     _("Open...\tC-o"), (int) '~', menu_cmd, (unsigned long) CK_Load,
+		     _("New\tC-n"), (int) '~', menu_cmd, (unsigned long) CK_New,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Save\tF2"), (int) '~', menu_cmd, (unsigned long) CK_Save,
+		     _("Save as...\tF12"), (int) '~', menu_cmd, (unsigned long) CK_Save_As,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Insert file...\tF15"), (int) '~', menu_cmd, (unsigned long) CK_Insert_File,
+		     _("Copy to file...\tC-f"), (int) '~', menu_cmd, (unsigned long) CK_Save_Block
 	);
 /* Tool hint */
     CSetToolHint (catstrs (ident, ".filemenu", 0), _("Disk operations and file indexing/searching"));
@@ -410,26 +414,26 @@ void CDrawEditMenuButtons (const char *ident, Window parent, Window focus_return
 
     CDrawMenuButton (catstrs (ident, ".editmenu", 0), parent, focus_return, x, y, AUTO_WIDTH, AUTO_HEIGHT, 20,
 		     _(" Edit "),
-		     _("Toggle mark\tF3"), '~', menu_cmd, (unsigned long) CK_Mark,
-		     _("Toggle mark columns\tC-b"), '~', menu_cmd, (unsigned long) CK_Column_Mark,
-		     "", ' ', (void *) 0, 0L,
-		     _("Toggle book mark\tC-M-Ins"), '~', menu_cmd, (unsigned long) CK_Toggle_Bookmark,
-		     _("Previous book mark\tC-M-Up"), '~', menu_cmd, (unsigned long) CK_Prev_Bookmark,
-		     _("Next book mark\tC-M-Down"), '~', menu_cmd, (unsigned long) CK_Next_Bookmark,
-		     _("Flush book marks"), '~', menu_cmd, (unsigned long) CK_Flush_Bookmarks,
-		     "", ' ', (void *) 0, 0L,
-		     _("Toggle insert/overwrite\tIns"), '~', menu_cmd, (unsigned long) CK_Toggle_Insert,
-		     "", ' ', (void *) 0, 0L,
-		     _("Copy block to cursor\tF5"), '~', menu_cmd, (unsigned long) CK_Copy,
-		     _("Move block to cursor\tF6"), '~', menu_cmd, (unsigned long) CK_Move,
-		     _("Delete block\tF8/C-Del"), '~', menu_cmd, (unsigned long) CK_Remove,
-		     "", ' ', (void *) 0, 0L,
-		     _("Copy block to clipbrd\tC-Ins"), '~', menu_cmd, (unsigned long) CK_XStore,
-		     _("Cut block to clipbrd\tS-Del"), '~', menu_cmd, (unsigned long) CK_XCut,
-		     _("Paste block from clipbrd\tS-Ins"), '~', menu_cmd, (unsigned long) CK_XPaste,
-		     _("Selection history\tM-Ins"), '~', menu_cmd, (unsigned long) CK_Selection_History,
-		     "", ' ', (void *) 0, 0L,
-		     _("Undo\tC-BackSpace"), '~', menu_cmd, (unsigned long) CK_Undo
+		     _("Toggle mark\tF3"), (int) '~', menu_cmd, (unsigned long) CK_Mark,
+		     _("Toggle mark columns\tC-b"), (int) '~', menu_cmd, (unsigned long) CK_Column_Mark,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Toggle book mark\tC-M-Ins"), (int) '~', menu_cmd, (unsigned long) CK_Toggle_Bookmark,
+		     _("Previous book mark\tC-M-Up"), (int) '~', menu_cmd, (unsigned long) CK_Prev_Bookmark,
+		     _("Next book mark\tC-M-Down"), (int) '~', menu_cmd, (unsigned long) CK_Next_Bookmark,
+		     _("Flush book marks"), (int) '~', menu_cmd, (unsigned long) CK_Flush_Bookmarks,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Toggle insert/overwrite\tIns"), (int) '~', menu_cmd, (unsigned long) CK_Toggle_Insert,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Copy block to cursor\tF5"), (int) '~', menu_cmd, (unsigned long) CK_Copy,
+		     _("Move block to cursor\tF6"), (int) '~', menu_cmd, (unsigned long) CK_Move,
+		     _("Delete block\tF8/C-Del"), (int) '~', menu_cmd, (unsigned long) CK_Remove,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Copy block to clipbrd\tC-Ins"), (int) '~', menu_cmd, (unsigned long) CK_XStore,
+		     _("Cut block to clipbrd\tS-Del"), (int) '~', menu_cmd, (unsigned long) CK_XCut,
+		     _("Paste block from clipbrd\tS-Ins"), (int) '~', menu_cmd, (unsigned long) CK_XPaste,
+		     _("Selection history...\tM-Ins"), (int) '~', menu_cmd, (unsigned long) CK_Selection_History,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Undo\tC-BackSpace"), (int) '~', menu_cmd, (unsigned long) CK_Undo
 	);
 /* Tool hint */
     CSetToolHint (catstrs (ident, ".editmenu", 0), _("Manipulating blocks of text"));
@@ -438,29 +442,30 @@ void CDrawEditMenuButtons (const char *ident, Window parent, Window focus_return
 
     CDrawMenuButton (catstrs (ident, ".searchmenu", 0), parent, focus_return, x, y, AUTO_WIDTH, AUTO_HEIGHT, 4,
 		     _(" Srch/Replce "),
-		     _("Search...\tF7"), '~', menu_cmd, (unsigned long) CK_Find,
-		     _("Search again\tF17"), '~', menu_cmd, (unsigned long) CK_Find_Again,
-		     _("Replace...\tF4"), '~', menu_cmd, (unsigned long) CK_Replace,
-		     _("Replace again\tF14"), '~', menu_cmd, (unsigned long) CK_Replace_Again
+		     _("Search...\tF7"), (int) '~', menu_cmd, (unsigned long) CK_Find,
+		     _("Search again\tF17"), (int) '~', menu_cmd, (unsigned long) CK_Find_Again,
+		     _("Replace...\tF4"), (int) '~', menu_cmd, (unsigned long) CK_Replace,
+		     _("Replace again\tF14"), (int) '~', menu_cmd, (unsigned long) CK_Replace_Again
 	);
 /* Tool hint */
     CSetToolHint (catstrs (ident, ".searchmenu", 0), _("Search for and replace text"));
 
     CGetHintPos (&x, &d);
 
-    CDrawMenuButton (catstrs (ident, ".commandmenu", 0), parent, focus_return, x, y, AUTO_WIDTH, AUTO_HEIGHT, 11,
+    CDrawMenuButton (catstrs (ident, ".commandmenu", 0), parent, focus_return, x, y, AUTO_WIDTH, AUTO_HEIGHT, 12,
 		     _(" Command "),
-		     _("Goto line...\tM-l"), '~', menu_cmd, (unsigned long) CK_Goto,
-		     "", ' ', (void *) 0, 0L,
-		     _("Start record macro\tC-r"), '~', menu_cmd, (unsigned long) CK_Begin_Record_Macro,
-		     _("Finish record macro...\tC-r"), '~', menu_cmd, (unsigned long) CK_End_Record_Macro,
-		     _("Execute macro...\tC-a, KEY"), '~', menu_ctrl_key, (unsigned long) XK_a,
-		     _("Delete macro...\t"), '~', menu_cmd, (unsigned long) CK_Delete_Macro,
-		     "", ' ', (void *) 0, 0L,
-		     _("Insert date/time\tC-d"), '~', menu_cmd, (unsigned long) CK_Date,
-		     _("Format paragraph\tM-p"), '~', menu_cmd, (unsigned long) CK_Paragraph_Format,
-		     "", ' ', (void *) 0, 0L,
-		     _("Refresh display\tC-l"), '~', menu_cmd, (unsigned long) CK_Refresh
+		     _("Goto line...\tM-l"), (int) '~', menu_cmd, (unsigned long) CK_Goto,
+		     _("Goto matching bracket\tM-b"), (int) '~', menu_cmd, (unsigned long) CK_Match_Bracket,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Start record macro\tC-r"), (int) '~', menu_cmd, (unsigned long) CK_Begin_Record_Macro,
+		     _("Finish record macro...\tC-r"), (int) '~', menu_cmd, (unsigned long) CK_End_Record_Macro,
+		     _("Execute macro...\tC-a, KEY"), (int) '~', menu_ctrl_key, (unsigned long) XK_a,
+		     _("Delete macro...\t"), (int) '~', menu_cmd, (unsigned long) CK_Delete_Macro,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Insert date/time\tC-d"), (int) '~', menu_cmd, (unsigned long) CK_Date,
+		     _("Format paragraph\tM-p"), (int) '~', menu_cmd, (unsigned long) CK_Paragraph_Format,
+		     "", (int) ' ', (void *) 0, 0L,
+		     _("Refresh display\tC-l"), (int) '~', menu_cmd, (unsigned long) CK_Refresh
 	);
 /* Tool hint */
     CSetToolHint (catstrs (ident, ".commandmenu", 0), _("Macros and internal commands"));
