@@ -1,5 +1,5 @@
 /* fieldedtextbox.c - for drawing a scrollable text window widget
-   Copyright (C) 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -44,9 +44,7 @@
 #include <stdlib.h>
 #include <stdarg.h>
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xresource.h>
+#include <X11/Intrinsic.h>
 #include <X11/Xatom.h>
 #include "lkeysym.h"
 
@@ -62,6 +60,8 @@
 
 #include "mad.h"
 
+extern struct look *look;
+
 extern int option_text_fg_normal;
 extern int option_text_fg_bold;
 extern int option_text_fg_italic;
@@ -78,7 +78,6 @@ extern int option_text_bg_highlighted;
 
 #define pixmap_width(x) 0
 
-extern unsigned char per_char[256];
 int eh_fielded_textbox (CWidget * w, XEvent * xevent, CEvent * cwevent);
 static int calc_text_pos_fielded_textbox (CWidget * w, long b, long *q, int l);
 
@@ -95,7 +94,7 @@ static int this_text_width (char *s)
 	else if (*p == '\f')
 	    l += *(++p);
 	else if (this_is_printable (*p))
-	    l += per_char[(unsigned char) *p];
+	    l += FONT_PER_CHAR((unsigned char) *p);
     }
     return l;
 }
@@ -214,7 +213,7 @@ static void compose_line (void *data, int line_number, unsigned char *line, int
     *line = 0;
 }
 
-unsigned char *compose_line_cached (void *data, int l, int *tab, char **(*get_line) (void *, int, int *, int *), int *tagged)
+static unsigned char *compose_line_cached (void *data, int l, int *tab, char **(*get_line) (void *, int, int *, int *), int *tagged)
 {
     static unsigned char line[4096];
     static int c_tagged, c_l = -1;
@@ -272,13 +271,13 @@ void link_scrollbar_to_fielded_textbox (CWidget * scrollbar, CWidget * textbox,
     count = count_fielded_textbox_lines (textbox);
     if (!count)
 	count = 1;
-    scrollbar->firstline = (double) 65535.0 *textbox->firstline / textbox->numlines;
-    scrollbar->numlines = (double) 65535.0 *count / textbox->numlines;
+    scrollbar->firstline = (double) 65535.0 *textbox->firstline / (textbox->numlines ? textbox->numlines : 1);
+    scrollbar->numlines = (double) 65535.0 *count / (textbox->numlines ? textbox->numlines : 1);
 }
 
 void link_h_scrollbar_to_fielded_textbox (CWidget * scrollbar, CWidget * textbox, XEvent * xevent, CEvent * cwevent, int whichscrbutton)
 {
-    int redrawtext = 0, count, c;
+    int redrawtext = 0, c;
     static int r = 0;
     if ((xevent->type == ButtonRelease || xevent->type == MotionNotify) && whichscrbutton == 3) {
 	redrawtext = CSetTextboxPos (textbox, TEXT_SET_COLUMN, (double) scrollbar->firstline * (textbox->column / FONT_MEAN_WIDTH) / 65535.0);
@@ -314,9 +313,6 @@ void link_h_scrollbar_to_fielded_textbox (CWidget * scrollbar, CWidget * textbox
 	    r = 0;
 	}
     }
-    count = count_fielded_textbox_lines (textbox);
-    if (!count)
-	count = 1;
     scrollbar->firstline = (double) 65535.0 *(textbox->firstcolumn * FONT_MEAN_WIDTH) / textbox->column;
     scrollbar->numlines = (double) 65535.0 *(textbox->width - 6) / textbox->column;
 }
@@ -381,6 +377,7 @@ static char *get_block (CWidget * w, long start_mark, long end_mark, int *type,
     unsigned char c, *t;
     long x, y, a, b;
     void *data;
+    CPushFont ("editor", 0);
 
     a = min (w->mark2, w->mark1);
     b = max (w->mark2, w->mark1);
@@ -418,7 +415,7 @@ static char *get_block (CWidget * w, long start_mark, long end_mark, int *type,
 		    break;
 		}
     		j = text[++x];
-		while ((j -= per_char[' ']) > 0)
+		while ((j -= FONT_PER_CHAR(' ')) > 0)
 		    pool_write (p, (unsigned char *) " ", 1);
 		pool_write (p, (unsigned char *) " ", 1);
 		continue;
@@ -426,7 +423,7 @@ static char *get_block (CWidget * w, long start_mark, long end_mark, int *type,
 	    if (c == '\v') {
 		int j;
     		j = pixmap_width (text[++x]);
-		while ((j -= per_char[' ']) > 0)
+		while ((j -= FONT_PER_CHAR(' ')) > 0)
 		    pool_write (p, (unsigned char *) " ", 1);
 		continue;
 	    }
@@ -436,6 +433,7 @@ static char *get_block (CWidget * w, long start_mark, long end_mark, int *type,
 	x = 0;
     }
   done:
+    CPopFont ();
 #ifdef HAVE_DND
     *type = DndText;
 #endif
@@ -505,6 +503,7 @@ CWidget *CDrawFieldedTextbox (const char *identifier, Window parent, int x, int
     int *tab, wf;
 
     int w, h, x_hint, y_hint;
+    CPushFont ("editor", 0);
 
     tab = get_field_sizes (data, &numlines, &wf, get_line);
 
@@ -547,7 +546,7 @@ CWidget *CDrawFieldedTextbox (const char *identifier, Window parent, int x, int
 
     if (w > 80 && width != AUTO_WIDTH) {
 	wdt->hori_scrollbar = CDrawHorizontalScrollbar (scroll = catstrs (identifier, ".hsc", 0), parent,
-			 x, y + h + WIDGET_SPACING, w, 12, 0, 0);
+			 x, y + h + WIDGET_SPACING, w, (*look->get_fielded_textbox_hscrollbar_width) (), 0, 65535);
 	CSetScrollbarCallback (wdt->hori_scrollbar->ident, wdt->ident, link_h_scrollbar_to_fielded_textbox);
 	CGetHintPos (0, &y_hint);
     } else {
@@ -555,6 +554,7 @@ CWidget *CDrawFieldedTextbox (const char *identifier, Window parent, int x, int
     }
 
     set_hint_pos (x_hint, y_hint);
+    CPopFont ();
     return wdt;
 }
 
@@ -565,6 +565,7 @@ CWidget *CRedrawFieldedTextbox (const char *identifier, int preserve)
     CWidget *wdt;
     int *tab, w;
 
+    CPushFont ("editor", 0);
     wdt = CIdent (identifier);
     tab = get_field_sizes (wdt->hook, &numlines, &w, wdt->get_line);
 
@@ -584,6 +585,7 @@ CWidget *CRedrawFieldedTextbox (const char *identifier, int preserve)
     CRectangle (wdt->winid, 3, 3, wdt->width - 3 - 1, wdt->height - 3 - 1);
     CExpose (identifier);
 
+    CPopFont ();
     return wdt;
 }
 
@@ -620,7 +622,7 @@ static int calc_text_pos_fielded_textbox (CWidget *w, long b, long *q, int l)
 	    break;
 	case '\b':
 	case '\r':
-	    xn = x + per_char[text[k + 1]];
+	    xn = x + FONT_PER_CHAR(text[k + 1]);
 	    if (xn > l) {
 		*q = b;
 		return x;
@@ -640,7 +642,7 @@ static int calc_text_pos_fielded_textbox (CWidget *w, long b, long *q, int l)
 	    b++;
 	    break;
 	default:
-	    xn = x + per_char[c];
+	    xn = x + FONT_PER_CHAR(c);
 	    break;
 	}
 	if (xn > l)
@@ -676,32 +678,33 @@ static void convert_text_fielded_textbox (CWidget * w, long q, cache_type *line,
 	k = 0;
 
     p = line;
-    *p = 0;
+    p->c.ch = p->style = 0;
     for (;;) {
 	c = text[k];
-	*(p + 1) = 0;
-	*p |= 0xFFFF0000UL;	/* default background colors */
+	p[1].style = p[1].c.ch = 0;
+	p->c.fg = p->c.bg = 0xFF;	/* default background colors */
 	if (highlight_this_line)
-	    *p |= MOD_HIGHLIGHTED * 256;
+	    p->c.style |= MOD_HIGHLIGHTED;
 	if (tagged)
-	    *p |= MOD_INVERSE * 256;
+	    p->c.style |= MOD_INVERSE;
 	if (q >= m1 && q < m2)
-	    *p |= MOD_MARKED * 256;
+	    p->c.style |= MOD_MARKED;
 	switch (c) {
 	case '\0':
 	case '\n':
-	    *p++ |= ' ';
+	    (p++)->c.ch = ' ';
 	    if (highlight_this_line || tagged) {
 		q--;
 		k--;
-		x += per_char[' '];
+		x += FONT_PER_CHAR(' ');
 	    } else
 		return;
 	    break;
 	case '\f':
 	    k++;
 	    q++;
-	    *p++ |= MOD_ABNORMAL * 256 | text[k];
+	    p->c.style |= MOD_TAB;
+	    (p++)->c.ch = text[k];
 	    x += text[k];
 	    break;
 	case '\b':
@@ -713,17 +716,17 @@ static void convert_text_fielded_textbox (CWidget * w, long q, cache_type *line,
 	case '\v':
 	    k++;
 	    q++;
-	    *p++ |= MOD_ABNORMAL * 256 | 128 | text[k];
+	    p->c.style |= MOD_TAB;
+	    (p++)->c.ch = text[k];
 	    x += pixmap_width (text[k]);
 	    break;
 	default:
-	    x += per_char[c];
-	    *p &= 0xFFFFFF00UL;
-	    *p |= c;
+	    x += FONT_PER_CHAR(c);
+	    p->c.ch = c;
 	    if (italic > 0)
-		*p |= MOD_ITALIC * 256;
+		p->c.style |= MOD_ITALIC;
 	    if (bold > 0)
-		*p |= MOD_BOLD * 256;
+		p->c.style |= MOD_BOLD;
 	    p++;
 	    break;
 	}
@@ -734,7 +737,7 @@ static void convert_text_fielded_textbox (CWidget * w, long q, cache_type *line,
 	q++;
 	k++;
     }
-    *p = 0;
+    p->c.ch = p->style = 0;
 }
 
 
@@ -763,20 +766,20 @@ static void fielded_text_print_line (CWidget * w, long b, int row)
 extern int EditExposeRedraw;
 extern int EditClear;
 extern unsigned long edit_normal_background_color;
-extern int mode_spacing;
 
 void render_fielded_textbox (CWidget * w, int redrawall)
 {
     int row, height, isfocussed, curs, i, x;
     static Window last_win = 0;
     static int last_firstcolumn = 0;
+    CPushFont ("editor", 0);
     if (redrawall) {
 	EditExposeRedraw = 1;
 	EditClear = 1;
     }
     if (last_win == w->winid && last_firstcolumn != w->firstcolumn) {
 	x = 0;
-	CSetColor (color_palette (option_text_bg_normal));
+        CSetColor (color_palette (option_text_bg_normal));
 	for (i = 0;; i++) {
 	    x += w->tab[i];
 	    if (x >= w->column)
@@ -786,7 +789,6 @@ void render_fielded_textbox (CWidget * w, int redrawall)
     }
     last_firstcolumn = w->firstcolumn;
     last_win = w->winid;
-    mode_spacing = MOD_ABNORMAL * 256;
 
     height = w->height / FONT_PIX_PER_LINE;
 
@@ -814,7 +816,7 @@ void render_fielded_textbox (CWidget * w, int redrawall)
     }
     if ((w->numlines - w->firstline) * FONT_PIX_PER_LINE < w->height) {
 	x = 0;
-	CSetColor (color_palette (option_text_bg_normal));
+        CSetColor (color_palette (option_text_bg_normal));
 	for (i = 0;; i++) {
 	    x += w->tab[i];
 	    if (x >= w->column)
@@ -822,18 +824,13 @@ void render_fielded_textbox (CWidget * w, int redrawall)
 	    CLine (w->winid, x + LINE_OFFSET - (w->firstcolumn * FONT_MEAN_WIDTH), (w->numlines - w->firstline) * FONT_PIX_PER_LINE + 3, x + LINE_OFFSET - (w->firstcolumn * FONT_MEAN_WIDTH), w->height - 3);
 	}
     }
-    mode_spacing = 0;
     EditExposeRedraw = 0;
     EditClear = 0;
 
-    if (isfocussed) {
-	render_bevel (w->winid, 0, 0, w->width - 1, w->height - 1, 3, 1);	/*most outer border bevel */
-    } else {
-	render_bevel (w->winid, 2, 2, w->width - 3, w->height - 3, 1, 1);	/*border bevel */
-	render_bevel (w->winid, 0, 0, w->width - 1, w->height - 1, 2, 0);	/*most outer border bevel */
-    }
-    CSetColor (edit_normal_background_color);
-    CLine (w->winid, 3, 3, 3, w->height - 4);
+    (*look->render_fielded_textbox_tidbits) (w, isfocussed);
+
+    CPopFont ();
+    return;
 }
 
 /*
@@ -852,7 +849,9 @@ static long count_fielded_textbox_lines (CWidget * wdt)
 
 static void fielded_text_mouse_mark (CWidget * w, XEvent * event, CEvent * ce)
 {
+    CPushFont ("editor", 0);
     mouse_mark (event, ce->double_click, w->funcs);
+    CPopFont ();
 }
 
 
@@ -885,6 +884,7 @@ int eh_fielded_textbox (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	w->mark1 = w->mark2 = 0;
 	break;
     case ButtonPress:
+	CPushFont ("editor", 0);
 	CFocus (w);
 	if (xevent->xbutton.button == Button1)
 	    w->cursor = (xevent->xbutton.y - BDR) / FONT_PIX_PER_LINE + w->firstline;
@@ -895,6 +895,7 @@ int eh_fielded_textbox (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	cwevent->ident = w->ident;
 	cwevent->xt = (xevent->xbutton.x - 7) / FONT_MEAN_WIDTH + w->firstcolumn;
 	cwevent->yt = w->cursor;
+	CPopFont ();
     case ButtonRelease:
     case MotionNotify:
 	if (!xevent->xmotion.state && xevent->type == MotionNotify)
@@ -937,8 +938,8 @@ int eh_fielded_textbox (CWidget * w, XEvent * xevent, CEvent * cwevent)
 
 /* now update the scrollbar position */
     if (w->vert_scrollbar && w->numlines) {
-	w->vert_scrollbar->firstline = (double) 65535.0 *w->firstline / w->numlines;
-	w->vert_scrollbar->numlines = (double) 65535.0 *count / w->numlines;
+	w->vert_scrollbar->firstline = (double) 65535.0 *w->firstline / (w->numlines ? w->numlines : 1);
+	w->vert_scrollbar->numlines = (double) 65535.0 *count / (w->numlines ? w->numlines : 1);
 	w->vert_scrollbar->options = 0;
 	render_scrollbar (w->vert_scrollbar);
     }
