@@ -1,6 +1,6 @@
 /* gtkedit.c - front end for gtk/gnome version
 
-   Copyright (C) 1996, 1997 the Free Software Foundation
+   Copyright (C) 1996-2000 the Free Software Foundation
 
    Authors: 1996, 1997 Paul Sheer
 
@@ -107,18 +107,18 @@ void gtk_edit_configure_font_dimensions (GtkEdit * edit)
     p = _ ("The Quick Brown Fox Jumps Over The Lazy Dog");
     for (i = ' '; i <= '~'; i++)
 	q[i - ' '] = i;
-    if (XTextWidth (f, "M", 1) == XTextWidth (f, "M", 1))
+    if (XmbTextEscapement (f, "M", 1) == XmbTextEscapement (f, "M", 1))
 	gtk_edit_fixed_font = 1;
     else
 	gtk_edit_fixed_font = 0;
     XTextExtents (f, q, '~' - ' ', &direction, &gtk_edit_option_font_ascent, &gtk_edit_option_font_descent, &s);
-    gtk_edit_option_font_mean_width = XTextWidth (f, p, strlen (p)) / strlen (p);
+    gtk_edit_option_font_mean_width = XmbTextEscapement (f, p, strlen (p)) / strlen (p);
     for (i = 0; i < 256; i++) {
 	t = (unsigned char) i;
 	if (i > f->max_char_or_byte2 || i < f->min_char_or_byte2) {
 	    gtk_edit_font_width_per_char[i] = 0;
 	} else {
-	    gtk_edit_font_width_per_char[i] = XTextWidth (f, (char *) &t, 1);
+	    gtk_edit_font_width_per_char[i] = XmbTextEscapement (f, (char *) &t, 1);
 	}
     }
 }
