@@ -1,5 +1,5 @@
 /* initapp.c - initialise X Server connection, X application, and widget library
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,34 +20,34 @@
 /* setup application */
 
 /*
-Colormap allocation:
+   Colormap allocation:
 
-A colormap is allocated as follows into the array color_pixels:
-('i' refers to color_pixels[i])
+   A colormap is allocated as follows into the array color_pixels:
+   ('i' refers to color_pixels[i])
 
-These are allocated in sequential palette cells, hence
-(for pseudocolor only) color_pixels[j + i] = color_pixels[j] + i,
-for j = 0,16,43. Obviously in TrueColor this is not true.
+   These are allocated in sequential palette cells, hence
+   (for pseudocolor only) color_pixels[j + i] = color_pixels[j] + i,
+   for j = 0,16,43. Obviously in TrueColor this is not true.
 
-,----------+------------------------------------------.
-|    i     |   colors                                 |
-+----------+------------------------------------------+
-|  0-15    | 16 levels of the widget colors that      |
-|          | make up button bevels etc. Starting from |
-|          | (i=0) black (for shadowed bevels), up    |
-|          | to (i=15) bright highlighted bevels      |
-|          | those facing up-to-the-left. These       |
-|          | are sequential (see next block).         |
-+----------+------------------------------------------+
-|  16-42   | 3^3 combinations of RGB, vis. (0,0,0),   |
-|          | (0,0,127), (0,0,255), (0,127,0), ...     |
-|          | ... (255,255,255).                       |
-+----------+------------------------------------------+
-|  43-106  | 64 levels of grey. (optional)            |
-+----------+------------------------------------------+
-|  107->   | For other colors. (Not used at present)  |
-`----------+------------------------------------------'
-*/
+   ,----------+------------------------------------------.
+   |    i     |   colors                                 |
+   +----------+------------------------------------------+
+   |  0-15    | 16 levels of the widget colors that      |
+   |          | make up button bevels etc. Starting from |
+   |          | (i=0) black (for shadowed bevels), up    |
+   |          | to (i=15) bright highlighted bevels      |
+   |          | those facing up-to-the-left. These       |
+   |          | are sequential (see next block).         |
+   +----------+------------------------------------------+
+   |  16-42   | 3^3 combinations of RGB, vis. (0,0,0),   |
+   |          | (0,0,127), (0,0,255), (0,127,0), ...     |
+   |          | ... (255,255,255).                       |
+   +----------+------------------------------------------+
+   |  43-106  | 64 levels of grey. (optional)            |
+   +----------+------------------------------------------+
+   |  107->   | For other colors. (Not used at present)  |
+   `----------+------------------------------------------'
+ */
 
 /* Thence macros are defined in coolwidgets.h for color lookup */
 
@@ -72,9 +72,14 @@ for j = 0,16,43. Obviously in TrueColor this is not true.
 #define  DEF_APP_GLOB		/* so that globals get defined not externed */
 
 #include "coolwidget.h"
+#include "xim.h"
 #include "stringtools.h"
 
 #include <signal.h>
+#include <sys/types.h>
+#if HAVE_SYS_WAIT_H
+# include <sys/wait.h>
+#endif
 
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -89,7 +94,8 @@ CInitData *given = 0;
 /* defaults */
 #define DEFAULT_DISPLAY			NULL
 #define DEFAULT_GEOM			NULL
-#define DEFAULT_FONT			"8x13bold"
+#define DEFAULT_FONT			"-*-fixed-bold-r-*--13-*-*-*-*-80-*"
+#define FONT_FALLBACK			"-*"
 #define DEFAULT_BG_COLOR		0 /* not used *************/ 
 #define DEFAULT_WIDGET_COLOR_R		"0.9"
 #define DEFAULT_WIDGET_COLOR_G		"1.1"
@@ -97,18 +103,28 @@ CInitData *given = 0;
 
 #define DEFAULT_BDWIDTH         1
 
+struct look *look = 0;
+#ifdef NEXT_LOOK
+extern struct look look_next;
+#endif
+extern struct look look_cool;
+extern struct look look_gtk;
+
 struct resource_param {
     char *name;
     char **value;
 };
 
 static char *init_display = DEFAULT_DISPLAY;
-       char *init_geometry = DEFAULT_GEOM;
-       char *init_font = DEFAULT_FONT;
-       char *init_bg_color = DEFAULT_BG_COLOR;
-static char *init_fg_color_red = DEFAULT_WIDGET_COLOR_R;
-static char *init_fg_color_green = DEFAULT_WIDGET_COLOR_G;
-static char *init_fg_color_blue = DEFAULT_WIDGET_COLOR_B;
+char *init_geometry = DEFAULT_GEOM;
+char *init_font = DEFAULT_FONT;
+char *init_widget_font = 0;
+char *init_bg_color = DEFAULT_BG_COLOR;
+char *init_fg_color_red = DEFAULT_WIDGET_COLOR_R;
+char *init_fg_color_green = DEFAULT_WIDGET_COLOR_G;
+char *init_fg_color_blue = DEFAULT_WIDGET_COLOR_B;
+char *init_look = "gtk";
+
 
 Atom ATOM_WM_PROTOCOLS, ATOM_WM_DELETE_WINDOW;
 Atom ATOM_WM_NAME, ATOM_WM_NORMAL_HINTS, ATOM_WM_TAKE_FOCUS;
@@ -120,6 +136,7 @@ struct resource_param resources[] =
     {"display", &init_display},
     {"geometry", &init_geometry},
     {"font", &init_font},
+    {"widget_font", &init_widget_font},
     {"fg_red", &init_fg_color_red},
     {"fg_blue", &init_fg_color_blue},
     {"fg_green", &init_fg_color_green},
@@ -129,7 +146,7 @@ struct resource_param resources[] =
 static void alloccolorerror (void)
 {
 /* Translations in initapp.c are of a lower priority than other files, since they only output when cooledit is run in verbose mode */
-    fprintf (stderr, _("Cannot allocate colors. Could be to many applications\ntrying to use the colormap. If closing other apps doesn't\nhelp, then your graphics hardware may be inadequite.\n"));
+    fprintf (stderr, _ ("Cannot allocate colors. Could be to many applications\ntrying to use the colormap. If closing other apps doesn't\nhelp, then your graphics hardware may be inadequite.\n"));
     exit (1);
 }
 
@@ -139,7 +156,7 @@ static void init_widgets (void)
     last_widget = 1;		/*widget[0] is never used since index 0 is used
 				   to indicate an error message */
     for (i = 0; i < MAX_NUMBER_OF_WIDGETS; i++)
-	CIndex (i) = NULL;		/*initialise */
+	CIndex (i) = NULL;	/*initialise */
 }
 
 static void open_display (char *app_name, int wait_for_display)
@@ -179,60 +196,16 @@ static void get_resources (void)
     }
 }
 
-unsigned char per_char[256] =
-{255};
-unsigned char per_char_descent[256] =
-{255};
-
-
 static void load_font (void)
 {
-    char *p;
-    char q[256];
-    unsigned char t;
-    int i, direction, m;
-    XCharStruct s;
-
-/* This is used to calculate the mean font width. It need not be translated for iso8859-1 character sets */
-    p = _ ("The Quick Brown Fox Jumps Over The Lazy Dog");
-    for (i = ' '; i <= '~'; i++)
-	q[i - ' '] = i;
-
-    if ((CFontStruct = XLoadQueryFont (CDisplay, init_font)) == NULL) {
-/* 'Display' is the X Server to show the application on */
-	fprintf (stderr, _ ("%s: display %s cannot load font %s\n"),
-		 CAppName, DisplayString (CDisplay), init_font);
+    char f[256];
+    if (CPushFont ("editor", init_font))
 	exit (1);
-    }
-    fixed_font = 1;
-    m = XTextWidth (CFontStruct, "M", 1);
-    if (m != XTextWidth (CFontStruct, "i", 1))
-	fixed_font = 0;
-    if (m != XTextWidth (CFontStruct, " ", 1))
-	fixed_font = 0;
-    if (m != XTextWidth (CFontStruct, ".", 1))
-	fixed_font = 0;
-    if (m != XTextWidth (CFontStruct, "1", 1))
-	fixed_font = 0;
-    if (m != XTextWidth (CFontStruct, "@", 1))
-	fixed_font = 0;
-    if (m != XTextWidth (CFontStruct, "~", 1))
-	fixed_font = 0;
-    XTextExtents (CFontStruct, q, '~' - ' ', &direction, &font_ascent, &font_descent, &s);
-    mean_font_width = CStringWidth (p) / strlen (p);
-    for (i = 0; i < 256; i++) {
-	t = (unsigned char) i;
-	if (i > CFontStruct->max_char_or_byte2 || i < CFontStruct->min_char_or_byte2) {
-	    per_char[i] = 0;
-	    per_char_descent[i] = 0;
-	} else {
-	    per_char[i] = XTextWidth (CFontStruct, (char *) &t, 1);
-	    per_char_descent[i] = CFontStruct->per_char[t - CFontStruct->min_char_or_byte2].descent;
-	    if (fixed_font && (per_char[i] != m)) {
-		per_char[i] = 0;
-		per_char_descent[i] = 0;
-	    }
-	}
+    sprintf (f, init_widget_font, FONT_HEIGHT - 1);
+    if (CPushFont ("widget", f)) {
+	fprintf (stderr, _ ("%s: falling back to font %s\n"), CAppName, FONT_FALLBACK);
+	if (CPushFont ("widget", FONT_FALLBACK))
+	    exit (1);
     }
 }
 
@@ -243,45 +216,45 @@ static void visual_comments (int class)
 	printf ("PseudoColor");
 	if (CDepth >= 7)
 /* 'Depth' is the number of color bits per graphics pixel */
-	    printf (_(" - depth ok, this will work.\n"));
+	    printf (_ (" - depth ok, this will work.\n"));
 	else
 /* 'Depth' is the number of color bits per graphics pixel */
-	    printf (_(" - depth low, this may not work.\n"));
+	    printf (_ (" - depth low, this may not work.\n"));
 	break;
     case GrayScale:
 	printf ("Grayscale -\n");
 /* 'Visual' is the hardware method of displaying graphics */
-	printf (_("Mmmmh, haven't tried this visual class, let's see what happens.\n"));
+	printf (_ ("Mmmmh, haven't tried this visual class, let's see what happens.\n"));
 	break;
     case DirectColor:
 	printf ("DirectColor -\n");
 /* 'Visual' is the hardware method of displaying graphics */
-	printf (_("Mmmmh, haven't tried this visual class, let's see what happens.\n"));
+	printf (_ ("Mmmmh, haven't tried this visual class, let's see what happens.\n"));
 	break;
     case StaticColor:
 	printf ("StaticColor - ");
 /* "Let us attempt to use this Visual even though it may not work" */
-	printf (_("lets give it a try.\n"));
+	printf (_ ("lets give it a try.\n"));
 	break;
     case StaticGray:
 	printf ("StaticGray - ");
 /* "Let us attempt to use this Visual even though it may not work" */
-	printf (_("lets give it a try.\n"));
+	printf (_ ("lets give it a try.\n"));
 	break;
     case TrueColor:
 	printf ("TrueColor - ");
 /* "Adequite" (with sarcasm) : i.e. it is actually the best kind of Visual " */
-	printf (_("fine.\n"));
+	printf (_ ("fine.\n"));
 	break;
     default:
 /* 'Visual' is the method hardware method of displaying graphics */
-	CError (_("?\nVisual class unknown.\n"));
+	CError (_ ("?\nVisual class unknown.\n"));
 	break;
     }
 }
 
 /* must be free'd */
-XColor * get_cells(Colormap cmap, int *size)
+XColor *get_cells (Colormap cmap, int *size)
 {
     int i;
     XColor *c;
@@ -340,88 +313,197 @@ int find_coolwidget_grey_scale (XColor * c, int ncells)
 
 
 void CAllocColorCells (Colormap colormap, Bool contig,
-	unsigned long plane_masks[], unsigned int nplanes,
-	unsigned long pixels[], unsigned int npixels)
+		       unsigned long plane_masks[], unsigned int nplanes,
+		       unsigned long pixels[], unsigned int npixels)
 {
-    if(!XAllocColorCells (CDisplay, colormap, contig,
-		plane_masks, nplanes, pixels, npixels))
+    if (!XAllocColorCells (CDisplay, colormap, contig,
+			   plane_masks, nplanes, pixels, npixels))
 	alloccolorerror ();
 }
 
-void CAllocColor (Colormap cmap, XColor *c)
+void CAllocColor (Colormap cmap, XColor * c)
 {
     if (!XAllocColor (CDisplay, cmap, c))
 	alloccolorerror ();
 }
 
-/* the 16 coolwidget widget colors: (provided the visual is sufficient) */
-static void get_button_color (XColor * color, int i)
+static void get_grey_colors (XColor * color, int i)
 {
-    double r, g, b, min_wc;
-
-    r = 1 / atof (init_fg_color_red);
-    g = 1 / atof (init_fg_color_green);
-    b = 1 / atof (init_fg_color_blue);
-
-    min_wc = min (r, min (g, b));
-
-    color->red = (float) 65535 *my_pow ((float) i / 20, r) * my_pow (0.75, -min_wc);
-    color->green = (float) 65535 *my_pow ((float) i / 20, g) * my_pow (0.75, -min_wc);
-    color->blue = (float) 65535 *my_pow ((float) i / 20, b) * my_pow (0.75, -min_wc);
+    color->red = color->green = grey_intense (i);
+    color->blue = grey_intense (i);
     color->flags = DoRed | DoBlue | DoGreen;
 }
 
-struct color_matrix_struct {
-    unsigned int R, G, B;
-} color_matrix[27] =
+static void get_button_color (XColor * color, int i)
 {
-    {0, 0, 0},
-    {0, 0, 128},
-    {0, 0, 255},
-    {0, 139, 0},
-    {0, 139, 139},
-    {0, 154, 205},
-    {0, 255, 0},
-    {0, 250, 154},
-    {0, 255, 255},
-    {139, 37, 0},
-    {139, 0, 139},
-    {125, 38, 205},
-    {139, 117, 0},
-    {127, 127, 127},
-    {123, 104, 238},
-    {127, 255, 0},
-    {135, 206, 235},
-    {127, 255, 212},
-    {238, 0, 0},
-    {238, 18, 137},
-    {238, 0, 238},
-    {205, 102, 0},
-/*    {255, 127, 80},*/
-    {248, 183, 183},
-    {224, 102, 255},
-    {238, 238, 0},
-    {238, 230, 133},
-    {248, 248, 255}
-};
+    (*look->get_button_color) (color, i);
+}
+
+int option_invert_colors = 0;
+int option_invert_crome = 0;
+int option_invert_red_green = 0;
+int option_invert_green_blue = 0;
+int option_invert_red_blue = 0;
+
+#define clip(x,a,b) ((x) >= (b) ? (b) : ((x) <= (a) ? (a) : (x)))
+#define cswap(a,b) {float t; t = (a); (a) = (b); (b) = t;}
+
+/* inverts the cromiance - this is for editor background colors that are very light */
+static int transform (int color)
+{
+    float r, g, b, y, y_max, c1, c1_max, c2, c2_max;
+    r = (float) ((color >> 16) & 0xFF);
+    g = (float) ((color >> 8) & 0xFF);
+    b = (float) ((color >> 0) & 0xFF);
+    y_max = 0.3 * 240.0 + 0.6 * 240.0 + 0.1 * 240.0;
+    if (option_invert_red_green)
+	cswap (r, g)
+	    if (option_invert_green_blue)
+	    cswap (g, b)
+		if (option_invert_red_blue)
+		cswap (r, b)
+		    y = 0.3000 * r + 0.6000 * g + 0.1000 * b;
+    c1 = -0.1500 * r - 0.3000 * g + 0.4500 * b;
+    c2 = 0.4375 * r - 0.3750 * g - 0.0625 * b;
+    c1_max = -0.1500 * 255.0 - 0.3000 * 255.0 + 0.4500 * 255.0;
+    c2_max = 0.4375 * 255.0 - 0.3750 * 255.0 - 0.0625 * 255.0;
+    if (option_invert_crome) {
+	c1 = c1_max - c1;
+	c2 = c2_max - c2;
+    }
+    if (option_invert_colors)
+	y = y_max - y;
+    r = 1.0 * y + 0.0000 * c1 + 1.6 * c2;
+    g = 1.0 * y - 0.3333 * c1 - 0.8 * c2;
+    b = 1.0 * y + 2.0000 * c1 + 0.0 * c2;
+    r = clip (r, 0.0, 255.0);
+    g = clip (g, 0.0, 255.0);
+    b = clip (b, 0.0, 255.0);
+    return (int) (((int) r) << 16) | (((int) g) << 8) | (((int) b) << 0);
+}
+
+/* colours */
+int option_color_0 = 0x080808;
+int option_color_1 = 0x000065;
+int option_color_2 = 0x0000FF;
+int option_color_3 = 0x008B00;
+int option_color_4 = 0x008B8B;
+int option_color_5 = 0x009ACD;
+int option_color_6 = 0x00FF00;
+int option_color_7 = 0x00FA9A;
+int option_color_8 = 0x00FFFF;
+int option_color_9 = 0x8B2500;
+int option_color_10 = 0x8B008B;
+int option_color_11 = 0x7D26CD;
+int option_color_12 = 0x8B7500;
+int option_color_13 = 0x7F7F7F;
+int option_color_14 = 0x7B68EE;
+int option_color_15 = 0x7FFF00;
+int option_color_16 = 0x87CEEB;
+int option_color_17 = 0x7FFFD4;
+int option_color_18 = 0xEE0000;
+int option_color_19 = 0xEE1289;
+int option_color_20 = 0xEE00EE;
+int option_color_21 = 0xCD6600;
+int option_color_22 = 0xF8B7B7;
+int option_color_23 = 0xE066FF;
+int option_color_24 = 0xEEEE00;
+int option_color_25 = 0xEEE685;
+int option_color_26 = 0xF8F8FF;
 
 /* takes 0-26 and converts it to RGB */
 static void get_general_colors (XColor * color, int i)
 {
-    color->red = (long) color_matrix[i].R << 8;
-    color->green = (long) color_matrix[i].G << 8;
-    color->blue = (long) color_matrix[i].B << 8;
+    unsigned long c = 0;
+    switch (i) {
+    case 0:
+	c = transform (option_color_0);
+	break;
+    case 1:
+	c = transform (option_color_1);
+	break;
+    case 2:
+	c = transform (option_color_2);
+	break;
+    case 3:
+	c = transform (option_color_3);
+	break;
+    case 4:
+	c = transform (option_color_4);
+	break;
+    case 5:
+	c = transform (option_color_5);
+	break;
+    case 6:
+	c = transform (option_color_6);
+	break;
+    case 7:
+	c = transform (option_color_7);
+	break;
+    case 8:
+	c = transform (option_color_8);
+	break;
+    case 9:
+	c = transform (option_color_9);
+	break;
+    case 10:
+	c = transform (option_color_10);
+	break;
+    case 11:
+	c = transform (option_color_11);
+	break;
+    case 12:
+	c = transform (option_color_12);
+	break;
+    case 13:
+	c = transform (option_color_13);
+	break;
+    case 14:
+	c = transform (option_color_14);
+	break;
+    case 15:
+	c = transform (option_color_15);
+	break;
+    case 16:
+	c = transform (option_color_16);
+	break;
+    case 17:
+	c = transform (option_color_17);
+	break;
+    case 18:
+	c = transform (option_color_18);
+	break;
+    case 19:
+	c = transform (option_color_19);
+	break;
+    case 20:
+	c = transform (option_color_20);
+	break;
+    case 21:
+	c = transform (option_color_21);
+	break;
+    case 22:
+	c = transform (option_color_22);
+	break;
+    case 23:
+	c = transform (option_color_23);
+	break;
+    case 24:
+	c = transform (option_color_24);
+	break;
+    case 25:
+	c = transform (option_color_25);
+	break;
+    case 26:
+	c = transform (option_color_26);
+	break;
+    }
+    color->red = ((c >> 16) & 0xFF) << 8;
+    color->green = ((c >> 8) & 0xFF) << 8;
+    color->blue = ((c >> 0) & 0xFF) << 8;
     color->flags = DoRed | DoBlue | DoGreen;
 }
 
-static void get_grey_colors (XColor *color, int i)
-{
-    color->red = color->green = grey_intense(i);
-    color->blue = grey_intense(i);
-    color->flags = DoRed | DoBlue | DoGreen;
-}
-
-
 void alloc_grey_scale (Colormap cmap)
 {
     XColor color;
@@ -448,7 +530,7 @@ static void setup_staticcolor (void)
     XColor color;
     int size, i, j, k, n, m = 0, num_greys, grey;
 
-    c = get_cells(CColormap, &size);
+    c = get_cells (CColormap, &size);
     grey_levels = CMalloc ((size + 2) * sizeof (unsigned short));
     num_greys = 0;
 
@@ -480,8 +562,8 @@ static void setup_staticcolor (void)
     num_greys = grey;
 
     if (num_greys <= 2) {	/* there's just no hope  :(   */
-	if(verbose_operation)
-	    printf (_("This will work, but it may look terrible.\n"));
+	if (verbose_operation)
+	    printf (_ ("This will work, but it may look terrible.\n"));
 	for (grey = 0; grey < 16; grey++) {
 	    color.flags = DoRed | DoGreen | DoBlue;
 	    color.red = grey * 65535 / 15;
@@ -515,7 +597,7 @@ static void setup_staticcolor (void)
 	}
     }
 
-    for(i=0;i<27;i++) {
+    for (i = 0; i < 27; i++) {
 	get_general_colors (&color, i);
 	m = CGetCloseColor (c, size, color, 0);
 	CAllocColor (CColormap, &(c[m]));
@@ -572,7 +654,7 @@ void store_grey_scale (Colormap cmap)
     int i;
     if (verbose_operation)
 /* "Grey scale" is a list of increasingly bright grey levels of color */
-	printf (_("Storing grey scale.\n"));
+	printf (_ ("Storing grey scale.\n"));
     if (!XAllocColorCells (CDisplay, cmap, 1, color_planes, 6, color_pixels + 43, 1))
 	alloccolorerror ();
     for (i = 0; i < 64; i++) {
@@ -598,14 +680,14 @@ void try_color (Colormap cmap, XColor * c, int size, XColor color, int i)
 	    XStoreColor (CDisplay, cmap, &color);
 	    if (verbose_operation)
 /* "Assign color" */
-		printf (_("Store,"));
+		printf (_ ("Store,"));
 	    return;
 	}
     }
     if (!XAllocColor (CDisplay, cmap, &closest))
 	if (verbose_operation)
 /* "Ignoring" means that the program will continue regardless of this error */
-	    printf (_("\nerror allocating this color - ignoring;"));	/* this should never happen since closest comes from the colormap */
+	    printf (_ ("\nerror allocating this color - ignoring;"));	/* this should never happen since closest comes from the colormap */
     color_pixels[i] = closest.pixel;
     if (verbose_operation)
 	printf ("%ld,", ((error == 0) ? 0 : 1) + ((error / (8 + 10 + 5)) >> (16 - BitsPerRGBofVisual (CVisual))));
@@ -637,7 +719,7 @@ static void setup_store_colors (void)
 	if (i >= 0) {
 	    if (verbose_operation)
 /* Not essential to translate */
-		printf (_("found grey scale\n"));
+		printf (_ ("found grey scale\n"));
 	    alloc_grey_scale (CColormap);
 	} else {
 #endif
@@ -648,7 +730,7 @@ static void setup_store_colors (void)
     }
     if (verbose_operation)
 /* This isn't very important, run cooledit -verbose to see how this works */
-	printf (_("Trying colors...\n( 'Store'=store my own color, Number=integer error with existing color )\n"));
+	printf (_ ("Trying colors...\n( 'Store'=store my own color, Number=integer error with existing color )\n"));
 
     for (i = 0; i < 16; i++) {
 	get_button_color (&color, i);
@@ -667,21 +749,21 @@ static void setup_store_colors (void)
 
 static void setup_colormap (int class)
 {
-    switch(class) {
-	case StaticColor:
-	case StaticGray:
-	    setup_staticcolor ();
-	    break;
-	case GrayScale:
-	    setup_alloc_colors (1);
-	    return;
-	case DirectColor:
-	case TrueColor:
-	    setup_alloc_colors (0);
-	    return;
-	case PseudoColor:
-	    setup_store_colors ();
-	    break;
+    switch (class) {
+    case StaticColor:
+    case StaticGray:
+	setup_staticcolor ();
+	break;
+    case GrayScale:
+	setup_alloc_colors (1);
+	return;
+    case DirectColor:
+    case TrueColor:
+	setup_alloc_colors (0);
+	return;
+    case PseudoColor:
+	setup_store_colors ();
+	break;
     }
 }
 
@@ -694,22 +776,22 @@ static int cursor_blink_rate;
    Aim1: Get the cursor to flash all the time:
 
    Aim2: Have coolwidgets send an alarm event, just like
-		any other event. For the application to use.
+   any other event. For the application to use.
 
    Requirements: XNextEvent must still be the blocker
-		so that the process doesn't hog when idle.
+   so that the process doesn't hog when idle.
 
    Problems: If the alarm handler sends an event using
-		XSendEvent it may hang the program.
+   XSendEvent it may hang the program.
 
    To solve, we put a pause() before XNextEvent so that it waits for 
    an alarm, and also define our own CSendEvent routine with
    its own queue. So that things don't slow down, we pause() only
    if no events are pending. Also make the alarm rate high (100 X per sec).
    (I hope this is the easiest way to do this  :|   )
-*/
+ */
 
-void CSetCursorBlinkRate(int b)
+void CSetCursorBlinkRate (int b)
 {
     if (b < 1)
 	b = 1;
@@ -767,7 +849,7 @@ static RETSIGTYPE alarmhandler (int x)
 	}
     } else {
 	xevent.type = AlarmEvent;
-	if (CQueueSize () < 128 && !block_push_event) {	/* say */
+	if (CQueueSize () < 128 && !block_push_event) {		/* say */
 	    CSendEvent (&xevent);
 	}
 	xevent.type = TickEvent;
@@ -782,7 +864,59 @@ static RETSIGTYPE alarmhandler (int x)
 #endif
 }
 
+#define CHILD_EXITED_MAX	32
 
+static struct {
+    pid_t pid;
+    int status;
+} children_exitted[CHILD_EXITED_MAX];
+
+#if (RETSIGTYPE==void)
+#define handler_return return
+#else
+#define handler_return return 1
+#endif
+
+static RETSIGTYPE childhandler (int x)
+{
+    pid_t p, save_errno = errno;
+    int i, status, j;
+    for (j = 0; j < 2; j++) {
+	do {
+	    errno = 0;
+	} while ((-1 == (p = waitpid (-1, NULL, WNOHANG))) && (errno == EINTR));
+	if (p > 0)
+	    for (i = 0; i < CHILD_EXITED_MAX; i++)
+		if (!children_exitted[i].pid) {
+		    children_exitted[i].pid = p;
+		    children_exitted[i].status = status;
+		    break;
+		}
+    }
+    errno = save_errno;
+    signal (SIGCHLD, childhandler);
+    handler_return;
+}
+
+/* returns non-zero on child exit */
+int CChildExitted (pid_t p, int *status)
+{
+    int i;
+    for (i = 0; i < CHILD_EXITED_MAX; i++)
+	if (p && children_exitted[i].pid == p) {
+	    if (status)
+		*status = children_exitted[i].status;
+	    children_exitted[i].pid = 0;
+	    return 1;
+	}
+    return 0;
+}
+
+static void set_child_handler (void)
+{
+    memset (children_exitted, 0, sizeof (children_exitted));
+    signal (SIGCHLD, childhandler);
+}
 
 static void set_alarm (void)
 {
@@ -791,18 +925,18 @@ static void set_alarm (void)
     xevent.xany.display = CDisplay;
     xevent.xany.send_event = 1;
 
-    CSetCursorBlinkRate(7); /* theta rhythms ? */
+    CSetCursorBlinkRate (7);	/* theta rhythms ? */
 
     signal (SIGALRM, alarmhandler);
     setitimer (ITIMER_REAL, &alarm_every, NULL);
 }
 
-void CEnableAlarm(void)
+void CEnableAlarm (void)
 {
-    set_alarm();
+    set_alarm ();
 }
 
-void CDisableAlarm(void)
+void CDisableAlarm (void)
 {
     setitimer (ITIMER_REAL, &alarm_off, NULL);
     signal (SIGALRM, 0);
@@ -814,13 +948,13 @@ void get_temp_dir (void)
 	return;
     temp_dir = getenv ("TEMP");
     if (temp_dir)
-	if(*temp_dir) {
+	if (*temp_dir) {
 	    temp_dir = (char *) strdup (temp_dir);
 	    return;
 	}
     temp_dir = getenv ("TMP");
     if (temp_dir)
-	if(*temp_dir) {
+	if (*temp_dir) {
 	    temp_dir = (char *) strdup (temp_dir);
 	    return;
 	}
@@ -843,7 +977,7 @@ void get_home_dir (void)
 	    home_dir = (char *) strdup (home_dir);
 	    return;
 	}
-    fprintf (stderr, _("%s: HOME environment variable not set and no passwd entry - aborting\n"), CAppName);
+    fprintf (stderr, _ ("%s: HOME environment variable not set and no passwd entry - aborting\n"), CAppName);
     abort ();
 }
 
@@ -883,29 +1017,68 @@ struct visual_priority {
     int depth_low;
     int depth_high;
 } visual_priority[] = {
-    {TrueColor, 15, 16},
-    {TrueColor, 12, 14},
-    {PseudoColor, 6, 999},
-    {DirectColor, 12, 999},
-    {TrueColor, 17, 999},
 
-    {DirectColor, 8, 11},
-    {TrueColor, 8, 11},
-    {StaticColor, 8, 999},
+    {
+	TrueColor, 15, 16
+    },
+    {
+	TrueColor, 12, 14
+    },
+    {
+	PseudoColor, 6, 999
+    },
+    {
+	DirectColor, 12, 999
+    },
+    {
+	TrueColor, 17, 999
+    },
 
-    {PseudoColor, 4, 5},
-    {DirectColor, 6, 7},
-    {TrueColor, 6, 7},
-    {StaticColor, 6, 7},
+    {
+	DirectColor, 8, 11
+    },
+    {
+	TrueColor, 8, 11
+    },
+    {
+	StaticColor, 8, 999
+    },
 
-    {DirectColor, 4, 5},
-    {TrueColor, 4, 5},
-    {StaticColor, 4, 5},
+    {
+	PseudoColor, 4, 5
+    },
+    {
+	DirectColor, 6, 7
+    },
+    {
+	TrueColor, 6, 7
+    },
+    {
+	StaticColor, 6, 7
+    },
 
-    {GrayScale, 6, 999},
-    {StaticGray, 6, 999},
-    {GrayScale, 4, 5},
-    {StaticGray, 4, 5},
+    {
+	DirectColor, 4, 5
+    },
+    {
+	TrueColor, 4, 5
+    },
+    {
+	StaticColor, 4, 5
+    },
+
+    {
+	GrayScale, 6, 999
+    },
+    {
+	StaticGray, 6, 999
+    },
+    {
+	GrayScale, 4, 5
+    },
+    {
+	StaticGray, 4, 5
+    },
 };
 
 #endif
@@ -932,15 +1105,15 @@ void get_preferred (XVisualInfo * v, int n, Visual ** vis, int *depth)
     if (option_preferred_visual)
 	if (!strcasecmp (option_preferred_visual, "help")
 	    || !strcasecmp (option_preferred_visual, "h")) {
-	    printf (_("%s:\n  The <visual-class> is the hardware technique used by the\n" \
-		    "computer to draw pixels to the screen. _Usually_ only one\n" \
-		    "visual is truly supported. This option is provided\n" \
-		    "if you would rather use a TrueColor than a PseudoColor\n" \
-		    "visual where you are short of color palette space.\n" \
-		    "The depth is the number of bits per pixel used by the hardware.\n" \
-		    "It is automatically selected using heuristics.\n"), \
+	    printf (_ ("%s:\n  The <visual-class> is the hardware technique used by the\n" \
+		       "computer to draw pixels to the screen. _Usually_ only one\n" \
+		 "visual is truly supported. This option is provided\n" \
+	     "if you would rather use a TrueColor than a PseudoColor\n" \
+		 "visual where you are short of color palette space.\n" \
+		       "The depth is the number of bits per pixel used by the hardware.\n" \
+		   "It is automatically selected using heuristics.\n"), \
 		    CAppName);
-	    printf (_("Available visuals on this system are:\n"));
+	    printf (_ ("Available visuals on this system are:\n"));
 	    for (i = 0; i < n; i++)
 		printf ("        class %s, depth %d\n", visual_name[v[i].class], v[i].depth);
 	    exit (1);
@@ -973,14 +1146,14 @@ void get_preferred (XVisualInfo * v, int n, Visual ** vis, int *depth)
 		}
     if (option_preferred_visual)
 /* We will select a visual in place of the one you specified, since yours is unavailable */
-	fprintf (stderr, _("%s: preferred visual not found, selecting...\n"), CAppName);
+	fprintf (stderr, _ ("%s: preferred visual not found, selecting...\n"), CAppName);
 
     if (def) {
 	*vis = def;
 	*depth = def_depth;
     } else {
 /* We will select the default visual, since the list didn't have a matching visual */
-	fprintf (stderr, _("%s: no known visuals found, using default...\n"), CAppName);
+	fprintf (stderr, _ ("%s: no known visuals found, using default...\n"), CAppName);
 	*vis = DefaultVisual (CDisplay, DefaultScreen (CDisplay));
 	*depth = DefaultDepth (CDisplay, DefaultScreen (CDisplay));
     }
@@ -1006,14 +1179,14 @@ static void get_preferred_visual_and_depth (void)
 static void assign_default_cmap (void)
 {
     if (verbose_operation)
-	printf (_("Using DefaultColormap()\n"));
+	printf (_ ("Using DefaultColormap()\n"));
     CColormap = DefaultColormap (CDisplay, DefaultScreen (CDisplay));
 }
 
 static void assign_own_cmap (void)
 {
     if (verbose_operation)
-	printf (_("Creating own colormap\n"));
+	printf (_ ("Creating own colormap\n"));
     CColormap = XCreateColormap (CDisplay,
 			 RootWindow (CDisplay, DefaultScreen (CDisplay)),
 				 CVisual, AllocNone);
@@ -1054,7 +1227,7 @@ static void get_colormap (void)
     if (XVisualIDFromVisual (CVisual)
 	== XVisualIDFromVisual (DefaultVisual (CDisplay, DefaultScreen (CDisplay)))) {
 	if (verbose_operation)
-	    printf (_("Default visual ID found\n"));
+	    printf (_ ("Default visual ID found\n"));
 	assign_default_cmap ();
 	return;
     }
@@ -1068,7 +1241,7 @@ static void get_colormap (void)
     if (DEFAULT_CMAPS == None) {
 	if (verbose_operation)
 /* "An Atom of name %s could not be found". 'Atom' is X terminology */
-	    printf (_("No Atom %s \n"), COLORMAP_PROPERTY);
+	    printf (_ ("No Atom %s \n"), COLORMAP_PROPERTY);
 	assign_check_colormap ();
 	return;
     } else {
@@ -1077,22 +1250,22 @@ static void get_colormap (void)
 	if (!XGetRGBColormaps (CDisplay, CRoot, &cmap, &n,
 			       DEFAULT_CMAPS)) {
 	    if (verbose_operation)
-		printf (_("XGetRGBColormaps(%s) failed\n"), COLORMAP_PROPERTY);
+		printf (_ ("XGetRGBColormaps(%s) failed\n"), COLORMAP_PROPERTY);
 	    assign_check_colormap ();
 	    return;
 	}
 	if (verbose_operation)
-	    printf (_("Choosing from %d 'XGetRGBColormaps' colormaps\n"), n);
+	    printf (_ ("Choosing from %d 'XGetRGBColormaps' colormaps\n"), n);
 	for (i = 0; i < n; i++) {
 	    if (XVisualIDFromVisual (CVisual) == cmap[i].visualid) {
 		if (verbose_operation)
-		    printf (_("Colormap %d matches visual ID\n"), i);
+		    printf (_ ("Colormap %d matches visual ID\n"), i);
 		CColormap = cmap[i].colormap;
 		return;
 	    }
 	}
 	if (verbose_operation)
-	    printf (_("No colormap found matching our visual ID\n"));
+	    printf (_ ("No colormap found matching our visual ID\n"));
     }
 #endif
 
@@ -1107,13 +1280,34 @@ int ignore_handler (Display * c, XErrorEvent * e)
 void init_cursors (void);
 
 /*-------------------------------------------------------------*/
-void CInitialise (CInitData *config_start)
+void CInitialise (CInitData * config_start)
 {
+    if (!config_start->look)
+	config_start->look = init_look;
+
+    if (!strncmp (config_start->look, "gtk", 3)) {
+	look = &look_gtk;
+    } else if (!strncmp (config_start->look, "next", 4)) {
+#ifdef NEXT_LOOK
+	look = &look_next;
+#else
+	fprintf (stderr, _ ("%s: NeXT look was not compiled into this binary\n"), config_start->name);
+	exit (1);
+#endif
+    } else if (!strncmp (config_start->look, "cool", 4)) {
+	look = &look_cool;
+    } else {
+	look = &look_gtk;
+    }
+
+    option_interwidget_spacing = (*look->get_default_interwidget_spacing) ();
+    init_widget_font = (*look->get_default_widget_font) ();
+
     given = config_start;
     verbose_operation = (given->options & CINIT_OPTION_VERBOSE);
 
-    if(verbose_operation)
-	printf ("sizeof(CWidget) = %d\n", (int) sizeof(CWidget));
+    if (verbose_operation)
+	printf ("sizeof(CWidget) = %d\n", (int) sizeof (CWidget));
 
     CAppName = given->name;
 
@@ -1133,6 +1327,8 @@ void CInitialise (CInitData *config_start)
 	init_geometry = given->geometry;
     if (given->font)
 	init_font = given->font;
+    if (given->widget_font)
+	init_widget_font = given->widget_font;
     if (given->bg)
 	init_bg_color = given->bg;
     if (given->fg_red)
@@ -1149,22 +1345,26 @@ void CInitialise (CInitData *config_start)
 /*  Initialise window manager atoms to detect a user close */
     wm_interaction_init ();
 
-/* Set up font */
-    load_font ();
-
 /* Now set up the visual and colors */
     get_preferred_visual_and_depth ();
 
-    if(verbose_operation) {
-	printf (_("Found a visual, depth = %d,\n       visual class = "), CDepth);
+    if (verbose_operation) {
+	printf (_ ("Found a visual, depth = %d,\n       visual class = "), CDepth);
 	visual_comments (ClassOfVisual (CVisual));
     }
-
     get_colormap ();
 
 /* Now setup that color map discribed above */
     setup_colormap (ClassOfVisual (CVisual));
 
+/* Set up font */
+    load_font ();
+
+#ifdef USE_XIM
+/* set the XIM locale */
+    init_xlocale ();
+#endif
+
 /* some special cursors */
     init_cursors ();
 
@@ -1177,6 +1377,11 @@ void CInitialise (CInitData *config_start)
     mouse_init ();
 #endif
 
+    XAaInit (CDisplay, CVisual, CDepth, CRoot);
+
+/* set child handler */
+    set_child_handler ();
+
 /* an alarm handler generates xevent of tyoe AlarmEvent every 1/4 second to flash the cursor */
     set_alarm ();
 }
