@@ -1,5 +1,5 @@
 /* textinput.c - draw an editable text widget
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,9 +24,7 @@
 #include <stdlib.h>
 #include <stdarg.h>
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xresource.h>
+#include <X11/Intrinsic.h>
 #include <X11/Xatom.h>
 #include "lkeysym.h"
 
@@ -46,6 +44,7 @@
 int eh_textinput (CWidget * w, XEvent * xevent, CEvent * cwevent);
 void input_mouse_mark (CWidget * w, XEvent * event, CEvent * ce);
 
+extern struct look *look;
 
 
 /* {{{  history stuff: draws a history of inputs a widget of the same ident */
@@ -72,13 +71,13 @@ static struct textinput_history *history_widgets[MAX_HIST_WIDGETS] =
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
 static int last = 0;
 
-static void add_to_history (struct textinput_history *h, const char *text)
+static void add_to_history (struct textinput_history *h, const char *text, int allow_blank_lines)
 {
     int i, j;
     char *t, *p;
     if (!text)
 	return;
-    if (!*text)
+    if (!*text && !allow_blank_lines)
 	return;
     t = (char *) strdup (text);
     if ((p = strchr (t, '\n')))
@@ -107,11 +106,13 @@ static void add_to_history (struct textinput_history *h, const char *text)
 static void add_to_widget_history (const char *ident, const char *text)
 {
     int i;
+    int allow_blank_lines = 0;
+    allow_blank_lines = (strchr (ident, '+') != 0);
     for (i = 0; i < MAX_HIST_WIDGETS; i++) {
 	if (!history_widgets[i])
 	    break;
 	if (!strcmp (history_widgets[i]->ident, ident)) {
-	    add_to_history (history_widgets[i], text);
+	    add_to_history (history_widgets[i], text, allow_blank_lines);
 	    return;
 	}
     }
@@ -119,7 +120,7 @@ static void add_to_widget_history (const char *ident, const char *text)
     history_widgets[last] = CMalloc (sizeof (struct textinput_history));
     memset (history_widgets[last], 0, sizeof (struct textinput_history));
     strcpy (history_widgets[last]->ident, ident);
-    add_to_history (history_widgets[last], text);
+    add_to_history (history_widgets[last], text, allow_blank_lines);
     last++;
 
     if (last == MAX_HIST_WIDGETS) {	/* shift up one */
@@ -303,11 +304,14 @@ static char *draw_text_input_history (CWidget * text_input)
 	return 0;
 
     x = text_input->x;
+    CPushFont ("editor", 0);
     columns = (text_input->width - WIDGET_SPACING * 3 - 4 - 6 - 20) / FONT_MEAN_WIDTH;
     w = CWidgetOfWindow (text_input->parentid);
 
-    if (!w)
+    if (!w) {
+    	CPopFont ();
 	return 0;
+    }
     if (text_input->y > w->height / 2) {
 	p = get_history_list (text_input->ident, 1, &num_lines);
 	lines = (text_input->y - 2 - WIDGET_SPACING * 2 - 4 - 6) / FONT_PIX_PER_LINE;
@@ -322,41 +326,57 @@ static char *draw_text_input_history (CWidget * text_input)
 	r = CTrivialSelectionDialog (w->winid, x, y, columns, lines, p, 0, 0);
     }
     free (p);
+    CPopFont ();
     return r;
 }
 
+#if 0
+wchar_t *mbstowcs_dup (unsigned char *s);
+int wchar_t_strlen (wchar_t * s);
+
+int wchar_t_columns (char *t, int x)
+{
+    wchar_t *s;
+    t = (char *) strdup ((char *) t);
+    t[x] = '\0';
+    s = mbstowcs_dup ((unsigned char *) t);
+    x = wchar_t_strlen (s);
+    free (t);
+    free (s);
+    return x;
+}
+#endif
+
 void render_passwordinput (CWidget * wdt)
 {
     int wc, k, l, w = wdt->width, h = wdt->height;
     Window win;
     char *password;
 
+    CPushFont ("editor", 0);
+
     win = wdt->winid;
 
     CSetBackgroundColor (COLOR_WHITE);
     CSetColor (COLOR_BLACK);
     password = (char *) strdup (wdt->text);
     memset (password, '*', strlen (wdt->text));
-    XDrawImageString (CDisplay, win, CGC,
-		      FONT_OFFSET_X + 3 + TEXTINPUT_RELIEF,
+    CImageString (win, FONT_OFFSET_X + 3 + TEXTINPUT_RELIEF,
 		      FONT_OFFSET_Y + 3 + TEXTINPUT_RELIEF,
-		      password, strlen (password));
+		      password);
     CSetColor (COLOR_WHITE);
-    l = XTextWidth (CFontStruct, password, strlen (password));
+    l = CImageStringWidth (password);
     k = min (l, w - 6);
+    memset (password, 0, strlen (password));
+    free (password);
     CRectangle (win, 3, 3, k, option_text_line_spacing + 1);
     CLine (win, 3, 4, 3, h - 5);
     CLine (win, 3, h - 4, k + 3, h - 4);
     CRectangle (win, k + 3, 3, w - 6 - k, h - 6);
-    free (password);
-    if (win == CGetFocus ()) {
-	render_bevel (win, 0, 0, w - 1, h - 1, 3, 1);
-    } else {
-	render_bevel (win, 2, 2, w - 3, h - 3, 1, 1);
-	render_bevel (win, 0, 0, w - 1, h - 1, 2, 0);
-    }
-    wc = 3 + TEXTINPUT_RELIEF + 1 + XTextWidth (CFontStruct, password, wdt->cursor);
-    set_cursor_position (win, wc, 5, 0, h - 5, CURSOR_TYPE_TEXTINPUT, 0, 0, 0);
+    (*look->render_passwordinput_tidbits) (wdt, win == CGetFocus ());
+    wc = 3 + TEXTINPUT_RELIEF + 1 + CImageTextWidth (password, wdt->cursor);
+    set_cursor_position (win, wc, 5, 0, h - 5, CURSOR_TYPE_TEXTINPUT, 0, 0, 0, 0);
+    CPopFont ();
     return;
 }
 
@@ -373,7 +393,8 @@ void render_textinput (CWidget * wdt)
 	render_passwordinput (wdt);
 	return;
     }
-
+    
+    CPushFont ("editor", 0);
     win = wdt->winid;
     isfocussed = (win == CGetFocus ());
 
@@ -384,9 +405,7 @@ void render_textinput (CWidget * wdt)
 	f = 0;
 /*wc is the position of the cursor from the left of the input window */
 	wc = 3 + TEXTINPUT_RELIEF + 1 +
-	    XTextWidth (CFontStruct, wdt->text +
-			wdt->firstcolumn,
-			wdt->cursor - wdt->firstcolumn);
+	    CImageTextWidth (wdt->text + wdt->firstcolumn, wdt->cursor - wdt->firstcolumn);
 
 	/*now lets make sure the cursor is well within the view */
 
@@ -414,12 +433,11 @@ void render_textinput (CWidget * wdt)
     s = wdt->text + wdt->firstcolumn;
     l = strlen (s);
     CSetColor (COLOR_WHITE);
-    k = min (XTextWidth (CFontStruct, s, l), w - h - 6);
+    k = min (CImageTextWidth (s, l), w - h - 6);
     CRectangle (win, 3, 3, k, option_text_line_spacing + 1);
     CLine (win, 3, 4, 3, h - 5);
     CLine (win, 3, h - 4, k + 3, h - 4);
     CRectangle (win, k + 3, 3, w - h - 6 - k, h - 6);
-
 /* now draw the visible part of the string */
     wdt->mark1 = min (wdt->mark1, l + wdt->firstcolumn);
     wdt->mark2 = min (wdt->mark2, l + wdt->firstcolumn);
@@ -429,54 +447,41 @@ void render_textinput (CWidget * wdt)
     if (m1 > wdt->firstcolumn) {
 	CSetBackgroundColor (COLOR_WHITE);
 	CSetColor (COLOR_BLACK);
-	XDrawImageString (CDisplay, win, CGC,
+	CImageText (win,
 			  FONT_OFFSET_X + 3 + TEXTINPUT_RELIEF,
 			  FONT_OFFSET_Y + 3 + TEXTINPUT_RELIEF,
 			  s, m1 - wdt->firstcolumn);
-	x += XTextWidth (CFontStruct, s, m1 - wdt->firstcolumn);
+	x += CImageTextWidth (s, m1 - wdt->firstcolumn);
 	s += m1 - wdt->firstcolumn;
     }
     if (x < w - h && m2 > wdt->firstcolumn) {
 	m1 = max (wdt->firstcolumn, m1);
 	CSetBackgroundColor (COLOR_BLACK);
 	CSetColor (COLOR_WHITE);
-	XDrawImageString (CDisplay, win, CGC,
+	CImageText (win,
 			  FONT_OFFSET_X + 3 + TEXTINPUT_RELIEF + x,
 			  FONT_OFFSET_Y + 3 + TEXTINPUT_RELIEF,
 			  s, m2 - m1);
-	x += XTextWidth (CFontStruct, s, m2 - m1);
+	x += CImageTextWidth (s, m2 - m1);
 	s += m2 - m1;
     }
     if (x < w - h) {
 	CSetBackgroundColor (COLOR_WHITE);
 	CSetColor (COLOR_BLACK);
-	XDrawImageString (CDisplay, win, CGC,
+	CImageString (win,
 			  FONT_OFFSET_X + 3 + TEXTINPUT_RELIEF + x,
 			  FONT_OFFSET_Y + 3 + TEXTINPUT_RELIEF,
-			  s, strlen (s));
-    }
-    if (isfocussed) {
-	render_bevel (win, 0, 0, w - h - 1, h - 1, 3, 1);	/*most outer border bevel */
-    } else {
-	render_bevel (win, 2, 2, w - h - 3, h - 3, 1, 1);	/*border bevel */
-	render_bevel (win, 0, 0, w - h - 1, h - 1, 2, 0);	/*most outer border bevel */
+			  s);
     }
 
-    /* history button to the right */
-    CSetColor (COLOR_FLAT);
-    if (wdt->options & BUTTON_PRESSED) {
-	CRectangle (win, w - h + 2, 2, h - 4, h - 4);
-	render_bevel (win, w - h, 0, w - 1, h - 1, 2, 1);
-    } else if (wdt->options & BUTTON_HIGHLIGHT) {
-	CRectangle (win, w - h + 1, 1, h - 2, h - 2);
-	render_bevel (win, w - h, 0, w - 1, h - 1, 1, 0);
-    } else {
-	CRectangle (win, w - h + 2, 2, h - 4, h - 4);
-	render_bevel (win, w - h, 0, w - 1, h - 1, 2, 0);
-    }
-    set_cursor_position (win, wc, 5, 0, h - 5, CURSOR_TYPE_TEXTINPUT, 0, 0, 0);
+    (*look->render_textinput_tidbits) (wdt, isfocussed);
+
+    set_cursor_position (win, wc, 5, 0, h - 5, CURSOR_TYPE_TEXTINPUT, 0, 0, 0, 0);
+    CPopFont();
+    return;
 }
 
+
 void text_input_destroy (CWidget * w)
 {
     CAddToTextInputHistory (w->ident, w->text);
@@ -494,9 +499,7 @@ static long cp (CWidget * wdt, int x, int y)
     i = wdt->firstcolumn;
     for (;;) {
 	int a;
-	a = XTextWidth (CFontStruct, wdt->text +
-			wdt->firstcolumn,
-			i - wdt->firstcolumn);
+	a = CImageTextWidth (wdt->text + wdt->firstcolumn, i - wdt->firstcolumn);
 	if (a > x) {
 	    if (i > 0)
 		return i - 1;
@@ -636,6 +639,7 @@ CWidget *CDrawTextInput (const char *identifier, Window parent, int x, int y,
     if (text == TEXTINPUT_LAST_INPUT)
 	text = CLastInput (identifier);
 
+    CPushFont ("editor", 0);
     if (!(wdt = CIdent (identifier))) {
 	int w, h;
 	if (width == AUTO_WIDTH || height == AUTO_HEIGHT)
@@ -648,7 +652,7 @@ CWidget *CDrawTextInput (const char *identifier, Window parent, int x, int y,
 	set_hint_pos (x + width + WIDGET_SPACING, y + height + WIDGET_SPACING);
 
 	wdt = CSetupWidget (identifier, parent, x, y,
-	 width, height, C_TEXTINPUT_WIDGET, INPUT_BUTTON, COLOR_FLAT, 1);
+	 width, height, C_TEXTINPUT_WIDGET, INPUT_KEY, COLOR_FLAT, 1);
 
 /* For the text input widget we need enough memory allocated to the label
    for it to grow to maxlen, so reallocate it */
@@ -679,6 +683,7 @@ CWidget *CDrawTextInput (const char *identifier, Window parent, int x, int y,
 	wdt->keypressed = 0;
 	render_textinput (wdt);
     }
+    CPopFont();
 
     return wdt;
 }
@@ -706,13 +711,19 @@ void selection_send (XSelectionRequestEvent * rq);
 
 int eh_textinput (CWidget * w, XEvent * xevent, CEvent * cwevent)
 {
-    int handled = 0;
+    int handled = 0, save_options;
     int cursor;
+    char *u;
 
     switch (xevent->type) {
+    case FocusIn:
+    case FocusOut:
+	render_textinput (w);
+	break;
     case SelectionRequest:
 	text_get_selection (w);
 	selection_send (&(xevent->xselectionrequest));
+	render_textinput (w);
 	return 1;
     case SelectionNotify:
 	cursor = w->keypressed ? w->cursor : 0;
@@ -720,6 +731,7 @@ int eh_textinput (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	xevent->xselection.requestor, xevent->xselection.property, True);
 	w->mark1 = w->mark2 = 0;
 	w->cursor = cursor;
+	render_textinput (w);
 	break;
     case ButtonPress:
 	resolve_button (xevent, cwevent);
@@ -740,10 +752,12 @@ int eh_textinput (CWidget * w, XEvent * xevent, CEvent * cwevent)
 		w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
 	    }
 	}
+	render_textinput (w);
 	CFocus (w);
     case ButtonRelease:
 	if (!(w->options & TEXTINPUT_PASSWORD))
 	    input_mouse_mark (w, xevent, cwevent);
+	render_textinput (w);
 	break;
     case Expose:
 	if (xevent->xexpose.count)
@@ -752,36 +766,50 @@ int eh_textinput (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
 	if (xevent->xbutton.x >= w->width - w->height)
 	    w->options |= BUTTON_HIGHLIGHT;
+	render_textinput (w);
 	break;
     case MotionNotify:
-	if (!xevent->xmotion.state)
-	    return 0;
-	w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
+	save_options = w->options;
+	w->options &= ~(BUTTON_PRESSED | BUTTON_HIGHLIGHT);
 	if (xevent->xmotion.x >= w->width - w->height) {
-	    if ((w->options & BUTTON_HIGHLIGHT) && !xevent->xmotion.state)
-		return 0;
 	    w->options |= BUTTON_HIGHLIGHT;
+	    if (save_options != w->options)
+		render_textinput (w);
+	    return 0;
 	} else {
-	    if (!w->options && !xevent->xmotion.state)
+	    if (!xevent->xmotion.state) {
+		if (save_options != w->options)
+		    render_textinput (w);
 		return 0;
+	    }
 	}
 	if (!(w->options & TEXTINPUT_PASSWORD))
 	    input_mouse_mark (w, xevent, cwevent);
+	render_textinput (w);
 	break;
     case LeaveNotify:
 	w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
+	render_textinput (w);
 	break;
     case KeyPress:
 	cwevent->ident = w->ident;
 	cwevent->state = xevent->xkey.state;
 	cursor = w->cursor;
 	if (cwevent->insert > 0) {
-	    input_insert (w, cwevent->insert);
+            if (!((w->options & TEXTINPUT_NUMBERS) && !(cwevent->insert >= '0' && cwevent->insert <= '9')))
+	        input_insert (w, cwevent->insert);
 	    handled = 1;
 	} else {
 	    unsigned char *intext;
 	    intext = (unsigned char *) w->text;
 	    switch (cwevent->command) {
+	    case CK_Insert_Unicode:
+		u = (char *) CGetUnichar (CRoot, "Unicode characters");
+		if (u)
+		    while (*u)
+			input_insert (w, *u++);
+		handled = 1;
+		break;
 	    case CK_XPaste:
 		if (!XGetSelectionOwner (CDisplay, XA_PRIMARY)) {
 		    cursor = w->cursor;
@@ -853,18 +881,21 @@ int eh_textinput (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	    }
 	    w->keypressed |= handled;
 	}
-	if (handled)
+	if (handled) {
 	    w->mark1 = w->mark2 = 0;
+	    render_textinput (w);
+	}
 	cwevent->text = w->text;
     }
 
-    render_textinput (w);
     return handled;
 }
 
 void input_mouse_mark (CWidget * w, XEvent * event, CEvent * ce)
 {
+    CPushFont ("editor", 0);
     mouse_mark (event, ce->double_click, w->funcs);
+    CPopFont ();
 }
 
 
