@@ -1,5 +1,5 @@
 /* menu.c - proper 3D menus
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,9 +24,7 @@
 #include <stdlib.h>
 #include <stdarg.h>
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xresource.h>
+#include <X11/Intrinsic.h>
 #include "lkeysym.h"
 
 #include "stringtools.h"
@@ -36,12 +34,12 @@
 
 #include "mad.h"
 
-#define MENU_USER_MODE 1
-#define MENU_DRAGGING_MODE 2
+extern struct look *look;
 
-int menu_mode = MENU_USER_MODE;
 CWidget *current_pulled_button = 0;
 
+int menu_grabbed = 0;
+
 int eh_menu (CWidget * w, XEvent * xevent, CEvent * cwevent);
 void CMenuSelectionDialog (CWidget * button);
 
@@ -60,44 +58,14 @@ void destroy_menu (CWidget * w)
     free (w->menu);
 }
 
-/* outermost bevel */
-#define BEVEL_MAIN	2
-/* next-outermost bevel */
-#define BEVEL_IN 	4
-#define BEVEL_OUT	5
-/* between items, and between items and next-outermost bevel */
-#define SPACING		4
-#define S		SPACING
-/* between window border and items */
-#define O		(BEVEL_OUT + SPACING)
-/* size of bar item */
-#define BAR_HEIGHT	8
-#define B		BAR_HEIGHT
-/* between items rectangle and text */
-#define RELIEF		(TEXT_RELIEF + 1)
-/* total height of an item */
-#define H		(FONT_PIX_PER_LINE + RELIEF * 2)
-
 /* 
    Gets y1 and y2 of the entry of a menu (i.e. a menu item). x1, y1,
    x2, y2 would describe the rectangle that encloses that entry,
    relative to the top left corner of the pulled-down menu window.
  */
-void get_menu_item_extents (int n, int j, struct menu_item m[], int *y1, int *y2)
+void get_menu_item_extents (int n, int j, struct menu_item m[], int *border, int *relief, int *y1, int *y2)
 {
-    int i, n_items = 0, n_bars = 0;
-    if (!n || j < 0) {
-	*y1 = O;
-	*y2 = *y1 + H;
-    } else {
-	for (i = 0; i < j; i++)
-	    if (m[i].text[2])
-		n_items++;
-	    else
-		n_bars++;
-	*y1 = O + n_items * (H + S) + n_bars * (B + S) + (m[j].text[2] ? 0 : 2);
-	*y2 = *y1 + (m[j].text[2] ? H : (B - 4));
-    }
+    (*look->get_menu_item_extents) (n, j, m, border, relief, y1, y2);
 }
 
 /*
@@ -107,22 +75,21 @@ void get_menu_item_extents (int n, int j, struct menu_item m[], int *y1, int *y2
  */
 int whereis_pointer (int x, int y, int w, int n, struct menu_item m[])
 {
-    int i, y1, y2;
-    if (x >= O && x < w - O)
-	for (i = 0; i < n; i++) {
-	    if (!m[i].text[2])
-		continue;
-	    get_menu_item_extents (n, i, m, &y1, &y2);
-	    if (y >= y1) {
-		if (y < y2)
-		    return i;
-	    } else
-		break;
-	}
+    int i, y1, y2, border, relief;
+    for (i = 0; i < n; i++) {
+	if (!m[i].text[2])
+	    continue;
+	get_menu_item_extents (n, i, m, &border, &relief, &y1, &y2);
+	if (y >= y1) {
+	    if (y < y2 && x >= border && x < w - border)
+		return i;
+	} else
+	    break;
+    }
     return -1;
 }
 
-static int find_menu_hotkey (struct menu_item m[], int this, int num)
+int find_menu_hotkey (struct menu_item m[], int this, int num)
 {
     unsigned char used_keys[256];
     int n = 0, j;
@@ -136,73 +103,32 @@ static int find_menu_hotkey (struct menu_item m[], int this, int num)
 
 void menu_draw (Window win, int w, int h, struct menu_item m[], int n, int light)
 {
-    int i, y1, y2;
-    static int last_light = 0, last_n = 0;
-    static Window last_win = 0;
-    render_bevel (win, 0, 0, w - 1, h - 1, BEVEL_MAIN, 0);
-    render_bevel (win, BEVEL_IN, BEVEL_IN, w - 1 - BEVEL_IN, h - 1 - BEVEL_IN, BEVEL_OUT - BEVEL_IN, 1);
-
-    if (last_win == win && last_n != n) {
-	XClearWindow (CDisplay, win);
-    } else if (last_light >= 0 && last_light < n) {
-	get_menu_item_extents (n, last_light, m, &y1, &y2);
-	CSetColor (COLOR_FLAT);
-	CRectangle (win, O - 1, y1 - 1, w - O * 2 + 2, y2 - y1 + 2);
-    }
-    last_win = win;
-    last_n = n;
-    for (i = 0; i < n; i++) {
-	get_menu_item_extents (n, i, m, &y1, &y2);
-	if (i == light && m[i].text[2]) {
-	    CSetColor (color_widget (11));
-	    CRectangle (win, O + 1, y1 + 1, w - O * 2 - 2, y2 - y1 - 2);
-	    render_bevel (win, O - 1, y1 - 1, w - O, y2, 2, 0);
-	} else {
-	    if (m[i].text[2])
-		render_bevel (win, O, y1, w - O - 1, y2 - 1, 1, 1);
-	    else
-		render_bevel (win, O + 6, y1, w - O - 1 - 6, y2 - 1, 2, 0);
-	}
-	if (m[i].text[2]) {
-
-	    char *u;
-	    u = strrchr (m[i].text, '\t');
-	    if (u)
-		*u = 0;
-	    CSetColor (COLOR_BLACK);
-	    if (m[i].hot_key == '~')
-		m[i].hot_key = find_menu_hotkey (m, i, n);
-	    drawstring_xy_hotkey (win, RELIEF + O - (i == light),
-				  RELIEF + y1 - (i == light), m[i].text, m[i].hot_key);
-	    if (u) {
-		drawstring_xy (win, RELIEF + O + (w - (O + RELIEF) * 2 - CStringWidth (u + 1)) - (i == light),
-			       RELIEF + y1 - (i == light), u + 1);
-		*u = '\t';
-	    }
-	}
-    }
-    last_light = light;
+    (*look->menu_draw) (win, w, h, m, n, light);
 }
 
 void render_menu (CWidget * w)
 {
-    int n, y1, y2, i;
+    int n, border, relief, y1, y2, i;
+    unsigned new_width, new_height;
     if (!w)
 	return;
     n = w->numlines;
-    get_menu_item_extents (n, n - 1, w->menu, &y1, &y2);
-    w->height = y2 + O;
-    w->width = 0;
+    get_menu_item_extents (n, n - 1, w->menu, &border, &relief, &y1, &y2);
+    new_height = y2 + border;
+    new_width = 0;
     for (i = 0; i < n; i++) {
 	int t;
-	t = CStringWidth (w->menu[i].text);
-	t += CStringWidth ("W");
-	if (w->width < t)
-	    w->width = t;
+	t = CImageStringWidth (w->menu[i].text);
+	t += CImageStringWidth ("W");
+	if (new_width < t)
+	    new_width = t;
+    }
+    new_width += (border + relief) * 2;
+    if (w->width != new_width || w->height != new_height) {
+	w->width = new_width;
+	w->height = new_height;
+	XResizeWindow (CDisplay, w->winid, w->width, w->height);
     }
-    w->width += (O + RELIEF) * 2;
-    XResizeWindow (CDisplay, w->winid, w->width, w->height);
-
 #if 0
     if (w->y + w->height > HeightOfScreen (DefaultScreenOfDisplay (CDisplay))) {
 	int y;
@@ -213,7 +139,7 @@ void render_menu (CWidget * w)
     }
 #endif
 #define BOUND_LIMITS 50
-    get_menu_item_extents (n, w->current, w->menu, &y1, &y2);
+    get_menu_item_extents (n, w->current, w->menu, &border, &relief, &y1, &y2);
     if (w->current >= 0) {
 	if (y2 + w->y + BOUND_LIMITS >= HeightOfScreen (DefaultScreenOfDisplay (CDisplay)))
 	    CSetWidgetPosition (w->ident, w->x, HeightOfScreen (DefaultScreenOfDisplay (CDisplay)) - y2 - BOUND_LIMITS);
@@ -274,6 +200,7 @@ void pull_up (CWidget * button)
 	button->droppedmenu = 0;
     }
     focus_stack_remove_window (button->winid);
+    render_menu_button (button);
 }
 
 static CWidget *last_menu = 0;
@@ -300,13 +227,17 @@ static CWidget *pull_down (CWidget * button) /* must create a new widget */
     if(button->droppedmenu)
 	return 0;
 
-    CSetLastMenu (button);
+    sib = CGetLastMenu ();
+    if (sib)
+	if (strcmp (button->ident, sib->ident))
+	    pull_up (sib);			/* pull up last menu if different */
 
     sib = button;
-
-    while((sib = CNextFocus (sib)) != button) /* pull up any other sibling menus */
+    while((sib = CNextFocus (sib)) != button)	/* pull up any other sibling menus */
 	pull_up (sib);
 
+    CSetLastMenu (button);
+
     n = button->numlines;
 
     CGetWindowPosition (button->winid, CRoot, &x, &y);
@@ -316,8 +247,9 @@ static CWidget *pull_down (CWidget * button) /* must create a new widget */
 
     x += button->firstcolumn;
 
-    menu = CSetupWidget (catstrs(button->ident, ".pull", 0), CRoot, x,
-	    y + button->height, width, height, C_MENU_WIDGET, INPUT_KEY, COLOR_FLAT, 0);
+    menu = CSetupWidget (catstrs (button->ident, ".pull", 0), CRoot, x,
+	   y + button->height, width, height, C_MENU_WIDGET, INPUT_KEY, COLOR_FLAT, 0);
+    menu->options |= (button->options & MENU_AUTO_PULL_UP);
 
     menu_hand_cursor (menu->winid);
 
@@ -328,6 +260,7 @@ static CWidget *pull_down (CWidget * button) /* must create a new widget */
     menu->droppedmenu = button;
     button->droppedmenu = menu;
     current_pulled_button = button;
+    render_menu_button (button);
     return menu;
 }
 
@@ -336,33 +269,23 @@ void CPullDown (CWidget * button)
     pull_down (button);
 }
 
+CWidget *get_pulled_menu (void)
+{
+    return current_pulled_button;
+}
+
 void CPullUp (CWidget * button)
 {
     pull_up (button);
 }
 
-static void grab_ungrab (Window win)
-{
-    XUngrabPointer (CDisplay, CurrentTime);
-}
-
-static void pull_down_ungrab (CWidget * w)
-{
-    CWidget *d;
-    d = pull_down (w);
-    if (!d)
-	return;
-    CFocus (w);
-    if (menu_mode == MENU_DRAGGING_MODE)
-	grab_ungrab (w->winid);
-}
-
 int execute_item (CWidget * menu, int item)
 {
     int r = 0;
     char ident[33];
     strcpy (ident, menu->ident);
     menu->droppedmenu->current = item;
+    XUngrabPointer (CDisplay, CurrentTime);
     XUnmapWindow (CDisplay, menu->winid);
     if (item >= 0 && item < menu->numlines)
 	if (menu->menu[item].call_back) {
@@ -379,32 +302,7 @@ int execute_item (CWidget * menu, int item)
 
 static void render_menu_button (CWidget * wdt)
 {
-    int w = wdt->width, h = wdt->height;
-    int x = 0, y = 0;
-
-    Window win = wdt->winid;
-
-    if (wdt->disabled)
-	goto disabled;
-    if (wdt->options & BUTTON_PRESSED) {
-	render_bevel (win, x, y, x + w - 1, y + h - 1, 2, 1);
-    } else if (wdt->options & BUTTON_HIGHLIGHT) {
-	CSetColor (COLOR_FLAT);
-	XDrawRectangle (CDisplay, win, CGC, x + 1, y + 1, w - 3, h - 3);
-	render_bevel (win, x, y, x + w - 1, y + h - 1, 1, 0);
-    } else {
-  disabled:
-	CSetColor (COLOR_FLAT);
-	XDrawRectangle (CDisplay, win, CGC, x, y, w - 1, h - 1);
-	XDrawRectangle (CDisplay, win, CGC, x + 1, y + 1, w - 3, h - 3);
-    }
-
-    if (!wdt->label)
-	return;
-    if (!(*(wdt->label)))
-	return;
-    CSetColor (COLOR_BLACK);
-    drawstring_xy_hotkey (win, x + 2 + BUTTON_RELIEF, y + 2 + BUTTON_RELIEF, wdt->label, wdt->hotkey);
+    (*look->render_menu_button) (wdt);
 }
 
 
@@ -417,8 +315,6 @@ int eh_menubutton (CWidget * w, XEvent * xevent, CEvent * cwevent)
     CWidget *f;
     switch (xevent->type) {
     case FocusOut:
-	if (w->droppedmenu)	/* hide the menu if the user switches focus to another application */
-	    XUnmapWindow (CDisplay, w->droppedmenu->winid);
     case FocusIn:
 	render_menu_button (w);
 	CExposeWindowArea (w->parentid, 0, w->x - WIDGET_FOCUS_RING, w->y - WIDGET_FOCUS_RING, w->width + WIDGET_FOCUS_RING * 2, w->height + WIDGET_FOCUS_RING * 2);
@@ -496,7 +392,6 @@ int eh_menubutton (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	    w->options |= BUTTON_PRESSED;
 	}
 	render_menu_button (w);
-	grab_ungrab (w->winid);	/* stops the pointer from being contained in the window */
 	if (!w->droppedmenu)
 	    CMenuSelectionDialog (w);
 	return 1;
@@ -507,8 +402,10 @@ int eh_menubutton (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	render_menu_button (w);
 	return 1;
     case MotionNotify:
-	if (!w->droppedmenu && menu_mode == MENU_DRAGGING_MODE)
-	    pull_down_ungrab (w);
+	if (!w->droppedmenu && menu_grabbed) {
+	    pull_down (w);
+	    CFocus (w);
+	}
 	return 1;
     case EnterNotify:
 	w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
@@ -575,16 +472,13 @@ CWidget *CDrawMenuButton (const char *ident, Window parent, Window focus_return,
 	height = h + 4 + BUTTON_RELIEF * 2;
 
     wdt = CSetupWidget (ident, parent, x, y,
-			width, height, C_MENU_BUTTON_WIDGET, INPUT_BUTTON | PointerMotionMask, COLOR_FLAT, 1);
+			width, height, C_MENU_BUTTON_WIDGET, INPUT_KEY | OwnerGrabButtonMask, COLOR_FLAT, 1);
+    wdt->options |= MENU_AUTO_PULL_UP;		/* default */
 
-#if 0
-    set_hint_pos (x + width - XTextWidth (CFontStruct, " ", 1), y + height + WIDGET_SPACING);
-#else
     set_hint_pos (x + width, y + height + WIDGET_SPACING);
-#endif
     wdt->label = (char *) strdup (label);
     wdt->hotkey = find_hotkey (wdt);
-    wdt->options |= WIDGET_TAKES_FOCUS_RING | WIDGET_HOTKEY_ACTIVATES;
+    wdt->options |= WIDGET_HOTKEY_ACTIVATES;
 
     m = CMalloc ((num_items ? num_items : 1) * sizeof (struct menu_item));
 
@@ -594,7 +488,7 @@ CWidget *CDrawMenuButton (const char *ident, Window parent, Window focus_return,
 	text = va_arg (ap, char *);
 	text = text ? text : "";
 	m[i].text = (char *) strdup (catstrs (" ", text, " ", 0));
-	m[i].hot_key = va_arg (ap, char);
+	m[i].hot_key = va_arg (ap, int);
 	m[i].call_back = va_arg (ap, callfn);
 	m[i].data = va_arg (ap, unsigned long);
     }
@@ -608,25 +502,19 @@ CWidget *CDrawMenuButton (const char *ident, Window parent, Window focus_return,
     return wdt;
 }
 
-void CAddMenuItem (const char *ident, const char *text, int hot_key, callfn call_back, unsigned long data)
+void insert_menu_item (CWidget * w, int i, const char *text, int hot_key, callfn call_back, unsigned long data)
 {
     struct menu_item *m;
-    CWidget *w;
-
-    w = CIdent (ident);
-    if (!w) {
-	CErrorDialog (0, 0, 0, _ (" Add Menu Item "), " %s: %s ", _ ("No such menu"), ident);
-	return;
-    }
 
     m = CMalloc ((w->numlines + 1) * sizeof (struct menu_item));
-    memcpy (m, w->menu, w->numlines * sizeof (struct menu_item));
+    memcpy (m, w->menu, i * sizeof (struct menu_item));
+    memcpy (m + i + 1, w->menu + i, (w->numlines - i) * sizeof (struct menu_item));
     free (w->menu);
     w->menu = m;
-    m[w->numlines].text = (char *) strdup (catstrs (" ", text, " ", 0));
-    m[w->numlines].hot_key = hot_key;
-    m[w->numlines].call_back = call_back;
-    m[w->numlines].data = data;
+    m[i].text = (char *) strdup (catstrs (" ", text, " ", 0));
+    m[i].hot_key = hot_key;
+    m[i].call_back = call_back;
+    m[i].data = data;
 
     w->numlines++;
 
@@ -638,12 +526,39 @@ void CAddMenuItem (const char *ident, const char *text, int hot_key, callfn call
     }
 }
 
+void CAddMenuItem (const char *ident, const char *text, int hot_key, callfn call_back, unsigned long data)
+{
+    CWidget *w;
+    w = CIdent (ident);
+    if (!w) {
+	CErrorDialog (0, 0, 0, _ (" Add Menu Item "), " %s: %s ", _ ("No such menu"), ident);
+	return;
+    }
+    insert_menu_item (w, w->numlines, text, hot_key, call_back, data);
+}
+
+void CInsertMenuItem (const char *ident, const char *after, const char *text, int hot_key, callfn call_back, unsigned long data)
+{
+    int i;
+    CWidget *w;
+    w = CIdent (ident);
+    if (!w) {
+	CErrorDialog (0, 0, 0, _ (" Insert Menu Item "), " %s: %s ", _ ("No such menu"), ident);
+	return;
+    }
+    i = CHasMenuItem (ident, after);
+    if (i < 0) {
+	CErrorDialog (0, 0, 0, _ (" Insert Menu Item "), " %s: %s ", _ ("No such item"), after);
+	return;
+    }
+    insert_menu_item (w, i, text, hot_key, call_back, data);
+}
 
 static void remove_item (CWidget * w, int i)
 {
     if (!w)
 	return;
-    if (i >= w->numlines)
+    if (i >= w->numlines || i < 0)
 	return;
     if (w->menu[i].text)
 	free (w->menu[i].text);
@@ -666,67 +581,113 @@ void CRemoveMenuItemNumber (const char *ident, int i)
     remove_item (w, i);
 }
 
+/*
+   Starts from the bottom of the menu and searches for the first
+   menu item containing text (strstr != NULL), and returns the
+   integer.
+ */
+int CHasMenuItem (const char *ident, const char *text)
+{
+    CWidget *w;
+    int i;
+    w = CIdent (ident);
+    if (!w)
+	return -1;
+    if (w->numlines)
+	for (i = w->numlines - 1; i >= 0; i--)
+	    if (strstr (w->menu[i].text, text) || !*text)
+		return i;
+    return -1;
+}
+
 /*
    Starts from the bottom of the menu and searches for the first
    menu item containing text (strstr != NULL), and deletes it.
-   (this is untested).
  */
 void CRemoveMenuItem (const char *ident, const char *text)
+{
+    remove_item (CIdent (ident), CHasMenuItem (ident, text));
+}
+
+void CReplaceMenuItem (const char *ident, const char *old_text, const char *new_text, int hot_key, callfn call_back, unsigned long data)
 {
     struct menu_item *m;
     CWidget *w;
     int i;
 
     w = CIdent (ident);
-
-    if (!w->numlines)
+    if (!w) {
+	CErrorDialog (0, 0, 0, _ (" Replace Menu Item "), " %s: %s ", _ ("No such menu"), ident);
 	return;
-
-    m = w->menu;
-
-    for (i = w->numlines - 1; i >= 0; i--) {
-	if (!text)
-	    goto remove;
-	if (strstr (m[i].text, text) || !*text) {
-	  remove:;
-	    remove_item (w, i);
-	    render_menu (w->droppedmenu);
-	    return;
-	}
     }
+    i = CHasMenuItem (ident, old_text);
+    if (i < 0) {
+	CErrorDialog (0, 0, 0, _ (" Replace Menu Item "), " %s: %s ", _ ("No such item"), old_text);
+	return;
+    }
+    m = w->menu;
+    free (m[i].text);
+    m[i].text = (char *) strdup (catstrs (" ", new_text, " ", 0));
+    m[i].hot_key = hot_key;
+    m[i].call_back = call_back;
+    m[i].data = data;
+
+    if (w->droppedmenu != 0)
+	render_menu (w->droppedmenu);
 }
 
 void CMenuSelectionDialog (CWidget * button)
 {
     CEvent cwevent;
+    XEvent xevent;
 
-    if (menu_mode == MENU_DRAGGING_MODE) {
-	pull_down_ungrab (button);
-	return;
-    }
+/* sanity check */
     if (!button)
 	return;
 
-    if (!button->droppedmenu)
-	pull_down_ungrab (button);
-    else
+/* drop the menu and focus on the button */
+    pull_down (button);
+    CFocus (button);
+
+/* we are already inside this function */
+    if (menu_grabbed)
 	return;
+    menu_grabbed = 1;
 
-    menu_mode = MENU_DRAGGING_MODE;
+/* grab the pointer - we want events even when pointer is over other apps windows */
+    XGrabPointer
+	(CDisplay, button->winid, True,
+	 ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | PointerMotionMask |
+	 EnterWindowMask | LeaveWindowMask | OwnerGrabButtonMask, GrabModeAsync,
+	 GrabModeAsync, None, CGetCursorID (CURSOR_MENU), CurrentTime);
 
+/* current_pulled_button could go to zero if some handler does a pull_up of the menu */
     while (current_pulled_button) {
-	CNextEvent (0, &cwevent);
-	if (cwevent.kind == C_MENU_BUTTON_WIDGET || cwevent.kind == C_MENU_WIDGET) {
-	    continue;
-	} else {
-	    if ((cwevent.type == ButtonPress || cwevent.type == ButtonRelease) && current_pulled_button) {
-		pull_up (current_pulled_button);
-		current_pulled_button = 0;
-	    }
+	CNextEvent (&xevent, &cwevent);
+/* terminate the loop if we have a button press/release as the user expects */
+	if (xevent.type == ButtonRelease || xevent.type == ButtonPress) {
+	    CWidget *w;
+	    w = CWidgetOfWindow (xevent.xbutton.window);
+	    if (!w)
+		break;
+	    if (w->kind != C_MENU_BUTTON_WIDGET && w->kind != C_MENU_WIDGET)
+		break;
+	    if (xevent.xbutton.x >= w->width || xevent.xbutton.x < 0 ||
+		xevent.xbutton.y >= w->height || xevent.xbutton.y < 0)
+		break;
 	}
     }
+
+/* check if any menu up */
+    if (current_pulled_button) {
+	pull_up (current_pulled_button);
+	current_pulled_button = 0;
+    }
+
+/* ungrab and return focus */
+    menu_grabbed = 0;
+    XUngrabPointer (CDisplay, CurrentTime);
     CFocusLast ();
-    menu_mode = MENU_USER_MODE;
 }
 
 
