@@ -1,5 +1,5 @@
 /* textwidget.c - for drawing a scrollable text window widget
-   Copyright (C) 1996, 1997 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,9 +24,7 @@
 #include <stdlib.h>
 #include <stdarg.h>
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xresource.h>
+#include <X11/Intrinsic.h>
 #include <X11/Xatom.h>
 #include "lkeysym.h"
 
@@ -41,11 +39,13 @@
 
 #include "mad.h"
 
+extern struct look *look;
+
 int option_text_fg_normal = 0;
-int option_text_fg_bold = 17;
+int option_text_fg_bold = 1;
 int option_text_fg_italic = 18;
 
-int option_text_bg_normal = 22;
+int option_text_bg_normal = 26;
 int option_text_bg_marked = 25;
 int option_text_bg_highlighted = 12;
 
@@ -53,7 +53,6 @@ void selection_clear (void);
 
 static long current;
 
-extern int calc_text_len2 (CWidget *w, long b, long upto);
 extern int calc_text_pos2 (CWidget * w, long b, long *q, int l);
 extern void convert_text2 (CWidget * w, long from, cache_type *line, int x, int x_max, int row);
 void edit_translate_xy (int xs, int ys, int *x, int *y);
@@ -217,6 +216,7 @@ CWidget *CDrawTextbox (const char *identifier, Window parent, int x, int y,
     CWidget *wdt;
 
     int w, h;
+    CPushFont ("editor", 0);
     if (width == AUTO_WIDTH || height == AUTO_HEIGHT)
 	CTextSize (&w, &h, text);
     if (width == AUTO_WIDTH)
@@ -256,6 +256,7 @@ CWidget *CDrawTextbox (const char *identifier, Window parent, int x, int y,
     } else {
 	set_hint_pos (x + width + WIDGET_SPACING, y + height + WIDGET_SPACING);
     }
+    CPopFont ();
     return wdt;
 }
 
@@ -277,6 +278,7 @@ CWidget *CRedrawTextbox (const char *identifier, const char *text, int preserve)
 	    free (w->text);
 	w->text = (char *) strdup (text);
     }
+    CPushFont ("editor", 0);
     w->textlength = strlen (w->text);
     numlines = strcountlines (text, 0, 1000000000, w->options & TEXTBOX_WRAP ? (w->width - TEXTBOX_BDR) / FONT_MEAN_WIDTH : 32000) + 1;
     w->numlines = numlines;
@@ -296,6 +298,7 @@ CWidget *CRedrawTextbox (const char *identifier, const char *text, int preserve)
 	CSetTextboxPos (w, TEXT_SET_CURSOR_LINE, cursor);
     }
     CExpose (identifier);
+    CPopFont ();
 
     return w;
 }
@@ -304,8 +307,12 @@ CWidget *CRedrawTextbox (const char *identifier, const char *text, int preserve)
 char *CGetTextBoxLine (CWidget * w, int i)
 {
     int width;
+    char *r;
+    CPushFont ("editor", 0);
     width = w->options & TEXTBOX_WRAP ? (w->width - TEXTBOX_BDR) / FONT_MEAN_WIDTH : 32000;
-    return strline (w->text, strmovelines (w->text, w->current, i - w->firstline, width));
+    r = strline (w->text, strmovelines (w->text, w->current, i - w->firstline, width));
+    CPopFont ();
+    return r;
 }
 
 /* clears the text box */
@@ -347,6 +354,7 @@ int CSetTextboxPos (CWidget * wdt, int which, long p)
     int width, i, j;
     if (p < 0)
 	p = 0;
+    CPushFont ("editor", 0);
 
     width = wdt->options & TEXTBOX_WRAP ? (wdt->width - TEXTBOX_BDR) / FONT_MEAN_WIDTH : 32000;
 
@@ -354,6 +362,7 @@ int CSetTextboxPos (CWidget * wdt, int which, long p)
     case TEXT_SET_COLUMN:
 	i = wdt->firstcolumn;
 	wdt->firstcolumn = p;
+	CPopFont ();
 	return (i != wdt->firstcolumn);
     case TEXT_SET_LINE:
 	i = wdt->firstline;
@@ -368,6 +377,7 @@ int CSetTextboxPos (CWidget * wdt, int which, long p)
 	    wdt->firstline += strcountlines (wdt->text, wdt->current, q - wdt->current, width);
 	    wdt->current = q;
 	}
+	CPopFont ();
 	return (i != wdt->firstline);
     case TEXT_SET_POS:
 	i = wdt->firstline;
@@ -377,6 +387,7 @@ int CSetTextboxPos (CWidget * wdt, int which, long p)
 	    wdt->firstline += strcountlines (wdt->text, wdt->current, p - wdt->current, width);
 	    wdt->current = p;
 	}
+	CPopFont ();
 	return (i != wdt->firstline);
     case TEXT_SET_CURSOR_LINE:
 	i = wdt->firstline;
@@ -388,16 +399,17 @@ int CSetTextboxPos (CWidget * wdt, int which, long p)
 	wdt->cursor = p;
 	if (p < wdt->firstline)
 	    CSetTextboxPos (wdt, TEXT_SET_LINE, p);
-	else if (p > wdt->firstline + wdt->height / FONT_PIX_PER_LINE - 1)
-	    CSetTextboxPos (wdt, TEXT_SET_LINE, p - wdt->height / FONT_PIX_PER_LINE + 1);
+	else if (p > wdt->firstline + (wdt->height - FONT_PIX_PER_LINE - 6) / FONT_PIX_PER_LINE)
+	    CSetTextboxPos (wdt, TEXT_SET_LINE, p - (wdt->height - FONT_PIX_PER_LINE - 6) / FONT_PIX_PER_LINE);
+	CPopFont ();
 	return (i != wdt->firstline || j != wdt->cursor);
     }
 /* NLS ? */
     CError ("settextpos: command not found.\n");
+    CPopFont ();
     return 0;
 }
 
-extern unsigned char per_char[256];
 
 static void text_print_line (CWidget * w, long b, int row)
 {
@@ -406,7 +418,7 @@ static void text_print_line (CWidget * w, long b, int row)
 		     (int (*)(void *, long, long *, int)) calc_text_pos2,
 			    -w->firstcolumn * FONT_MEAN_WIDTH, w->winid,
 			    w->width, b, row, row * FONT_PIX_PER_LINE + EDIT_TEXT_VERTICAL_OFFSET, 0,
-			    per_char[' '] * TAB_SIZE);
+			    FONT_PER_CHAR(' ') * TAB_SIZE);
 }
 
 
@@ -431,6 +443,7 @@ long render_textbox (CWidget * w, int redrawall)
     long b;
     int c = 0, r = 0, row, height, isfocussed, wrap_width = 32000,
      curs, lines_drawn = 0;
+    CPushFont ("editor", 0);
     if (w->options & TEXTBOX_WRAP) {
 	wrap_width = (w->width - TEXTBOX_BDR) / FONT_MEAN_WIDTH;
 	if (w->resized) {	/* a resize will change the number lines if text is wrapping */
@@ -478,15 +491,10 @@ long render_textbox (CWidget * w, int redrawall)
 
     EditExposeRedraw = 0;
     EditClear = 0;
-
-    if (isfocussed) {
-	render_bevel (w->winid, 0, 0, w->width - 1, w->height - 1, 3, 1);	/*most outer border bevel */
-    } else {
-	render_bevel (w->winid, 2, 2, w->width - 3, w->height - 3, 1, 1);	/*border bevel */
-	render_bevel (w->winid, 0, 0, w->width - 1, w->height - 1, 2, 0);	/*most outer border bevel */
-    }
+    (*look->render_textbox_tidbits) (w, isfocussed);
     CSetColor (edit_normal_background_color);
     CLine (w->winid, 3, 3, 3, w->height - 4);
+    CPopFont ();
     return lines_drawn;
 }
 
@@ -503,6 +511,7 @@ long count_textbox_lines (CWidget * wdt, int all)
     int wrap_mode;
     unsigned char *text;
 
+    CPushFont ("editor", 0);
     nroff = (wdt->options & TEXTBOX_MAN_PAGE);
     wrap_mode = (wdt->options & TEXTBOX_WRAP);
     if (nroff)
@@ -534,6 +543,7 @@ long count_textbox_lines (CWidget * wdt, int all)
 	}
 	col++;
     }
+    CPopFont ();
     return row + 1;
 }
 
@@ -541,6 +551,7 @@ long count_textbox_lines (CWidget * wdt, int all)
 int CTextboxCursorMove (CWidget * wdt, KeySym key)
 {
     int handled = 0;
+    CPushFont ("editor", 0);
 /* when text is highlighted, the cursor must be off */
     if (wdt->options & TEXTBOX_NO_CURSOR || wdt->mark1 != wdt->mark2) {
 	int to_move = 0;
@@ -619,12 +630,15 @@ int CTextboxCursorMove (CWidget * wdt, KeySym key)
 	}
 	CSetTextboxPos (wdt, TEXT_SET_CURSOR_LINE, wdt->cursor);	/* just does some checks */
     }
+    CPopFont ();
     return handled;
 }
 
 static void text_mouse_mark (CWidget * w, XEvent * event, CEvent * ce)
 {
+    CPushFont ("editor", 0);
     mouse_mark (event, ce->double_click, w->funcs);
+    CPopFont ();
 }
 
 /* gets selected text into selection structure, stripping nroff */
@@ -668,6 +682,7 @@ int eh_textbox (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	break;
     case ButtonPress:
 	CFocus (w);
+	CPushFont ("editor", 0);
 	if (xevent->xbutton.button == Button1)
 	    w->cursor = (xevent->xbutton.y - TEXTBOX_BDR) / FONT_PIX_PER_LINE + w->firstline;
 	if (w->cursor > w->numlines - 1)
@@ -677,6 +692,7 @@ int eh_textbox (CWidget * w, XEvent * xevent, CEvent * cwevent)
 	cwevent->ident = w->ident;
 	cwevent->xt = (xevent->xbutton.x - 7) / FONT_MEAN_WIDTH + w->firstcolumn;
 	cwevent->yt = w->cursor;
+	CPopFont ();
     case ButtonRelease:
     case MotionNotify:
 	if (!xevent->xmotion.state && xevent->type == MotionNotify)
@@ -715,6 +731,7 @@ void link_scrollbar_to_textbox (CWidget * scrollbar, CWidget * textbox, XEvent *
 {
     int redrawtext = 0, count = -1, c;
     static int r = 0;
+    CPushFont ("editor", 0);
     if ((xevent->type == ButtonRelease || xevent->type == MotionNotify) && whichscrbutton == 3) {
 	redrawtext = CSetTextboxPos (textbox, TEXT_SET_LINE, (double) scrollbar->firstline * textbox->numlines / 65535.0);
     } else if (xevent->type == ButtonPress && (cwevent->button == Button1 || cwevent->button == Button2)) {
@@ -755,6 +772,8 @@ void link_scrollbar_to_textbox (CWidget * scrollbar, CWidget * textbox, XEvent *
 	count = 1;
     scrollbar->firstline = (double) 65535.0 *textbox->firstline / textbox->numlines;
     scrollbar->numlines = (double) 65535.0 *count / textbox->numlines;
+    CPopFont ();
+    return;
 }
 
 
