@@ -1,5 +1,5 @@
 /* xdnd.c, xdnd.h - C program library for handling the Xdnd protocol
-   Copyright (C) 1998 Paul Sheer
+   Copyright (C) 1996-2000 Paul Sheer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,26 +22,51 @@
    Released 1998-08-07
    Changes:
 
+   2000-08-08: INCR protocol implemented.
 
 */
 
 /*
+    DONE:
+     - INCR protocol now implemented
+
     TODO:
      - action_choose_dialog not yet supported (never called)
      - widget_delete_selection not yet supported and DELETE requests are ignored
      - not yet tested with applications that only supported XDND 0 or 1
-     - INCR protocol not yet implemented
 */
 
+#include <config.h>
 #include <X11/Xlib.h>
 #include <X11/X.h>
 #include <X11/Xatom.h>
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
+
+#if TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+#include <sys/types.h>
+#ifdef HAVE_UNISTD_H
+#   include <unistd.h>
+#endif
+
+#ifdef HAVE_SYS_SELECT_H
+#  include <sys/select.h>
+#endif
+
 #include "xdnd.h"
 #include "mad.h"
 
+
 /* just to remind us : */
 
 #if 0
@@ -67,10 +92,23 @@ XClientMessageEvent xclient;
 #define xdnd_xfree(x) {if (x) { free (x); x = 0; }}
 
 #ifdef DND_DEBUG
-#define dnd_debug1(a)       printf("%s: %d: " a "\n", __FILE__, __LINE__)
-#define dnd_debug2(a,b)     printf("%s: %d: " a "\n", __FILE__, __LINE__ , b)
-#define dnd_debug3(a,b,c)   printf("%s: %d: " a "\n", __FILE__, __LINE__ , b, c)
-#define dnd_debug4(a,b,c,d) printf("%s: %d: " a "\n", __FILE__, __LINE__ , b, c, d)
+
+#include <sys/time.h>
+#include <unistd.h>
+
+char *xdnd_debug_milliseconds (void)
+{
+    struct timeval tv;
+    static char r[22];
+    gettimeofday (&tv, 0);
+    sprintf (r, "%.2ld.%.3ld", tv.tv_sec % 100L, tv.tv_usec / 1000L);
+    return r;
+}
+
+#define dnd_debug1(a)       printf("%s: %d: %s: " a "\n", __FILE__, __LINE__, xdnd_debug_milliseconds ())
+#define dnd_debug2(a,b)     printf("%s: %d: %s: " a "\n", __FILE__, __LINE__, xdnd_debug_milliseconds (), b)
+#define dnd_debug3(a,b,c)   printf("%s: %d: %s: " a "\n", __FILE__, __LINE__, xdnd_debug_milliseconds (), b, c)
+#define dnd_debug4(a,b,c,d) printf("%s: %d: %s: " a "\n", __FILE__, __LINE__, xdnd_debug_milliseconds (), b, c, d)
 #else
 #define dnd_debug1(a)       
 #define dnd_debug2(a,b)     
@@ -80,8 +118,7 @@ XClientMessageEvent xclient;
 
 #define dnd_warning(a) fprintf (stderr, a)
 
-#define dnd_version_at_least(a,b) ((a) <= (b))
-
+#define dnd_version_at_least(a,b) ((a) >= (b))
 
 static unsigned char dnd_copy_cursor_bits[] =
 {
@@ -195,6 +232,7 @@ void xdnd_reset (DndClass * dnd)
     dnd->rectangle.x = dnd->rectangle.y = 0;
     dnd->rectangle.width = dnd->rectangle.height = 0;
     dnd->dropper_window = 0;
+    dnd->dropper_toplevel = 0;
     dnd->dragger_window = 0;
     dnd->dragger_typelist = 0;
     dnd->desired_type = 0;
@@ -273,13 +311,33 @@ static int array_length (Atom * a)
 
 void xdnd_set_dnd_aware (DndClass * dnd, Window window, Atom * typelist)
 {
-    XChangeProperty (dnd->display, window, dnd->XdndAware, XA_ATOM, 32, PropModeReplace, (unsigned char *) &dnd->version, 1);
-    if (typelist) {
+    Window root_return, parent;
+    unsigned int nchildren_return;
+    Window *children_return = 0;
+    int r, s;
+    if (dnd->widget_exists)
+	if (!(*dnd->widget_exists) (dnd, window))
+	    return;
+    s = XChangeProperty (dnd->display, window, dnd->XdndAware, XA_ATOM, 32, PropModeReplace,
+			 (unsigned char *) &dnd->version, 1);
+#if 0
+    dnd_debug4 ("XChangeProperty() = %d, window = %ld, widget = %s", s, window,
+		(char *) CWidgetOfWindow (window));
+#endif
+    if (s && typelist) {
 	int n;
 	n = array_length (typelist);
 	if (n)
-	    XChangeProperty (dnd->display, window, dnd->XdndAware, XA_ATOM, 32, PropModeAppend, (unsigned char *) typelist, n);
+	    s = XChangeProperty (dnd->display, window, dnd->XdndAware, XA_ATOM, 32, PropModeAppend,
+				 (unsigned char *) typelist, n);
     }
+    r =
+	XQueryTree (dnd->display, window, &root_return, &parent, &children_return,
+		    &nchildren_return);
+    if (children_return)
+	XFree (children_return);
+    if (r)
+	xdnd_set_dnd_aware (dnd, parent, typelist);
 }
 
 int xdnd_is_dnd_aware (DndClass * dnd, Window window, int *version, Atom * typelist)
@@ -497,6 +555,9 @@ int xdnd_choose_action_dialog (DndClass * dnd, Atom * actions, char **descriptio
 
 static void xdnd_send_event (DndClass * dnd, Window window, XEvent * xevent)
 {
+    dnd_debug4 ("xdnd_send_event(), window = %ld, l[0] = %ld, l[4] = %ld",
+    window, xevent->xclient.data.l[0], xevent->xclient.data.l[4]);
+    dnd_debug2 ("xdnd_send_event(), from widget widget %s", (char *) CWidgetOfWindow (xevent->xclient.data.l[0]));
     XSendEvent (dnd->display, window, 0, 0, xevent);
 }
 
@@ -628,7 +689,7 @@ void xdnd_send_finished (DndClass * dnd, Window window, Window from, int error)
 /* returns non-zero on error - i.e. no selection owner set. Type is of course the mime type */
 int xdnd_convert_selection (DndClass * dnd, Window window, Window requester, Atom type)
 {
-    if (window != XGetSelectionOwner (dnd->display, dnd->XdndSelection)) {
+    if (!(window = XGetSelectionOwner (dnd->display, dnd->XdndSelection))) {
 	dnd_debug1 ("xdnd_convert_selection(): XGetSelectionOwner failed");
 	return 1;
     }
@@ -665,6 +726,7 @@ void xdnd_selection_send (DndClass * dnd, XSelectionRequestEvent * request, unsi
     xdnd_send_event (dnd, request->requestor, &xevent);
 }
 
+#if 0
 /* respond to a notification that a primary selection has been sent */
 int xdnd_get_selection (DndClass * dnd, Window from, Atom property, Window insert)
 {
@@ -693,6 +755,89 @@ int xdnd_get_selection (DndClass * dnd, Window from, Atom property, Window inser
     } while (remaining);
     return error;
 }
+#endif
+
+static int paste_prop_internal (DndClass * dnd, Window from, Window insert, unsigned long prop, int delete_prop)
+{
+    long nread = 0;
+    unsigned long nitems;
+    unsigned long bytes_after;
+    int error = 0;
+    do {
+	Atom actual_type;
+	int actual_fmt;
+	unsigned char *s = 0;
+	if (XGetWindowProperty (dnd->display, insert, prop,
+				nread / 4, 65536, delete_prop,
+				AnyPropertyType, &actual_type, &actual_fmt,
+				&nitems, &bytes_after, &s) != Success) {
+	    XFree (s);
+	    return 1;
+	}
+	nread += nitems;
+	if (dnd->widget_insert_drop && !error)
+	    error = (*dnd->widget_insert_drop) (dnd, s, nitems, bytes_after, insert, from, actual_fmt);
+	XFree (s);
+    } while (bytes_after);
+    if (!nread)
+	return 1;
+    return 0;
+}
+
+/*
+ * Respond to a notification that a primary selection has been sent (supports INCR)
+ */
+int xdnd_get_selection (DndClass * dnd, Window from, Atom prop, Window insert)
+{
+    struct timeval tv, tv_start;
+    long nread;
+    unsigned long bytes_after;
+    Atom actual_type;
+    int actual_fmt;
+    unsigned long nitems;
+    unsigned char *s = 0;
+    if (prop == None)
+	return 1;
+    nread = 0;
+    if (XGetWindowProperty
+	(dnd->display, insert, prop, 0, 8, False, AnyPropertyType, &actual_type, &actual_fmt,
+	 &nitems, &bytes_after, &s) != Success) {
+	XFree (s);
+	return 1;
+    }
+    XFree (s);
+    if (actual_type != XInternAtom (dnd->display, "INCR", False))
+	return paste_prop_internal (dnd, from, insert, prop, True);
+    XDeleteProperty (dnd->display, insert, prop);
+    gettimeofday (&tv_start, 0);
+    for (;;) {
+	long t;
+	fd_set r;
+	XEvent xe;
+	if (XCheckMaskEvent (dnd->display, PropertyChangeMask, &xe)) {
+	    if (xe.type == PropertyNotify && xe.xproperty.state == PropertyNewValue) {
+/* time between arrivals of data */
+		gettimeofday (&tv_start, 0);
+		if (paste_prop_internal (dnd, from, insert, prop, True))
+		    break;
+	    }
+	} else {
+	    tv.tv_sec = 0;
+	    tv.tv_usec = 10000;
+	    FD_ZERO (&r);
+	    FD_SET (ConnectionNumber (dnd->display), &r);
+	    select (ConnectionNumber (dnd->display) + 1, &r, 0, 0, &tv);
+	    if (FD_ISSET (ConnectionNumber (dnd->display), &r))
+		continue;
+	}
+	gettimeofday (&tv, 0);
+	t = (tv.tv_sec - tv_start.tv_sec) * 1000000L + (tv.tv_usec - tv_start.tv_usec);
+/* no data for five seconds, so quit */
+	if (t > 5000000L)
+	    return 1;
+    }
+    return 0;
+}
 
 
 int outside_rectangle (int x, int y, XRectangle * r)
@@ -716,13 +861,16 @@ static float xdnd_sqrt (float x)
     return ans;
 }
 
+#define print_marks print_win_marks(from,__FILE__,__LINE__);
+
 /* returns action on success, 0 otherwise */
 Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 {
     XEvent xevent, xevent_temp;
     Window over_window = 0, last_window = 0;
 #if XDND_VERSION >= 3
-    Windiow last_toplevel_window = 0, toplevel_window = 0;
+    Window last_dropper_toplevel = 0;
+    int internal_dropable = 1;
 #endif
     int n;
     DndCursor *cursor;
@@ -782,6 +930,7 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		      cursor->cursor, CurrentTime) != GrabSuccess)
 	dnd_debug1 ("Unable to grab pointer");
 
+
     while (xevent.xany.type != ButtonRelease) {
 	XAllowEvents (dnd->display, SyncPointer, CurrentTime);
 	XNextEvent (dnd->display, &xevent);
@@ -802,10 +951,10 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 	    break;
 	case MotionNotify:
 	    dnd_aware = 0;
+	    dnd->dropper_toplevel = 0;
 	    memcpy (&xevent_temp, &xevent, sizeof (xevent));
-	    if (!xevent.xmotion.subwindow) {
-		xevent.xmotion.subwindow = xevent.xmotion.window;
-	    } else {
+            xevent.xmotion.subwindow = xevent.xmotion.window;
+            {
 		Window root_return, child_return;
 		int x_temp, y_temp;
 		unsigned int mask_return;
@@ -813,12 +962,13 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 				      &x_temp, &y_temp, &xevent.xmotion.x,
 				      &xevent.xmotion.y, &mask_return)) {
 #if XDND_VERSION >= 3
-		    if (!dnd_aware)
+		    if (!dnd_aware) {
 			if ((dnd_aware = xdnd_is_dnd_aware (dnd, xevent.xmotion.subwindow, &dnd->dragging_version, typelist))) {
-			    toplevel_window = xevent.xmotion.subwindow;
+			    dnd->dropper_toplevel = xevent.xmotion.subwindow;
 			    xevent.xmotion.x_root = x_temp;
 			    xevent.xmotion.y_root = y_temp;
 			}
+                    }
 #else
 		    xevent.xmotion.x_root = x_temp;
 		    xevent.xmotion.y_root = y_temp;
@@ -830,11 +980,18 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		break;
 	    }
 	  found_descendent:
+
 /* last_window is just for debug purposes */
 	    if (last_window != xevent.xmotion.subwindow) {
 		dnd_debug2 ("window crossing to %ld", xevent.xmotion.subwindow);
 		dnd_debug2 ("  current window is %ld", over_window);
+		dnd_debug3 ("     last_window = %ld, xmotion.subwindow = %ld", last_window, xevent.xmotion.subwindow);
+#if XDND_VERSION >= 3
+		dnd_debug3 ("     dropper_toplevel = %ld, last_dropper_toplevel.subwindow = %ld", dnd->dropper_toplevel, last_dropper_toplevel);
+#endif
+		dnd_debug3 ("     dnd_aware = %d, dnd->options & XDND_OPTION_NO_HYSTERESIS = %ld", dnd_aware, (long) dnd->options & XDND_OPTION_NO_HYSTERESIS);
 	    }
+
 #if XDND_VERSION < 3
 /* is the new window dnd aware? if not stay in the old window */
 	    if (over_window != xevent.xmotion.subwindow &&
@@ -845,8 +1002,13 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		    (dnd->options & XDND_OPTION_NO_HYSTERESIS)
 		))
 #else
-	    if (toplevel_window != last_toplevel_window &&
-		last_window != xevent.xmotion.subwindow &&
+            internal_dropable = 1;
+            if (dnd->widget_exists && (*dnd->widget_exists) (dnd, xevent.xmotion.subwindow))
+                if (!xdnd_is_dnd_aware (dnd, xevent.xmotion.subwindow, &dnd->dragging_version, typelist))
+                    internal_dropable = 0;
+	    dnd_debug3 ("dnd->dropper_toplevel = %ld, last_dropper_toplevel = %ld\n", dnd->dropper_toplevel, last_dropper_toplevel);
+	    if ((dnd->dropper_toplevel != last_dropper_toplevel ||
+		last_window != xevent.xmotion.subwindow) && internal_dropable &&
 		(
 		    (dnd_aware)
 		    ||
@@ -868,7 +1030,11 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 #if XDND_VERSION < 3
 			    xdnd_send_leave (dnd, over_window, from);
 #else
-			    xdnd_send_leave (dnd, last_toplevel_window, from);
+                            if (dnd->dropper_toplevel != last_dropper_toplevel) {
+			        xdnd_send_leave (dnd, last_dropper_toplevel, from);
+			    } else {
+				dnd_debug1 ("    not sending leave --> dnd->dropper_toplevel == last_dropper_toplevel");
+			    }
 #endif
 			}
 			dnd->internal_drag = 0;
@@ -880,9 +1046,6 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		}
 /* entering window we are currently over */
 		over_window = xevent.xmotion.subwindow;
-#if XDND_VERSION >= 3
-		last_toplevel_window = toplevel_window;
-#endif
 		if (dnd_aware) {
 		    dnd_debug1 ("  is dnd aware");
 		    dnd->stage = XDND_DRAG_STAGE_ENTERED;
@@ -895,7 +1058,8 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 #if XDND_VERSION < 3
 			xdnd_send_enter (dnd, over_window, from, typelist);
 #else
-			xdnd_send_enter (dnd, toplevel_window, from, typelist);
+                        if (dnd->dropper_toplevel != last_dropper_toplevel)
+			    xdnd_send_enter (dnd, dnd->dropper_toplevel, from, typelist);
 #endif
 		    }
 		    dnd->want_position = 1;
@@ -906,6 +1070,9 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		    XSendEvent (dnd->display, from, 0, ButtonMotionMask, &xevent_temp);
 		    XSync (dnd->display, 0);
 		}
+#if XDND_VERSION >= 3
+		last_dropper_toplevel = dnd->dropper_toplevel;
+#endif
 /* we are now officially in a new window */
 	    } else {
 /* got here, so we are just moving `inside' the same window */
@@ -914,7 +1081,7 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		    dnd_debug1 ("got motion at right stage");
 		    dnd->x = xevent.xmotion.x_root;
 		    dnd->y = xevent.xmotion.y_root;
-		    if (dnd->want_position && outside_rectangle (dnd->x, dnd->y, &dnd->rectangle)) {
+		    if (dnd->want_position || outside_rectangle (dnd->x, dnd->y, &dnd->rectangle)) {
 			dnd_debug1 ("  want position and outside rectangle");
 			if (dnd->internal_drag) {
 			    dnd_debug1 ("    our own widget");
@@ -928,15 +1095,18 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 			    }
 			    dnd_debug2 ("      return action=%ld", dnd->supported_action);
 			} else {
-			    dnd_debug2 ("    not our own widget - sending XdndPosition to %d", (int) over_window);
 #if XDND_VERSION < 3
+			    dnd_debug3 ("    not our own widget - sending XdndPosition to %ld, action %ld", over_window, action);
 			    xdnd_send_position (dnd, over_window, from, action, dnd->x, dnd->y, xevent.xmotion.time);
 #else
-			    xdnd_send_position (dnd, toplevel_window, from, action, dnd->x, dnd->y, xevent.xmotion.time);
+			    dnd_debug3 ("    not our own widget - sending XdndPosition to %ld, action %ld", dnd->dropper_toplevel, action);
+			    xdnd_send_position (dnd, dnd->dropper_toplevel, from, action, dnd->x, dnd->y, xevent.xmotion.time);
 #endif
 			}
+		    } else if (dnd->want_position) {
+			dnd_debug1 ("  inside rectangle");
 		    } else {
-			dnd_debug1 ("  doesn't want position or inside rectangle");
+			dnd_debug1 ("  doesn't want position");
 		    }
 		}
 	    }
@@ -946,7 +1116,11 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 	    dnd_debug1 ("ClientMessage recieved");
 	    if (xevent.xclient.message_type == dnd->XdndStatus && !dnd->internal_drag) {
 		dnd_debug1 ("  XdndStatus recieved");
-		if (dnd->stage == XDND_DRAG_STAGE_ENTERED && XDND_STATUS_TARGET_WIN (&xevent) == dnd->dropper_window) {
+		if (dnd->stage == XDND_DRAG_STAGE_ENTERED 
+#if XDND_VERSION < 3
+                        && XDND_STATUS_TARGET_WIN (&xevent) == dnd->dropper_window
+#endif
+                ) {
 		    dnd_debug1 ("    XdndStatus stage correct, dropper window correct");
 		    dnd->want_position = XDND_STATUS_WANT_POSITION (&xevent);
 		    dnd->ready_to_drop = XDND_STATUS_WILL_ACCEPT (&xevent);
@@ -965,9 +1139,15 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		    }
 		    dnd_debug3 ("      rectangle = (x=%d, y=%d, ", dnd->rectangle.x, dnd->rectangle.y);
 		    dnd_debug4                               ("w=%d, h=%d), want_position=%d\n", dnd->rectangle.width, dnd->rectangle.height, dnd->want_position);
-		} else {
-		    dnd_debug1 ("    XdndStatus stage incorrect or dropper window incorrect");
 		}
+#if XDND_VERSION < 3
+                else if (XDND_STATUS_TARGET_WIN (&xevent) != dnd->dropper_window) {
+		    dnd_debug3 ("    XdndStatus XDND_STATUS_TARGET_WIN (&xevent) = %ld, dnd->dropper_window = %ld", XDND_STATUS_TARGET_WIN (&xevent), dnd->dropper_window);
+		}
+#endif
+                else {
+		    dnd_debug2 ("    XdndStatus stage incorrect dnd->stage = %d", dnd->stage);
+                }
 	    }
 	    break;
 	case SelectionRequest:{
@@ -1011,7 +1191,11 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 	    }
 	} else {
 	    xdnd_set_selection_owner (dnd, from, dnd->desired_type);
+#if XDND_VERSION < 3
 	    xdnd_send_drop (dnd, dnd->dropper_window, from, time);
+#else
+	    xdnd_send_drop (dnd, dnd->dropper_toplevel, from, time);
+#endif
 	}
 	if (!dnd->internal_drag)
 	    for (;;) {
@@ -1019,11 +1203,15 @@ Atom xdnd_drag (DndClass * dnd, Window from, Atom action, Atom * typelist)
 		XNextEvent (dnd->display, &xevent);
 		if (xevent.type == ClientMessage && xevent.xclient.message_type == dnd->XdndFinished) {
 		    dnd_debug1 ("XdndFinished");
+#if XDND_VERSION < 3
 		    if (XDND_FINISHED_TARGET_WIN (&xevent) == dnd->dropper_window) {
+#endif
 			dnd_debug2 ("  source correct - exiting event loop, action=%ld", dnd->supported_action);
 			result = dnd->supported_action;		/* success - so return action to caller */
 			break;
+#if XDND_VERSION < 3
 		    }
+#endif
 		} else if (xevent.type == Expose) {
 		    if (dnd->handle_expose_events)
 			(*dnd->handle_expose_events) (dnd, &xevent);
@@ -1069,8 +1257,14 @@ int xdnd_handle_drop_events (DndClass * dnd, XEvent * xevent)
 	    error = xdnd_get_selection (dnd, dnd->dragger_window, xevent->xselection.property, xevent->xany.window);
 /* error is not actually used, i think future versions of the protocol maybe should return 
    an error status to the calling window with the XdndFinished client message */
-	    if (dnd_version_at_least (dnd->dragging_version, 2))
+	    if (dnd_version_at_least (dnd->dragging_version, 2)) {
+#if XDND_VERSION >= 3
+		xdnd_send_finished (dnd, dnd->dragger_window, dnd->dropper_toplevel, error);
+#else
 		xdnd_send_finished (dnd, dnd->dragger_window, dnd->dropper_window, error);
+#endif
+	        dnd_debug1 ("    sending finished");
+            }
 	    xdnd_xfree (dnd->dragger_typelist);
 	    xdnd_reset (dnd);
 	    dnd->stage = XDND_DROP_STAGE_IDLE;
@@ -1079,16 +1273,23 @@ int xdnd_handle_drop_events (DndClass * dnd, XEvent * xevent)
 	    dnd_debug1 ("  property is not Xdnd_NON_PROTOCOL_ATOM - ignoring");
 	}
     } else if (xevent->type == ClientMessage) {
-	dnd_debug1 ("got ClientMessage");
+	dnd_debug2 ("got ClientMessage to xevent->xany.window = %ld", xevent->xany.window);
 	if (xevent->xclient.message_type == dnd->XdndEnter) {
-	    dnd_debug1 ("  message_type is XdndEnter");
+	    dnd_debug2 ("  message_type is XdndEnter, version = %ld", XDND_ENTER_VERSION (xevent));
 #if XDND_VERSION >= 3
 	    if (XDND_ENTER_VERSION (xevent) < 3)
 		return 0;
 #endif
 	    xdnd_reset (dnd);
 	    dnd->dragger_window = XDND_ENTER_SOURCE_WIN (xevent);
+#if XDND_VERSION >= 3
+            dnd->dropper_toplevel = xevent->xany.window;
+	    dnd->dropper_window = 0;     /* enter goes to the top level window only,
+                                            so we don't really know what the
+                                            sub window is yet */
+#else
 	    dnd->dropper_window = xevent->xany.window;
+#endif
 	    xdnd_xfree (dnd->dragger_typelist);
 	    if (XDND_ENTER_THREE_TYPES (xevent)) {
 		dnd_debug1 ("    three types only");
@@ -1104,6 +1305,10 @@ int xdnd_handle_drop_events (DndClass * dnd, XEvent * xevent)
 	    dnd->dragging_version = XDND_ENTER_VERSION (xevent);
 	    result = 1;
 	} else if (xevent->xclient.message_type == dnd->XdndLeave) {
+#if XDND_VERSION >= 3
+            if (xevent->xany.window == dnd->dropper_toplevel && dnd->dropper_window)
+                xevent->xany.window = dnd->dropper_window;
+#endif
 	    dnd_debug1 ("  message_type is XdndLeave");
 	    if (dnd->dragger_window == XDND_LEAVE_SOURCE_WIN (xevent) && dnd->stage == XDND_DROP_STAGE_ENTERED) {
 		dnd_debug1 ("    leaving");
@@ -1112,15 +1317,43 @@ int xdnd_handle_drop_events (DndClass * dnd, XEvent * xevent)
 		dnd->stage = XDND_DROP_STAGE_IDLE;
 		xdnd_xfree (dnd->dragger_typelist);
 		result = 1;
+                dnd->dropper_toplevel = dnd->dropper_window = 0;
 	    } else {
 		dnd_debug1 ("    wrong stage or from wrong window");
 	    }
 	} else if (xevent->xclient.message_type == dnd->XdndPosition) {
-	    dnd_debug1 ("  message_type is XdndPosition");
+	    dnd_debug2 ("  message_type is XdndPosition to %ld", xevent->xany.window);
 	    if (dnd->dragger_window == XDND_POSITION_SOURCE_WIN (xevent) && dnd->stage == XDND_DROP_STAGE_ENTERED) {
 		int want_position;
 		Atom action;
 		XRectangle rectangle;
+                Window last_window;
+                last_window = dnd->dropper_window;
+#if XDND_VERSION >= 3
+/* version 3 gives us the top-level window only. WE have to find the child that the pointer is over: */
+                if (1 || xevent->xany.window != dnd->dropper_toplevel || !dnd->dropper_window) {
+                    Window parent, child, new_child = 0;
+                    dnd->dropper_toplevel = xevent->xany.window;
+                    parent = dnd->root_window;
+                    child = dnd->dropper_toplevel;
+                    for (;;) {
+                        int xd, yd;
+                        new_child = 0;
+                        if (!XTranslateCoordinates (dnd->display, parent, child, 
+                                    XDND_POSITION_ROOT_X (xevent), XDND_POSITION_ROOT_Y (xevent),
+                                    &xd, &yd, &new_child))
+                            break;
+                        if (!new_child)
+                            break;
+                        child = new_child;
+                    }
+                    dnd->dropper_window = xevent->xany.window = child;
+	            dnd_debug2 ("   child window translates to %ld", dnd->dropper_window);
+                } else if (xevent->xany.window == dnd->dropper_toplevel && dnd->dropper_window) {
+                    xevent->xany.window = dnd->dropper_window;
+	            dnd_debug2 ("   child window previously found: %ld", dnd->dropper_window);
+                }
+#endif
 		action = dnd->XdndActionCopy;
 		dnd->supported_action = dnd->XdndActionCopy;
 		dnd->x = XDND_POSITION_ROOT_X (xevent);
@@ -1130,18 +1363,33 @@ int xdnd_handle_drop_events (DndClass * dnd, XEvent * xevent)
 		    dnd->time = XDND_POSITION_TIME (xevent);
 		if (dnd_version_at_least (dnd->dragging_version, 1))
 		    action = XDND_POSITION_ACTION (xevent);
+#if XDND_VERSION >= 3
+                if (last_window && last_window != xevent->xany.window)
+		    if (dnd->widget_apply_leave)
+		        (*dnd->widget_apply_leave) (dnd, last_window);
+#endif
 		dnd->will_accept = (*dnd->widget_apply_position) (dnd, xevent->xany.window, dnd->dragger_window,
 		action, dnd->x, dnd->y, dnd->time, dnd->dragger_typelist,
 								  &want_position, &dnd->supported_action, &dnd->desired_type, &rectangle);
 		dnd_debug2 ("    will accept = %d", dnd->will_accept);
-		dnd_debug1 ("    sending status");
+#if XDND_VERSION >= 3
+		dnd_debug2 ("    sending status of %ld", dnd->dropper_toplevel);
+		xdnd_send_status (dnd, dnd->dragger_window, dnd->dropper_toplevel, dnd->will_accept,
+				  want_position, rectangle.x, rectangle.y, rectangle.width, rectangle.height, dnd->supported_action);
+#else
+		dnd_debug2 ("    sending status of %ld", xevent->xany.window);
 		xdnd_send_status (dnd, dnd->dragger_window, xevent->xany.window, dnd->will_accept,
 				  want_position, rectangle.x, rectangle.y, rectangle.width, rectangle.height, dnd->supported_action);
+#endif
 		result = 1;
 	    } else {
 		dnd_debug1 ("    wrong stage or from wrong window");
 	    }
 	} else if (xevent->xclient.message_type == dnd->XdndDrop) {
+#if XDND_VERSION >= 3
+            if (xevent->xany.window == dnd->dropper_toplevel && dnd->dropper_window)
+                xevent->xany.window = dnd->dropper_window;
+#endif
 	    dnd_debug1 ("  message_type is XdndDrop");
 	    if (dnd->dragger_window == XDND_DROP_SOURCE_WIN (xevent) && dnd->stage == XDND_DROP_STAGE_ENTERED) {
 		dnd->time = CurrentTime;
@@ -1155,8 +1403,13 @@ int xdnd_handle_drop_events (DndClass * dnd, XEvent * xevent)
 		    dnd->stage = XDND_DROP_STAGE_CONVERTING;
 		} else {
 		    dnd_debug1 ("    will_accept is false - sending finished");
-		    if (dnd_version_at_least (dnd->dragging_version, 1))
+		    if (dnd_version_at_least (dnd->dragging_version, 2)) {
+#if XDND_VERSION >= 3
+			xdnd_send_finished (dnd, dnd->dragger_window, dnd->dropper_toplevel, 1);
+#else
 			xdnd_send_finished (dnd, dnd->dragger_window, xevent->xany.window, 1);
+#endif
+                    }
 		    xdnd_xfree (dnd->dragger_typelist);
 		    xdnd_reset (dnd);
 		    dnd->stage = XDND_DROP_STAGE_IDLE;
